{{
FifoBig.spin -- Spin code for a Ring buffer (Fifo) of "large" objects

Author:         Tom Dinger (Ding-Batty)
Release Date:   2024/02/21

Copyright 2024 Thomas Dinger

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

  A FIFO buffer manages a fixed-size of entries, implemented with a ring
  buffer, with a separate header that describes the fifo, and spin code that
  implements insertions and deletions.

  The implementation assumes Fifo entries are fixed size, and the header
  and ring buffer can be stored separately, so the ring buffer memory can
  be allocated over hub memory "reclaimed" from PASM code loaded into cogs.
  The implementation uses no locks, as long as a single buffer has exaclty
  one cog that writes to it, and one that reads from it.

  This particular Fifo version assumes "large" entries, larger than 4 BYTEs or
  one LONG. It has insert and remove functions that can copy the entries into
  or out of the Fifo, but the expected use-case will be for the Fifo to provide
  the address of the entry to read or write, and the caller will fill in
  or extract the data.

  The Fifo uses the convention that when the insert and remove pointers are the
  same, that the Fifo is empty. This means that the Fifo is full when the
  insert pointer, when incremented, becomes equal to the remove pointer, which
  means that one entry in the Fifo is "wasted".

  Because hub addresses fit into 16 bits, the Fifo header fields are all
  WORD entries to save a bit of memory for no real slow-down for Spin code.
  It contains the following fields:

    InsertHubAddress   WORD   0  The address of the start of the next entry
                                 into which is written the entry being
                                 inserted, when there is room.
    RemoveHubAddress   WORD   0  The address of the start of the next entry
                                 from which is read the entry being removed,
                                 when the fifo is not empty.
    StartHubAddress    WORD   0  The lowest address for the entry storage
    LimitHubAddress    WORD   0  The address just after the highest address for
                                 entry storage, used to "wrap around" the
                                 insert and remove addresses.
    EntrySizeBytes     WORD   0  The size of one entry, in bytes.
                                 Usually this is a multiple of 4 bytes, so
                                 that entries are on a LONG boundary.

  Empty/Not Empty/Full Checking
  =============================

  This Fifo uses the simple version of empty and full:

  The Fifo is empty if:
    InsertHubAddress == RemoveHubAddress

  The Fifo is full if
    AdvancePtr(InsertHubAddress) == RemoveHubAddress

  Normal insert and remove routines only need to know if the Fifo is full or
  empty: If full, insertion will stall; if empty, removal will stall.

  It is not simple to determine the number of free entries in the Fifo at any
  given time; the simplest way might be:
    if InsertHubAddress > RemoveHubAddress
      num_free_entries "= ((RemoveHubAddress + LimitHubAddress - StartHubAddress - InsertHubAddress) / EntrySizeBytes) - 1
    else
      num_free_entries := ((RemoveHubAddress - InsertHubAddress) / EntrySizeBytes) - 1

  Similarly, the number of entries present in the Fifo is given by:
    if InsertHubAddress <= RemoveHubAddress
      num_used_entries "= (InsertHubAddress + LimitHubAddress - StartHubAddress - RemoveHubAddress) / EntrySizeBytes
    else
      num_free_entries := (InsertHubAddress - RemoveHubAddress) / EntrySizeBytes

  General insert and remove process descriptions
  ==============================================

  To insert a new entry into the Fifo:
    1. Wait until there is room for a new entry.
    2. Get the insertion pointer
    3. Copy the entry data to the buffer starting at the insertion pointer
    4. Add the size of the entry to the insertion pointer
       4.1. Wrap the insertion pointer back to the start address if needed
    5. Update the Fifo Header to reflect the insertion:
       5.1. Update the InsertHubAddress in the header to the updated
            insertion pointer from step 4

  To remove an entry from the Fifo:
    1. Wait until at least one entry is present in the fifo.
    2. Get the remove pointer
    3. Copy the entry data from the ring buffer starting at the remove pointer.
    4. Add the size of the entry to the rmeove pointer
       4.1. Wrap the remove pointer back to the start address if needed.
    5. Update the Fifo header to refplect the removal:
       5.1. Update the RemoveHubAddress with the updated remove pointer

The definition of the offsets for the FiFo header are in a separate file,
FiFoBig_Header.spin, so that PASM modules using those offsets do not need to
depend on the FiFo Spin code, so that we can better control the load order
of the modules in hub memory, for better memory use.
}}

OBJ
  constants : "FifoBig_Header"

CON
  ' Import the offsets to the fields of the above structure that describes
  ' the FIFO buffer. These are WORD fields.
  INSERT_HUBADDR_OFFSET   = constants#INSERT_HUBADDR_OFFSET
  REMOVE_HUBADDR_OFFSET   = constants#REMOVE_HUBADDR_OFFSET
  START_HUBADDR_OFFSET    = constants#START_HUBADDR_OFFSET
  LIMIT_HUBADDR_OFFSET    = constants#LIMIT_HUBADDR_OFFSET
  ENTRY_SIZE_BYTES_OFFSET = constants#ENTRY_SIZE_BYTES_OFFSET

  HEADER_SIZE_BYTES = constants#HEADER_SIZE_BYTES
  HEADER_SIZE_WORDS = constants#HEADER_SIZE_WORDS
  HEADER_SIZE_LONGS = constants#HEADER_SIZE_LONGS


'=====================================================================
PUB Init(header, start_hubaddr, entry_size_bytes, num_entries)
{{
  Initialize the FIFO header with the location and shape of the buffer.
  This specifies both the number of entries, and the size in bytes of each
  entry.

  Arguments:
    header -- the address in hub memory of the five-WORD structure that
        will describe the FIFO:
            WORD insert_hubaddr
            WORD remove_hubaddr
            WORD start_hubaddr
            WORD limit_hubaddr
            WORD entry_size_bytes
    start_hubaddr -- the address in hub memory for the FIFO buffer
    entry_size_bytes -- the size of each FIF0 entry in bytes. This is used
        by the generic get and put routines, which use BYTEMOVE() for the
        data transfers. More efficient get/put routines are defined for
        BYTE/WORD/LONG sized entries.

        NOTE that the current implementation expects the entry_size_bytes to
        be a multiple of four (bytes in a LONG).

    num_entries -- the number of entries (capacity) of the FIFO.

  Return value:
    None.
}}
  word[header +    START_HUBADDR_OFFSET] := start_hubaddr
  word[header +   INSERT_HUBADDR_OFFSET] := start_hubaddr
  word[header +   REMOVE_HUBADDR_OFFSET] := start_hubaddr
  word[header + ENTRY_SIZE_BYTES_OFFSET] := entry_size_bytes
  ' compute the limit address
  word[header +    LIMIT_HUBADDR_OFFSET] := {
}                         start_hubaddr + (num_entries * entry_size_bytes)
  ' Initialize the buffer, if needed.
  ' BYTEFILL(start_hubaddr, 0, num_entries * entry_size_bytes)
  ' ' or
  ' LONGFILL(start_hubaddr, 0, num_entries * entry_size_bytes >> 2)

'=====================================================================
PRI GetInsertPtr(header)
{{
  Get the Insert pointer value from a FiFo header.

  Arguments:
    header -- the address in hub memory of the five-WORD FiFo header

  Return value:
    The address in hub memory in the FiFo to next insert an entry.
}}
  return word[header + INSERT_HUBADDR_OFFSET]

'=====================================================================
PRI GetRemovePtr(header)
{{
  Get the Remove pointer value from a FiFo header.

  Arguments:
    header -- the address in hub memory of the five-WORD FiFo header

  Return value:
    The address in hub memory in the FiFo of the next entry to remove.
}}
  return word[header + REMOVE_HUBADDR_OFFSET]

'=====================================================================
PRI AdvancePtr(header, ptr)
{{
  Advance the ptr value (either the insert_hubaddr or the remove_hubaddr)
  correctly in the FIFO, wrapping around at the end of the buffer.

  Arguments:
    header -- the address in hub memory of the five-WORD FiFo header
    ptr -- the hub address in the FiFo to advance by one entry, wrapping
           around in the buffer if needed.

  Return value:
    The address in hub memory in the FiFo of the next entry to remove.
}}
  ptr += word[header + ENTRY_SIZE_BYTES_OFFSET]
  if ptr => word[header + LIMIT_HUBADDR_OFFSET]
    ptr := word[header + START_HUBADDR_OFFSET]
  return ptr

'=====================================================================
PRI AdvanceInsertPtr(header)
{{
  Advance the ptr value for the insert_hubaddr in the FiFo header.

  Arguments:
    header -- the address in hub memory of the five-WORD FiFo header

  Return value:
    none.
}}
  word[header + INSERT_HUBADDR_OFFSET] := {
    }                   AdvancePtr(header, GetInsertPtr(header))
 
'=====================================================================
PUB IsFull(header)
{{
  Determine if there is any room in the FIFO.

  Arguments:
    header -- the address in hub memory of the five-WORD FiFo header.

  Return value:
    Non-zero (true) if the FIFO is full, i.e. a Put() operation would stall
        waiting for room.
    Zero (false) if the FiFo has room for another entry.
}}
  return (AdvancePtr(header, GetInsertPtr(header)) == GetRemovePtr(header))

'=====================================================================
PUB IsEmpty(header)
{{
  Determine if there is any data in the FIFO.

  Arguments:
    header -- the address in hub memory of the five-WORD FiFo header.

  Return value:
    Non-zero (true) if the FiFo is empty, i.e. a Get() operation would stall
        waiting for data.
    Zero (false) if entries are present ready to be removed.
}}
  return GetInsertPtr(header) == GetRemovePtr(header)
  
'=====================================================================
PUB PutIfRoom(header, ptr) | insertptr, next_insertptr
{{
  Append the data to the end of the FIFO, if there is room in the FIFO.

  Arguments:
    header -- the address in hub memory of the five-WORD FiFo header.
    ptr -- address of the data to be written. (NOT the data itself)

  Return value:
    Non-zero (true) if the data was inserted into the FIFO.
    Zero (false) if there was no room in the FiFo for the entry.
}}
  ' "Simple" way to do the work.
  ' if not IsFull(header)
  '   BYTEMOVE(GetInsertPtr(header), ptr, word[header + ENTRY_SIZE_BYTES_OFFSET])
  '   AdvanceInsertPtr(header)
  '   return 1
  ' return 0

  ' Perhaps a more efficient way to do it:
  insertptr := GetInsertPtr(header)
  next_insertptr := AdvancePtr(header, insertptr) 

  if next_insertptr == GetRemovePtr(header)
    return 0  ' Buffer is full 

  ' We know the buffer is not full.
  'BYTEMOVE(insertptr, ptr, word[header + ENTRY_SIZE_BYTES_OFFSET])
  LONGMOVE(insertptr, ptr, word[header + ENTRY_SIZE_BYTES_OFFSET] >> 2)
  
  word[header + INSERT_HUBADDR_OFFSET] := next_insertptr
  return 1

'=====================================================================
PUB Put(header, ptr) | insertptr, next_insertptr
{{
  Append the data to the end of the FIFO, waaiting for room for the entry if
  necessary.

  Arguments:
    header -- the address in hub memory of the five-WORD FiFo header.
    ptr -- address of the data to be written. (NOT the data itself)

  Return value:
    Non-zero (true) always (data was inserted into the FIFO).
}}
  ' Simple way:
  '  repeat while IsFull(header)
  '    ' Stall, waiting for room.
  '  BYTEMOVE(GetInsertPtr(header), ptr, word[header + ENTRY_SIZE_BYTES_OFFSET])
  '  AdvanceInsertPtr(header)

  ' Perhaps a more efficient way to do it:
  insertptr := GetInsertPtr(header)
  next_insertptr := AdvancePtr(header, insertptr) 

  ' We don't control the remove pointer, so that has to be polled.
  repeat while insertptr == GetRemovePtr(header)
    ' Stall, waiting for room

  ' We know the buffer is not full.
  'BYTEMOVE(insertptr, ptr, word[header + ENTRY_SIZE_BYTES_OFFSET])
  LONGMOVE(insertptr, ptr, word[header + ENTRY_SIZE_BYTES_OFFSET] >> 2)
  word[header + INSERT_HUBADDR_OFFSET] := next_insertptr
  return 1

'=====================================================================
PUB PutAddress(header) | insertptr, next_insertptr
{{
  Get the address in the FiFo for the next entry to be inserted, waiting
  for room for the entry if necessary.

  After the entry has been inserted, the caller should then call
  FinishPut(header) to update the FiFo header fields.

  Arguments:
    header -- the address in hub memory of the five-WORD FiFo header.

  Return value:
    The address in the FiFo to receive the entry.
}}
  ' Get the address for the next insertion, when there is room for it.
  ' This will stall, waiting for room.

  insertptr := GetInsertPtr(header)
  next_insertptr := AdvancePtr(header, insertptr)

  ' We don't control the remove pointer, so that has to be polled.
  repeat while next_insertptr == GetRemovePtr(header)
    ' Stall, waiting for room

  ' There is room now.
  return insertptr

'=====================================================================
PUB PutAddressIfRoom(header) | insertptr, next_insertptr
{{
  Get the address in the FiFo for the next entry to be inserted, if there is
  room for another entry, otherwise return zero. This does not wait for room.

  After an entry has been inserted, the caller should then call
  FinishPut(header) to update the FiFo header fields.

  Arguments:
    header -- the address in hub memory of the five-WORD FiFo header.

  Return value:
    Zero -- There is no room for the entry.
    Non-zero -- The address in the FiFo to receive the entry.
}}
  insertptr := GetInsertPtr(header)
  next_insertptr := AdvancePtr(header, insertptr)

  if next_insertptr == GetRemovePtr(header)
    return 0  ' Buffer is full

  ' There is room now.
  return insertptr

'=====================================================================
PUB FinishPut(header)
{{
  Update the FiFo header fields to indicate that another entry has been
  added to the FiFo (after a call to PutAddress() or PutAddressIfRoom()).

  Arguments:
    header -- the address in hub memory of the five-WORD FiFo header.

  Return value:
    None.
}}
  ' Advance the insertion pointer and count, after an entry was added
  AdvanceInsertPtr(header)

'=====================================================================
PUB Get(header, ptr) | removeptr
{{
  Copies the next available entry from the FiFo into the supplied buffer. This
  will wait for an entry if the FiFo is empty at the time of the call.

  Arguments:
    header -- the address in hub memory of the five-WORD FiFo header.
    ptr -- address of the buffer to receive the entry.

  Return value:
    None.
}}
  removeptr := GetRemovePtr(header)
  repeat while removeptr == GetInsertPtr(header)
    ' Nothing to get yet

  'BYTEMOVE(ptr, removeptr, word[header + ENTRY_SIZE_BYTES_OFFSET])
  LONGMOVE(ptr, removeptr, word[header + ENTRY_SIZE_BYTES_OFFSET] >> 2)
  word[header + REMOVE_HUBADDR_OFFSET] := AdvancePtr(header, removeptr)

'=====================================================================
PUB GetIfAny(header, ptr) | removeptr
{{
  If the FiFo is not empty, copies the next available entry from the FiFo into
  the supplied buffer. If the FiFo is empty, this call will return immedately.

  Arguments:
    header -- the address in hub memory of the five-WORD FiFo header.
    ptr -- address of the buffer to receive the entry.

  Return value:
    Zero -- There is no available entry in the FiFo.
    Non-zero -- The next entry was retrieved.
}}
  removeptr := GetRemovePtr(header)
  if removeptr == GetInsertPtr(header)
    return 0  ' Nothing present to get

  'BYTEMOVE(ptr, removeptr, word[header + ENTRY_SIZE_BYTES_OFFSET])
  LONGMOVE(ptr, removeptr, word[header + ENTRY_SIZE_BYTES_OFFSET] >> 2)
  word[header + REMOVE_HUBADDR_OFFSET] := AdvancePtr(header, removeptr)
  return 1

'=====================================================================
PUB GetAddress(header) | removeptr
{{
  Gets the address in the FiFo of the next available entry. If the FiFo is
  empty at the time of the call, this will wait until an entry becomes available.

  After the entry has been retrieved, the caller should then call
  FinishGet(header) to update the FiFo header fields.

  Arguments:
    header -- the address in hub memory of the five-WORD FiFo header.

  Return value:
    The address in the FiFo of the entry to read.
}}
  removeptr := GetRemovePtr(header)
  repeat while removeptr == GetInsertPtr(header)
    ' Nothing to get yet
  return removeptr

'=====================================================================
PUB GetAddressIfAny(header) | removeptr
{{
  If the FiFo is not empty, gets the address in the FiFo of the next available
  entry. If the FiFo is empty at the time of the call, this returns immediately.

  After an entry has been retrieved, the caller should then call
  FinishGet(header) to update the FiFo header fields.

  Arguments:
    header -- the address in hub memory of the five-WORD FiFo header.

  Return value:
    Zero -- There was no entry available.
    Non-zero -- The address in the FiFo of the entry to read.
}}
  removeptr := GetRemovePtr(header)
  if removeptr == GetInsertPtr(header)
    return 0  ' Nothing present to get
  return removeptr

'=====================================================================
PUB FinishGet(header)
{{
  Update the FiFo header fields to indicate that another entry has been
  removed from the FiFo (after a call to GetAddress() or GetAddressIfRoom()).

  Arguments:
    header -- the address in hub memory of the five-WORD FiFo header.

  Return value:
    None.
}}
  ' Advance the removal pointer, after an entry was read.
  word[header + REMOVE_HUBADDR_OFFSET] := {
    }     AdvancePtr(header, GetRemovePtr(header))

