{{
Display linear and gamma-adjusted intensities, for one or more HUB75 displays.

Author:         Tom Dinger (Ding-Batty)
Release Date:   2022/04/10

Copyright 2024 Thomas E. Dinger

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

--------------------------------------------------------------------------

This program shows the difference between linear intensities and gamma-
corrected intensities, for various numbers of color bit depth, and various
gamma values. The program will display the linear intensity range on the
top half of the display, and the gamma-adjusted intensities on the bottom
half of the display, to provide a visual display of the gradient difference.

You can get a "dump" of the gamma lookup table to gamma-correct intensities
in other applications, for various color bit depths and gamma values.

I found this site helpful in understanding how gamma correction works:
https://www.cambridgeincolour.com/tutorials/gamma-correction.htm
(There are lots of other helpful tutorials on that site.)

Cog use
=======
Cog 0 -- the main program: launches all the other Cogs, prepares the state for
         digit display; collects and processes the single-character commands;
         reports memory and stack use.
Cog 1 -- the Serial port driver cog
Cog 2 -- the HUB75 display refresh cog
Cog 3 -- the display buffer write helper cog

}}

OBJ
  ' Pick up the static display configuration first.
  Config: "hub75_config.spin"

  Ser    : "Db_fullDuplexSerial"
  Umath  : "umath2.spin"
  ' The FiFo is used to enqueue write commands for the Helper to send to
  ' a dedicated PASM writer cog that updates the Display buffer.
  Fifo   : "FifoBig.spin"
  ' The Panel is the wrapper for the PASM cog that send the data to the
  ' display, and contains Spin-based write routines, such as Point, Box,
  ' HorizontalLine, etc. 
  Panel  : "hub75_display.spin"
  ' The Helper contains wrapper functions for writing to the display buffer
  ' by sending the commands to a helper PASM cog, in order to have faster
  ' display updates. It has the same write functions, e.g. Point, Box, etc.
  Helper : "hub75_helper.spin"

  ' The init object (below) and all objects it includes, will be used as
  ' additional buffer allocation space after they are finished during init.
  ' This helps to maximize hub memory available for large buffer use.
  Init : "gamma_explorer_init.spin"

CON { General constants }
  _CLKMODE = XTAL1 + PLL16X 
  _XINFREQ = 5_000_000

  { Configuration for serial connection. }
  CR = $0D
  LF = $0A

  { LED panel(s) configuration }
  ' Save some typing later, by reflecting a few constants here.
  DISPLAY_ROWS       = Config#DISPLAY_ROWS
  REFRESH_ROWS       = Config#REFRESH_ROWS
  DISPLAY_COLUMNS    = Config#DISPLAY_COLUMNS
  MAX_DISPLAY_ROW    = Config#MAX_DISPLAY_ROW
  MAX_DISPLAY_COLUMN = Config#MAX_DISPLAY_COLUMN

  RED_BIT            = Config#RED_BIT
  GRN_BIT            = Config#GREEN_BIT
  BLU_BIT            = Config#BLUE_BIT  

  ' Limit the range of the gamma values that can be used.
  ' Gamma == 1.0 is no gamma adjustment -- linear intensity display.
  MIN_GAMMA_TIMES_10 = 10   ' == 1.0
  MAX_GAMMA_TIMES_10 = 50   ' == 5.0

  ' The UpdateDisplay() function can update either the linear gradient shown
  ' in the top half of the screen, or the gamma-adjusted gradient in the bottom
  ' half, or both.
  SHOW_LINEAR = $01
  SHOW_GAMMA  = $02
  SHOW_BOTH   = SHOW_LINEAR | SHOW_GAMMA

  ' ROM math table addresses
  LOG_TABLE_BASE      = $C000
  ANTI_LOG_TABLE_BASE = $D000

  ' Constants for stack tracking.

  ' The address of the location that stores the base address of the main stack,
  ' which runs up form there to $7FFF
  STACK_BASE_ADDR = $000A
  END_OF_STACK    = Init#END_OF_STACK  ' $0000_8000

  ' This is the distance from the last local variable address in the
  ' function that fills main memory that is safe to fill with the
  ' UNLIKELY_VALUE.
  OFFSET_FOR_MAIN_STACK_FILL = Init#OFFSET_FOR_MAIN_STACK_FILL  ' $0010
  UNLIKELY_VALUE             = Init#UNLIKELY_VALUE              ' $DEADBEEF

VAR
  ' Variables used in Spin, that are initialized at runtime (not by the compiler)

  ' These must be consecutive in this order, so that the Init object can fill
  ' them in.
  ' We use other cogs as well but they run Spin code, not PASM.
  WORD    SerialCog   ' 0 if not started, else CogID + 1
  WORD    DisplayCog  ' 0 if not started, else CogID + 1
  WORD    HelperCog   ' 0 if not started, else CogID + 1

  ' We allocate two buffers from the free space:
  ' - Command FiFo for display write-helper cog
  ' - Display Buffer
  ' They are allocated from a combination of reserved space and reuse of the
  ' initialization code including the PASM cog code from Hub memory.
  ' Since the largest buffer is the display buffer, and it (usually) is much
  ' larger than the init code that is recovered, we don't need to dynamically
  ' allocate the serial driver buffers, which makes init a bit simpler.
  WORD    BufferInfo [2 * 2]  ' two buffers to allocate.

  ' The color model being used here separates the number of bits to represent
  ' color intensities, and the number of bits used for the color depth of the
  ' display. Usually, these will be the same in an application, but it may be
  ' helpful to see the effects of varying the number of color bits.

  ' DisplayColorDepth is the number of bits for each color in the display
  ' buffer. It is chosen during application initialization as the maximum
  ' depth, up to the maximum supported by this driver (10 bits), for which
  ' the display buffer will fit in available hub memory, and for which the
  ' refresh rate is 60 HZ or higher, based on the current display
  ' configuration (columns and rows).
  
  LONG    DisplayColorDepth
  LONG    MaxDisplayColorValue  ' Set to (|< DisplayColorDepth) - 1

  ' ColorDepth is the number of bits used for color intensities before gamma
  ' correction. This can range from 2 to 10. The program will calculate the
  ' gamma-adjusted values for these using the current gamma value when the
  ' bottom half of the display is drawn.

  LONG    ColorDepth
  LONG    MaxColorValue    ' Set to (|< ColorDepth) - 1

  ' The gamma value used for the intensity adjustments, multiplied by 10,
  ' because the calculation is done as a rational number:
  '   GammaTimes10 / 10
  ' The gamma value can range from 1.0 to 5.0, so this variable can range
  ' from 10 to 50.

  LONG    GammaTimes10

  ' The index into the color table used for the gradient displays. It is
  ' possible for an application to use different gamma values for the red,
  ' green and blue intensities, depending on the performance of the LEDs,
  ' although most non-video applications can use one table for all three
  ' colors.
  LONG    ColorIndex

DAT
  ' We allocate the storage for the command FiFo entries dynamically,
  ' overlaying the PASM (and Spin) code at the end of the program, which is
  ' only used during early initialization.
  CmdFifoDataAddr   LONG    0

  ' Storage for debug output from the helper driver
  ' See that driver for details on the contents.
  ' Should be at least 10 longs
  HelperDebugOutput LONG    0[16]

  ' Define a table of display colors that we will cycle through to use for the
  ' displayed gradients.
  ' We define the colors using 10 bits per color, and let the intensity calculation
  ' for the gradients handle the scaling details.
  ' The intensity calculations are done for a white display (all three colors
  ' present), then the entry from this table is used as a mask to keep just the
  ' wanted color(s).

  ColorTableSize    LONG     4
  '                                  R                       G                       B
  ColorTable
  ColorTableWhite   LONG    %%0_33333_00000_00000 | %%0_00000_33333_00000 | %%0_00000_00000_33333
  ColorTableRed     LONG    %%0_33333_00000_00000
  ColorTableGreen   LONG                            %%0_00000_33333_00000
  ColorTableBlue    LONG                                                    %%0_00000_00000_33333
'  ColorTableYellow  LONG    %%0_33333_00000_00000 | %%0_00000_33333_00000
'  ColorTableMagenta LONG    %%0_33333_00000_00000                         | %%0_00000_00000_33333
'  ColorTableCyan    LONG                            %%0_00000_33333_00000 | %%0_00000_00000_33333

  ' The display buffer stores all color layers for all displays.
  ' Someday it will overlap the PASM objects that are initilaized (loaded into
  ' cogs) once at the start of operations, to maximize the amount of memory
  ' available for displays.
  '
  ' One byte of memory stores one color bit layer for two pixels.

  DisplayBufferAddr LONG    0   ' Gets a dynamically-allocated buffer

' ========================================================================
PUB Main | avail, left, needed, firstfree
  ' Collect some numbers for assessing memory use.
  avail := Init.AvailableBytes
  firstfree := Init.GetNextFree

  ' First-stage init:
  ' - Collect the digit bitmaps
  ' - reserve memory for the (large-ish) buffers
  ' - load the PASM cogs.
  ' After this call, the init code should no longer be needed, so the reserved
  ' buffers can be overwritten and the objects using the PASM cogs can finish
  ' their initializations.
  needed := Init.Init(@SerialCog, @BufferInfo, Ser.GetInitVars,       {
         }            Panel.GetDisplayBufAddrVar, Panel.GetRefreshClocksAddrHub, {
         }            Helper.GetCmdFifoHeader, @HelperDebugOutput)
  left := Init.AvailableBytes
  Ser.FinishSecondInit(SerialCog)  ' SecondInit because serial output was used
                                   ' in the Init() code.

  ' How did we do with buffer reservations?
  ReportHex(string("VBASE: "), WORD[8])
  ReportDec(string("Initial memory (bytes):   "), avail)
  ReportDec(string("Recovered from init code: "), avail - Init#RESERVED_SIZE_BYTES)
  ReportHex(string("First free:               "), firstfree)
  ReportHex(string("FiFo buffer size:         "), BufferInfo[0])
  ReportHex(string("FiFo buffer address:      "), BufferInfo[1])
  ReportHex(string("Display buffer size:      "), BufferInfo[2])
  ReportHex(string("Display buffer address:   "), BufferInfo[3])
  ReportDec(string("Memory needed:            "), needed)
  ReportDec(string("Memory remaining (bytes): "), left)

  if needed > 0
    ' We were not able to get as much buffer memory as we wanted.
    if BufferInfo[1] == 0
      ReportStr(string("Not enough memory for Command FiFo or Display Buffer"))
    else
      ReportStr(string("Not enough memory for Display Buffer"))
    'ReportDec(string("Memory needed (bytes):    "), needed)
  else
    ' We have our buffers -- continue with initialization and run all the tests.
    Main2

  ' Here when done.
  ReportStackUse
  repeat
    ' wait forvever.

PUB Main2 | c, save_style
{{
  Complete top-level init, and start operations.
}}
  CmdFifoDataAddr   := BufferInfo[1]
  DisplayBufferAddr := BufferInfo[3]
  ReportDec(string("Started display cog: "), DisplayCog - 1)
  ReportDec(string("Started helper cog: "), HelperCog - 1)
  ' Determine the color depth for the display from the size of the display
  ' buffer (so we don't have to come up with another way to pass the display
  ' color depth back from the init code).
  DisplayColorDepth := BufferInfo[2] / constant(DISPLAY_COLUMNS * REFRESH_ROWS)
  ReportDec(string("Display bits per color: "), DisplayColorDepth)

  ' Tell the drivers about the two cogs and allocated buffers.
  ' Note that telling the display driver about the address of the display
  ' buffer will unblock that cog _and_ the helper cog at the same time, so
  ' that one should be done last.
  LONGFILL(DisplayBufferAddr, 0, BufferInfo[2] >> 2)  ' Clear the buffer.
  Helper.FinalHelperInit(HelperCog, DisplayBufferAddr, DisplayColorDepth, {
           }             CmdFifoDataAddr, BufferInfo[0])
  Panel.FinalDisplayInit(DisplayCog, DisplayBufferAddr, DisplayColorDepth)

  ' Prepare our starting configuration:
  ' - display color depth == color space depth
  ' - start with white display
  ' - start with gamma == 2.4 (typical value)
  MaxDisplayColorValue := (|< DisplayColorDepth) - 1
  ColorDepth := DisplayColorDepth
  MaxColorValue := (|< ColorDepth) - 1
  GammaTimes10  := 22 '  MIN_GAMMA_TIMES_10
  ColorIndex := 0  ' start with white

  UpdateDisplay(SHOW_BOTH)

  ' Wait for a command character, and do it.
  ShowHelpText
  c := -1
  repeat while c <> "X" and c <> "x"
    c := Ser.Rx ' Get command character
    case c
       "?":     ' Show help
         ShowHelpText
      "f":
        DumpFifo(string("Fifo state"))
      "K":
        ReportStackUse
      "b":   ' increment color bits
        if ColorDepth < 10  ' MaxColorDepth
          MaxColorValue := (|< ++ColorDepth) - 1
          UpdateDisplay(SHOW_BOTH)
          ReportDec(string("Color depth: "), ColorDepth)
      "B":   ' decrement color depth
        if ColorDepth > 2
          MaxColorValue := (|< --ColorDepth) - 1
          UpdateDisplay(SHOW_BOTH) 
          ReportDec(string("Color depth: "), ColorDepth)
      "c":   ' increment color, with wrap-around
        if ++ColorIndex => ColorTableSize
          ColorIndex := 0
        UpdateDisplay(SHOW_BOTH)
      "C":   ' decrement color, with wrap-around
        if --ColorIndex < 0
          ColorIndex := ColorTableSize - 1
        UpdateDisplay(SHOW_BOTH)
      "d", "D":  ' Dump the current gamma mapping for the input and output depth
        DumpGamma(GammaTimes10, ColorDepth, DisplayColorDepth)
      "i", "I":  ' Dump a gamma table for the current input depth, ignoring the
                 ' color depth for the display
        DumpGamma(GammaTimes10, ColorDepth, ColorDepth)
      "+", "g":  ' increment gamma value by 0.1
        if GammaTimes10 < MAX_GAMMA_TIMES_10
          GammaTimes10 := ++GammaTimes10
          UpdateDisplay(SHOW_GAMMA)
          ReportGamma(GammaTimes10)
      "-", "G":  ' decrement gamma value by 0.1
        if GammaTimes10 > MIN_GAMMA_TIMES_10
          GammaTimes10 := --GammaTimes10
          UpdateDisplay(SHOW_GAMMA)
          ReportGamma(GammaTimes10)
      other: ' do nothing

  ReportStr(string("Stopping..."))
  Panel.Stop
  ReportStr(string("Display cog stopped"))
  Helper.Stop
  ReportStr(string("Helper cog stopped"))

' ------------------------------------------------------------------------
PUB ShowHelpText
  Ser.Str(@HelpText)

DAT
HelpText
    BYTE "Command characters take immediate effect:", CR
    BYTE  "  ? -- Show this help text", CR
    BYTE "Debugging commands:", CR
    BYTE "  f -- Dump the FiFo Header state (insert and remove pointers)", CR
    BYTE "Display operations:", CR
    BYTE "  b and B -- Increment or decrement the color depth", CR 
    BYTE "  c and C -- Colors: cycle through a set of predefined colors", CR
    BYTE "  d or D -- Dump the gamma table for the current bit depths", CR
    BYTE "  K -- Free stacK: report the amount of used and unused stack", CR
    BYTE "  + or g -- Increase the gamma value by 0.1 up to 5.0", CR
    BYTE "  - or G -- Decrease the gamma value by 0.1 down to 1.0", CR
    BYTE "  x or X -- Exit; Shut down.", CR
    BYTE 0

' ========================================================================

PUB UpdateDisplay(show) | num_colors, increment, width, intensity, column
{{
  Update the requested portions of the display, using the current configuration.

  The top half of the display shows the linear intensity gradient in the
  current color, from black on the left to brightest on the right. The bottom
  half shows the gamma-adjusted intensity gradient in the same color so they
  can be compared.

  The two halves can be updated separately, because some adjustment commands
  only affect the bottom half of the display, e.g. changing the gamma value.

  Arguments:
    show -- a bit field indicating which portion of the display to update:
            SHOW_LINEAR = $01 -- update the top half of the display
            SHOW_GAMMA  = $02 -- update the bottom half of the display
            SHOW_BOTH will update both halves of the display.

  Return value:
    None.
}}
  if not show
    return  ' Neither half was selected to be updated.

  ' Determine the number of columns, and the width of each column, based
  ' on the ColorDepth value and the DISPLAY_COLUMNS value.
  ' If there are more intensity steps than display columns, we skip some
  ' of the intensities (when "increment" is > 1). If there are more
  ' display columns than intensity steps, multiple columns are used for
  ' each intensity.
  num_colors := |< ColorDepth
  if num_colors =< DISPLAY_COLUMNS
    ' All the color values will fit on the display
    increment := 1      ' show every intensity
    width := DISPLAY_COLUMNS / num_colors
  else
    ' Not all the color values will fit on the display, so skip some
    width := 1
    increment := num_colors / DISPLAY_COLUMNS

  ' Iterate over all the intensities for the current color
  column := 0
  repeat intensity from 0 to MaxColorValue step increment
    if show & SHOW_LINEAR
      Helper.Box(column,            0, width, REFRESH_ROWS, LinearColor(intensity))
    if show & SHOW_GAMMA
      Helper.Box(column, REFRESH_ROWS, width, REFRESH_ROWS, GammaColor(intensity))
    column += width

PRI LinearColor(intensity)
{{
  Adjust the intensity value from the ColorDepth range to fit into the display
  color depth (because they can be different), using a linear intensity
  mapping. The mapping makes sure 0 maps to 0, and the highest value maps to
  the matching highest value. Then build the RGB color with the calculated
  intensity, masked for the current display color.

  It applies the rounded function:
    adjusted = in_max * intensity / out_max

  Arguments:
    intensity -- the intensity value, in the range 0..MaxColorValue, which is
                 not checked or limited.

  Return value:
    The rounded adjusted intensity value in the range 0..MaxDisplayColorValue.
}}
  intensity := (intensity * MaxDisplayColorValue + (MaxColorValue >> 1)) / MaxColorValue
  return Panel.RgbColor(intensity, intensity, intensity) & LONG[@ColorTable + (ColorIndex << 2)]

PRI GammaColor(intensity)
{{
  Adjust the intensity value from the ColorDepth range to fit into the display
  color depth (because they can be different), using a gammal-adjusted
  intensity value. The mapping makes sure 0 maps to 0, and the highest value
  maps to the matching highest value. Then build the RGB color with the
  calculated intensity, masked for the current display color.

  See the comments for CalcGamma(intensity) for a discussion of the gamma
  function as applied to the intensity calculation.

  Arguments:
    intensity -- the intensity value, in the range 0..MaxColorValue, which is
                 not checked or limited.

  Return value:
    The gamma-adjusted intensity value in the range 0..MaxDisplayColorValue.
}}
  intensity := CalcGamma(intensity, MaxColorValue, MaxDisplayColorValue, GammaTimes10, 10)
  return Panel.RgbColor(intensity, intensity, intensity) & LONG[@ColorTable + (ColorIndex << 2)]

PRI DumpGamma(gamma_times_10, input_depth, output_depth) {
      }         | max_input_value, max_output_value, i, gi
{{
  Write the current settings, and a gamma-adjusted intensity mapping table to
  serial ouptut.

  This table contains the same values as is used in the gradients currently
  displayed, so the number of color bits for the input and output values might
  be different.

  The generated table uses WORD values, because the maximum output value could
  be as large as 1023. Also, there will be a trailing comma

  Arguments:
    gamma_times_10 -- the gamma * 10 value, so it is in tenths.
    input_depth -- the number of bits for the input intensity -- this
                   determines the number of table entries. This should be in
                   the range 2..10, not checked.
    output_depth -- the number of bits for the calculated intensities. This
                    should be in the range 2..10, not checked.

  Return value:
    None.
}}
  ReportStr(string("Gamma table and settings"))
  ReportGamma(gamma_times_10)
  ReportDec(string("Input  color depth: "), input_depth)
  ReportDec(string("Output color depth: "), output_depth)
  ' The color names here should match the order in the ColorTable, above.
  case ColorIndex
    0: ReportStr(string("Color: White", CR))
    1: ReportStr(string("Color: Red", CR))
    2: ReportStr(string("Color: Green", CR))
    3: ReportStr(string("Color: Blue", CR))

  max_input_value  := (|<  input_depth) - 1
  max_output_value := (|< output_depth) - 1
  repeat i from 0 to max_input_value
    ' Write four values per line.
    if (i & $03) == 0
      Ser.Str(string("    WORD    "))
    gi := CalcGamma(i, max_input_value, max_output_value, gamma_times_10, 10)
    Ser.Tx("$")
    Ser.Hex(gi, 4)
    if (i & $03) == $03     ' The final value on the line
      Ser.Tx(CR)
    else
      Ser.Str(string(", ")) ' Not the final value on the line.
  Ser.Tx(CR)

' ========================================================================

PUB CalcGamma(v, in_max, out_max, g_numerator, g_denominator) | part
{{
  Calculate the gamma-adjusted intensity value for the input value "v".

  The gamma function for the adjustment is defined as:

      Vout = Vin ** gamma

  For values of Vin and Vout in the 0 to 1 range.

  For our purposes, the incoming value is in the 0..in_max range as an integer,
  and we want the output value to be in the 0..out_max range as an integer.
  Also, we define the gamma value as a rational number to make the calculation
  somewhat easier in fixed point.

  So we want to calculate:
      Out = out_max * ( (V / in_max) ** (g_numerator / g_demnominator) )

  The Propeller doesn't have exponentiation, but it does have lookup tables
  for calculating log base 2 and anti-log base 2 (which I will call log2 and
  pow2), so we rewrite the above as:

    Out = out_max * pow2( (g_numerator / g_denominator) * log2(v / in_max) )

  Using some transforms, we can simplify this for calculation purposes.
    log2( a * b ) == log2(a) + log2(b)
    log2( 1 / b ) == -log2(b)
  So
    log2(v / in_max) == log2(v) - log2(in_max)
  giving:

    Out = out_max * pow2( (log2(v) - log2(in_max)) * g_numerator / g_denominator)

  A limitation of the Propeller anti-log tables is that it gives an integer value,
  and the argument to the pow2() function above is between 0 and 1. So we move the
  out_max multiplication into the pow2() operation:

    Out = pow2( log2(out_max) + ((log2(v) - log2(in_max)) * g_numerator / g_denominator) )

  (We also have to round the division by g_denominator by adding (g_denoninator >> 1) to
  the numerator value before the division.)

  We also use interpoltion between the anti-log table entries, to use the left-over
  five bits of the fractional input to the anti-log. Furthermore, we round all
  divisions to the nearest integer, to try to reduce the errors that will propagate
  through the calculations.

  By doing all that, I have verified that the calculation as implemented does match
  the equivalent calculation done on a PC (in Perl, but also Python or C/C++ could
  do the same), using double-precision arithmetic, at (nearly) every value for the
  input:

    Out = int((out_max * ( (v / in_max) ** (g_numerator / g_denominator) )) + 0.5)

  Arguments:
    v -- the value to be raised to the gamma power, in the range of 0..in_max
    in_max -- the maximum input value of "v", to convert "v' to a value in the 0..1
              range.
    out_max -- the maximum output value, to scale the 0..1 calculation to the range
               0..out_max
    g_numerator -- The gamma numerator
    g_denominator -- The gamma denominator

  Return value:
    The gamma-adjusted value, in the range of 0..out_max
}}
  ' This function does rounding for all division operations, and
  ' interpolates for the anti-log look-up between table entries.
  ' This gets extremely close to the PC calculation, for values of
  ' v in the 0..1023 range (up to 10 bits for the value -- there are only
  ' two or three values for which this computation differs from the PC by 1.)

  if v == 0
    return 0   ' Since we cannot calculate log2(0).

  part := log2(out_max) + (((g_numerator * (log2(v) - log2(in_max))) + (g_denominator >> 1)) / g_denominator)
  return pow2_interpolate_round(part)

PUB pow2_interpolate_round(v) | ix, m1, m2, adjust, mantissa
{{
  Calculate the integer value of 2 ** v, for v as a fixed-point value in the
  range of 0.0 to 31.99999..., using 16 bits for the fractional part, as
  described in the Propeller manual appendix on the log and anti-log tables.
  Calculate the gamma-adjusted intensity value for the input value "v".

  The anti-log table lookup uses only 11 of the 16 fractional bits, so this
  code uses the remaining five bits to interpolate linearly between the table
  entries to get a slight error reduction -- this makes a very small, but
  measurable difference.

  Also, the shift operations as documented in the manual are really division
  operations in spirit, and so we round them to the nearest integer, which also
  makes a very small but measurable difference.

  Arguments:
    v -- the fixed-point exponent to use. It has 16 low-order bits for the
         fractional value, and the high order bits are the integer portion which
         are in the range 0..31 (only 5 out of 16 bits used).
         "v" is permitted to be negative, which means the result will be < 1.

  Return value:
    A very close approximation to: floor(2 ** v + 0.5)
}}
  ' For negative values of v, 2 ** v == 1 / (2 ** (-v))
  ' So if 2 ** (-v) is > 2, it rounds to zero, otherwise to 1.
  ' So if -v > 1, it rounds to zero, otherwise to 1. 
  if (v < 0)
    if ((-v) > $10000)  ' $10000 == 1.0 in the fixed-point representation.
        return 0
    return 1

  ix := (v >> 4) & $0FFE    ' Isolate the 11 high-order bits of the fraction, and
                            ' make it a WORD table offset.
  m1 := WORD[ANTI_LOG_TABLE_BASE + ix]
  ' Also get the next entry, watching out for overrunning the table, for
  ' use in interpolation.
  if (ix < constant($0FFE))
    m2 := WORD[ANTI_LOG_TABLE_BASE + ix + 2]
  else
    m2 := $10000   ' m1 was the final table entry. This is the limit value
                   ' for the range of the final entry.

  ' Linear interpolation between these two entries, using the rest of the bits
  ' that were not used in the lookup. If you don't want the interpolation
  ' step, leave out the m2 lookup, and just set adjust to zero.
  '
  ' Linear interpolation works here because the base function (2 ** v) is
  ' continuous (and smooth); it also happens to be monotonic and
  ' non-decreasing.
  adjust := (((m2 - m1) * (v & $1F)) + 16) >> 5  ' divide-by-32, rounded.
  'adjust := 0

  mantissa := (m1 + adjust + $10000) << 15
  ' The right-shift operation of (31 - (v >> 16)) bits is a divide operation, so
  ' we do rounding for the result as well.
  ' For this application, all the values are < 2048, so the subtraction will
  ' never be negative (the integer part will always be < 12, which is only four
  ' bits out of the high 16 bits).
  'return  mantissa >> (31 - ((v >> 16) & $1F));
  return ((mantissa >> (30 - ((v >> 16) & $1F))) + 1) >> 1

PUB log2(v) | i, ix
{{
  Calculate the fixed-point value of log(v) in base 2.

  This is done following the algorithm from the Propeller manual in the
  appendix on using the ROM log table.

  Arguments:
    v -- the integer value whose log is calcualted. This is in the range
         1..1023, not checked. (Zero _is_ checked, because it cannot really
         be used in the calculation.)

  Return value:
    A very close approximation to: log2(v), as fixed-point value with 16
    bits of fractional value.
}}
  ' Protect against v == 0, for safety. Strictly speaking, log2(0) cannot be
  ' calculated -- it would be something like negative infinity.
  if (v == 0)
    return 0

  ' The bit number of the high bit is the integer portion of the log2.
  i := (>| v) - 1
  ' The next 11 bits are used as the lookup index into the log2 table.
  ix := v & ((|< i) - 1)  ' Remove the high bit just identified.
  ' Adjust the masked value so that it is just the top 11 bits of the
  ' masked value
  if (i > 11)
    ix >>= i - 11
  elseif (i < 11)
    ix <<= 11 - i

  ' The table entry is the low order 16 bits of the result, and the
  ' value of i is the high order 16 bits.
  return (i << 16) + WORD[LOG_TABLE_BASE + (ix << 1)] 

' ========================================================================

PUB CheckStack(n, p, l) | lused, pend
  ' Examine the stack starting at p, that contains L longs, and determine
  ' how much of it was used, and how much was untouched.
  pend := p + (l << 2) - 4
  repeat while (pend => p) and (long[pend] == UNLIKELY_VALUE)
    pend -= 4

  ' Here pend is either p - 4, or the address of the last value
  ' used.
  '   longs_used = ((pend + 4) - p) >> 2
  '   longs_free = l - longs_used
  lused := ((pend + 4) - p) >> 2
  Ser.Str(n)
  ReportDec(string(" used longs: "), lused)
  Ser.Str(n)
  ReportDec(string(" free longs: "), l - lused)

PUB CheckMainStack
  CheckStack(string("Main"), WORD[STACK_BASE_ADDR], {
    }        (END_OF_STACK - WORD[STACK_BASE_ADDR]) >> 2)
 
' ========================================================================
' Routines to help with testing.

PUB udec32tostr(v, buf)
  ' Returns the length of the string written into buf
  return Umath.udec64tostr(0, v, buf)

PUB sdec32tostr(v, buf) | p, len, c, i
  ' Returns the length of the string written into buf
  ' One negative value cannot be converted to a positive value by
  ' negation: $8000_0000, but it is its own negation, so we can just
  ' handle it like any other negative number
  if v < 0
    buf := "-"
    return Umath.udec64tostr(0, -v, buf + 1) + 1
  return Umath.udec64tostr(0, v, buf)

PUB ReportStr(s)
  Ser.Str(s)
  Ser.Tx(CR)

PUB ReportDec(label, v)
  Ser.Str(label)
  Ser.Dec(v)
  Ser.Tx(CR)

PUB ReportHex(label, v)
  Ser.Str(label)
  Ser.Hex(v, 8)
  Ser.Tx(CR)

PUB ReportBin(label, v)
  Ser.Str(label)
  Ser.Bin(v, 32)
  Ser.Tx(CR)

PUB ReportTenths(label, v)
  Ser.Str(label)
  Ser.Dec(v / 10)
  Ser.Tx(".")
  Ser.Tx("0" + (v // 10))
  Ser.Tx(CR)

PUB ReportGamma(g)
  ReportTenths(string("Gamma: "), g)

PUB DumpFifo(label) | p
  ReportStr(label)
  p := Helper.GetCmdFifoHeader
  
  ReportHex(string("FiFo   ptr: "), p)
  ReportHex(string("Insert ptr: "), word[p + Fifo#INSERT_HUBADDR_OFFSET])
  ReportHex(string("Remove ptr: "), word[p + Fifo#REMOVE_HUBADDR_OFFSET])
  ReportHex(string("Start  ptr: "), word[p + Fifo#START_HUBADDR_OFFSET])
  ReportHex(string("Limit  ptr: "), word[p + Fifo#LIMIT_HUBADDR_OFFSET])
  ReportDec(string("Entry size: "), word[p + Fifo#ENTRY_SIZE_BYTES_OFFSET])

PUB ReportStackUse
  CheckMainStack  ' So we can see if we can shrink the main stack, and increase
                  ' the amount of memory for dynamic buffers.

PUB PowerOfTen(p) | r
  r := 1
  repeat while p--
    r *= 10
  return r
