{{
HUB75 panel driver obect,  PASM display refresh cog using video output.

Author:         Tom Dinger (Ding-Batty)
Release Date:   2024/04/10

Copyright 2024 Thomas E. Dinger

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

--------------------------------------------------------------------------

This file contains the Display refresh code for HUB75 display panel(s).
It is written in PASM for sufficient speed to avoid flicker, using the
video generator to write the parallel bit streams, and CTRB to generate
the HUB75 shift clock.

This is mostly statically configured, using settings from the hub75_config.spin
file; the final initialization is deferred until after the cog is started,
so that the hub memory that contained the PASM code loaded into the cog can
be recovered and used for the display buffer, and other buffers.

For more information:
---------------------

See the file "hub75_info.txt" for some details on the display interface
called "HUB75", with some useful links.

Display Buffer
==============

This driver assumes the following memory organization:

Each BYTE of the display buffer contains 3 color bits for each of two RGB LEDs
in the same column, but REFRESH_ROWS apart. The six color bits are stored in
the low-order bits of the byte: %xx_RGB_rgb (for whichever color bit order is
configured). These are written to OUTA using the video generator in the cog
to write out the six bis using the VGA output mode.

The color values for one row are stored in consecutive bytes, with the lower-
address byte containing the lower columns. Rows are also stored consecutively,
with the rows at the top of the display at the lower addresses, so the LED
for row == 0 and column == 0 is the upper-left corner LED.

A Columns x Rows display will use (Columns + Rows) / 2 bytes of storage.
So, a 32 x 32 display will use 512 bytes of storage, and a 64 x 64 display
will use 2048 bytes of storage, for one color bit.

For color depth > 1 (more than 1 bit per color), the bits are stored in
"layers" with one layer for each color bit. These layers are also stored
consecutively, with the lowest-order color bit first (this can change.)

The memory layout for the color bits makes it easier to refresh the display,
but likely requires help with writing into the display buffer from a write-
helper cog, as Spin is too slow for writes of any large numbers of pixels.

Configuration and Initialization
================================

Most of this object is initialized "statically" from the configuration
constants. But two settings are received at runtime:
- The color bit depth, which is from 1 to 8 (I might be able to raise the upper
  limit for smaller panels).
- The base address of the display buffer.

The color bit depth is not a property of the hardware, but configured by the
application in the Init call and patched into the PASM cog before it is started.
So, the color bit depth cannot change for the run of the cog.

The base address of the display buffer is passed to the cog code indirectly
through the PAR register only after the cog has been started. The PAR register
points to a hub LONG that will contain the starting address of the display
buffer; the cog wait for it to become non-zero, and then uses that display
buffer address.

This initialization delay permits the application to load the COG, and then
lay out the memory buffers including the display buffer, and the app is able
to include the HUB memory contains the cog PASM code during the layout
decisions.

When the driver initializes, it configures the PLL for the video generator to
run at the same rate as the system clock (e.g. 80MHz), so that the clocks are
pretty well synchronized (with about a half-cycle, with some jitter). It then
configures the video generator to use the PLL clock, and use 8 PLL clocks per
video pixel, and 8 * 4 == 32 PLL clocks per frame (4 pairs of led bits).

The driver then measures the alignment of hub access against the "waitvid
hand-off point" (WHOP), and adjusts the relationship for "optimal" alignment,
as discussed some in the following thread:
https://forums.parallax.com/discussion/131254/fyi-waitvid-vs-hub-window

@kuroneko also published some demonstration video drivers usin this technique,
for example in:
https://forums.parallax.com/discussion/135844/poc-most-useless-piece-of-code-ever-the

(There is discussion about this in a number of other threads as well, but they
predate the latest forum migration and lots of links are broken, so they take
quite a bit of work to find.)

Once the hub access point (every 16 clocks) and the WHOP point that reloads the
next for pixels of data are alignemd properly, the code then walks over the
display buffer one LONG at a time and passes it to the video at the WHOP. For
each displayed row, it also generates the HUB75 shift clock signal by running
CTRB configured as an NCO for sysclk/8 (which will keep it aligned with the
hub accesses and WHOPs), and turns it on at the start of each row and off at the
end. This requires careful cycle-counting, as is noted in the comments.

Also, the hub/WHOP alignment is tracked as well during the "overhead" code
between rows and color bits -- all the overhead is padded to take exactly the
same number of clocks/instructions to maintain the synchrnization, and as a result,
all row updates, no matter what overhead is between them, takes
8 * DISPLAY_COLUMNS + 64 system clocks to complete, making the refresh rate "easy"
to compute. 

Performance
===========

In Spin only, we can get about 17 to 21 panel refreshes per second, with only
one-bit color depth for a 64 x 64 display. This results in a lot of flickering
on the display.

This version of the display driver uses the video generator to write the bits
for each pair of bytes in parallel, and CTRB is used to generate the
synchronized shift clock pulses to advance the six display shift registers.
The driver uses 8 sysclks per pixel pair, and for a pair of 64 x 64 panels
(128 x 64) and color bit depth of 3, it reaches >300 refreshes per second.

Revisions:
==========
- 2024/04/10 -- Added ACTUAL_ROW_NUMBER_MASK; revised refresh measurement; added
                EstimateRefreshClocks();
- 2024/02/26 -- Original release

}}

OBJ
  ' Pick up the static configuration.
  Config: "hub75_config.spin"

CON
  ' Import the needed configuration settings from the Config object, for
  ' code brevity.

  REFRESH_ROWS         = Config#REFRESH_ROWS
  MAX_REFRESH_ROW      = Config#MAX_REFRESH_ROW
  ROW_NUMBER_BITS      = Config#ROW_NUMBER_BITS
  DISPLAY_COLUMNS      = Config#DISPLAY_COLUMNS
  MAX_DISPLAY_COLUMN   = Config#MAX_DISPLAY_COLUMN
  SHIFT_CLOCK_PIN      = Config#SHIFT_CLOCK_PIN
  SHIFT_LATCH_PIN      = Config#SHIFT_LATCH_PIN
  OE_INVERTED_PIN      = Config#OE_INVERTED_PIN
  SHIFT_CLOCK_MASK     = Config#SHIFT_CLOCK_MASK
  SHIFT_LATCH_MASK     = Config#SHIFT_LATCH_MASK
  OE_INVERTED_MASK     = Config#OE_INVERTED_MASK
  ROW_NUMBER_MASK      = Config#ROW_NUMBER_MASK
  ROW_NUMBER_INCREMENT = Config#ROW_NUMBER_INCREMENT
  COLORS_BASE_PIN      = Config#COLORS_BASE_PIN
  COLORS_MASK          = Config#COLORS_MASK

  RED_BIT              = Config#RED_BIT
  GREEN_BIT            = Config#GREEN_BIT
  BLUE_BIT             = Config#BLUE_BIT

  REFRESH_ROWS_SHIFT   = >| MAX_REFRESH_ROW

  ' Define the order to output the bytes for the four pixels per long:
  ' low-order bits are first both for the pixel index (below) and the
  ' bytes in the LONG containing the pixel colors. This is used as the
  ' "Pixels" parameter for the WAITVID (or the equivalent video generator
  ' load operation), to select the four bit patterns for the color bits
  ' from the "Colors" WAITVID parameter.
  PIXEL_ORDER = %%3210

  ' We want to control the offset of the WHOP (and the appearance of
  ' the first output bit of the data provided in the WHOP) from the hub access
  ' point, so we can have predicatable timing behavior.
  ' So, we want to know a few things:
  ' - Where in the WAITVID is the WHOP?
  ' - Where in the WAITVID (or after) is the first bit available?
  ' We control the timing of the finish of the hub access and the following
  ' WAITVID completion, in the range of 4..20, and measure the timing of various
  ' elements.
  ' We also want to know if there is a "tolerance" of a range of clocks for
  ' picking up the two parameters in the WHOP.
  '
  ' For 8 clocks per pixel, it looks like the "sweet spot" for range of control
  ' is 7..10 for WHOP_OFFSET, which puts the WHOP on the instruction immediately
  ' after the hub access. The current value is selected, along with the phase
  ' of CTRB generating the HUB75 shoft clocks, to put the shift clock rising edge
  ' in the middle of the pixel level hold from the video generator.
  WHOP_OFFSET = 8
      '  6 == pcZero
      '  7 == pcData1
      '  8 == pcData1
      '  9 == pcData1
      ' 10 == pcData1
      ' 11 == pcData2
      ' 12 == pcData2
      ' 13 == not clear (junk?)
      ' 14 == not clear (junk?)
      ' 15 == pcZero
      ' 16 == pcZero
      ' 17 == pcZero
      ' 18 == pcZero
      ' 19 == pcZero
      ' 20 == a few extra pulses
      ' 21 == a few extra pulses
      ' 22 == a few extra pulses
      ' 23 == pcData1 (offset 200 ns from WHOP+12)
      ' 24 == pcData1 (offset 200 ns from WHOP+12)
      ' 25 == pcData1 (offset 228 ns from WHOP+12)
      ' 26 == pcData1 (offset 244 ns from WHOP+12)
      ' 27 == pcData2 (offset 252 ns from WHOP+12)
      ' 28 == pcData2 (offset 268 ns from WHOP+12?)
      ' 29 == junk? 

  ' For an adjustment window of 8 (actually 7..11), we have the following
  ' "slots":
  '    +00       WHOP
  '    +04       Start or stop shift clocks in CTRB wth PSHB == $4000_0000 or 0
  '              by setting FRQB to %0010_00000 << 23
  '    +08, +12  Hub access aligned
  '    +16
  '    +20
  '    +24, +28  Hub access aligned
  '
  ' For an adjustment of 12 above, we have the following "slots" in the 32-clock
  ' window for each WHOP:
  '    +00       WHOP
  '    +04       Start or stop shift clocks in CTRB wth PSHB == $2000_0000 or 0
  '              by setting FRQB to %0010_00000 << 23
  '    +04, +08  Hub access (overlaps with above)
  '    +12
  '    +16
  '    +20, +24  Hub access
  '    +28

  ' COLOR_PINS_VIDEO_GROUP is 0..3
  COLOR_PINS_VIDEO_GROUP = (COLORS_BASE_PIN >> 3) & 3

'=========================================================================
' Validate the various configuration settings:
' - Pins are valid, and do not conflict with each other
' - display configuration "makes sense"
'
' The validation is done using FIT statements to make logical assertions
' about the config values, such as "is a power of two", "has one of a few
' specific values", etc.
' ------------------------------------------------------
' Verify the above settings.
'
' Now, use a "trick" to get the Spin compiler to check that the values for
' CACHE_LINE_LENGTH and CACHE_SIZE_BYTES are acceptable:
' - Make sure they both are powers of two.
' - Make sure they both are in the supported range
' - Make sure the pair is a supported combination.

DAT
  ' If the condition is false (==0) then the "org 1" means that the
  ' fit statement will generate an error. So we set the conditionals up
  ' so that a unsupported configurations generate a false (==0) value.

  org 1

  ' The REFRESH_ROWS must be a power of two.
  ' Error means REFRESH_ROWS is not a power of 2
  fit ||((REFRESH_ROWS == |< ((>| REFRESH_ROWS)-1)))

  ' The DISPLAY_COLUMNS must be a multiple of four, because we use four
  ' pixels per LONG of hub memory.
  ' Error means DISPLAY_COLUMNS is not a multiple of 4
  fit ||((DISPLAY_COLUMNS & 3) == 0)

  ' The COLORS_BASE_PIN must be one of 0, 8, 16, or 24, as the base pin of
  ' a (VGA) video pin group.
  ' Error means the COLORS_BASE_PIN was given an unspported value.
  ' For some reason FlexSpin doesn't like the "OR" operator:
  '   "error: syntax error, unexpected instruction"
'  fit ||((COLORS_BASE_PIN ==  0) OR (COLORS_BASE_PIN ==  8) OR {
'}        (COLORS_BASE_PIN == 16) OR (COLORS_BASE_PIN == 24))
  ' The following seems to work FlexProp, and the other compilers:
  fit ||((COLORS_BASE_PIN ==  0) + (COLORS_BASE_PIN ==  8) + {
}        (COLORS_BASE_PIN == 16) + (COLORS_BASE_PIN == 24))

  ' This version of the code assumes that the first color layer in the buffer
  ' is the least-significant color bit, so we make sure that the config file
  ' is set up correctly.
  fit ||((Config#DBUF_HI2LO_LAYER_DELTA & $80000000) <> 0)
  '
  ' For most-significant bit in the first layer, use:
  'fit (Config#DBUF_LO2HI_LAYER_DELTA & $80000000) <> 0

  ' We could do a lot more pin assignment collision testing. Caveat Emptor.

'=========================================================================

PUB NoOp
{{
  Every Spin Object must have at least one Spin PUB function, and the first
  one may be run automatically, so we make sure the first one is "safe."
  This one does nothing.

  Arguments:
    None

  Return Value:
    None
}}
  ' Do nothing

'=========================================================================
PUB LoadDisplayCog(ColorDepth, DisplayBufAddrVar, RefreshClocksAddrHub)
{{
  Add the runtime-configured parameters to the hub image for the cog PASM code
  before loading the cog:
  - Color Depth (bumber of bits in the each color, i.e. the number of "layers"
  - Address of the hub variable that will eventually get the address of the
    display buffer, to allow delayed initialization of the cog, so the
    application can reclaim the hub copy of the cog code.
  - The address of the location in the hub to get the count of sysclks used for
    a single full display refresh, so the refresh rate can be reported.
  Then the cog is started, which will load the cog, and then the cog code will
  wait until the display buffer address is supplied.

  Arguments:
    ColorDepth -- the numner of bits in each color value, which is the number
                  of color "layers" in the display buffer. E.g. "24-bit" color,
                  which is 8:8:8, would have a depth of 8.
    DisplayBufAddrVar -- the hub addres of the variable that will eventually
                         contain the actual hub address of the display buffer.
                         This can start out as zero, which will cause the
                         cog initialization code to stall, waiting for a non-
                         zero address to be supplied. This is a "two-stage"
                         initialization for the cog.
    RefreshClocksAddrHub -- the hub address of the LONG variable that will
                            receive the number of sysclks needed for one full
                            refresh of the display, so that the refresh rate
                            can be calculated and reported.

  Return Value:
    The Cog ID of the display refresh cog, plus one, so that it is non-zero.
}}
  ' Patch the color depth into the Cog code, and load the cog, passing the
  ' pointer to the place the cog will (eventually) get the display buffer
  ' address.
  '
  ' Returns the cog# + 1 of the cog started.

  ' Put the final few values into the cog.
  pcFullRefreshClocksAddrHub := RefreshClocksAddrHub
  pcColorDepthBits           := ColorDepth

  ' Start the cog, and pass the address in Hub memory that will eventually get
  ' the hub address of the display buffer. The cog, after starting, waits until
  ' that location contains a non-zero hub address, and uses that address as the
  ' start of the display buffer.
  return 1 + COGNEW(@pPasmStart, DisplayBufAddrVar)

'=========================================================================
PUB EstimateRefreshClocks(depth) | tmp, clks
{{
  Based on the current implementation, and the dimensions of the configured
  display, estimate the number of clock cycles the one refresh would take.

  This is useful to determine the maximum color depth that would have an
  acceptably high (=> 60 Hz) refresh rate at the current clock frequency.

  Arguments:
    depth -- the numner of bits in each color value, which is the number
             of color "layers" in the display buffer. E.g. "24-bit" color,
             which is 8:8:8, would have a depth of 8.

  Return Value:
    The estimated number of clocks for one full refresh of the display.

  Notes:
    refresh_clocks = 
            CPP * cols * rows * (2**depth - 1)
          + REPO * rows * (2**depth - 1 - depth)
          + NRO  * (rows - 1) * depth
          + LLRO * (bits - 1)
          + FRO
    with
      CPP  ==   8      Clocks Per Pixel
      REPO ==  64      Overhead for a REPeated row display
      NRO  ==  64      Next Row Overhead for the last repeat of a row, but not
                       for the last row of a layer
      LLRO ==  64      Last Layer Row Overhead for the last repeat of the last
                       row of a layer.
      FRO  ==  96      Full Refresh Overhead for the last repeat of the last
                       row of the last layer. This includes 32 clocks spent
                       measuring the actual refresh clocks.

  In fact, it reduces to
  
      clks := (((cols * rows) + 8) * 8) * (2**bits - 1) + 32

  which is
      rows_mult_shift = constant((>| bits) - 1)
      tmp  := (cols << (rows_mult_shift + 3)) + 64
      clks := (tmp << constant(|< bits) - tmp) + 32
  and the 32 can be dropped if the measurement code is removed.
}}
  '' This calculation is correct
  'return ( 8 * DISPLAY_COLUMNS * REFRESH_ROWS * ((|< depth) - 1)) {
  '  }  + (64 * REFRESH_ROWS * ((|< depth) - 1 - depth))          {
  '  }  + (64 * (REFRESH_ROWS - 1) * depth)                       {
  '  }  + (64 * (depth - 1))                                      {
  '  }  + 96

  ' This would be calculated quite differently in PASM to avoid the
  ' multiplies.
  ' Remove the final "+ 32" if the PASM timing code is removed from the
  ' refresh loop'
  tmp := (DISPLAY_COLUMNS + 8) << (ROW_NUMBER_BITS + 3)
  'return (tmp << depth) - tmp + 32
  return (tmp << depth) - tmp


' ========================================================================
'
' PASM code for display refresh
'
' Style notes:
' I use special prefixes for code labels, variables and constants stored in
' the pasm code, so that names make clear what is being referred to.
'
' "pCodeLabel" -- A label on code (possibly with a leading ":")
' "pvVariable" -- A label on a variable stored in a register in the cog,
'                 that can change over the course of cog execution
' "pcConstant" -- A label on a value that is effectively constant for the
'                 lifetime of the cog -- it may be set at initialization time,
'                 but will not be changed later.
' "piInstruction" -- A label on an instruction that may be the target of
'                    runtime instruction modification.

DAT
                    org 0
pPasmStart
                    jmp     #pInitContinue

' We store all the constants and some operating variables up front, generally
' the ones needing compile-time initialization.

' pcBufAddrHub contains the hub address of the first row and column of the
' lowest-significance color bit of the display buffer.
' Set by the initialization code that starts the cog.

pcBufAddrHub        long    0    ' Starting address of the entire buffer

' Hub addresses for the refresh clock count. Set by the initialization code
' that starts the cog. 
pcFullRefreshClocksAddrHub  long   0

' pcColorDepthBits is the number of bits per color. Set by the initialization
' code that starts the cog.
pcColorDepthBits    long     3

' pcDisplayColumns is the number of display columns per row; this is _not_ the
'                  number of longs per row, and might not even be the number of
'                  _bytes_ per row, depending on how the data is packed.
'
' It is unlikely, but possible, for these to exceed 511 -- for example, 8
' panels in a series, each of which has 64 columns, would reach 512 columns.
' If we need cog space, we could get rid of them and limit the column count.

pcDisplayColumns    long    DISPLAY_COLUMNS
pcColumnCountLongs  long    DISPLAY_COLUMNS >> 2  ' Number of LONGs in a row

' Control pin and mask definitions.

pcOEInvertedMask    long    OE_INVERTED_MASK

' Instead of using a normal counter for the current display row number being
' updated, and shifting it into position to write it to the display, we
' manage the row number value in its already-shifted position, to save a few
' instructions in the row processing loop.

pcRowNumberMask     long    ROW_NUMBER_MASK
pcRowNumberIncr     long    ROW_NUMBER_INCREMENT

' pcDiraMask is the mask of output bits to set in DIRA when the cog starts.

pcDiraMask          long    COLORS_MASK | Config#ACTUAL_ROW_NUMBER_MASK | {
  }                         SHIFT_CLOCK_MASK | SHIFT_LATCH_MASK | {
  }                         OE_INVERTED_MASK

' pcOEInvertedShiftLatchMask is used during row number update to eliminate
' a runtime instruction. We can treat the shift latch and Output enabled
' as a single bit (for the ICN2037 controller chip, at least), because th
' two signals are inverted w.r.t. each other, and trigger on opposite edges. 

pcOEInvertedShiftLatchMask long OE_INVERTED_MASK | SHIFT_LATCH_MASK

' Simple constants
pcZero               long     0
pc_4000_0000         long    $4000_0000

' Both counters in NCO mode on the same pin, so one counter can "gate"
' pulses from the other counter.
pcCtrbInit
pcCtraInit           long    %00100 << 26 | SHIFT_CLOCK_PIN

' The number of times to repeat display of each row, to get the right
' brightness for each layer.
pcMaxLayerRowRepeatCountLess1  long    0   ' (|< (pcColorDepthBits - 1)) - 1

'                        v: x -- don't care
'                        | vvvvv: ctr mode %00001: PLL, no output pins
'                        | ||||| vvv: PLLDIV for PLL / 1
pcCtraMode      LONG    %0_00001_111_00000000_000000_000_000000

'                         v: x -- don't care
'                         | vvvvv: ctr mode %00100: NCO, single-ended.
'                         | |||||              vvvvv: CTRB output A Pin
pcCtrbMode      LONG    (%0_00100_000 << 23) | SHIFT_CLOCK_PIN

'                        v: 8 pllclk per pixel
'                        |          vv: 4 pixels per "frame" == 4 * 8 pllclks
pcVidSCL        LONG    (8 << 12) | 32

'                        v: don't care
'                        | vv: Video Mode %01: VGA
'                        | || v: CMode: 4-color (2 bpp)
'                        | || | vv_vvv: Chroma0/1 + aural sub -- don't care.
'                        | || | || |||         vvv: Vgroup -- e.g. 1 for pins 15..8
'                        | || | || |||         |||                            vvvvvvvv: output pins in group
pcVidCFG        LONG    %0_01_1_00_000 << 23 | COLOR_PINS_VIDEO_GROUP << 9 | %00111111
pvDummy         LONG    0
pvDelay         LONG    0

' The following variables can use static initialization, which saves one runtime
' instruction to set them; otherwise they could use RES.
pvColumnLoopCount     LONG    DISPLAY_COLUMNS >> 2
pvLayerRowRepeatCount LONG    1
pvRowRepeatLoopCnt    LONG    1
pvPrevRow             LONG    0
pvCurrentRow          LONG    0

' To stop the cog, the main program can also set the OEInverted pin for
' output, and can write a 1 to it to override anything this cog can do
' (output state is OR'd amoung all the cogs); then this cog can be
' stopped.

'-------------------------------------------------------------------------

' The code continues here...
pInitContinue
                    ' Wait for the app to pass the display buffer
                    ' address to us.
:pInitWaitLoop
                    rdlong  pcBufAddrHub, PAR  wz
    if_z            jmp     #:pInitWaitLoop

' Things to do during initialization:
' - init OEInverted in OUTA to disable the display when we change the
'   pin to output
' - Set DIRA for the output pins we want.
' - select movs/movd instructions for the display refresh loop

                    ' Calculate some helpful constants
                    mov     pcMaxLayerRowRepeatCountLess1, #1
                    shl     pcMaxLayerRowRepeatCountLess1, pcColorDepthBits
                    shr     pcMaxLayerRowRepeatCountLess1, #1
                    sub     pcMaxLayerRowRepeatCountLess1, #1

                    ' Set OEInverted so that the output starts disabled.
                    ' All other output bits are zero.
                    mov     OUTA, pcOEInvertedMask

                    ' Do some start-of-refresh-pass initializations here, so that
                    ' they aren't needed once we start counting WHOP clocks.
                    mov     pvHubFetchAddr, pcBufAddrHub

                    mov     pvColumnLoopCount, pcColumnCountLongs
                    mov     pvLayerRowRepeatCount, #1
                    mov     pvRowRepeatLoopCnt, pvLayerRowRepeatCount
                    mov     pvFullRefreshClocks, #0

                    '' Instead of measuring the number of clocks for a full
                    '' refresh of all layers, we can simply calculate that here,
                    '' because of the strict timing constraints on writing row
                    '' bits:
                    '' Each row write has 64 clocks of overhead, so one row for one
                    '' color bit takes
                    ''    (8 * pcDisplayColumns) + 64  clocks
                    '' we have REFRESH_ROWS rows:
                    ''    ((8 * pcDisplayColumns) + 64) * REFRESH_ROWS clock
                    '' for color bits, we repeat each row
                    ''    (|< pcColorDepthBits) - 1 times
                    'mov     pvTemp, pcDisplayColumns
                    'shl     pvTemp, #3   ' * 8
                    'add     pvTemp, #64  ' + 64 == clocks for one row
                    'shl     pvTemp, #REFRESH_ROWS_SHIFT   ' * 32
                    'mov     pvFullRefreshClocks, pvTemp
                    'shl     pvFullRefreshClocks, pcColorDepthBits
                    'sub     pvFullRefreshClocks, pvTemp
                    'wrlong  pvFullRefreshClocks, pcFullRefreshClocksAddrHub
                    mov     pvTemp, pcDisplayColumns
                    add     pvTemp, #8
                    shl     pvTemp, #(REFRESH_ROWS_SHIFT + 3)
                    mov     pvFullRefreshClocks, pvTemp
                    shl     pvFullRefreshClocks, pcColorDepthBits
                    sub     pvFullRefreshClocks, pvTemp
                    ' if we remove the refresh timing code, we can also remove this
                    ' additional 32 clocks.
                    'add     pvFullRefreshClocks, #32
                    wrlong  pvFullRefreshClocks, pcFullRefreshClocksAddrHub

                    ' Set up CTRA for PLL to generate sysclk/1
                    ' (This assumes FRQA and CTRA were already init'd to zero)
                    ' The FRQ divides sysclk by 16, the PLL multiplies it by 16,
                    ' and we use the PLLDIV of PLL/1.
                    movi    FRQA, #%0001_00000  ' 80MHz / 16 == 5 MHz
                                                ' vid PLL does * 16 back to 80MHz
                    ' We use PLL tap PLL / 1 to get 80 MHz, or 8 PLL clocks per
                    ' pixel == 8 sysclks per pixel.
                    mov     CTRA, pcCtraMode

                    ' Set up CTRB to generate a sysclk/8 signal in NCO mode.
                    ' We adjust the phase (by setting FRQB) when we know the
                    ' WHOP point, and we enable/disable it by setting or zeroing
                    ' the FRQB value.
                    ' We set the starting phase for the generated clocks here --
                    ' we turn the pulses on and off at the exact same point
                    ' within the pulses so the phase is always correct when we
                    ' turn it on again.
                    mov     FRQB, #0             ' Set later to the right value
                                                 ' for sysclk / 8
                    mov     PHSB, pc_4000_0000   ' Correct phase relative to the WHOP
                    mov     CTRB, pcCtrbMode

                    ' Set up video generator for 8 clock per pixel,
                    ' 32 clocks (4 pixels) per long, 2 bits per pixel.
                    ' (VGA mode).
                    mov     VSCL, pcVidSCL 
                    mov     VCFG, pcVidCFG

                    ' Wait 1 to 3 ms, to let the PLL settle down.
                    rdlong  CNT, #0   ' Get the current sysclk frequency
                                      ' Put it in the CNT shadow register
                    shr     CNT, #9   ' 1/512th of a second, about 2 ms
                    add     CNT, CNT  ' shadow-cnt has future CNT value
                    waitcnt CNT, #0   ' about 2 ms delay

                    ' Now, measure the sysclk relationship between the end of
                    ' WAITVID and the end of the next hub access. We then
                    ' "stretch" the VSCL frame length for one frame so that
                    ' the WHOP falls 5 to 8 sysclks after the end of the hub
                    ' access.
                    waitvid pcZero, #0
                    rdlong  pvDummy, #0     ' don't care about the data.
                    neg     pvTemp, CNT     ' the end of the hub access
                    waitvid pcZero, #0
                    ' We assume the WHOP is here, or before here.
                    add     pvTemp, CNT     ' now the delay amount

                    ' Based on the measurement, figure out how much to adjust
                    ' te frame size to delay the WHOP until it has the right
                    ' relationship with the hub access.
                    ' The WHOP_OFFSET value was calculated by trying different
                    ' values, and looking at the video pin output and the
                    ' shift-clock pin output on a dual-trace 'scope, until we
                    ' found the value that worked the best.
                    mov     pvDelay, #(WHOP_OFFSET + 32)
                    sub     pvDelay, pvTemp
                    and     pvDelay, #31

                    ' Stretch one frame
                    add     VSCL, pvDelay  ' one frame is stretched.
                    waitvid pcZero, #0 
                    sub     VSCL, pvDelay  ' next frame is back to normal
                    waitvid pcZero, #0
                    ' Here we are at WHOP + 8 (approximately)

                    ' From here on, we need to keep track of the WHOP points and
                    ' hub accesses points to keep things aligned. We do not use
                    ' WAITVID after this point (although we could, I think).
                    ' Also the hub access point alignment is not quite reliable
                    ' because there are two of them in each 32-clock WHOP
                    ' period.
                    '
                    ' Also, we don't really care about the video output bits if
                    ' we are not generating shift-clocks, so we don't try to
                    ' control the video output state when we are not generating
                    ' shift-clocks.
                    mov     DIRA, pcDiraMask                ' WHOP + 8
                    neg     pvRefreshPassStartCnt, CNT      ' WHOP +12
                    nop                                     ' WHOP +16
                    jmp     #pColumnLoopContinued           ' WHOP +20

'-------------------------
pNextRowStart
                    mov     pvRowRepeatLoopCnt, pvLayerRowRepeatCount 'WHOP + 4
                    nop                                     ' WHOP + 8 
                    jmp     #pColumnLoop                    ' WHOP +12

'-------------------------
pRepeatRow
                    sub     pvHubFetchAddr, pcDisplayColumns 'WHOP +24
                    nop                                     ' WHOP +28 
                    nop                                     ' WHOP   (data ignored)
                    nop                                     ' WHOP + 4 
                    nop                                     ' WHOP + 8 
                    nop  ' jmp     #pColumnLoop             ' WHOP +12

'-------------------------

pColumnLoop
                    ' Put the ORInverted and ShiftLatch bit(s) into
                    ' the pvPrevRow for the end of the row. This will
                    ' be XOR'd into OUTA, and since those bits are zero
                    ' in OUTA at the time, they will be set.
                    ' It doesn't hurt to do this multiple times, and we have
                    ' an otherwise unused instruction slot here.
                    or      pvPrevRow, pcOEInvertedShiftLatchMask ' WHOP +16
                    jmp     #pColumnLoopContinued           ' WHOP +20

'-------------------------
pColorBitLayerStart
                    ' Fix the special case of COLOR_DEPTH=1 for the low-order
                    ' color layer.
    if_c_and_z      mov     pvLayerRowRepeatCount, #1       ' WHOP +12
                    shl     pvLayerRowRepeatCount, #1       ' WHOP +16
                    mov     pvRowRepeatLoopCnt, pvLayerRowRepeatCount 'WHOP +20

pColumnLoopContinued
                    rdlong  pvColors, pvHubFetchAddr        ' WHOP +24, +28
                    cmp     pvColors, #PIXEL_ORDER          ' WHOP

                    ' Start/restart the shift-clock from CTRB
                    ' It doesn't hurt to do this every frame, because the counter
                    ' and the instructions stay in-phase every time. But the first
                    ' time through this loop body will start generating the
                    ' shift-clocks for the first four pixel pairs.
                    movi    FRQB, #%0010_00000              ' WHOP + 4
                    add     pvHubFetchAddr, #4              ' WHOP + 8

                    djnz    pvColumnLoopCount, #pColumnLoop ' WHOP +12
                    ' 4 more clocks here when _not_ looping ' WHOP +16

                    mov     pvColumnLoopCount, pcColumnCountLongs ' WHOP +20
                    sub     pvRowRepeatLoopCnt, #1 wz       ' WHOP +24

                    ' If Z is clear, then this is _not_ the last time through
                    ' this row on the layer, so we skip over the row number
                    ' increment, and we back up the pvHubFetchAddr to be at the
                    ' start of the same row again, and we repeat the row output.
                    '
                    ' If Z is set, this is the end of the last time through this
                    ' row for this layer, so we _do_ increment the row number,
                    ' and we do not back up pvHubFetchAddr, so that is continues
                    ' with the next row.
                    ' But... We cannot change the row number or disable output
                    ' until we have finished the shift-clocks, which happens
                    ' after the final WHOP instruction.

                    ' Latch the row data, set the row number, enable output,
                    ' and advance the row number
                    ' We XOR-change the row number bits in OUTA to get an
                    ' "atomic" update of the row number without (briefly)
                    ' setting the row number to zero. My assumption is that
                    ' the row number is _not_ latched in any way.

                    ' This version triggers the ShiftLatch and disables output
                    ' in the same instruction that the row bits are changed,
                    ' which should not cause any ghosting. The second latch,
                    ' associated with the Output Enable pin, will not collect
                    ' the row settings until OEInverted is set to zero in the
                    ' next instruction.
                    ' (I suppose it is possible that OEInverted is slower to
                    ' propagate than the row number bits, but that should only
                    ' be a very few ns (or less time) of wrong-row display.)

                    xor     pvPrevRow, pvCurrentRow         ' WHOP +28
                    mov     pvPrevRowTemp, pvPrevRow        ' WHOP   (data ignored)
                    mov     FRQB, #0  ' end shift-clocks    ' WHOP + 4
                    mov     pvPrevRow, pvCurrentRow         ' WHOP + 8
                                                                      
                    xor     OUTA, pvPrevRowTemp             ' WHOP +12
                    andn    OUTA, pcOEInvertedShiftLatchMask 'WHOP +16

    if_nz           jmp     #pRepeatRow                     ' WHOP +20

                    ' Here we advance to the next row, and see if we are done
                    ' with the layer (row number wrapped around).

                    add     pvCurrentRow, pcRowNumberIncr   ' WHOP +24
                    and     pvCurrentRow, pcRowNumberMask wz 'WHOP +28  ' did we wrap around?

                    ' If Z is clear, then we are proceeding to the
                    ' next row, so we need to reset the row repeat loop counter,
                    ' and jump.
    if_nz           jmp     #pNextRowStart                  ' WHOP   (data ignored)

' End-of-layer
                    ' All the rows for this color bit have been shown.
                    ' After each color bit, we adjust the row repeat count
                    ' (doubling it) so the next bit is shown for twice as long,
                    ' if not starting a new refresh pass.

                    ' See if we have just finished the most significant bit of
                    ' the color depth:
                    ' - As a special case: if COLOR_DEPTH = 2, we have
                    '   pcMaxLayerRowRepeatCountLess1 == 1, which is also
                    '   the repeat count for the first of the two layers, so
                    '   the CMPSUB will set the C bit and set the repeat count
                    '   for the next row to zero, which is not what we want.
                    '   Detect this by also checking for Z -- if C is set _and_
                    '   Z is set, this is _not_ the final row, and we need to
                    '   "fix" the pvLayerRowRepeatCount by setting it back to 1
                    '   before it is doubled for the next row.
                    ' - If we have finished the highest bit (the largest repeat
                    '   count), the "cmpsub" instruction will set the C flag, and
                    '   change the repeat count back to 1.
                    ' - If we have not finished the highest bit, then the
                    '   repeat count will be unchanges and C clear, and we should
                    '   shift the count left one bit and do the next layer.
                    cmpsub  pvLayerRowRepeatCount, pcMaxLayerRowRepeatCountLess1  wc, wz ' WHOP + 4
     if_nc_or_z     jmp     #pColorBitLayerStart            ' WHOP + 8

                    ' Start the next refresh pass
                    ' Here on WHOP + 12

'' This section, if enabled, measures the actual sysclks for one refresh, but
'' it takes an extra 32 sysclks (one WHOP period) to accomplish. It is available
'' for validating the predicted refresh clock calculated earlier.
'                    cmp     pvFullRefreshClocks, #0  wz                ' WHOP +12
'                    mov     pvFullRefreshClocks, pvRefreshPassStartCnt ' WHOP +16
'                    neg     pvRefreshPassStartCnt, CNT                 ' WHOP +20
'                    sub     pvFullRefreshClocks, pvRefreshPassStartCnt ' WHOP +24
'
'                    mov     pvHubFetchAddr, pcBufAddrHub               ' WHOP +28
'                    mov     pvRowRepeatLoopCnt, pvLayerRowRepeatCount  ' WHOP (data ignored)
'                    nop     ' optional for hub access alignment        ' WHOP + 4
' 
'                    ' Copy back to hub
'     if_nz          wrlong  pvFullRefreshClocks, pcFullRefreshClocksAddrHub ' WHOP + 8, +12
'
'                    nop                                                ' WHOP +16
'                    jmp     #pColumnLoopContinued                      ' WHOP +20

' This section is used when not measuring refresh clocks. 
                    mov     pvHubFetchAddr, pcBufAddrHub               ' WHOP +12
                    mov     pvRowRepeatLoopCnt, pvLayerRowRepeatCount  ' WHOP +16
                    jmp     #pColumnLoopContinued                      ' WHOP +20

'-------------------------------------------------------------------------
' Runtime-initialized variables.

' Stores the number of system clocks needed for one full refresh of all color
' layers for the display. This is copied back into hub memory so that the main
' program can report the performance of the driver, in refreshes per second.
pvFullRefreshClocks  res     1
pvRefreshPassStartCnt res     1

pvTemp               res     1    ' temporary variable
                     res     1
' These are variables to manage the refresh operation: loop counters,
' hub address pointers, etc. (Could be RES at the end...)
pvColors             res     1  ' Gets the current row/col color bits
pvHubFetchAddr       res     1  ' The next BYTE to read
pvPrevRowTemp        res     1

'====================================================================

PasmEnd             fit  $1F0

