{{
HUB75 panel driver obect,  PASM display refresh cog.

Author:         Tom Dinger (Ding-Batty)
Release Date:   2024/04/10

Copyright 2024 Thomas E. Dinger

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

--------------------------------------------------------------------------

This file contains Display refresh code for HUB75 display panel(s).
It is written in PASM for sufficient speed to avoid flicker,, bit-banging the
output using a fast MOVS/MOVD inner loop for four pixel pairs per LONG.

This is mostly statically configured, using settings from the hub75_config.spin
file; the final initialization is deferred until after the cog is started,
so that the hub memory that contained the PASM code loaded into the cog can
be recovered and used for the display buffer, and other buffers.

For more information:
---------------------

See the file "hub75_info.txt" for some details on the display interface
called "HUB75", with some useful links.

Display Buffer
==============

This driver assumes the following memory organization:

Each BYTE of the display buffer contains 3 color bits for each of two RGB LEDs
in the same column, but REFRESH_ROWS apart. The six color bits are stored in
the low-order bits of the byte: %xx_RGB_rgb (for whichever color bit order is
configured). These are written to OUTA using either the MOVS or MOVD instruction
depending on the COLORS_BASE_PIN setting (this is configured once when the cog
starts up.

The color values for one row are stored in consecutive bytes, with the lower-
address byte containing the lower columns. Rows are also stored consecutively,
with top rows at the lower addresses, so the LED for row == 0 and column == 0
is the upper-left corner LED (by convention).

A Columns x Rows display will use (Columns + Rows) / 2 bytes of storage.
So, a 32 x 32 display will use 512 bytes of storage, and a 64 x 64 display
will use 2048 bytes of storage, for one color bit.

For color depth > 1 (more than 1 bit per color), the bits are stored in
"layers" with one layer for each color bit. These layers are also stored
consecutively, with the lowest-order color bit first (this might change.)

The memory layout for the color bits makes it easier to refresh the display,
but likely reuqires help with writing into the display buffer from a write-
helper cog, as Spin is too slow for writes of any large numbers of pixels.

Configuration and Initialization
================================

Most of this object is initialized "statically" from the configuration
constants. But two settings are received at runtime:
- The color bit depth, which is from 1 to 8 (I might be able to raise the upper
  limit for smaller panels).
- The base address of the display buffer.

The color bit depth is not a property of the hardware, but configured by the
application in the Init call and patched into the PASM cog before it is started.
So, the color bit depth cannot change for the run of the cog.

The base address of the display buffer is passed to the cog code indirectly
through the PAR register only after the cog has been started. The PAR register
points to a hub LONG that will contain the starting address of the display
buffer; the cog wait for it to become non-zero, and then uses that display
buffer address.

This initialization delay permits the application to load the COG, and then
lay out the memory buffers including the display buffer, and the app is able
to include the HUB memory contains the cog PASM code during the layout
decisions.

The only significant initialization is to set up the two cog counters: CTRA
is configured to generate the shift-clock signal every 8 system clocks, and
CTRB is configured to allow a limited number of shift-clock pulses in a
burst (4 in this case, one for each byte of the LONG).

Performance
===========

In Spin only, we can get about 17 to 21 panel refreshes per second, with only
one-bit color depth. This is pretty poor, and resuls in a lot of flickering
on the display.

This version of the PASM driver uses waitcnt to adjust the display time for
the different order color bits, and fetches groups of 4 pairs of pixel bits
at a time (RDLONG instead of RDBYTE), and uses both counter in the cog to
generate the ShiftClock pulses for the data. It is able to get about
3000 refreshes per second for one-bit color depth for a 64x64 panel; 6 bits
would get about 48 refreshes per second for 64x64. (TODO: Update these numbers
with the "final" version of the WAITCNT-based code.)

Revisions:
==========
- 2024/04/10 -- Added ACTUAL_ROW_NUMBER_MASK; revised refresh measurement; added
                EstimateRefreshClocks();
- 2024/02/26 -- Original release

}}

OBJ
  ' Pick up the static configuration.
  Config: "hub75_config.spin"

CON
  ' Import the needed configuration settings from the Config object, for
  ' code brevity.

  REFRESH_ROWS         = Config#REFRESH_ROWS
  MAX_REFRESH_ROW      = Config#MAX_REFRESH_ROW
  ROW_NUMBER_BITS      = Config#ROW_NUMBER_BITS
  DISPLAY_COLUMNS      = Config#DISPLAY_COLUMNS
  MAX_DISPLAY_COLUMN   = Config#MAX_DISPLAY_COLUMN
  SHIFT_CLOCK_PIN      = Config#SHIFT_CLOCK_PIN
  SHIFT_LATCH_PIN      = Config#SHIFT_LATCH_PIN
  OE_INVERTED_PIN      = Config#OE_INVERTED_PIN
  SHIFT_CLOCK_MASK     = Config#SHIFT_CLOCK_MASK
  SHIFT_LATCH_MASK     = Config#SHIFT_LATCH_MASK
  OE_INVERTED_MASK     = Config#OE_INVERTED_MASK
  ROW_NUMBER_MASK      = Config#ROW_NUMBER_MASK
  ROW_NUMBER_INCREMENT = Config#ROW_NUMBER_INCREMENT
  COLORS_BASE_PIN      = Config#COLORS_BASE_PIN
  COLORS_MASK          = Config#COLORS_MASK

  RED_BIT              = Config#RED_BIT
  GREEN_BIT            = Config#GREEN_BIT
  BLUE_BIT             = Config#BLUE_BIT


'=========================================================================
' Validate the various configuration settings:
' - Pins are valid, and do not conflict with each other
' - display configuration "makes sense"
'
' The validation is done using FIT statements to make logical assertions
' about the config values, such as "is a power of two", "has one of a few
' specific values", etc.
' ------------------------------------------------------
' Verify the above settings.
'
' Now, use a "trick" to get the Spin compiler to check that the values for
' CACHE_LINE_LENGTH and CACHE_SIZE_BYTES are acceptable:
' - Make sure they both are powers of two.
' - Make sure they both are in the supported range
' - Make sure the pair is a supported combination.

DAT
  ' If the condition is false (==0) then the "org 1" means that the
  ' fit statement will generate an error. So we set the conditionals up
  ' so that a unsupported configurations generate a false (==0) value.

  org 1

  ' The REFRESH_ROWS must be a power of two.
  ' Error means REFRESH_ROWS is not a power of 2
  fit ||((REFRESH_ROWS == |< ((>| REFRESH_ROWS) - 1)))

  ' The DISPLAY_COLUMNS must be a multiple of four, because we use four
  ' pixels per LONG of hub memory.
  ' Error means DISPLAY_COLUMNS is not a multiple of 4
  fit ||((DISPLAY_COLUMNS & 3) == 0)

  ' The COLORS_BASE_PIN must be either 0 or 9, because the code uses a
  ' MOVS or MOVD for the six color bits, for speed.
  ' Error means the COLORS_BASE_PIN was given an unspported value.
  ' For some reason FlexSpin doesn't like the "OR" operator:
  '   "error: syntax error, unexpected instruction"
  'fit ||((COLORS_BASE_PIN == 0) OR (COLORS_BASE_PIN == 9))
  fit ||((COLORS_BASE_PIN == 0) + (COLORS_BASE_PIN == 9))

  ' This version of the code assumes that the first color layer in the buffer
  ' is the least-significant color bit, so we make sure that the config file
  ' is set up correctly.
  fit ||((Config#DBUF_HI2LO_LAYER_DELTA & $80000000) <> 0)
  '
  ' For most-significant bit in the first layer, use:
  'fit ||((Config#DBUF_LO2HI_LAYER_DELTA & $80000000) <> 0)

  ' We could do a lot more pin assignment collision testing. Caveat Emptor.

'=========================================================================

PUB NoOp
{{
  Every Spin Object must have at least one Spin PUB function, and the first
  one may be run automatically, so we make sure the first one is "safe."
  This one does nothing.

  Arguments:
    None

  Return Value:
    None
}}
  ' Do nothing

'=========================================================================
PUB LoadDisplayCog(ColorDepth, DisplayBufAddrVar, RefreshClocksAddrHub)
{{
  Add the runtime-configured parameters to the hub image for the cog PASM code
  before loading the cog:
  - Color Depth (bumber of bits in the each color, i.e. the number of "layers"
  - Address of the hub variable that will eventually get the address of the
    display buffer, to allow delayed initialization of the cog, so the
    application can reclaim the hub copy of the cog code.
  - The address of the location in the hub to get the count of sysclks used for
    a single full display refresh, so the refresh rate can be reported.
  Then the cog is started, which will load the cog, and then the cog code will
  wait until the display buffer address is supplied.

  Arguments:
    ColorDepth -- the numner of bits in each color value, which is the number
                  of color "layers" in the display buffer. E.g. "24-bit" color,
                  which is 8:8:8, would have a depth of 8.
    DisplayBufAddrVar -- the hub addres of the variable that will eventually
                         contain the actual hub address of the display buffer.
                         This can start out as zero, which will cause the
                         cog initialization code to stall, waiting for a non-
                         zero address to be supplied. This is a "two-stage"
                         initialization for the cog.
    RefreshClocksAddrHub -- the hub address of the LONG variable that will
                            receive the number of sysclks needed for one full
                            refresh of the display, so that the refresh rate
                            can be calculated and reported.

  Return Value:
    The Cog ID of the display refresh cog, plus one, so that it is non-zero.
}}
  ' Put the final few values into the cog.
  pcFullRefreshClocksAddrHub := RefreshClocksAddrHub
  pcColorDepthBits           := ColorDepth

  ' Start the cog, and pass the address in Hub memory that will eventually get
  ' the hub address of the display buffer. The cog, after starting, waits until
  ' that location contains a non-zero hub address, and uses that address as the
  ' start of the display buffer.
  return 1 + COGNEW(@pPasmStart, DisplayBufAddrVar)

'=========================================================================
PUB EstimateRefreshClocks(depth) | tmp, clks
{{
  Based on the current implementation, and the dimensions of the configured
  display, estimate the number of clock cycles the one refresh would take.

  This is useful to determine the maximum color depth that would have an
  acceptably high (=> 60 Hz) refresh rate at the current clock frequency.

  Arguments:
    depth -- the numner of bits in each color value, which is the number
             of color "layers" in the display buffer. E.g. "24-bit" color,
             which is 8:8:8, would have a depth of 8.

  Return Value:
    The estimated number of clocks for one full refresh of the display.

  Notes:
    refresh_clocks = 
            CPP * cols * rows * (2**depth - 1)
          + REPO * rows * (2**depth - 1 - depth)
          + NRO  * (rows - 1) * depth
          + LLRO * (bits - 1)
          + FRO
    with
      CPP  ==  12      Clocks Per Pixel
      REPO ==  96      Overhead for a REPeated row display
      NRO  ==  96      Next Row Overhead for the last repeat of a row, but not
                       for the last row of a layer
      LLRO ==  96      Last Layer Row Overhead for the last repeat of the last
                       row of a layer.
      FRO  == 160      Full Refresh Overhead for the last repeat of the last
                       row of the last layer. This includes ?? clocks spent
                       measuring the actual refresh clocks.
}}
  '' This calculation is correct
  'return (12 * DISPLAY_COLUMNS * REFRESH_ROWS * ((|< depth) - 1)) {
  '  }  + (96 * REFRESH_ROWS * ((|< depth) - 1 - depth))          {
  '  }  + (96 * (REFRESH_ROWS - 1) * depth)                       {
  '  }  + (96 * (depth - 1))                                      {
  '  }  + 160

  'return (12 * DISPLAY_COLUMNS * REFRESH_ROWS * ((|< depth) - 1)) {
  '  }  + 96 * ( (REFRESH_ROWS * ((|< depth) - 1 - depth))          {
  '  }  +      + ((REFRESH_ROWS - 1) * depth)                       {
  '  }  +      + (depth - 1) + 1)                                   {
  '  }  + 64

  'return (12 * DISPLAY_COLUMNS * REFRESH_ROWS * ((|< depth) - 1)) {
  '  }  + 96 * (REFRESH_ROWS * ((|< depth) - 1))          {
  '  }  + 64

  'return ((12 * DISPLAY_COLUMNS + 96) * REFRESH_ROWS * ((|< depth) - 1)) {
  '  }  + 64

  ' This would be calculated quite differently in PASM to avoid the
  ' multiplies.
  ' Change the final "+ 64" to "+ 32" if the PASM timing code is removed from the
  ' refresh loop'
  tmp := ((((DISPLAY_COLUMNS << 1) + DISPLAY_COLUMNS) << 2) + 96) << ROW_NUMBER_BITS
  'return ((tmp << depth) - tmp) + 64
  return ((tmp << depth) - tmp) + 32


' ========================================================================
'
' PASM code for display refresh
'
' Style notes:
' I use special prefixes for code labels, variables and constants stored in
' the pasm code, so that names make clear what is being referred to.
'
' "pCodeLabel" -- A label on code (possibly with a leading ":")
' "pvVariable" -- A label on a variable stored in a register in the cog,
'                 that can change over the course of cog execution
' "pcConstant" -- A label on a value that is effectively constant for the
'                 lifetime of the cog -- it may be set at initialization time,
'                 but will not be changed later.
' "piInstruction" -- A label on an instruction that may be the target of
'                    runtime instruction modification.
'

DAT
                    org 0
pPasmStart
                    jmp     #pInitContinue

' We store all the constants and some operating variables up front, generally
' the ones needing compile-time initialization.

' pcBufAddrHub contains the hub address of the first row and column of the
' lowest-significance color bit of the display buffer.
' Set by deferred initialization after the cog starts.

pcBufAddrHub        long    0    ' Starting address of the entire buffer

' Hub addresses for the refresh clock count. Set by the initialization code
' that starts the cog. 
pcFullRefreshClocksAddrHub  long   0

' pcColorDepthBits is the number of bits per color. Set by the initialization
' code that starts the cog.
pcColorDepthBits    long     1

' pcRefreshRows is the number of rows used to refresh the display; may be 32, 16,
'               or even 8 for some displays.
' pcMaxRefreshRow is the upper limit for refresh rows -- sometimes useful.
'
' (These could be simple constants as they are unlikely to be > 511, but for
' now, leave them as cog registers. We can get rid of them if we need the
' cog space.

pcRefreshRows       long    REFRESH_ROWS
pcMaxRefreshRow     long    MAX_REFRESH_ROW

' pcDisplayColumns is the number of display columns per row; this is _not_ the
'                  number of longs per row, and might not even be the number of
'                  _bytes_ per row, depending on how the data is packed.
' pcMaxDisplayColumn is the upper limit for column numbers, sometimes useful.
'
' It is unlikely, but possible, for these to exceed 511 -- for example, 8
' panels in a series, each of which has 64 columns, would reach 512 columns.
' If we need cog space, we could get rid of them and limit the column count.

pcDisplayColumns    long    DISPLAY_COLUMNS
pcMaxDisplayColumn  long    MAX_DISPLAY_COLUMN
pcColumnCountLongs  long    DISPLAY_COLUMNS >> 2  ' Number of LONGs in a row
pcBytesPerColorLayer  long    DISPLAY_COLUMNS * REFRESH_ROWS

' Control pin and mask definitions.

pcColorsBasePin     long    Config#COLORS_BASE_PIN

pcShiftClockPin     long    SHIFT_CLOCK_PIN
pcShiftLatchPin     long    SHIFT_LATCH_PIN
pcOEInvertedPin     long    OE_INVERTED_PIN
pcShiftClockMask    long    SHIFT_CLOCK_MASK
pcShiftLatchMask    long    SHIFT_LATCH_MASK
pcOEInvertedMask    long    OE_INVERTED_MASK

' Instead of using a normal counter for the current display row number being
' updated, and shifting it into position to write it to the display, we
' manage the row number value in its already-shifted position, to save a few
' instructions in the row processing loop.

pcRowNumberMask     long    ROW_NUMBER_MASK
pcRowNumberIncr     long    ROW_NUMBER_INCREMENT

' pcDiraMask is the mask of output bits to set in DIRA when the cog starts.

pcDiraMask          long    COLORS_MASK | Config#ACTUAL_ROW_NUMBER_MASK | {
  }                         SHIFT_CLOCK_MASK | SHIFT_LATCH_MASK | {
  }                         OE_INVERTED_MASK

' pcOEInvertedShiftLatchMask is used during row number update to eliminate
' a runtime instruction. We can treat the shift latch and Output enabled
' as a single bit (for the ICN2037 controller chip, at least), because th
' two signals are inverted w.r.t. each other, and trigger on opposite edges. 

pcOEInvertedShiftLatchMask long OE_INVERTED_MASK | SHIFT_LATCH_MASK

' Simple constants
pcZero               long     0
pcOne                long     1
pcMinusThree         long    -3 ' simple constant
pc_2000_0000         long    $2000_0000
pc_4000_0000         long    $4000_0000
pc_8000_0000         long    $8000_0000
pc_C000_0000         long    $C000_0000

' Both counters in NCO mode on the same pin, so one counter can "gate"
' pulses from the other counter.
pcCtrbInit
pcCtraInit           long    %00100 << 26 | SHIFT_CLOCK_PIN

' A constant to load into PHSB to gate 4 shift-clocks out of CTRA.
pc_4_shiftClks       long    $8000_0000 - (4 * 8) ' allow 4 pulses

' The colors shift determines how much to shift a LONG read from the display
' buffer to get to the next pair of color bits. For four pixel pairs in one
' LONG (one pair per BYTE), the shift is 8. There may be some display buffer
' encodings in the future that put 5 pairs of pixels into the 32-bit LONG in
' the future.

pcColorsShift        long    8  ' amount to shift the long per color

pcStallCountDeltaInit long    8    ' starting value for pStallCountDelta for
                                   ' each full pass (all color bits)

' To stop the cog, the main program can also set the OEInverted pin for
' output, and can write a 1 to it to override anything this cog can do
' (output state is OR'd amoung all the cogs); then this cog can be
' stopped.

'-------------------------------------------------------------------------

' The code continues here...
pInitContinue
                    ' Wait for the app to pass the display buffer
                    ' address to us.
:pInitWaitLoop
                    rdlong  pcBufAddrHub, PAR  wz
    if_z            jmp     #:pInitWaitLoop

' Things to do during initialization:
' - init OEInverted in OUTA to disable the display when we change the
'   pin to output
' - Set DIRA for the output pins we want.
' - select movs/movd instructions for the display refresh loop

                    ' Set OEInverted so that the output starts disabled.
                    ' All other output bits are zero.
                    mov     OUTA, pcOEInvertedMask

                    ' Set up our output pins
                    mov     DIRA, pcDiraMask

                    mov     pvFullRefreshClocks, #0

                    ' Use either MOVS or MOVD, depending on the value
                    ' of Config#COLOR_BASE_PIN, which is either 0 (for MOVS)
                    ' or 9 (for MOVD)
                    '    movi   instruction, #%010101_001  for MOVD
                    '    movi   instruction, #%010100_001  for MOVS
                    ' When loaded, the instructions are all MOVS
                    ' We change them to MOVD if the colors base is pin 9.
                    cmp     pcColorsBasePin, #0 wz
    if_nz           movi    piMovs0, #%010101_001
    if_nz           movi    piMovs1, #%010101_001
    if_nz           movi    piMovs2, #%010101_001
    if_nz           movi    piMovs3, #%010101_001

                    ' Calculate the number clocks needed for one full refresh
                    mov     pv_t, pcDisplayColumns
                    shl     pv_t, #1
                    add     pv_t, pcDisplayColumns
                    shl     pv_t, #2  ' pv_t := pcDisplayColumns * 12
                    add     pv_t, #96
                    shl     pv_t, #ROW_NUMBER_BITS
                    mov     pvFullRefreshClocks, pv_t
                    shl     pvFullRefreshClocks, pcColorDepthBits
                    sub     pvFullRefreshClocks, pv_t
                    ' If the refresh timing code is removed from the display
                    ' loop, Change the +64 to +32, because the timing code adds
                    ' two additional hub access periods
                    add     pvFullRefreshClocks, #32
                    'add     pvFullRefreshClocks, #64
                    wrlong  pvFullRefreshClocks, pcFullRefreshClocksAddrHub

                    ' Prepare the counters for generating a train of four
                    ' ShiftClock pulses on-demand, aligned with the hub read
                    ' of a long for four columns.
                    '
                    ' CTRA:
                    ' - NCO Mode; increment of $2000_0000 for a period of 8
                    '   clocks (4 high, 4 low, no errors)
                    ' - Phase is set at the start of each row
                    ' CTRB:
                    ' - NCO Mode; increment of 1
                    ' - PHSB set to $8000_0000 initially; set to
                    '   $8000_0000 - 4*8 to allow 4 pulses (4 rising edges
                    '   in a burst.

                    mov     PHSB, pc_8000_0000
                    mov     FRQB, #1
                    mov     CTRB, pcCtrbInit

                    mov     PHSA, #1
                    mov     FRQA, pC_2000_0000
                    mov     CTRA, pcCtraInit

                    mov     pvCurrentRow, #0   ' The current row being written
                    mov     pvPrevRow, #0
                    'mov     pvRefreshPassStartCnt, CNT
                    neg     pvRefreshPassStartCnt, CNT

pRefreshPassStart
                    mov     pvHubFetchAddr, pcBufAddrHub
                    mov     pvStallCountDelta, pcStallCountDeltaInit

                    ' Loop through all the color bits;
                    ' For now, only one bit, so no looping needed.
                    mov     pvColorBitLayerLoopCnt, pcColorDepthBits

pColorBitLayerStart
                    ' Loop through all the rows, until pCurrentRow becomes
                    ' zero again:
                    ' - Loop through all the columns
                    '   - RDBYTE from pHubFetchAddr
                    '   - ++pHubFetchAddr
                    '   - movd  OUTA, the_byte
                    ' - Increment the pCurrentRow, and compute the XOR for
                    '   the row number change
                    ' - pOEInverted set high to disable output
                    ' - toggle pShiftLatch
                    ' - Change (xor) the row number in OUTA
                    ' - pOEInverted set low to enable output
                    ' if the new row number is not yet zero again, loop for the
                    ' next row.

pRowStart
                    ' We reinitialize PHSA to get the rising edges properly
                    ' aligned with the code and hub alignment -- it is already
                    ' configured for 8-clock per period (4 high, 4 low).
                    ' CTRB is already configured for NCO, and it is already
                    ' steady-high (disabling low-to-high transitions on the
                    ' ShiftClock pin), and we allow it to "gate" four periods
                    ' (about 32 clocks) for the four bytes.

                    ' We arrange the relative phase of the two counters, so that
                    ' the gating signal opens in the middle of the high phase of
                    ' the repeating clock, so the rising edge is 6 cycles after
                    ' the write cycle opening the gate.
                    ' This will put the rising edge of the ShiftClock two sysclks
                    ' after the write of the color bits.

                    rdlong  pvColors, pvHubFetchAddr  ' Get the first set of colors       ' Get the first long
                    ' Loop through all the columns
                    mov     pvColumnLoopCount, pcColumnCountLongs
                    nop     ' Stall for explicit hub alignment, before adjusting
                            ' the counter phases

                    mov     PHSA, pc_C000_0000                 ' shiftclk l2h
                    jmp     #pWrbyte0                          ' shiftclk h2l

pColumnLoop

                    rdlong  pvColors, pvHubFetchAddr           ' shiftclk l2h, h2l
pWrbyte0
                    mov     PHSB, pc_4_shiftClks               ' shiftclk l2h
piMovs0             movs    OUTA, pvColors           ' byte 0  ' shiftclk h2l
                    shr     pvColors, pcColorsShift            ' shiftclk L2H  ' hub access
piMovs1             movs    OUTA, pvColors           ' byte 1  ' shiftclk H2L
                    shr     pvColors, pcColorsShift            ' shiftclk L2H
piMovs2             movs    OUTA, pvColors           ' byte 2  ' shiftclk H2L
                    shr     pvColors, pcColorsShift            ' shiftclk L2H  ' hub access
piMovs3             movs    OUTA, pvColors           ' byte 3  ' shiftclk H2L
                    add     pvHubFetchAddr, #4                 ' shiftclk L2H
                    djnz    pvColumnLoopCount, #pColumnLoop   ' shiftclk h2l
                                                                               ' hub access
                    ' Latch the row data, set the row number, enable output,
                    ' and advance the row number
                    ' We XOR-change the row number bits in OUTA to get an
                    ' "atomic" update of the row number without (briefly)
                    ' setting the row number to zero.

                    ' This version triggers the ShiftLatch and disables output
                    ' in the same instruction that the row bits are changed,
                    ' which should not cause any ghosting. The second latch,
                    ' associated with the Output Enable pin, will not collect
                    ' the row settings until OEInverted is set to zero in the
                    ' next instruction.
                    ' (I suppose it is possible that OEInverted is slower to
                    ' propagate than the row number bits, but that should only
                    ' be a very few ns (or less) of wrong-row display.)
                    xor     pvPrevRow, pvCurrentRow
                    or      OUTA, pcOEInvertedShiftLatchMask
                    xor     OUTA, pvPrevRow
                    andn    OUTA, pcOEInvertedShiftLatchMask           ' hub access
                    mov     pvPrevRow, pvCurrentRow
                    add     pvCurrentRow, pcRowNumberIncr
                    ' Moved down after the row CNT calculations, so we don't
                    ' need to preserve the Z flag for a long time.
                    ' and     pvCurrentRow, pcRowNumberMask  wz

                    ' Stall here; higher-order color bits have longer stalls,
                    ' to increase the brightness.
                    ' Be sure to leave the Z bit untouched (or reorder the
                    ' instructions).
                    mov      pv_t, pvStallCountDelta
                    add      pv_t, cnt                                 ' hub access
                    waitcnt  pv_t, #0

                    ' Calculate the number of clocks used for this row.
                    ' This value will be added to the pStallCountDelta
                    ' when display of the next significant color bit starts.
                    mov      pvRowDurationCnt, pvRowStartCnt
                    mov      pvRowStartCnt, CNT
                    sub      pvRowDurationCnt, pvRowStartCnt
                    neg      pvRowDurationCnt, pvRowDurationCnt

                    and     pvCurrentRow, pcRowNumberMask  wz
                    ' Z is set if the pCurrentRow has returned to zero, meaning
                    ' we are done one pass over the rows (for this color bit).
    if_nz           jmp     #pRowStart

                    ' All the rows for this color bit have been shown.
                    ' After each color bit, we adjust the stall time, to
                    ' the next higher-order color bit show for twice as long
                    ' as the previous one, by adding the time for one row of
                    ' the previous bit to the stall.
                    ' Update the stall duration, and start the next color bit
                    ' layer.
                    add     pvStallCountDelta, pvRowDurationCnt
                    djnz    pvColorBitLayerLoopCnt, #pColorBitLayerStart

'                    ' After all color bits have been done, we should report
'                    ' (by writing to the hub) the row, color-bit, and refresh
'                    ' timings.
'                    'mov     pv_t, pvRefreshPassStartCnt
'                    'mov     pvRefreshPassStartCnt, CNT
'                    'sub     pv_t, pvRefreshPassStartCnt
'                    'neg     pvFullRefreshClocks, pv_t
'                    cmp     pvFullRefreshClocks, #0  wz  ' see if first time
'                    mov     pvFullRefreshClocks, pvRefreshPassStartCnt
'                    neg     pvRefreshPassStartCnt, CNT
'                    sub     pvFullRefreshClocks, pvRefreshPassStartCnt
'                    ' Copy back to hub
''    if_nz           wrlong  pvFullRefreshClocks, pcFullRefreshClocksAddrHub
'                    nop
'                    nop
                    jmp     #pRefreshPassStart

'-------------------------------------------------------------------------
' Runtime-initialized variables.

' Stores the number of system clocks needed for one full refresh of all color
' layers for the display. This is copied back into hub memory so that the main
' program can report the performance of the driver, in refreshes per second.
pvFullRefreshClocks  res     1


pvStallCountDelta    res     1    ' amount to stall for each row.
pv_t                 res     1    ' temporary variable
pvRefreshPassStartCnt res     1

' These are variables to manage the refresh operation: loop counters,
' hub address pointers, etc. (Could be RES at the end...)
pvHubFetchAddr       res     1  ' The next BYTE to read
pvCurrentRow         res     1  ' The row currently being written.
pvPrevRow            res     1  ' The previous row number, for calculating
                                ' the XOR for the row.
pvColumnLoopCount    res     1  ' The column loop control variable
pvColors             res     1  ' Gets the current row/col color bits

' Storage for measuring the row duration in clocks, so that higher color
' bits can use a longer delay.
pvRowDurationCnt     res     1  ' The sysclk count for one row of the current
                                ' color bit
pvRowStartCnt        res     1  ' The starting CNT value for the row.

' Loop variable for the color bit layers.
pvColorBitLayerLoopCnt res     1

'====================================================================

PasmEnd             fit  $1F0

