{{
HUB75 write helper obect, to speed up writing to the display.

Author:         Tom Dinger (Ding-Batty)
Release Date:   2024/02/26

Copyright 2024 Thomas E. Dinger

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

--------------------------------------------------------------------------

This file contains write routines for updating the display buffer for one
or more hub75-based display panels. It uses a PASM code implementation
(in a separate file: hub75_helper_pasm.spin) to perform the low-level updates,
and uses a command FiFo circular buffer to send the write commands to the
PASM cog.

The write routines in this file share the same API (PUB names and arguments)
as the pure Spin ones in hub75_display.spin (with one exception), to make
substituting one for the other simpler.
}}

OBJ
  ' Pick up the static configuration.
  Config: "hub75_config.spin"
  Hconfig: "hub75_helper_config.spin"

CON
  ' Import the needed configuration settings from the Config object, for
  ' code brevity.
  REFRESH_ROWS         = Config#REFRESH_ROWS
  MAX_REFRESH_ROW      = Config#MAX_REFRESH_ROW
  DISPLAY_ROWS         = Config#DISPLAY_ROWS
  MAX_DISPLAY_ROW      = Config#MAX_DISPLAY_ROW
  DISPLAY_COLUMNS      = Config#DISPLAY_COLUMNS
  MAX_DISPLAY_COLUMN   = Config#MAX_DISPLAY_COLUMN
  SHIFT_CLOCK_PIN      = Config#SHIFT_CLOCK_PIN
  SHIFT_LATCH_PIN      = Config#SHIFT_LATCH_PIN
  OE_INVERTED_PIN      = Config#OE_INVERTED_PIN
  SHIFT_CLOCK_MASK     = Config#SHIFT_CLOCK_MASK
  SHIFT_LATCH_MASK     = Config#SHIFT_LATCH_MASK
  OE_INVERTED_MASK     = Config#OE_INVERTED_MASK
  ROW_NUMBER_MASK      = Config#ROW_NUMBER_MASK
  ROW_NUMBER_INCREMENT = Config#ROW_NUMBER_INCREMENT

  BYTES_PER_COLOR_LAYER = REFRESH_ROWS * DISPLAY_COLUMNS

  RED_BIT      = Config#RED_BIT
  GREEN_BIT    = Config#GREEN_BIT
  BLUE_BIT     = Config#BLUE_BIT

  RED_BIT_SHIFT   = Config#RED_BIT_SHIFT
  GREEN_BIT_SHIFT = Config#GREEN_BIT_SHIFT
  BLUE_BIT_SHIFT  = Config#BLUE_BIT_SHIFT

OBJ
  Fifo : "FifoBig.spin"

VAR
  LONG    HelperCog             ' 1 + CogId for the write helper cog
  LONG    ColorDepthBits        ' Set at initialization time.
  LONG    BufAddrHub            ' the starting address of the display buffer.



DAT
' The main program sets this to zero to allow the cog to proceed to the next
' pin test and reset it to zero.
'DebugFlag           long    1

Zero                long    0   ' used for line drawing.

' The following can be used for debug output -- pass the address as the
' result pointer, and the first entry will be the "done" result, and the
' following entries will be the debug output from internal processing, if
' that is included for the operation.
DebugResult         long    0 [16]

' CmdFifoHeader is the FiFo buffer description block, that points to the
' actual storage for the FiFo (which could be allocated separately).
' The layout is defined in FifoBig_header.spin                                     

'CmdFifoHeader  WORD 0 [Fifo#HEADER_SIZE_WORDS]
CmdFifoHeader
CmdFifoHeader_insert_hubadr   word    0
CmdFifoHeader_remove_hubadr   word    0
CmdFifoHeader_start_hubadr    word    0
CmdFifoHeader_limit_hubadr    word    0
CmdFifoHeader_entry_size      word    0

' ========================================================================
PUB Stop
{{
  Stops the PASM cog, if it is running.

  Arguments:
    None

  Return Value:
    None
}}
  if HelperCog
    COGSTOP(HelperCog~ - 1)

' ========================================================================
PUB GetCmdFifoHeader
{{
  Returns the address of the FiFo Header, describing the insert and remove
  locations for the FiFo, and the entry size.

  Arguments:
    None

  Return Value:
    The address of the header in hub memory.
}}
  return @CmdFifoHeader

' ========================================================================
PUB GetDebugResult
{{
  Returns the address of the debug buffer. If this is enabled for a particular
  command (by setting the RESULT_PTR to @DebugResult in the command),
  this function will return the address of the 16-LONG buffer containing
  the result code and any other debugging information included for that
  command.

  Arguments:
    None

  Return Value:
    The address of the debug output buffer.
}}
  return @DebugResult

' ========================================================================
PUB FinalHelperInit(CogPlusOne, BufAdr, ColorDepth, FifoHubAdr, FifoSizeBytes)
{{
  Complete the initialization for the helper cog:
  - Record the cog ID, display buffer address, and color depth
  - Initialize the command FiFo header to be ready to accept commands.

  This needs to be completed after the write helper cog has been started
  (in hub75_helper_pasm.spin), but before that helper cog (and the display
  refresh cog) are told the address of the display buffer, which unblocks
  those PASM cogs' operations.

  Arguments:
    CogPlusOne -- 1 + CogID of the write helper PASM cog
    BufAdr -- the display buffer address.
    ColorDepth -- the number of bits for each color ("layers" for the display
                  buffer). In the range 1..8, not checked.
    FifoHubAdr -- the address of the actual FiFo that stores the command
                  entries as a ring buffer. This address is used to initialize
                  the command FiFo header.
    FifoSizeBytes -- the size of the actual FiFo, used to initialize the
                     command FiFo header. This is not checked, but it should
                     be larger than CMDBUF_SIZE_BYTES * 2, because one FiFo
                     entry is "wasted" for the buffer-full condition. 

  Return Value:
    None.
}}
  HelperCog := CogPlusOne
  ColorDepthBits := ColorDepth   ' Needed?
  BufAddrHub := BufAdr
  InitCmdFifo(FifoHubAdr, FifoSizeBytes)

' ========================================================================
PUB InitCmdFifo(FifoHubAdr, SizeInBytes)
{{
  Initialize the command FiFo header.

  Arguments:
    FifoHubAdr -- the address of the FiFo buffer.
    SizeInBytes -- the size of the FiFo buffer. This is not checked, but it
                   should be larger than CMDBUF_SIZE_BYTES * 2, because one
                   FiFo entry is "wasted" for the buffer-full condition.  

  Return Value:
    None.
}}
  Fifo.Init(@CmdFifoHeader, FifoHubAdr, HConfig#CMDBUF_SIZE_BYTES, {
    }       SizeInBytes / HConfig#CMDBUF_SIZE_BYTES)

' ========================================================================
' Display Buffer Write operations
'
' All these functions (with a few exceptions) correspond to the
' same functions in hub75_display.spin, taking the same arguments.
'
' Exceptions:
' - Bitmap1Bpp takes one additional argument, skip_pat, not supported by
'   the Spin version.
' - DrawPixelWithBitmap is a debugging function, and not really practical
'   because it will have much more overhead than the Pixel() function.
' - NoOpAndWait issues a No-Op command, and waits for it to complete. This
'   is a synchronization helper: it can be used after a series of commands
'   have been issued/enqueued, to block until all of them have completed.
'
' ---------------------------------------
PUB Pixel(column, row, color) | entry, l
{{
  Enqueue the command to set a single pixel to a particular color.
  This blocks waiting for a command slot to be available in the FiFo, but
  does not block waiting for the command to complete.

  Arguments:
    col -- the column number, zero-based. The leftmost column is column 0.
    row -- the row number, zero-based, The topmost row is row 0.
    color -- the color to set the pixel. It may either be the RGB color or
             the display color

  Return Value:
    Returns the address in the FiFo buffer of the command entry just added.
    This can be useful for debugging.
}}
  entry := Fifo.PutAddress(@CmdFifoHeader)
  word[entry + HConfig#PIXELCMD_CMD       ] := HConfig#CMD_PIXEL
  word[entry + HConfig#PIXELCMD_RESULT_PTR] := 0  ' @DebugResult
  word[entry + HConfig#PIXELCMD_COLUMN    ] := column
  word[entry + HConfig#PIXELCMD_ROW       ] := row
  long[entry + HConfig#PIXELCMD_COLOR     ] := color
  Fifo.FinishPut(@CmdFifoHeader)

  ' If debugging -- wait for the operation to complete
  ' repeat while DebugResult == 0
  '   ' Stall
  return entry

' ---------------------------------------
PUB DrawPixelWithBitmap(col, row, color) | entry ' , done, l
{{
  Enqueu a command to set a single pixel to a particular color. This version
  uses the 1bpp bitmap operation as a test, which has huge overhead for a
  single pixel.
  This blocks waiting for a command slot to be available in the FiFo, but
  does not block waiting for the command to complete.

  Arguments:
    col -- the column number, zero-based. The leftmost column is column 0.
    row -- the row number, zero-based, The topmost row is row 0.
    color -- the color to set the pixel. It may either be the RGB color or
             the display color

  Return Value:
    Returns the address in the FiFo buffer of the command entry just added.
    This can be useful for debugging.
}}
  ' Draw a 1 x 1 pixel bitmap (that is always zero), so it will get the background
  ' color.
  ' This is only useful for testing bitmap output.

  'DebugResult := 0
  entry := Fifo.PutAddress(@CmdFifoHeader)
  word[entry + HConfig#BITMAPCMD_CMD           ] := HConfig#CMD_BITMAP_1BPP
  word[entry + HConfig#BITMAPCMD_RESULT_PTR    ] := 0 ' @DebugResult
  word[entry + HConfig#BITMAPCMD_COLUMN        ] := col
  word[entry + HConfig#BITMAPCMD_ROW           ] := row
  word[entry + HConfig#BITMAPCMD_WIDTH         ] := 1
  word[entry + HConfig#BITMAPCMD_HEIGHT        ] := 1
  word[entry + HConfig#BITMAPCMD_LTRIM         ] := 0
  word[entry + HConfig#BITMAPCMD_TTRIM         ] := 0
  word[entry + HConfig#BITMAPCMD_BITMAP_HUBADDR] := @Zero
  word[entry + HConfig#BITMAPCMD_ROWADDR_INCR  ] := 0
  long[entry + HConfig#BITMAPCMD_BGCOLOR       ] := color
  long[entry + HConfig#BITMAPCMD_FGCOLOR       ] := 0
  long[entry + HConfig#BITMAPCMD_SKIP_PAT      ] := 0
  Fifo.FinishPut(@CmdFifoHeader)

  ' If debugging -- wait for the operation to complete
  ' repeat while DebugResult == 0
  '   ' Stall
  return entry

' ---------------------------------------
PUB VerticalLine(column, row, len, color) | entry ' , done, l
{{
  Enqueue a command to draw a vertical line.
  This blocks waiting for a command slot to be available in the FiFo, but
  does not block waiting for the command to complete.

  The line is cropped to the visible display, if it extends off the top or
  bottom of the display.

  Arguments:
    column -- the column number for the vertical line, zero-based. The
              leftmost column is column 0. It is cropped to the range
              0..MAX_DISPLAY_COLUMN.
    row -- the row number for the top of the vertical line, zero-based, The
           topmost display row is row 0. It is cropped to the range
           0..MAX_DISPLAY_ROW
    len -- the length of the line in pixels.
    color -- the color to set the pixel. It may either be the RGB color or
             the display color

  Return Value:
    Returns the address in the FiFo buffer of the command entry just added.
    This can be useful for debugging.
}}
  ' Draw a 1 x len pixel bitmap (that is always zero), so it will get the background
  ' color. This draws a vertical line.

  'DebugResult := 0
  entry := Fifo.PutAddress(@CmdFifoHeader)
  word[entry + HConfig#BITMAPCMD_CMD           ] := HConfig#CMD_BITMAP_1BPP
  word[entry + HConfig#BITMAPCMD_RESULT_PTR    ] := 0 ' @DebugResult
  word[entry + HConfig#BITMAPCMD_COLUMN        ] := column
  word[entry + HConfig#BITMAPCMD_ROW           ] := row
  word[entry + HConfig#BITMAPCMD_WIDTH         ] := 1
  word[entry + HConfig#BITMAPCMD_HEIGHT        ] := len
  word[entry + HConfig#BITMAPCMD_LTRIM         ] := 0
  word[entry + HConfig#BITMAPCMD_TTRIM         ] := 0
  word[entry + HConfig#BITMAPCMD_BITMAP_HUBADDR] := @Zero
  word[entry + HConfig#BITMAPCMD_ROWADDR_INCR  ] := 0  ' stay on Zero for all rows
  long[entry + HConfig#BITMAPCMD_BGCOLOR       ] := color
  long[entry + HConfig#BITMAPCMD_FGCOLOR       ] := 0
  long[entry + HConfig#BITMAPCMD_SKIP_PAT      ] := 0
  Fifo.FinishPut(@CmdFifoHeader)

  ' If debugging -- wait for the operation to complete
  ' repeat while DebugResult == 0
  '   ' Stall
  return entry
  
' ---------------------------------------
PUB HorizontalLine(column, row, len, color) | entry ' , done, l
{{
  Enqueue a command to draw a horizontal line.
  This blocks waiting for a command slot to be available in the FiFo, but
  does not block waiting for the command to complete.

  The line is cropped to the visible display, if it extends off the left or
  right edge of the display.

  Arguments:
    column -- the column number for the left end of the horizontal line,
              zero-based. The leftmost column is column 0. It is cropped to
              0..MAX_DISPLAY_COLUMN.
    row -- the row number, zero-based, The topmost row is row 0. It is
           cropped to 0..MAX_DISPLAY_ROW
    len -- the length of the line in pixels.
    color -- the color to set the pixel. It may either be the RGB color or
             the display color

  Return Value:
    Returns the address in the FiFo buffer of the command entry just added.
    This can be useful for debugging.
}}
  ' Draw a 1 x len pixel bitmap (that is always zero), so it will get the background
  ' color. This draws a horizontal line in the background color.

  'DebugResult := 0
  entry := Fifo.PutAddress(@CmdFifoHeader)
  word[entry + HConfig#BITMAPCMD_CMD           ] := HConfig#CMD_BITMAP_1BPP
  word[entry + HConfig#BITMAPCMD_RESULT_PTR    ] := 0 ' @done
  word[entry + HConfig#BITMAPCMD_COLUMN        ] := column
  word[entry + HConfig#BITMAPCMD_ROW           ] := row
  word[entry + HConfig#BITMAPCMD_WIDTH         ] := len
  word[entry + HConfig#BITMAPCMD_HEIGHT        ] := 1
  word[entry + HConfig#BITMAPCMD_LTRIM         ] := 0
  word[entry + HConfig#BITMAPCMD_TTRIM         ] := 0
  word[entry + HConfig#BITMAPCMD_BITMAP_HUBADDR] := @Zero
  word[entry + HConfig#BITMAPCMD_ROWADDR_INCR  ] := 0
  long[entry + HConfig#BITMAPCMD_BGCOLOR       ] := color
  long[entry + HConfig#BITMAPCMD_FGCOLOR       ] := 0
  long[entry + HConfig#BITMAPCMD_SKIP_PAT      ] := 0
  Fifo.FinishPut(@CmdFifoHeader)

  ' If debugging -- wait for the operation to complete
  ' repeat while DebugResult == 0
  '   ' Stall
  return entry

' ---------------------------------------
PUB Box(column, row, width, height, color) | entry ' , done, l
{{
  Enqueue a command to draw a box of a uniform color.
  This blocks waiting for a command slot to be available in the FiFo, but
  does not block waiting for the command to complete.

  Arguments:
    column -- the column number for upper-left corner, zero-based. The
              leftmost column is column 0. It is cropped to
              0..MAX_DISPLAY_COLUMN.
    row -- the row number for the upper-left corner, zero-based, The
           topmost display row is row 0. It is cropped to
           0..MAX_DISPLAY_ROW
    width -- the width, in columns, of the box.
    height -- the height, in rows, of the box.
    color -- the color for the box. It may either be the RGB color or
             the display color

  Return Value:
    Returns the address in the FiFo buffer of the command entry just added.
    This can be useful for debugging.
}}
  ' Draw a width x height pixel bitmap (that is always zero), so it will get the
  ' background color. This draws a box.

  'DebugResult := 0
  entry := Fifo.PutAddress(@CmdFifoHeader)
  word[entry + HConfig#BITMAPCMD_CMD           ] := HConfig#CMD_BITMAP_1BPP
  word[entry + HConfig#BITMAPCMD_RESULT_PTR    ] := 0 ' @done
  word[entry + HConfig#BITMAPCMD_COLUMN        ] := column
  word[entry + HConfig#BITMAPCMD_ROW           ] := row
  word[entry + HConfig#BITMAPCMD_WIDTH         ] := width
  word[entry + HConfig#BITMAPCMD_HEIGHT        ] := height
  word[entry + HConfig#BITMAPCMD_LTRIM         ] := 0
  word[entry + HConfig#BITMAPCMD_TTRIM         ] := 0
  word[entry + HConfig#BITMAPCMD_BITMAP_HUBADDR] := @Zero
  word[entry + HConfig#BITMAPCMD_ROWADDR_INCR  ] := 0
  long[entry + HConfig#BITMAPCMD_BGCOLOR       ] := color
  long[entry + HConfig#BITMAPCMD_FGCOLOR       ] := 0
  long[entry + HConfig#BITMAPCMD_SKIP_PAT      ] := 0
  Fifo.FinishPut(@CmdFifoHeader)

  ' If debugging -- wait for the operation to complete
  ' repeat while DebugResult == 0
  '   ' Stall
  return entry
  
' ---------------------------------------
PUB Bitmap1Bpp(bitmap_ptr, bitmap_ptr_row_incr, col, row, width, height, {
      }            ltrim, ttrim, bgcolor, fgcolor, skip_pat) | entry, done
{{
  Enqueue a command to draw a 1 bit-per-pixel (1bpp) rectangular bitmap in an
  area of the display.
  This blocks waiting for a command slot to be available in the FiFo, but
  does not block waiting for the command to complete.

  Only a portion of the bitmap may be displayed: a number of rows and/or
  columns of the butmap may be skipped over using "ltrim" and "ttrim"

  The character is clipped to the display, so it may appear to "hang off" of
  the edge(s) of the display.

  This only handles bitmaps up to 32 columns (bits) wide.
  A bitmap look like a vector of values (BYTE, WORD, OR LONG). Each value is
  the bit pattern for the pixels of one row of the bitmap, topmost row first.
  The bits in each row value have the leftmost pixels in the loworder bits.
  The code can handle BYTE, WORD or LONG storage for a bitmap by examining the
  alignment of the bitmap_ptr, and uses the "bitmap_ptr_row_incr" to advance
  the correct distance to the next bitmap row value.

  If the width is larger than the number of bits in the value (BYTE, WORD, or
  LONG), the additional width is filled with zeros, i.e. the background color.

  Arguments:
    bitmap_ptr -- the (hub) address of the top row of the bitmap. This may be
                  BYTE, WORD or LONG aligned; if for example the pointer is for
                  an odd address it will be considered a BYTE-aligned bitmap
                  and only 8 bitmap bits will be available in that row.
    bitmap_ptr_row_incr -- the distance to advance the "bitmap_ptr" value to
                           move to the next row; typically the values 1, 2, or
                           4, but may be larger.
    col -- the column number for upper-left corner, zero-based. The leftmost
           display column is column 0. It is clipped to the actual columns of
           the display, so a bitmap can appear to "hang off" of the edges
           of the display.
    row -- the row number for the upper-left corner, zero-based, The
           topmost display row is row 0. It is clipped to the actual rows of
           the display so a bitmap can appear to "hang off" of the edges
           of the display.
    width -- the width, in columns, of the display area for the bitmap.
    height -- the height, in rows, of the display area for the bitmap.
    ltrim -- the number of columns on the left of the bitmap to skip over.
    ttrim -- the number of rows on the top of the bitmap to skip over
    bgcolor -- the color to display for empty (zero) pixels.
    fgcolor -- the color to display for non-empty (one) pixels.
    skip_pat -- the patterns of columns to "skip over":
                For full display of the bitmap, this should be zero,
                if non-zero, 1 bits in the value correspond to columns
                of the bitmap to skip over, with the low-order bit being
                the leftmost visible column.  

  Return Value:
    Returns the address in the FiFo buffer of the command entry just added.
    This can be useful for debugging.
}}
  'DebugResult := 0
  entry := Fifo.PutAddress(@CmdFifoHeader)
  word[entry + HConfig#BITMAPCMD_CMD           ] := HConfig#CMD_BITMAP_1BPP
  word[entry + HConfig#BITMAPCMD_RESULT_PTR    ] := 0  ' @done
  word[entry + HConfig#BITMAPCMD_COLUMN        ] := col
  word[entry + HConfig#BITMAPCMD_ROW           ] := row
  word[entry + HConfig#BITMAPCMD_WIDTH         ] := width
  word[entry + HConfig#BITMAPCMD_HEIGHT        ] := height
  word[entry + HConfig#BITMAPCMD_LTRIM         ] := ltrim
  word[entry + HConfig#BITMAPCMD_TTRIM         ] := ttrim
  word[entry + HConfig#BITMAPCMD_BITMAP_HUBADDR] := bitmap_ptr
  word[entry + HConfig#BITMAPCMD_ROWADDR_INCR  ] := bitmap_ptr_row_incr
  long[entry + HConfig#BITMAPCMD_BGCOLOR       ] := bgcolor
  long[entry + HConfig#BITMAPCMD_FGCOLOR       ] := fgcolor
  long[entry + HConfig#BITMAPCMD_SKIP_PAT      ] := skip_pat
  Fifo.FinishPut(@CmdFifoHeader)

  ' If debugging -- wait for the operation to complete
  ' repeat while DebugResult == 0
  '   ' Stall
  return entry

' ---------------------------------------
PUB RomCharacter(chr, col, row, width, height, ltrim, ttrim, bgcolor, fgcolor) {
      }         | entry, done
{{
  Enqueue a command to draw a ROM Font character.
  This blocks waiting for a command slot to be available in the FiFo, but
  does not block waiting for the command to complete.

  The character is clipped to the display, so it may appear to "hang off" of
  the edge(s) of the display.

  It may also be clipped when displayed, by using ltrim and rtrim. The
  character bitmap is "trimmed" before it is placed, so the "row" and "col"
  position will show the upper-left _untrimmed_ pixel of the character.

  "Trimming" (ltrim and ttrim) is a way to "skip over" the blank rows and
  columns around the edges of many of the characters.

  Arguments:
    ch -- the Character code (ASCII) to draw.
    col -- the column number for upper-left corner, zero-based. The leftmost
           display column is column 0. It is clipped to the actual columns of
           the display, so a character can appear to "hang off" of the edges
           of the display.
    row -- the row number for the upper-left corner, zero-based, The
           topmost display row is row 0. It is clipped to the actual rows of
           the display so a character can appear to "hang off" of the edges
           of the display.
    width -- the width, in columns, of the display area for the character.
    height -- the height, in rows, of the display area for the character.
    ltrim -- the number of columns on the left of the character bitmap to
             skip over.
    ttrim -- the number of rows on the top of the character bitmap to skip over
    bgcolor -- the color to display for empty (zero) character pixels.
    fgcolor -- the color to display for non-empty (one) character pixels.  

  Return Value:
    Returns the address in the FiFo buffer of the command entry just added.
    This can be useful for debugging.
}}
  'DebugResult := 0
  entry := Fifo.PutAddress(@CmdFifoHeader)
  word[entry + HConfig#ROMFONTCMD_CMD           ] := HConfig#CMD_ROM_FONT
  word[entry + HConfig#ROMFONTCMD_RESULT_PTR    ] := 0  ' @done
  word[entry + HConfig#ROMFONTCMD_COLUMN        ] := col
  word[entry + HConfig#ROMFONTCMD_ROW           ] := row
  word[entry + HConfig#ROMFONTCMD_WIDTH         ] := width
  word[entry + HConfig#ROMFONTCMD_HEIGHT        ] := height
  word[entry + HConfig#ROMFONTCMD_LTRIM         ] := ltrim
  word[entry + HConfig#ROMFONTCMD_TTRIM         ] := ttrim
  word[entry + HConfig#ROMFONTCMD_ROM_CHAR      ] := chr
  long[entry + HConfig#ROMFONTCMD_BGCOLOR       ] := bgcolor
  long[entry + HConfig#ROMFONTCMD_FGCOLOR       ] := fgcolor
  long[entry + HConfig#ROMFONTCMD_SKIP_PAT      ] := 0
  Fifo.FinishPut(@CmdFifoHeader)

  ' If debugging -- wait for the operation to complete
  ' repeat while DebugResult == 0
  '   ' Stall
  return entry

' ---------------------------------------
PUB NoOpAndWait | entry, done
{{
  Issue a No-Op operation, and wait for it to complete.

  This is a simple synchronization operation: the caller can send (enqueue)
  a series of write commands in the FiFo, then call this function which will
  send a NoOp command and block until it completes.

  Arguments:
    None.

  Return Value:
    None.
}}
  done := 0
  entry := Fifo.PutAddress(@CmdFifoHeader)
  word[entry + HConfig#NOOPCMD_CMD       ] := HConfig#CMD_NOOP
  word[entry + HConfig#NOOPCMD_RESULT_PTR] := @done
  Fifo.FinishPut(@CmdFifoHeader)
  repeat while done == 0
    ' Stall
  return entry

