{{
HUB75 panel write-helper obect, PASM code.

Author:         Tom Dinger (Ding-Batty)
Release Date:   2024/02/26

Copyright 2024 Thomas E. Dinger

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

--------------------------------------------------------------------------

This file contains the PASM code used to speed up writing to the display
buffer for the hub75 display. It understands the details of the memory
layout, as described in other files for the display driver.

Communication between the Spin wrapper functions and this PASM cog is done
with a command FiFo buffer, essentially a command mailbox circular buffer,
which allows the calling code to queue up a number of operations at once
without stalling.

Configuration and Initialization
================================

Most of this object is initialized "statically" from the configuration
constants. But several settings are received at runtime:
- The color bit depth, which is from 1 to 8 (I might be able to raise the upper
  limit for smaller panels).
- The base address of the display buffer.
- The initialized FiFo descriptor.

The color bit depth is not a property of the hardware, but configured by the
application in the Init call and patched into the PASM cog before it is started.
So, the color bit depth cannot change for the run of the cog.

The base address of the display buffer is passed to the cog code indirectly
through the PAR register only after the cog has been started. The PAR register
points to a hub LONG that will contain the starting address of the display
buffer; the cog wait for it to become non-zero, and then uses that display
buffer address.

The Descriptor for the command FiFo is patched into the cog code before it is
launched, but the cog does not actually use it until it has started up and
gets the display buffer base address, so the memory pointers in the FiFo
descriptor are filled in after the cog is launched, but before it it allowed
to complete initialization after starting.

The cog initialization delay permits the application to load the COG, and then
lay out the memory buffers including the display buffer and command FiFo, and
the app is able to include the HUB memory contains the cog PASM code during
the layout decisions.

Revisions:
==========
- 2024/04/10 -- Change from 8-bit color max, to 10-bit color max.
- 2024/02/26 -- Original release

}}

OBJ
  ' Pick up the static configuration for the display -- mostly for the
  ' row and column counts.
  Config: "hub75_config.spin"

  ' Get the config settings for the command FiFo
  HConfig: "hub75_helper_config.spin"

  ' We only need the Fifo header offset constants for the PASM code.
  Fifo: "FifoBig_Header"

CON
  ' Import the needed configuration settings from the display Config object,
  ' for code brevity.

  DISPLAY_ROWS         = Config#DISPLAY_ROWS
  MAX_DISPLAY_ROW      = Config#MAX_DISPLAY_ROW
  REFRESH_ROWS         = Config#REFRESH_ROWS
  MAX_REFRESH_ROW      = Config#MAX_REFRESH_ROW
  DISPLAY_COLUMNS      = Config#DISPLAY_COLUMNS
  MAX_DISPLAY_COLUMN   = Config#MAX_DISPLAY_COLUMN
  SHIFT_CLOCK_PIN      = Config#SHIFT_CLOCK_PIN
  SHIFT_LATCH_PIN      = Config#SHIFT_LATCH_PIN
  OE_INVERTED_PIN      = Config#OE_INVERTED_PIN
  SHIFT_CLOCK_MASK     = Config#SHIFT_CLOCK_MASK
  SHIFT_LATCH_MASK     = Config#SHIFT_LATCH_MASK
  OE_INVERTED_MASK     = Config#OE_INVERTED_MASK
  ROW_NUMBER_MASK      = Config#ROW_NUMBER_MASK
  ROW_NUMBER_INCREMENT = Config#ROW_NUMBER_INCREMENT
  COLORS_MASK          = Config#COLORS_MASK
  DBUF_LO2HI_LAYER_DELTA = Config#DBUF_LO2HI_LAYER_DELTA
  DBUF_HI2LO_LAYER_DELTA = Config#DBUF_HI2LO_LAYER_DELTA

  RED_BIT              = Config#RED_BIT
  GREEN_BIT            = Config#GREEN_BIT
  BLUE_BIT             = Config#BLUE_BIT

  '' Debugging help -- LEDs on the Board of Education
  'DEBUG_LED_PIN1 = 26
  'DEBUG_LED_PIN2 = 27


'=========================================================================
' Validate the various configuration settings:
' - Pins are valid, and do not conflict with each other
' - display configuration "makes sense"
'
' The validation is done using FIT statements to make logical assertions
' about the config values.

DAT
' Now, use a "trick" to get the Spin compiler to check that the values for
' various settings, because the code depends on the settings having specific
' perperties, e.g. being a power-of-two.

  ' If the condition is false (==0) then the "org 1" means that the
  ' fit statement will generate an error. So we set the conditionals up
  ' so that a unsupported configurations generate a false (==0) value.

  org 1

  ' The DISPLAY_COLUMNS must be a multiple of four, because we use four
  ' pixels per LONG of hub memory.
  ' Error means DISPLAY_COLUMNS is not a multiple of 4
  fit ||((DISPLAY_COLUMNS & 3) == 0)


'=========================================================================

PUB NoOp
{{
  Every Spin Object must have at least one Spin PUB function, and the first
  one may be run automatically, so we make sure the first one is "safe."
  This one does nothing.

  Arguments:
    None

  Return Value:
    None
}}
  ' Do nothing

'=========================================================================
PUB LoadHelperCog(ColorDepth, DisplayBufAddrVar, CmdFifoHeaderAddr, {
      }           DebugFifoOutputAddr)
{{
  Add the runtime-configured parameters to the hub image for the cog PASM code
  before loading the cog:
  - Color Depth (bumber of bits in the each color, i.e. the number of "layers"
  - Address of the hub variable that will eventually get the address of the
    display buffer, to allow delayed initialization of the cog, so the
    application can reclaim the hub copy of the cog code.
  - Address of the Command FiFo Header, that will (eventually) point to the
    portion of Hub memory used as the FiFo ring buffer.
  - The address of a memory area to get debug information, when needed,
    from the various command implementations. 
  Then the cog is started, which will load the cog, and then the cog code will
  wait until the display buffer address is supplied.

  Arguments:
    ColorDepth -- the numner of bits in each color value, which is the number
                  of color "layers" in the display buffer. E.g. "24-bit" color,
                  which is 8:8:8, would have a depth of 8.
    DisplayBufAddrVar -- the hub addres of the variable that will eventually
                         contain the actual hub address of the display buffer.
                         This can start out as zero, which will cause the
                         cog initialization code to stall, waiting for a non-
                         zero address to be supplied. This is a "two-stage"
                         initialization for the cog.
    CmdFiFoHeaderAddr -- the hub address of the FiFo header, that will contain
                         the Ring buffer/FiFo management information for the
                         (queued) commands to the helper.
    DebugFifoOutputAddr -- the hub address of ten LONG values to receive
                           debugging information from command processing.
                           This may not be needed if the debugging code is
                           disabled.

  Return Value:
    The Cog ID of the display refresh cog, plus one, so that it is non-zero.
}}
  ' Patch the color depth into the Cog code, and load the cog, passing the
  ' pointer to the place the cog will (eventually) get the display buffer
  ' address.
  '
  ' Returns the cog# + 1 of the cog started.

  ' Put the final two values into the cog.
  pcColorDepthBits := ColorDepth
  pcCmdFifoAddr    := CmdFifoHeaderAddr
  pcDebugFifoOutputHubAddr := DebugFifoOutputAddr

  ' Start the cog, and pass the address in Hub memory that will eventually get
  ' the hub address of the display buffer. The cog, after starting, waits until
  ' that location contains a non-zero hub address, and uses that address as the
  ' start of the display buffer.
  return 1 + COGNEW(@pPasmStart, DisplayBufAddrVar)

' ========================================================================
'
' PASM code for writing to the display buffer
'
' Style notes:
' I use special prefixes for code labels, variables and constants stored in
' the pasm code, so that names make clear what is being referred to.
'
' "pCodeLabel" -- A label on code (possibly with a leading ":")
' "piCodeLabel" -- A label on code for a target of instruction modification.
' "pvVariable" -- A label on a variable stored in a register in the cog,
'                 that can change over the course of cog execution
'                 ("pasm variable")
' "pcConstant" -- A label on a value that is effectively constant for the
'                 lifetime of the cog -- it may be set at initialization time,
'                 but will not be changed later.
'                 ("pasm constant")
' "piInstruction" -- A label on an instruction that may be the target of
'                    runtime instruction modification.
'

DAT
                    org     0
pPasmStart
                    jmp     #pInitContinue

' We store all the constants and some operating variables up front -- no real
' need to do so, and the variables might be moved to the end and become RES
' values
'
' Variables that CANNOT be RES:
' - ones than need compile-time init
' - ones that are set by the initialization code just before launching the PASM
' cog.

' Local storage of parameters needed for operation:
' - Buffer address
' - distance to advance the buffer address per row (usually just column count).
' - row count (usually 32, but might be 16 or 8 for some panels)
' - column count (here 64 per chained panel, but might be another value
'   The driver might need to calculate the number of longs in a row, if
'   it uses that optimization (fetch a long, and shift bits into place).
' - Bits and masks for the shift clock, the shift latch, output enable,
'   and the row number bits (a mask and a shift amount).

' pcBufAddrHub contains the hub address of the first row and column of the
' lowest-significance color bit of the display buffer.
' Set by deferred initialization after the cog starts.

pcBufAddrHub              long    0    ' Starting address of the display buffer

' These store the start of the low-order color bit layer, and the high order
' color bit layer, so that the code can write to them in the correct order
' depending on the order of color bits in the display color format (which has
' changed during the implementation period).
pcBufLowColorLayerAddrHub long    0
pcBufHighColorLayerAddrHub long    0

' pcColorDepthBits is the number of bits per color.
' This is set at runtime just before the cog is started.

pcColorDepthBits          long    1

' pcCmdFifoAddr is the address in hub memory of the Command Fifo descriptor
' (see Fifo.spin for details), that contains pointers to the actual fifo entry
' memory managed by the descriptor. These pointers may not be useable until
' after the display buffer address pas been passed to the running cog.

pcCmdFifoAddr             long     0

' The following should be at least 10 longs in size
' This is set just before the cog is started.
pcDebugFifoOutputHubAddr  long    0

' pcRefreshRows is the number of rows used to refresh the display; may be 32, 16,
'               or even 8 for some displays.
' pcMaxRefreshRow is the upper limit for refresh rows -- sometimes useful.
'
' (These could be simple constants as they are unlikely to be > 511, but for
' now, leave them as cog registers. We can get rid of them if we need the
' cog space.

pcDisplayRows       long    DISPLAY_ROWS
pcMaxDisplayRow     long    MAX_DISPLAY_ROW
pcRefreshRows       long    REFRESH_ROWS
pcMaxRefreshRow     long    MAX_REFRESH_ROW

' pcDisplayColumns is the number of display columns per row; this is _not_ the
'                  number of longs per row, and might not even be the number of
'                  _bytes_ per row, depending on how the data is packed.
' pcMaxDisplayColumn is the upper limit for column numbers, sometimes useful.
'
' It is unlikely, but possible, for these to exceed 511 -- for example, 8
' panels in a series, each of which has 64 columns, would reach 512 columns.
' If we need cog space, we could get rid of them and limit the column count.

pcDisplayColumns    long    DISPLAY_COLUMNS
pcMaxDisplayColumn  long    MAX_DISPLAY_COLUMN
'pcColumnCountLongs  long    DISPLAY_COLUMNS >> 2  ' Number of LONGs in a row

pcColorLayerBytes   long    DISPLAY_COLUMNS * REFRESH_ROWS
'pcDisplayBufSizeBytes long    0   ' pcColorLayerBytes * pcColorDepthBits

pcDisplayColorBitOrder long    Config#DISPLAY_COLOR_BIT_ORDER
 
pcRomFontBaseAddr   long    $8000

' These control how to walk through the color layers in the display buffer
pcBufferLowToHighAddrDelta  long    DBUF_LO2HI_LAYER_DELTA
pcBufferHighToLowAddrDelta  long    DBUF_HI2LO_LAYER_DELTA


' Simple constants
pcZero              long     0
pcOne               long     1
pcMinusOne          long    -1
pcMinusThree        long    -3 ' simple constant

'pcDebugLed1         long    |< DEBUG_LED_PIN1
'pcDebugLed2         long    |< DEBUG_LED_PIN2


' The code continues here...
pInitContinue
                    ' Wait for the app to pass the display buffer
                    ' address to us.
:pInitWaitLoop
                    rdlong  pcBufAddrHub, PAR  wz
    if_z            jmp     #:pInitWaitLoop

                    ' At this point, the FiFo pointers should all be
                    ' correct, so we collect the FiFo information for
                    ' faster access. We need to _write_ to the FiFo
                    ' entries for remove_count, remove_pointer, and
                    ' we need to _read_ directly from the FiFo for
                    ' the insert_pointer and insert_count.
                    ' The values for start_ptr, limit_ptr, and
                    ' entry size are fixed for the life of the FiFo.

                    ' Get all the pointers into the FiFo descriptor that we need
                    mov     pcFifoInsertPtrHubAddr, pcCmdFifoAddr
                    add     pcFifoInsertPtrHubAddr, #Fifo#INSERT_HUBADDR_OFFSET
                    mov     pcFifoRemovePtrHubAddr, pcCmdFifoAddr
                    add     pcFifoRemovePtrHubAddr, #Fifo#REMOVE_HUBADDR_OFFSET

                    ' Now, collect our local copies of the values that only
                    ' this code sets or that do not change, so we don't need to
                    ' read from hub memory repeatedly for them:
                    ' - remove_count
                    ' - remove_ptr
                    ' - start_ptr
                    ' - limit_ptr
                    ' - entry_size
                    rdword  pvFifoRemovePtr, pcFifoRemovePtrHubAddr
                    add     pcCmdFifoAddr, #Fifo#START_HUBADDR_OFFSET
                    rdword  pcFifoStartPtr, pcCmdFifoAddr
                    add     pcCmdFifoAddr, #(Fifo#LIMIT_HUBADDR_OFFSET - Fifo#START_HUBADDR_OFFSET)
                    rdword  pcFifoLimitPtr, pcCmdFifoAddr
                    add     pcCmdFifoAddr, #(Fifo#ENTRY_SIZE_BYTES_OFFSET - Fifo#LIMIT_HUBADDR_OFFSET)
                    rdword  pcFifoEntrySizeBytes, pcCmdFifoAddr
                    sub     pcCmdFifoAddr, #Fifo#ENTRY_SIZE_BYTES_OFFSET

                    mov     pcColorDepthBitsTimes3, pcColorDepthBits
                    add     pcColorDepthBitsTimes3, pcColorDepthBits
                    add     pcColorDepthBitsTimes3, pcColorDepthBits

                    ' Calculate a few address values that depend on the
                    ' number of color layers, and their order, either low-to-
                    ' high, or high-to-low, in memory:
                    ' - The size of the display buffer
                    ' - The address of the lowest order color layer;
                    ' - The address of the highest order color layer
                    ' - The distance to "back up" the pointer into the display
                    '   after a pixel update of all the layers. This might be
                    '   the size of the entire buffer, or the negated size,
                    '   depending on the order of bits in the display color
                    '   format, and the order of layers that the refresh cog
                    '   wants. 
                    ' Since the color depth is only 1..8 (or thereabouts), we
                    ' use a simple addition loop. It would also be pretty easy
                    ' to implement a three-bit multiplication loop -- faster,
                    ' but a few more instructions.
                    mov     pcDisplayBufSizeBytes, #0
                    mov     pvTemp, pcColorDepthBits
:pBufSizeLoop       add     pcDisplayBufSizeBytes, pcColorLayerBytes
                    djnz    pvTemp, #:pBufSizeLoop

                    ' If the low-to-high delta is < 0, then the first layer is
                    ' the highest-order color bit, otherwise the first layer is
                    ' the lowest-order color bit.
                    cmps    pcBufferLowToHighAddrDelta, #0  wc
    if_c            mov     pcBufHighColorLayerAddrHub, pcBufAddrHub
    if_c            mov     pcBufLowColorLayerAddrHub, pcBufAddrHub
    if_c            add     pcBufLowColorLayerAddrHub, pcDisplayBufSizeBytes
    if_c            sub     pcBufLowColorLayerAddrHub, pcColorLayerBytes
    if_c            neg     pcDisplayBufSizePixelDelta, pcDisplayBufSizeBytes
    if_nc           mov     pcBufLowColorLayerAddrHub, pcBufAddrHub
    if_nc           mov     pcBufHighColorLayerAddrHub, pcBufAddrHub
    if_nc           add     pcBufHighColorLayerAddrHub, pcDisplayBufSizeBytes
    if_nc           sub     pcBufHighColorLayerAddrHub, pcColorLayerBytes
    if_nc           mov     pcDisplayBufSizePixelDelta, pcDisplayBufSizeBytes

'                    ' Debugging help
'                    mov     DIRA, pcDebugLed1
'                    or      OUTA, pcDebugLed1  ' starts ON.
'                    mov     DIRA, pcDebugLed2
'                    andn    OUTA, pcDebugLed2  ' starts OFF.

pWaitForNextCmd
                    ' If the buffer is not empty, we have a command.
                    ' Simple Fifo version: the Fifo is empty if the insert
                    ' and remove pointers are equal
                    rdword   pvCmd, pcFifoInsertPtrHubAddr
                    cmp      pvCmd, pvFifoRemovePtr  wz
    if_z            jmp      #pWaitForNextCmd

                    ' Here we know that we have a command to process in the FiFo.
                    ' Pick up the command from the FiFo
                    ' check that it is in-range
                    '   if not, jump to the end-of-command logic that updates
                    '   the Fifo to remove the command:
                    '   - advance and wrap the remove_ptr
                    '   - increment the remove_count; must be last.

                    rdword  pvCmd, pvFifoRemovePtr
                    and     pvCmd, #HConfig#CMD_MASK
                    ' Check for an out-of-range command -- convert it to a no-op
                    cmp     pvCmd, #HConfig#MAX_CMD  wc, wz
                    ' if c bit is clear, pvCmd is => MAX_CMD_PLUS_ONE
    if_nc_and_nz    mov     pvCmd, #0  ' out of range --> No-op
    if_nc_and_nz    wrword  pvCmd, pvFifoRemovePtr  ' update the command in the FiFo

                    ' Dispatch on the command
                    mov     pvResult, pcOne  ' assume success
                    mov     pvResultHubAddr, #0  ' assume no return value

                    '' Each time we dispatch a command, we toggle the debug
                    '' led1
                    'xor     OUTA, pcDebugLed1    ' debug
                    add     pvCmd, #pCmdDispatchTable
                    jmp     pvCmd    ' Jumps into the following table.
pCmdDispatchTable   jmp     #pCmdFinishedFailure  ' $0000 -- No-op with failure
                    jmp     #pCmdSetColor         ' $0001 -- SetColor (single pixel)
                    jmp     #pCmdBitmap1Bpp       ' $0002 -- Bitmap1bpp
                    jmp     #pCmdRomFontOutput    ' $0003 -- ROM font char output
                    jmp     #pCmdNoOp             ' $0004 -- No-Op with success

pCmdFinishedFailure
                    mov     pvResult, pcMinusOne                    
pCmdFinishedSuccess
                    cmp     pvResultHubAddr, #0  wz
    if_nz           wrlong  pvResult, pvResultHubAddr
pCmdFinished
                    '' Each time we finish a command, we toggle the debug
                    '' led2
                    'xor     OUTA, pcDebugLed2

                    ' To finish (remove) the current command:
                    ' - Advance the RemovePtr by the entry size, write back to
                    '   the Fifo Descriptor
                    ' - Remember to wrap the remove ptr
                    add     pvFifoRemovePtr, pcFifoEntrySizeBytes
                    cmp     pvFifoRemovePtr, pcFifoLimitPtr wc
    if_nc           mov     pvFifoRemovePtr, pcFifoStartPtr
                    wrword  pvFifoRemovePtr, pcFifoRemovePtrHubAddr
                    jmp     #pWaitForNextCmd

'-------------------------------------------------------------------

' Perhaps name this pConvertColorsToDisplayBitOrder ?
pConvertColorsToDisplay
                    ' Check to see if we should convert anything.
                    test    pvColor, pcDisplayColorBitOrder wz
    if_nz           jmp     #pConvertColorsToDisplay_ret

                    ' Convert pvColor from ##0RrrrrGggggBbbbb (10-bit
                    ' bit fields) to interleaved color bits rbRGBrgbRGB...
                    ' The low-order bits of the output are the high-order
                    ' color bits.
                    mov     pvTemp, #0
                    mov     pvColorBitLayerLoopCnt, pcColorDepthBits
:pConvertColorsLoop
                    rol     pvTemp, #3
                    ror     pvColor, #10  wc  ' get the next blue bit
                    muxc    pvTemp, #BLUE_BIT
                    ror     pvColor, #10  wc  ' get the next green bit
                    muxc    pvTemp, #GREEN_BIT
                    ror     pvColor, #13  wc
                    muxc    pvTemp, #RED_BIT

                    djnz    pvColorBitLayerLoopCnt, #:pConvertColorsLoop

                    mov     pvColor, pvTemp
                    or      pvColor, pcDisplayColorBitOrder

pConvertColorsToDisplay_ret
                    ret

'-------------------------------------------------------------------

' Define a table of multiples of the number of DISPLAY_COLUMNS, in order to
' get a fast multiply operation, for multipliers in the range 0..31. This
' lets use get the correct display row byte offset from the row number.
'
' DISPLAY_COLUMNS is not a power-of-two because displays can be chained
' together.

pcDisplayColumnsMultiples
                    LONG     0
                    LONG     1 * DISPLAY_COLUMNS
                    LONG     2 * DISPLAY_COLUMNS
                    LONG     3 * DISPLAY_COLUMNS
                    LONG     4 * DISPLAY_COLUMNS
                    LONG     5 * DISPLAY_COLUMNS
                    LONG     6 * DISPLAY_COLUMNS
                    LONG     7 * DISPLAY_COLUMNS
                    LONG     8 * DISPLAY_COLUMNS
                    LONG     9 * DISPLAY_COLUMNS
                    LONG    10 * DISPLAY_COLUMNS
                    LONG    11 * DISPLAY_COLUMNS
                    LONG    12 * DISPLAY_COLUMNS
                    LONG    13 * DISPLAY_COLUMNS
                    LONG    14 * DISPLAY_COLUMNS
                    LONG    15 * DISPLAY_COLUMNS
                    LONG    16 * DISPLAY_COLUMNS
                    LONG    17 * DISPLAY_COLUMNS
                    LONG    18 * DISPLAY_COLUMNS
                    LONG    19 * DISPLAY_COLUMNS
                    LONG    20 * DISPLAY_COLUMNS
                    LONG    21 * DISPLAY_COLUMNS
                    LONG    22 * DISPLAY_COLUMNS
                    LONG    23 * DISPLAY_COLUMNS
                    LONG    24 * DISPLAY_COLUMNS
                    LONG    25 * DISPLAY_COLUMNS
                    LONG    26 * DISPLAY_COLUMNS
                    LONG    27 * DISPLAY_COLUMNS
                    LONG    28 * DISPLAY_COLUMNS
                    LONG    29 * DISPLAY_COLUMNS
                    LONG    30 * DISPLAY_COLUMNS
                    LONG    31 * DISPLAY_COLUMNS

pCalcStartingBuf
                    ' Calculate the correct byte in the display buffer:
                    ' assuming this would be for the highest-order color bit:
                    '    pcBufHighColorLayerAddrHub + (pvRow * DISPLAY_COLUMNS) + pvCol
                    ' TODO: We can constrain pvRow to be 0..31, so we could
                    ' use a special-purpose multiply routine, which would be
                    ' about 18 instructions, instead of just the one, sigh.
                    ' Or, we could use a 32-entry lookup table, containing
                    ' precalculated values for multiples of DISPLAY_COLUMNS
                    ' which would only be about 3 instructions, but more
                    ' cog memory.

                    add     :piMultLookup, pvRow
                    sub     :piMultLookup, pvRow  ' applied after next instruction
                                                  ' is fetched.
:piMultLookup       mov     pvBuf, pcDisplayColumnsMultiples + 0
                    add     pvBuf, pvColumn
                    add     pvBuf, pcBufHighColorLayerAddrHub  ' pcBufAddrHub
pCalcStartingBuf_ret
                    ret

'-------------------------------------------------------------------

pSetColorHelper
                    ' pvColor is the color to use for the layers
                    ' pvColorShift, and pvColorMask are used to isolate the
                    ' correct bits for the row.
                    ' pvBuf is the address in the first layer to update.
                    '
                    ' Uses pvTemp and pvByte, and pvColorBitLayerLoopCnt
                    ' Destroys pvColor.

                    mov     pvColorBitLayerLoopCnt, pcColorDepthBits
:pSetColorLayerLoop
                    mov     pvTemp, pvColor
                    shl     pvTemp, pvColorShift
                    and     pvTemp, pvColorMask
                    rdbyte  pvByte, pvBuf
                    andn    pvByte, pvColorMask
                    or      pvByte, pvTemp
                    wrbyte  pvByte, pvBuf
                    shr     pvColor, #3
                    add     pvBuf, pcBufferHighToLowAddrDelta
                    djnz    pvColorBitLayerLoopCnt, #:pSetColorLayerLoop

                    ' restore the starting pvBuf value
                    add     pvBuf, pcDisplayBufSizePixelDelta
pSetColorHelper_ret ret

'-------------------------------------------------------------------

pCmdNoOp
                    ' Used to sync to operation-complete status.
                    mov     pvParams, pvFifoRemovePtr
                    add     pvParams, #HConfig#PIXELCMD_RESULT_PTR
                    rdword  pvResultHubAddr, pvParams
                    jmp     #pCmdFinishedSuccess

'-------------------------------------------------------------------

pCmdSetColor
                    ' Set a single pixel to a specific color.
                    mov     pvParams, pvFifoRemovePtr
                    add     pvParams, #HConfig#PIXELCMD_RESULT_PTR
                    rdword  pvResultHubAddr, pvParams
                    add     pvParams, #(HConfig#PIXELCMD_COLUMN - HConfig#PIXELCMD_RESULT_PTR)
                    rdword  pvColumn, pvParams
                    shl     pvColumn, #16  ' sign extend -- see below
                    add     pvParams, #(HConfig#PIXELCMD_ROW - HConfig#PIXELCMD_COLUMN)
                    rdword  pvRow, pvParams

                    ' Sign-extend the 16-bit row and column
                    ' (interleave these in the fetch logic)
                    'shl     pvColumn, #16
                    sar     pvColumn, #16
                    shl     pvRow, #16
                    sar     pvRow, #16

                    ' Check that the row and column are in range, and if not,
                    ' do nothing.
                    ' Each comparison sets the Carry flag if pRow or pColumn
                    ' is out of range in that direction.
                    cmps    pvRow, #0  wc
    if_nc           cmp     pcMaxDisplayRow, pvRow  wc
    if_nc           cmps    pvColumn, #0  wc
    if_nc           cmp     pcMaxDisplayColumn, pvColumn  wc
    if_c            jmp     #pCmdFinishedFailure

                    add     pvParams, #(HConfig#PIXELCMD_COLOR - HConfig#PIXELCMD_ROW)
                    rdlong  pvColor, pvParams
                    ' Maybe convert the color representation.
                    mov     pvOrigColor, pvColor
                    call    #pConvertColorsToDisplay

                    ' If the row is > REFRESH_ROWS, we have to subtract
                    ' REFRESH_ROWS from the row, and shift the three bit
                    ' rgb bit values to the left.
                    mov     pvColorShift, #0
                    mov     pvColorMask, #%000_111

                    cmpsub     pvRow, pcRefreshRows  wc
                    ' If C is set, pvRow was => REFRESH_ROWS.
    if_c            mov     pvColorShift, #3
    if_c            mov     pvColorMask, #%111_000

                    ' Calculate the correct byte in the display buffer:
                    '    pcBufAddrHub + (pvRow * DISPLAY_COLUMNS) + pvCol
                    call    #pCalcStartingBuf

                    '' Debugging -- send back the pvBuf and pvCmd values
                    'tjz     pvResultHubAddr, #:pCmdSetColorDebugDone
                    'add     pvResultHubAddr, #4
                    'wrlong  pvBuf, pvResultHubAddr
                    'add     pvResultHubAddr, #4
                    'wrlong  pvColor, pvResultHubAddr
                    'add     pvResultHubAddr, #4
                    'wrlong  pvOrigColor, pvResultHubAddr
                    'sub     pvResultHubAddr, #12
:pCmdSetColorDebugDone

                    'wrlong  pcBufAddrHub, pcDebugFifoOutputHubAddr
                    'add     pcDebugFifoOutputHubAddr, #4
                    'wrlong  pvBuf, pcDebugFifoOutputHubAddr
                    'sub     pcDebugFifoOutputHubAddr, #4

                    ' Now, loop over the color bits:
                    ' for each color bit
                    '   Get the rgb bits from pvColor
                    '   shift them and mask them
                    '   Fetch the byte from the display buffer
                    '   Update the three bit field in the byte
                    '   Write the updated byte back to the display buffer
                    '   shift pvColor right three bits for the next color layer
                    '   advance pvBuf by the size of a color layer, which is
                    '     REFRESH_ROWS * DISPLAY_COLUMNS

                    call    #pSetColorHelper
                    jmp     #pCmdFinishedSuccess
                   
'-------------------------------------------------------------------

pCmdCommonBitmapParams
                    ' Fetch the common parameters for a number of functions
                    ' that operate on bitmaps.
                    ' The Column and Row parameters are treated as 16-bit
                    ' signed values, and so have to be sign-extended.

                    mov     pvParams, pvFifoRemovePtr
                    add     pvParams, #2
                    'mov     pvResultHubAddr, #0
                    rdword  pvResultHubAddr, pvParams
                    add     pvParams, #2
                    rdword  pvColumn, pvParams
                    add     pvParams, #2
                    ' Remember to sign-extend pvColumn (and pvRow)
                    shl     pvColumn, #16
                    rdword  pvRow, pvParams
                    add     pvParams, #2
                    shl     pvRow, #16
                    rdword  pvWidth, pvParams
                    add     pvParams, #2
                    sar     pvColumn, #16
                    rdword  pvHeight, pvParams
                    add     pvParams, #2
                    sar     pvRow, #16
                    rdword  pvLTrim, pvParams
                    add     pvParams, #2
                    rdword  pvTTrim, pvParams
                    add     pvParams, #2   ' leave pvParams pointing to the next param

pCmdCommonBitmapParams_ret
                    ret
                    
'-------------------------------------------------------------------

pCmdRomFontOutput
                    ' Write a portion, or all, of a character from the P1 ROM
                    ' font to the display, using the supplied colors.
                    ' The character bitmap can be trimmed left/right/top/bottom,
                    ' and it can be clipped to the display boundaries ("hang
                    ' off the edge")
                    '
                    ' First, pick up all our many parameters.
                    ' This code "knows" the order and size of the parameters in
                    ' the command buffer.
                    '
                    ' We also sign-extend the pvRow and pvColumn.
                    ' The instructions for this ("shl pv, #16" then "sar pv, #16")
                    ' are intreleaved with the hub reads in the delay slots.

                    call    #pCmdCommonBitmapParams

                    ' On return, pvTemp points to the next parameter position
                    rdword  pvRomChar, pvParams
                    add     pvParams, #(2 + 2)  ' also skip the padding word
                    rdlong  pvColor, pvParams
                    add     pvParams, #4
                    mov     pvOrigColor, pvColor  ' DEBUG
                    call    #pConvertColorsToDisplay
                    mov     pvPalette0, pvColor
                    rdlong  pvColor, pvParams
                    add     pvParams, #4
                    mov     pvOrigColor2, pvColor  ' DEBUG
                    rdlong  pvColumnSkipPattern, pvParams
                    'add     pvParams, #4
                    call    #pConvertColorsToDisplay
                    mov     pvPalette3, pvColor

                    ' Get the base address of the character
                    ' Each entry is two characters, 32 longs long, so we
                    ' offset the base address by 32 * 4 * pvRomChar / 2, then
                    ' remove the bottom bit to get long alignment.
                    mov     pvBitmapHubAddr, pvRomChar
                    andn    pvBitmapHubAddr, #1     ' remove the low bit
                    shl     pvBitmapHubAddr, #6  ' * 32 * 4 / 2
                    add     pvBitmapHubAddr, pcRomFontBaseAddr
                    ' Remember that the pvRomChar has been shifted left...
                    mov     pvBitmapRowIncr, #4  ' the char rows are 4 bytes.

                    ' Prepare the rest of the palette, to properly select the
                    ' foreground and background colors for the two possible
                    ' characters.
                    ' If the char value is even, then we use bit zero to select
                    ' fg or bg. If the value is odd, then we use bit one to
                    ' select fg or bg.
                    test    pvRomChar, #1  wz       ' see if odd or even 
    if_z            mov     pvPalette2, pvPalette0  ' bg in palette[0] and palette[2]
    if_z            mov     pvPalette1, pvPalette3  ' fg in palette[1] and palette[3]
    if_nz           mov     pvPalette1, pvPalette0  ' bg in palette[0] and palette[1]
    if_nz           mov     pvPalette2, pvPalette3  ' fg in palette[2] and palette[3]

                    ' This uses two bits per pixel.
                    mov     pvBitsPerPixel, #2
                    mov     pvPixelMask, #%11

                    '' Debugging -- pass back our bitmap address
                    'tjz     pvResultHubAddr, #:pCmdRomFontDebugDone
                    'add     pvResultHubAddr, #4
                    'wrlong  pvBitmapHubAddr, pvResultHubAddr
                    'add     pvResultHubAddr, #4
                    'wrlong  pvRomChar, pvResultHubAddr
                    'sub     pvResultHubAddr, #8  ' put it back.
:pCmdRomFontDebugDone

                    jmp     #pBitmapOutput

'-------------------------------------------------------------------

pCmdBitmap1Bpp
                    ' Write a portion, or all, of a 1 bit-per-pixel bitmap to
                    ' the display, using the supplied foreground and
                    ' colors.
                    ' The bitmap can be trimmed left/right/top/bottom, and it
                    ' can be clipped to the display boundaries ("hang off the
                    ' edge")
                    '
                    ' The bitmap pointer can be byte, word, or long aligned,
                    ' but only one bitmap hub read will be done per row,
                    ' which is useful for drawing horizontal and vertical lines
                    ' and boxes of arbitrary size, but it means that large
                    ' bitmaps have to be drawn 32 columns at a time.

                    ' First, pick up all our many parameters.
                    ' This code "knows" the order and size of the parameters in
                    ' the command buffer.
                    '
                    ' We also sign-extend the pvRow, pvColumn, and rowaddr_incr.
                    ' The instructions for this ("shl pv, #16" then "sar pv, #16")
                    ' are intreleaved with the hub reads in the delay slots.

                    call    #pCmdCommonBitmapParams

                    ' On return, pvTemp points to the bitmap address parameter
                    rdword  pvBitmapHubAddr, pvParams
                    add     pvParams, #2
                    rdword  pvBitmapRowIncr, pvParams
                    add     pvParams, #2
                     shl     pvBitmapRowIncr, #16  ' sign extend increment
                    rdlong  pvColor, pvParams
                    add     pvParams, #4
                     sar     pvBitmapRowIncr, #16  ' sign extend increment
                    mov     pvOrigColor, pvColor  ' DEBUG
                    call    #pConvertColorsToDisplay
                    mov     pvBGColor, pvColor
                    rdlong  pvColor, pvParams
                    add     pvParams, #4
                    mov     pvOrigColor2, pvColor  ' DEBUG
                    rdlong  pvColumnSkipPattern, pvParams
                    'add     pvParams, #4
                    call    #pConvertColorsToDisplay
                    mov     pvFGColor, pvColor
                    'jmp     #pBitmapOutput

                    '' Debugging -- pass back our bitmap address
                    'tjz     pvResultHubAddr, #:pCmdBitmap1BppDebugDone
                    'add     pvResultHubAddr, #4
                    'wrlong  pvBGColor, pvResultHubAddr
                    'add     pvResultHubAddr, #4
                    'wrlong  pvFGColor, pvResultHubAddr
                    'add     pvResultHubAddr, #4
                    'wrlong  pvOrigColor, pvResultHubAddr
                    'add     pvResultHubAddr, #4
                    'wrlong  pvOrigColor2, pvResultHubAddr
                    'sub     pvResultHubAddr, #16  ' put it back.
:pCmdBitmap1BppDebugDone

                    ' This uses one bit per pixel.
                    mov     pvBitsPerPixel, #1
                    mov     pvPixelMask, #1

' Common Bitmap processing code.
' This code handles all the range-checks, trimming, and advancing through
' the rows and columns of the bitmap at multiple bpp.
pBitmapOutput

                    cmp     pvResultHubAddr, #0  wz
    if_nz           wrlong  pcZero, pvResultHubAddr


                    ' The rest of the code is pretty generic for all the
                    ' bitmap operations, so we could factor this (or just
                    ' jump into it after picking up the parameters).

                    ' If the bitmap would hang off the edge of the screen,
                    ' incorporate that into the pv?Trim variables, and
                    ' adjust pvRow, pvColumn, pvWidth, and pvHeight to
                    ' take that into account. If pvWidth or pvHeight
                    ' become zero or negative, then nothing will be
                    ' drawn.

:pCheckTop          cmps    pvRow, #0  wc
    if_nc           jmp     #:pCheckBottom
                    neg     pvRow, pvRow    ' the number of rows hanging off the display
                    add     pvTTrim, pvRow
                    sub     pvHeight, pvRow  wc, wz  ' reduces the height, might make
                                                     ' height zero or negative
    if_c_or_z       jmp     #pCmdFinishedFailure     ' nothing left to display

':pCheckTopAdjustPtr add     pvBitmapHubAddr, pvBitmapRowIncr
'                    djnz    pvRow, #:pCheckTopAdjustPtr
                    mov     pvRow, #0
                    ' Now pvRow is zero 
                    
:pCheckBottom       ' if (row + height - display_rows) > 0, then the bitmap is
                    ' hanging off the bottom of the screen by that number of
                    ' rows.
                    mov     pvTemp, pvRow
                    add     pvTemp, pvHeight
                    sub     pvTemp, pcDisplayRows wc, wz
    if_c_or_z       jmp     #:pCheckLeft
                    sub     pvHeight, pvTemp  wc, wz
    if_c_or_z       jmp     #pCmdFinishedFailure  ' nothing left to display

:pCheckLeft         cmps    pvColumn, #0  wc
    if_nc           jmp     #:pCheckRight
                    neg     pvColumn, pvColumn
                    add     pvLTrim, pvColumn          ' Trim more on the left
                    sub     pvWidth, pvColumn  wc, wz  ' reduces the width, might make
                                                       ' width zero or negative
    if_c_or_z       jmp     #pCmdFinishedFailure
                    ' No need to adjust the bitmap pointer
                    mov     pvColumn, #0

:pCheckRight
                    ' if (column + width - columns) > 0, then the bitmap is
                    ' hanging off the right edge of the screen by that number of
                    ' columns.
                    mov     pvTemp, pvColumn
                    add     pvTemp, pvWidth
                    sub     pvTemp, pcDisplayColumns wc, wz
    if_c_or_z       jmp     #:pChecksDone
                    sub     pvWidth, pvTemp  wc, wz
    if_c_or_z       jmp     #pCmdFinishedFailure  ' nothing left to display

:pChecksDone
                    '' Debugging -- pass back some information.
                    'wrlong  pvResultHubAddr, pvFifoRemovePtr
                    'add     pvFifoRemovePtr, #4
                    'wrlong  pvBGColor, pvFifoRemovePtr
                    'sub     pvFifoRemovePtr, #4 ' put it back.
                    '
                    'jmp     #pCmdFinishedSuccess  ' DEBUG

' Init the pointer into the display buffer
' For each row
'   get the row bits, trimmed on the right
'   For each column in the row
'     Get the pixel color index
'     Get the color
'     draw the color for that position to all color layers.
'     (advance to the next column)
'   advance to the next row
'
' If the starting row is < pvRefreshRows, we start with
' pvColorShift := 0 and pvColorMask := %111
' As soon as pvRow becomes == pvRefreshRows, we change pvColorShift := 3,
'   pvColorMask := %111_000, and we subtract the layer size from pvBuf.

                    ' If the starting row is >= 32, we have to adjust
                    ' pvColorShift, pvColorMask, and pvBuf.
                    ' When the row nnumber crosses into the bottom half of the
                    ' display, we do the adjustment after that crossing, but if
                    ' if it doesn't cross that boundary, we need to make the
                    ' adjustment here.
                    mov     pvColorShift, #0
                    mov     pvColorMask, #%111
                    cmpsub     pvRow, pcRefreshRows  wc
    if_c            mov     pvColorShift, #3
    if_c            mov     pvColorMask, #%111_000
                    call    #pCalcStartingBuf ' does not touch C bit
    if_c            add     pvRow, pcRefreshRows

                    ' If we advance into the bottom half of the display while
                    ' writing the rows, we have to do the following then:
                    '        add pvRow, #1
                    '        cmp pvRow, pcRefreshRows wz
                    ' if_z   mov pvColorShift, #3
                    ' if_z   mov pvColorMask, #%111_000
                    ' if_z   sub pvBuf, pcColorLayerBytes

                    ' Adjust pvLTrim to be the correct shift amount for the
                    ' number of bits per pixel.
                    ' We really want:
                    '   pvLTrim *= pvBitsPerPixel
                    ' for 'small" values of pvBitsPerPixel
                    ' This logic only works for 1bpp and 2bpp -- all other values
                    ' will be WRONG.
                    mov     pvLTrimPixels, pvLTrim
                    cmp     pvBitsPerPixel, #2 wz
    if_z            shl     pvLTrimPixels, #1

                    ' Adjust the starting point in the bitmap, based on pvTTrim.
                    ' repeat while pvTTrim-- > 0
                    '   pvBitmapHubAddr += pvBitmapRowIncr
                    ' We could use a multiply routine, or we can repeatedly add.
                    ' Which is faster depends on the typical values for pvRow
                    ' and pvBitmapRowIncr -- if pvBitmapRowIncr is a power of
                    ' two, then it is just a shift operation on (-pvRow).
                    ' But when drawing a wide bitmap, the row increment might
                    ' also be large, and not a power-of-two.
                    tjz     pvTTrim, #:pRowLoop
:pTTrimLoop         add     pvBitmapHubAddr, pvBitmapRowIncr
                    djnz    pvTTrim, #:pTTrimLoop 

:pRowLoop
                    ' Get the pixel data.
                    test    pvBitmapHubAddr, #1  wc  ' C set if rdbyte needed
                    test    pvBitmapHubAddr, #2  wz  ' Z clear if rdword needed
    if_c            rdbyte  pvPixels, pvBitmapHubAddr
    if_nc_and_nz    rdword  pvPixels, pvBitmapHubAddr
    if_nc_and_z     rdlong  pvPixels, pvBitmapHubAddr
                    add     pvBitmapHubAddr, pvBitmapRowIncr ' advance early (could move this instruction)

                    ' Trim the left edge of the bitmap for this row
                    shr     pvPixels, pvLTrimPixels
                    mov     pvSkip, pvColumnSkipPattern
                    shr     pvSkip, pvLTrim

                    ' For each pixel in the row (column loop)
                    '   Get the pixel
                    '   Get the color for that pixel
                    '   write that pixel to all layers (call a helper)

                    mov     pvPixelColumnLoopCnt, pvWidth
:pColumnLoop
                    ' Skip any column that we shouldn't touch.
                    shr     pvSkip, #1  wc
   if_c             jmp     #:pColumnLoopContinue

                    mov     pvPixelIndex, pvPixels
                    and     pvPixelIndex, pvPixelMask
                    add     :piLookupColor, pvPixelIndex
                    sub     :piLookupColor, pvPixelIndex  ' Update happens after the
                                                          ' read in the next instruction
:piLookupColor      mov     pvColor, pvColorPalette + 0
                    ' pvBuf is ready, pvColor is ready.
                    call    #pSetColorHelper

:pColumnLoopContinue
                    ' Advance to the next pixel in the row
                    add     pvBuf, #1
                    shr     pvPixels, pvBitsPerPixel
                    djnz    pvPixelColumnLoopCnt, #:pColumnLoop
                    sub     pvBuf, pvWidth  ' put back for advance to next row

                    ' advance to next row
                    add     pvBuf, pcDisplayColumns
                    add     pvRow, #1
                    ' Check for wrapping back to the high-order color position for
                    ' the bottom half of the display.
                    cmp     pvRow, pcRefreshRows  wz
    if_z            mov     pvColorShift, #3
    if_z            mov     pvColorMask, #%111_000
                    ' Since pvBuf is in the "first row of the bottom half", we
                    ' can just back up the size of one layer
    if_z            sub     pvBuf, pcColorLayerBytes

                    djnz    pvHeight, #:pRowLoop

                    jmp     #pCmdFinishedSuccess

'-------------------------------------------------------------------------
' RES variables -- initialized by the cog after it starts.

pcColorDepthBitsTimes3    res     1  ' calculated during init.

' The following are used for managing fetching commands from the FiFo, and
' indicating that commands have been removed from the Fifo.
pcFifoInsertPtrHubAddr    res     1
pcFifoRemovePtrHubAddr    res     1

pvFifoRemovePtr           res     1
pcFifoStartPtr            res     1
pcFifoLimitPtr            res     1
pcFifoEntrySizeBytes      res     1

' pvCmd vRawColors are set from the first WORD of the command.
pvCmd                     res     1
pvResultHubAddr           res     1
pvResult                  res     1
pvColumn                  res     1
pvRow                     res     1
pvWidth                   res     1
pvHeight                  res     1
pvLTrim                   res     1
pvTTrim                   res     1
pvColorShift              res     1
pvColorMask               res     1   ' either %000_111 or %111_000
pvBitmapHubAddr           res     1
pvBitmapRowIncr           res     1   ' number of bytes to advance to the next
                                      ' row of the bitmap
pvBitsPerPixel            res     1
pvPixelMask               res     1
pvBuf                     res     1
pvBufSave                 res     1
pvPixels                  res     1
pvPixelIndex              res     1
pvByte                    res     1
pvColor                   res     1
pvOrigColor               res     1
pvOrigColor2              res     1
pvParams                  res     1
pvRomChar                 res     1
pvColumnSkipPattern       res     1  ' %101_0101_0101_0101_0101_0101_0101_01010
pvSkip                    res     1
pvLTrimPixels             res     1

' Cog-local color palette,handles up to 2bpp
pvColorPalette
pvBGColor
pvPalette0                res     1
pvFGColor
pvPalette1                res     1
pvPalette2                res     1
pvPalette3                res     1
' Additional color slots up to the max bpp we permit.
pcDisplayBufSizeBytes     res     1   ' pcColorLayerBytes * pcColorDepthBits
pcDisplayBufSizePixelDelta res     1 

pv_t                      res     1    ' temporary variable
pvTemp                    res     1

' Loop variable for the color bit layers.
pvColorBitLayerLoopCnt    res     1

' Loop variable for the columns of a row for bitmap writing
pvPixelColumnLoopCnt      res     1


'====================================================================

PasmEnd             fit  $1F0

