{{
  Measure the refresh rate for all the COLOR_DEPTH values.

Author:         Tom Dinger (Ding-Batty)
Release Date:   2022/04/10

Copyright 2024 Thomas Dinger

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

--------------------------------------------------------------------------

This test program loads and starts the HUB75 display refresh driver once for
each value of COLOR_DEPTH from 1 to 8, and measures/collects the refresh
clocks for each, and report the refresh rate for the given configuration.

This does not need actual display hardware connected, but it can give a visual
indication of how it is working.

This should be run for each display refresh driver, and possibly for each
display resolution of interest.

For more information:
---------------------

See the file "hub75_info.txt" for some details on the display interface
called "HUB75", with some useful links.


Revisions:
==========
- 2024/04/10 -- Updated refresh rate reporting;
- 2024/02/26 -- Original release

}}

OBJ
  ' Pick up the static configuration for the display(s).
  Config: "hub75_config.spin"

  Ser    : "Db_FullDuplexSerial"
  Umath  : "umath2.spin"

  ' We can't recover the hub memory copy of the PASM code, since we have to
  ' restart the driver multiple times.
  'PanelPasm : "hub75_display_pasm_repeat.spin"
  PanelPasm : "hub75_display_pasm_video_4clock.spin"
  'PanelPasm : "hub75_display_pasm_video_8clock.spin"
  'PanelPasm : "hub75_display_pasm_waitcnt.spin"

  ' The init object (below) and all objects it includes, will be used as
  ' additional buffer allocation space after they are finished during init.
  ' This helps to maximize hub memory available for large buffer use.
  Init : "measure_refresh_rate_init.spin"


CON { General constants }
  _CLKMODE = XTAL1 + PLL16X
  _XINFREQ = 5_000_000

  { Configuration for test serial connection. }
  BAUDRATE = Init#BAUDRATE  ' 115_200
  RXPIN    = Init#RXPIN     ' 31    ' The usual pins.
  TXPIN    = Init#TXPIN     ' 30
  ' For this version of FDS, the sizes do not need to be powers-of-two. The
  ' Spin code is a little slower, but the PASM code is not slowed by this.
  RX_BUFFER_SIZE = 28
  TX_BUFFER_SIZE = 200

  CR = $0D
  LF = $0A

  ' Define the time, in seconds, each color depth display is shown.
  ' If this is zero, the user needs to press a key (not shift :) to tell
  ' the code to advance to the next color depth. This is provided so that
  ' the display at each color depth can be examined. If you only want the
  ' performance numbers, set this to one.
  DISPLAY_SECONDS = 1

  ' Save some typing later, by reflecting a few constants here.
  DISPLAY_ROWS       = Config#DISPLAY_ROWS
  REFRESH_ROWS       = Config#REFRESH_ROWS
  DISPLAY_COLUMNS    = Config#DISPLAY_COLUMNS
  MAX_DISPLAY_ROW    = Config#MAX_DISPLAY_ROW
  MAX_DISPLAY_COLUMN = Config#MAX_DISPLAY_COLUMN

  DC8 = DISPLAY_COLUMNS / 8

  RED_BIT            = Config#RED_BIT
  GRN_BIT            = Config#GREEN_BIT
  BLU_BIT            = Config#BLUE_BIT

  ' Set the maximum size of the display buffer.
  ' This should allow a "reasonable" amount of stack space for this program.
  ' This will limit the number of layers that can be initialized for the
  ' display buffer.
  '
  ' The display buffer memory is not a limitation on the number of color
  ' layers that can be tested, since we don't actually need a display to
  ' be present just to collect refresh sysclk counts.
  '
  ' We'd like to be able to show a 5-clock-per-pixel image for a 128x64
  ' display, which needs 20KB for the display buffer.
  ' BST works with 26KiB, PropTool 1.3.2 only 24KiB
  ' DISPLAY_BUFFER_SIZE_BYTES = 1024 * 24
  ' Constants for stack tracking.

  ' The address of the location that stores the base address of the stack,
  ' which runs up form there to $7FFF
  STACK_BASE_ADDR = $000A
  END_OF_STACK    = Init#END_OF_STACK  ' $0000_8000

  ' This is the distance from the last local variable address in the
  ' function that fills main memory that is safe to fill with the
  ' UNLIKELY_VALUE.
  OFFSET_FOR_MAIN_STACK_FILL = Init#OFFSET_FOR_MAIN_STACK_FILL  ' $0010
  UNLIKELY_VALUE             = Init#UNLIKELY_VALUE   ' $DEADBEEF

VAR
  ' This receives the address of the display buffer, which unblocks the
  ' continued operation of the display refresh PASM cog.
  LONG    DisplayBufAddr

  ' This receives the number of sysclks used for one refresh pass of the
  ' display.
  LONG    RefreshSysClks

  WORD    SerialCog   ' 0 if not started, else CogID + 1

  ' We allocate one buffer from the free space:
  ' - Display Buffer
  ' It is allocated from a combination of reserved space and reuse of the
  ' initialization code including the PASM cog code from Hub memory.
  ' Since the largest buffer is the display buffer, and it (usually) is much
  ' larger than the init code that is recovered, we don't need to dynamically
  ' allocate the serial driver buffers, which makes init a bit simpler.
  WORD    BufferInfo [1 * 2]  ' one buffer to allocate.

' Define the variables that contain the operating parameters.
' We use a DAT section so we can initialize them, which means there can be only
' one instance of this object; if we used a VAR section, we'd have to devote
' code to initializing the VARs, which might happen later, depending on how the
' code develops.
DAT
' Because they are "small" we don't dynamically allocate the serial object
' buffers.
Rx_Buffer          LONG    0[(RX_BUFFER_SIZE + 3) >> 2]
Tx_Buffer          LONG    0[(TX_BUFFER_SIZE + 3) >> 2]

' The Init Colors here are used to initialize the display buffer before it is
' first displayed. The code that fills it in "knows" the memory layout for
' the display buffer, and fills in a different region of the display buffer
' for each color bit layer, so each region shows a different intensity for
' all the colors.
DisplayBufferInitColors
                BYTE    ((                          0) << 3) | (                          0)  ' Black
                BYTE    ((RED_BIT                    ) << 3) | (RED_BIT                    )  ' Red
                BYTE    ((          GRN_BIT          ) << 3) | (          GRN_BIT          )  ' Green
                BYTE    ((                    BLU_BIT) << 3) | (                    BLU_BIT)  ' Blue
                BYTE    ((RED_BIT           | BLU_BIT) << 3) | (RED_BIT           | BLU_BIT)  ' Magenta
                BYTE    ((RED_BIT | GRN_BIT          ) << 3) | (RED_BIT | GRN_BIT          )  ' Yellow
                BYTE    ((          GRN_BIT | BLU_BIT) << 3) | (          GRN_BIT | BLU_BIT)  ' Cyan
                BYTE    ((RED_BIT | GRN_BIT | BLU_BIT) << 3) | (RED_BIT | GRN_BIT | BLU_BIT)  ' White

VAR
  ' Storage for the sysclk counts for one full refresh, for each COLOR_DEPTH.
  LONG    RefreshClocks[9]  ' 1..8, plus zero for simplicity

  ' The maximum number of color layers for which we can write to the
  ' display buffer -- this might be smaller that the color depth being tested.
  LONG    MaxColorDepthForInit

' ========================================================================
PUB Main | avail, left, needed, firstfree
  ' Collect some numbers for assessing memory use.
  avail := Init.AvailableBytes
  firstfree := Init.GetNextFree

  ' First-stage init:
  ' - Collect the digit bitmaps
  ' - reserve memory for the (large-ish) buffers
  ' - load the PASM cogs.
  ' After this call, the init code should no longer be needed, so the reserved
  ' buffers can be overwritten and the objects using the PASM cogs can finish
  ' their initializations.
  needed := Init.Init(@SerialCog, @BufferInfo, Ser.GetInitVars,       {
         }            @DisplayBufAddr, @RefreshSysClks)
  left := Init.AvailableBytes
  Ser.FinishInit(serialcog, @Rx_Buffer, RX_BUFFER_SIZE, @Tx_Buffer, TX_BUFFER_SIZE)

  ' How did we do with buffer reservations?
  ReportHex(string("VBASE: "), WORD[8])
  ReportDec(string("Initial memory (bytes):   "), avail)
  ReportDec(string("Recovered from init code: "), avail - Init#RESERVED_SIZE_BYTES)
  ReportHex(string("First free:               "), firstfree)
  ReportHex(string("Display buffer size:      "), BufferInfo[0])
  ReportHex(string("Display buffer address:   "), BufferInfo[1])
  ReportHex(string("Memory needed:            "), needed)
  ReportDec(string("Memory remaining (bytes): "), left)

  if needed > 0
    ' We were not able to get as much buffer memory as we wanted.
    ReportStr(string("Not enough memory for Display Buffer"))
    ReportDec(string("Memory needed (bytes):    "), needed)
  else
    ' We have a display buffer -- tell the rest of the code the
    ' max COLOR_DEPTH for buffer init.
    MaxColorDepthForInit := BufferInfo[0] / constant(DISPLAY_COLUMNS * REFRESH_ROWS)
    ' We have our buffer -- continue with initialization and run all the tests.
    Main2

  ' Here when done.
  CheckMainStack  ' So we can see if we can shrink the main stack, and increase
                  ' the amount of memory for dynamic buffers.
  repeat
    ' wait forvever.

PUB Main2 | depth, buf, dcog, r, r10, max_layers
  ReportStr(string("Testing refresh rate for all color depths"))
  ReportDec(string("Columns: "), DISPLAY_COLUMNS)
  ReportDec(string("Rows:    "), DISPLAY_ROWS)

  ' Calculate the maximum number of color layers we can safely initialize
  ' in the available display buffer.
  'max_layers := DISPLAY_BUFFER_SIZE_BYTES / (DISPLAY_COLUMNS * REFRESH_ROWS)
  ReportDec(string("Max color layers in buffer: "), MaxColorDepthForInit)
  buf := BufferInfo[1]
  'LONGFILL(buf, 0, BufferInfo[0] >> 2)

  repeat depth from 1 to 8
    ' Run a single test:
    ' - Report test starting for depth value and dimensions (cols x rows)
    ' - Start the display cog
    ' - Finish the display cog init
    ' - Capture the refresh rate.
    ' - Stop the display cog.

    ReportDec(string(CR, "COLOR_DEPTH: "), depth)
    dcog := StartDisplayCog(depth)
    ' We may want to change this to pass both the test depth, and the max for
    ' init, since the layout on the screen for different test depths may be
    ' different.
    InitDisplayBuffer(buf, depth <# MaxColorDepthForInit)
    DisplayBufAddr := buf  ' unblock the display refresh cog.
    ReportDec(string("Started display cog: "), dcog - 1)

    ' Collect the refresh rate.
    DelaySec(DISPLAY_SECONDS)
    ' Wait for the refresh clocks value to be available -- probably not
    ' needed if we are waiting above.
    repeat while RefreshSysClks == 0
      ' Stall, waiting for the count to be published.
    RefreshClocks[depth]  := RefreshSysClks
    ShowRefreshRate(RefreshSysClks, depth)

    COGSTOP(dcog~ - 1)
    ReportStr(string("Display cog stopped"))

  ' We have run all the tests, report the results in a table.
  OUTA[Config#OE_INVERTED_PIN] := 1  ' Disable the display
  DIRA[Config#OE_INVERTED_PIN] := 1
  ReportStr(string("Summary:", CR, "Depth   Rate"))
  repeat depth from 1 to 8
    r := Umath.multdiv_rounded(CLKFREQ, 10, RefreshClocks[depth])
    Ser.Str(string("  "))
    Ser.Dec( depth)
    Ser.Str(string("    "))
    r10 := r / 10
    ' Line up the decimal points.
    if r10 < 10000
      Ser.Tx(" ")
      if r10 < 1000
        Ser.Tx(" ")
        if r10 < 100
          Ser.Tx(" ")
          if r10 < 10
            Ser.Tx(" ")
    Ser.Dec(r / 10) ' Fixed width output would be nice...
    Ser.Tx(".")
    Ser.Tx("0" + (r // 10)) ' One digit to the right of the decimal point.
    Ser.Tx(CR)

PUB ShowRefreshRate(clks, depth)
  ' Wait for the first report of the refresh time.
  ReportDec(string("One refresh, predicted   sysclks: "), PanelPasm.EstimateRefreshClocks(depth))
  ReportDec(string("One refresh, approximate sysclks: "), clks)
  ReportTenths(string("Refresh rate, per second: "), {
    }          Umath.multdiv_rounded(CLKFREQ, 10, clks))

PUB DelaySec(s)
{{
  Pause for a specified number of seconds, or until a key press is received
  from the serial port.

  Arguments
    s -- seconds to pause; if zero, pause until a byte is received from the
         serial port

  Return value
    None
}}
  if s
    waitcnt(CNT + (CLKFREQ * s))
  else
    Ser.Rx  ' Stall, and discard the character

PUB StartDisplayCog(depth)
  DisplayBufAddr := 0
  RefreshClocks := 0
  return PanelPasm.LoadDisplayCog(depth, @DisplayBufAddr, @RefreshSysClks)

' ========================================================================

PUB InitDisplayBuffer(bp, depth) | ls, w
  ' Fill in the display buffer with a known color pattern, divided into
  ' separate regions for each color bit layer.
  ' - For one bit, the entire display is uniform
  ' - For two bits, the top half is dim, and the bottom is bright
  ' - For three and four bits, the display is in quadrants.
  ' - For five to eight bits, the display is in eight parts: four on the top,
  '   and four on the bottom.
  ls := DISPLAY_COLUMNS * REFRESH_ROWS  ' Layer Size, in bytes
  LONGFILL(bp, 0, (ls >> 2) * depth)
  case depth
    1:
      ' Top and bottom half are the same colors for the one layer
      FillLayer(bp, %111_111, DISPLAY_COLUMNS)
    2:
      ' Top half is one layer, bottom half is the other
      FillLayer(bp + (ls * 0), %000_111, DISPLAY_COLUMNS)
      FillLayer(bp + (ls * 1), %111_000, DISPLAY_COLUMNS)
    3, 4:
      'First two layers are on top, next two are on the bottom
      w := DISPLAY_COLUMNS >> 1
      FillLayer(bp               , %000_111, w)
      FillLayer(bp +  ls      + w, %000_111, w)
      FillLayer(bp + (ls * 2)    , %111_000, w)
      if depth > 3
        FillLayer(bp + (ls * 3) + w, %111_000, w)
    5..8:
      ' First four layers are on top, the rest on the bottom
      w := DISPLAY_COLUMNS >> 2
      FillLayer(bp                 , %000_111, w)
      FillLayer(bp +  ls      + w  , %000_111, w)
      FillLayer(bp + (ls * 2) + w*2, %000_111, w)
      FillLayer(bp + (ls * 3) + w*3, %000_111, w)
      FillLayer(bp + (ls * 4)      , %111_000, w)
      if depth > 5
        FillLayer(bp + (ls * 5) + w  , %111_000, w)
        if depth > 6
          FillLayer(bp + (ls * 6) + w*2, %111_000, w)
          if depth > 7
            FillLayer(bp + (ls * 7) + w*3, %111_000, w)

PUB FillLayer(bp, mask, width) | r
  ' Fill the layer for width columns starting at bute address bp.
  ' This always writes REFRESH_ROWS rows, repeatedly using the masked
  ' values from DisplayBufferInitColors
  repeat r from 0 to constant(REFRESH_ROWS - 1)
    BYTEFILL(bp, BYTE[@DisplayBufferInitColors + (r & 7)] & mask, width)
    bp += DISPLAY_COLUMNS


' ========================================================================

PUB CheckStack(n, p, l) | lused, pend
  ' Examine the stack starting at p, that contains L longs, and determine
  ' how much of it was used, and how much was untouched.
  pend := p + (l << 2) - 4
  repeat while (pend => p) and (long[pend] == UNLIKELY_VALUE)
    pend -= 4

  ' Here pend is either p - 4, or the address of the last value
  ' used.
  '   longs_used = ((pend + 4) - p) >> 2
  '   longs_free = l - longs_used
  lused := ((pend + 4) - p) >> 2
  Ser.Str(n)
  ReportDec(string(" used longs: "), lused)
  Ser.Str(n)
  ReportDec(string(" free longs: "), l - lused)

PUB CheckMainStack
  CheckStack(string("Main"), WORD[STACK_BASE_ADDR], (END_OF_STACK - WORD[STACK_BASE_ADDR]) >> 2)

' ========================================================================
PUB ReportStr(s)
  Ser.Str(s)
  Ser.Tx(CR)

PUB ReportDec(label, v)
  Ser.Str(label)
  Ser.Dec(v)
  Ser.Tx(CR)

PUB ReportHex(label, v)
  Ser.Str(label)
  Ser.Hex(v, 8)
  Ser.Tx(CR)

Pub ReportTenths(label, v)
  Ser.Str(label)
  Ser.Dec(v / 10)
  Ser.Tx(".")
  Ser.Tx("0" + (v // 10))
  Ser.Tx(CR)

' ========================================================================


