{{
Test a clock display, using digit bitmaps, for one or more HUB75 displays.

WARNING: Only adapted to new serial and memory system down to line 865 or so!

Author:         Tom Dinger (Ding-Batty)
Release Date:   2022/04/10

Copyright 2024 Thomas E. Dinger

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

--------------------------------------------------------------------------

This is a demonstration/test program for a digital clock that can use one or
more daisy-changes HUB75 displays of 32 rows or more. It can show either a
four-digit display (HH:MM) or six digit display (HH:MM:SS) depending on the
dimensions of the display(s) connected. It uses a copy of the ROM font for
the display, and ca use all of the available digit change animations chosen
pseudo-randomly for each digit change.

The digit bitmaps are captured from the ROM Fonts, but the program could be
modified to use different bitmaps in other character dimensions. The ROM font
digits are 13x22 on a 16x32 field.

Cog use
=======
Cog 0 -- the main program: launches all the other Cogs, prepares the state for
         digit display; collects and processes the single-character commands;
         reports memory and stack use.
Cog 1 -- the Serial port driver cog
Cog 2 -- the HUB75 display refresh cog
Cog 3 -- the display buffer write helper cog
Cog 4 -- the "resolution" cog: watches for differences between the current
         display state (digit values, colon state) and the desired state, and
         effects the changes needed. Generally, the only cog that writes to
         the display buffer via the write helper. This cog "owns" the digit
         change animations.
Cog 5 -- the timing cog -- measures time in half-second increments (to animate
         the colon that turns off then on once a second), and updates the
         desired state of the digits for the display, and is the only cog
         that changes the display state while the clock is running. This is
         disabled when any command is going to change the display state:
         time, color, intensity, mode (12/24), animation style.


Revisions:
==========
- 2024/04/10 -- Change from 8-bit color max, to 10-bit color max;
                change for FDS in the init code;
- 2024/02/28 -- Original release

}}

OBJ
  ' Pick up the static display configuration first.
  Config: "hub75_config.spin"

  Ser    : "Db_fullDuplexSerial"  ' uses a cog
  Umath  : "umath2.spin"
  ' The FiFo is used to enqueue write commands for the Helper to send to
  ' a dedicated PASM writer cog that updates the Display buffer.
  Fifo   : "FifoBig.spin"
  ' The Panel is the wrapper for the PASM cog that send the data to the
  ' display, and contains Spin-based write routines, such as Point, Box,
  ' HorizontalLine, etc. 
  Panel  : "hub75_display.spin"
  ' The Helper contains wrapper functions for writing to the display buffer
  ' by sending the commands to a helper PASM cog, in order to have faster
  ' display updates. It has the same write functions, e.g. Point, Box, etc.
  Helper : "hub75_helper.spin"

  ' The init object (below) and all objects it includes, will be used as
  ' additional buffer allocation space after they are finished during init.
  ' This helps to maximize hub memory available for large buffer use.
  Init : "test_clock_bitmap_init.spin"

CON { General constants }
  _CLKMODE = XTAL1 + PLL16X 
  _XINFREQ = 5_000_000

  { Configuration for serial connection. }
  BAUDRATE = Init#BAUDRATE  ' 115_200
  RXPIN    = Init#RXPIN     ' 31    ' The usual pins.
  TXPIN    = Init#TXPIN     ' 30
  ' For this version of FDS, the sizes do not need to be powers-of-two. The
  ' Spin code is a little slower, but the PASM code is not slowed by this.
  RX_BUFFER_SIZE = 28
  TX_BUFFER_SIZE = 200

  CR = $0D
  LF = $0A

  { LED panel(s) configuration }
  COLOR_DEPTH     =  Init#COLOR_DEPTH
                            ' Number of bits of resolution per color, up to
                            ' 8 bits, which is the limit for the implementation
                            ' details. The limits are available memory and
                            ' flickering from slower refresh rates.
  MAX_COLOR_VALUE = (|< COLOR_DEPTH) - 1

  ' Save some typing later, by reflecting a few constants here.
  DISPLAY_ROWS       = Config#DISPLAY_ROWS
  REFRESH_ROWS       = Config#REFRESH_ROWS
  DISPLAY_COLUMNS    = Config#DISPLAY_COLUMNS
  MAX_DISPLAY_ROW    = Config#MAX_DISPLAY_ROW
  MAX_DISPLAY_COLUMN = Config#MAX_DISPLAY_COLUMN

  RED_BIT            = Config#RED_BIT
  GRN_BIT            = Config#GREEN_BIT
  BLU_BIT            = Config#BLUE_BIT  

  ' Instead of a single mailbox, we use a FiFo of commands (a mailbox queue)
  ' to enqueue multiple commands for the display buffer write helper cog to
  ' execute for the main program. We have a large queue because for some
  ' operations, we want the updates to go as fast as possible with little
  ' time for stalling.
  '
  ' If we need even faster display update, we could use multiple writer
  ' cogs, each with a separate command FiFo (or even a shared FiFo with the
  ' proper locks) to process write requests in parallel. There can be
  ' write interactions, however, because display buffer updates are
  ' read-modify-write, and are not locked, so using two writers is best
  ' avoided.  
  COMMAND_FIFO_SIZE = 32

  ' For some display tests showing and changing digits, we need bitmaps for
  ' the digits themselves. The ROM stores two characters interleaved in one
  ' group of 32 LONG values, but for our purposes it is better to store the
  ' digits separately. We also can treat a column of digits as an extended
  ' height bitmap, which can be used for a column-of-digits scrolling effect
  ' with very little code.
  '
  ' We extract the digits from the ROM Font. They are 16 x 32, but some of the
  ' rows and columns are blank, so we only keep 13 x 22 -- every digit (and
  ' the capital letter "O", used for zero) fits into 13-pixels wide.
  { Digit "shapes" }
  DIGIT_WIDTH  = Init#DIGIT_WIDTH   ' 13
  DIGIT_HEIGHT = Init#DIGIT_HEIGHT  ' 22

  ' The number of display columns for the colon between time fields.
  COLON_WIDTH = 2

  ' Values for the ColonState
  COLON_OFF = 0
  COLON_ON = 1

  ' Calculate some limits on digit display
  '
  ' For a 4-digit display on a 64-column display, we need 52 columns just for
  ' the digits (without considering the spacing) plus at least 2 columns for
  ' a colon in the HH:MM display. For wider displays we will most likely have
  ' plenty of display columns.
  '
  ' 64 - ((13 * 4) + 2) == 10, and we need 4 separations, so we can use a two-
  ' column spacing between digits, and 6 columns (2 + 2 + 2) for the colon.
  ' This will use:
  '   13 + 2 + 13 + 6 + 13 + 2 + 13 == 62 columns.
  ' So, for a 4-digit display on a 64-column panel, we'd have one pixel column
  ' blank on each side of the display.
  SEPARATOR_WIDTH       = 2
  DIGIT_COLUMN_DELTA    = DIGIT_WIDTH + SEPARATOR_WIDTH
  MAX_DIGIT_COLUMNS     = DISPLAY_COLUMNS / DIGIT_COLUMN_DELTA

  ' Center the row of digits vertically
  DIGIT_TOP_ROW = (DISPLAY_ROWS - DIGIT_HEIGHT) / 2

  ' For scrolling alternate pixel columns in opposite directions, we need a
  ' pair of bitmap masks to skip every other bit.
  SKIP_UP   = %0101_0101_0101_0101_0101_0101_0101_0101
  SKIP_DOWN = ! SKIP_UP
  SKIP_NONE = 0

  ' Styles for the display.
  ' This specifies the layout and update style of the digit display.
  ' This test program doesn't use the ROM font routines.
  ' STYLE_1ROW_ROM_WIPE_RANDOM   = 1  ' Select each digit animation randomly from
  '                                   ' the ones below.
  ' 
  ' MIN_STYLE_1ROW_ROM_WIPE            =  2
  ' STYLE_1ROW_ROM_REPLACE             =  2  ' immediate replacement
  ' STYLE_1ROW_ROM_WIPE_DOWN           =  3
  ' STYLE_1ROW_ROM_WIPE_UP             =  4
  ' STYLE_1ROW_ROM_WIPE_RIGHT          =  5
  ' STYLE_1ROW_ROM_WIPE_LEFT           =  6
  ' STYLE_1ROW_ROM_WIPE_TB2CTR         =  7
  ' STYLE_1ROW_ROM_WIPE_TB_F_CTR       =  8
  ' STYLE_1ROW_ROM_WIPE_LR_2_CTR       =  9
  ' STYLE_1ROW_ROM_WIPE_LR_F_CTR       = 10
  ' STYLE_1ROW_ROM_WIPE_F_CTR          = 11
  ' STYLE_1ROW_ROM_WIPE_TO_CENTER      = 12
  ' ' "Scrolls" are variations on "Wipes"
  ' STYLE_1ROW_ROM_SCROLL_LEFT         = 13
  ' STYLE_1ROW_ROM_SCROLL_RIGHT        = 14
  ' STYLE_1ROW_ROM_SCROLL_UP           = 15
  ' STYLE_1ROW_ROM_SCROLL_DOWN         = 16
  ' MAX_STYLE_1ROW_ROM_WIPE            = 16

  STYLE_1ROW_BMP_WIPE_RANDOM         = 17  ' Select each digit animation randomly from
                                              ' the ones below.

  MIN_STYLE_1ROW_BMP_WIPE            = 18
  STYLE_1ROW_BMP_REPLACE             = 18  ' immediate replacement
  STYLE_1ROW_BMP_WIPE_DOWN           = 19
  STYLE_1ROW_BMP_WIPE_UP             = 20
  STYLE_1ROW_BMP_WIPE_RIGHT          = 21
  STYLE_1ROW_BMP_WIPE_LEFT           = 22
  STYLE_1ROW_BMP_WIPE_TB2CTR         = 23
  STYLE_1ROW_BMP_WIPE_TB_F_CTR       = 24
  STYLE_1ROW_BMP_WIPE_LR_2_CTR       = 25
  STYLE_1ROW_BMP_WIPE_LR_F_CTR       = 26
  STYLE_1ROW_BMP_WIPE_F_CTR          = 27
  STYLE_1ROW_BMP_WIPE_TO_CENTER      = 28
  ' "Slides" are variations on "Wipes"
  STYLE_1ROW_BMP_SLIDE_DOWN          = 29
  STYLE_1ROW_BMP_SLIDE_UP            = 30
  STYLE_1ROW_BMP_SLIDE_LEFT          = 31
  STYLE_1ROW_BMP_SLIDE_RIGHT         = 32
  ' "Scrolls" are variations on "Wipes"
  STYLE_1ROW_BMP_SCROLL_LEFT         = 33
  STYLE_1ROW_BMP_SCROLL_RIGHT        = 34
  STYLE_1ROW_BMP_SCROLL_UP           = 35
  STYLE_1ROW_BMP_SCROLL_DOWN         = 36
  STYLE_1ROW_BMP_SCROLL_UP_DN        = 37

  STYLE_1ROW_BMP_FADE_TO_BLACK       = 38
  STYLE_1ROW_BMP_FADE_TO_WHITE       = 39
  MAX_STYLE_1ROW_BMP_WIPE            = 39

  ' Some helper constants for a few styles:
  '
  ' For WIPE_FROM_CENTER and WIPE_TO_CENTER, we precompute the difference
  ' between the DIGIT_WIDTH and HEIGHT, always assuming that
  ' DIGIT_WIDTH <= DIGIT_HEIGHT, so we calculate the additional number of
  ' rows for the ((DIGIT_WIDTH + 1) >> 1) column updates.
  ' Which would be:
  '    extra_rows := DIGIT_HEIGHT - (((DIGIT_WIDTH + 1) >> 1) << 1)
  ' extra_rows will always be => 0 if DIGIT_HEIGHT => DIGIT_WIDTH.
  ' Then we calculate the starting and ending column, and the starting and
  ' ending row from the column range by adding the extra_rows.
  ' Then, we just draw the "center" of the new digit over the old.
  ' This will redundantly write the center as it grows, but it is faster than
  ' doing four writes (I think).
  WIPE_BOTH_EXTRA_ROWS = DIGIT_HEIGHT - (((DIGIT_WIDTH + 1) >> 1) << 1)
  
  ' For scrolling, we need to decide how many blank rows or columns to use
  ' between the outgoing and incoming digits.
  DIGIT_SCROLL_H_SPACING = 2  ' two columns between scrolling digits
  DIGIT_SCROLL_V_SPACING = Init#DIGIT_SCROLL_V_SPACING   ' 3  ' three rows between scrolling digits.

  ' Counting direction.
  ' Changing direction of counting for some display styles can be complicated.
  COUNT_DIRECTION_UP   = 0
  COUNT_DIRECTION_DOWN = 1

  ' For fades, this is the shift to use to make the fade the right length
  ' when using the random digit animation, to cooperate with the animation
  ' delays of animations with a lot more steps.
  ' Strictly speaking, this should vary with the current Intensity value,
  ' not the MaxIntensity value -- but we start with this.
  '   Color Bits   Max Fade steps   Scaling (bits)   Maybe 
  '       1              2           => 8 (3)         16 (4)
  '       2              4           => 4 (2)          8 (3)
  '       3              8           => 2 (1)          4 (2)
  '       4             16           => 1 (0)          2 (1)
  '       5+            32+             1 (0)          1 (0)

  FADE_ANIMATION_SHIFT = ((5 - COLOR_DEPTH) #> 0)

  ' Constants for stack tracking.

  ' The address of the location that stores the base address of the stack,
  ' which runs up form there to $7FFF
  STACK_BASE_ADDR = $000A
  END_OF_STACK    = Init#END_OF_STACK  ' $0000_8000

  ' This is the distance from the last local variable address in the
  ' function that fills main memory that is safe to fill with the
  ' UNLIKELY_VALUE.
  OFFSET_FOR_MAIN_STACK_FILL = Init#OFFSET_FOR_MAIN_STACK_FILL  ' $0010
  UNLIKELY_VALUE             = Init#UNLIKELY_VALUE              ' $DEADBEEF

  ' Stack sizes for the Spin-based cogs
  ' Once we have measurements for how much the cogs actually use, we may be
  ' able to shrink these.

  ' First measurement of stack use suggests the resolution cog needs at least
  ' 79 longs.
  RESOLUTION_COG_STACK_LONGS = 100
  ' The timing cog used about 45 longs.
  TIME_COG_STACK_LONGS       =  64

VAR
  ' Variables used in Spin, that are initialized at runtime (not by the compiler)

  ' These must be consecutive in this order, so that the Init object can fill
  ' them in.
  ' We use other cogs as well but they run Spin code, not PASM.
  WORD    SerialCog   ' 0 if not started, else CogID + 1
  WORD    DisplayCog  ' 0 if not started, else CogID + 1
  WORD    HelperCog   ' 0 if not started, else CogID + 1

  ' We allocate two buffers from the free space:
  ' - Command FiFo for display write-helper cog
  ' - Display Buffer
  ' They are allocated from a combination of reserved space and reuse of the
  ' initialization code including the PASM cog code from Hub memory.
  ' Since the largest buffer is the display buffer, and it (usually) is much
  ' larger than the init code that is recovered, we don't need to dynamically
  ' allocate the serial driver buffers, which makes init a bit simpler.
  WORD    BufferInfo [2 * 2]  ' two buffers to allocate.

  ' Display management
  ' The approach to display update is to have the display update routine maintain
  ' the current state of the display, including partial updates, and have
  ' the other code set the "target" or desired "final" state of the display.
  '
  ' The display update routine "owns" the current display state variables, and
  ' updates them as the only writer for them. Other code may read them, but as
  ' they can change at any time, they are advisory.
  '
  ' The target display state is only read by the display update code, and can
  ' change at any time during update, so the display update code has to be able
  ' to handle changes at arbitrary times.

  ' The leftmost pixel column for each digit (or column of digits).
  ' Set when the display layout is initialized, based on the desired number
  ' of digit positions on the display.
  WORD    DigitColumn [MAX_DIGIT_COLUMNS]

  ' There are either one or two colons for the time display.
  WORD    ColonColumn [2]

  ' These describe the position of the "middle" row of digits: the display
  ' row for the top of the digits, and the height of the digits.
  BYTE    CenterRowTop
  BYTE    CenterRowHeight

  ' The style of display: how to update the display, the number of digits,
  ' and the direction for counting-style scrolling (not in this version of
  ' the test program).
  BYTE    DisplayStyle
  BYTE    DigitCount
  'BYTE    CountDirection
  ' The number of colons to show; 4 digit display has 1, 6 digit has 2.
  BYTE    ColonCount

  ' The list of the currently displayed values for each digit.
  ' The values may be 0..N or ASCII codes, depending on the style of
  ' update animation.
  BYTE    CurrentDigitValue [MAX_DIGIT_COLUMNS]
  BYTE    CurrentColonState [2]

  ' The list of the values that should be displayed for the digits.
  ' The display update cog will make the display changes based on the
  ' transformation of the current value displayed to the target value.
  BYTE    TargetDigitValue [MAX_DIGIT_COLUMNS]
  BYTE    TargetColonState [2]

  ' The following are the working (private) storage for managing the digit
  ' display updates.
  '
  ' We can have a different update style for each digit
  BYTE    WorkingDisplayStyle [MAX_DIGIT_COLUMNS]
  ' The target digit value, captured to keep it constant during the update
  BYTE    WorkingDigitValue [MAX_DIGIT_COLUMNS]
  ' The following tracks the progress through the digit update animation.
  LONG    WorkingChangeProgress [MAX_DIGIT_COLUMNS]

  ' The number of sysclks to delay per pixel row update for digit change
  ' animations. This will be around the value of CLKFREQ * 0.6 / 28
  LONG    AnimationDelay
  LONG    AnimationDelaySave
  ' For random selection of Wipe style per digit change.
  LONG    WipeRandom

  ' Variables for managing the display style: key presses can be used to cycle
  ' through a list of display styles.
  LONG    DisplayStyleIndex

  LONG    BoldColorRgb       ' current intensity
  LONG    BoldColorDisplay   ' Filled in at runtime.
  'LONG    MaxColorRgb        ' Max intensity.
  LONG    ColorIndex         ' For cycling through colors in ColorGradient[].
  LONG    Intensity

  ' In order to do fades efficiently, for a given color (ratio of R:G:B), we
  ' precompute the display colors based on intensity level, so the animation
  ' code only needs to do a lookup for each animation step.
  LONG    ColorGradient[MAX_COLOR_VALUE + 1]

  ' Time tracking storage
  LONG    Hours     ' 0..23, converted to 12-hour format when displayed, if needed
  LONG    Minutes   ' 0..59
  LONG    Seconds   ' 0..59

  ' Flags to control the time update cog:
  '
  ' TimeUpdateRunning is true if the time updating cog is running and updating
  ' the time and digit states, so no one else should make similar changes.
  '
  ' If TimeUpdateEnabled is true, the time updating cog will run, and make
  ' changes to the digits state. If it is false, the time update cog will
  ' (eventually) stop, so others can make time changes.
  BYTE    TimeUpdateEnabled
  BYTE    TimeUpdateRunning

DAT
  ' We allocate the storage for the command FiFo entries dynamically,
  ' overlaying the PASM (and Spin) code at the end of the program, which is
  ' only used during early initialization.
  CmdFifoDataAddr   LONG    0

  ' Storage for debug output from the helper driver
  ' See that driver for details on the contents.
  ' Should be at least 10 longs
  HelperDebugOutput LONG    0[16]

  ' Our 1bpp bitmap for a colon. 22 display rows, 2 display columns
  ColonBitmap       BYTE    %00
                    BYTE    %00
                    BYTE    %00
                    BYTE    %00
                    BYTE    %00
                    BYTE    %11
                    BYTE    %11
                    BYTE    %00
                    BYTE    %00
                    BYTE    %00
                    BYTE    %00
                    BYTE    %00
                    BYTE    %00
                    BYTE    %00
                    BYTE    %00
                    BYTE    %11
                    BYTE    %11
                    BYTE    %00
                    BYTE    %00
                    BYTE    %00
                    BYTE    %00
                    BYTE    %00

  ' Make it easier to convert 0..59 into two digits without multiply/divide.
  ' The value will have the tens digit in the high nibble, and the units in the
  ' low nibble. We could extend this for 0..99, if needed.
  TwoDigits         BYTE    $00, $01, $02, $03, $04, $05, $06, $07, $08, $09
                    BYTE    $10, $11, $12, $13, $14, $15, $16, $17, $18, $19
                    BYTE    $20, $21, $22, $23, $24, $25, $26, $27, $28, $29
                    BYTE    $30, $31, $32, $33, $34, $35, $36, $37, $38, $39
                    BYTE    $40, $41, $42, $43, $44, $45, $46, $47, $48, $49
                    BYTE    $50, $51, $52, $53, $54, $55, $56, $57, $58, $59

  ' Define a table of display colors that we will cycle through to select the
  ' current display color at the current intensity.
  ' We define the colors using 10 bits per color, and let the intensity calculation
  ' handle the scaling details.
  ColorTableSize    LONG     7
  '                                  R                       G                       B
  ColorTable
  ColorTableRed     LONG    %%0_33333_00000_00000
  ColorTableGreen   LONG                            %%0_00000_33333_00000
  ColorTableBlue    LONG                                                    %%0_00000_00000_33333
  ColorTableYellow  LONG    %%0_33333_00000_00000 | %%0_00000_33333_00000
  ColorTableMagenta LONG    %%0_33333_00000_00000                         | %%0_00000_00000_33333
  ColorTableCyan    LONG                            %%0_00000_33333_00000 | %%0_00000_00000_33333
  ColorTableWhite   LONG    %%0_33333_00000_00000 | %%0_00000_33333_00000 | %%0_00000_00000_33333
  ' Other colors are also possible, by varying the ratios.

  MaxIntensity      LONG     MAX_COLOR_VALUE
  DefaultIntensity  LONG     (MAX_COLOR_VALUE + 1) >> 1  ' roughly 50% intensity

  ' Digit storage
  ' We capture the bitmaps of the digits from the ROM font (and we can replace
  ' the "0" with the "O" if we don't want a slash in the zero), which will allow
  ' a few more interesting animations that just using the ROM font.
  ' We include space for blank rows between digits, that are used when doing
  ' vertical scroll animations.
  ' To map from 0..9 to digit, to get the bitmap address:
  '    @Digits_0[d * DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING]
  ' To map from "0".."9" to digit:
  '    @Digits_0[(d - "0") * DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING]
  ' 
  ' Note that this is filled in by the Init object, which has hard-coded the
  ' digit layout in this memory.
  Digits10          WORD    0 [               DIGIT_SCROLL_V_SPACING]
  Digits_0          WORD    0 [DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING]
  Digits_1          WORD    0 [DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING]
  Digits_2          WORD    0 [DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING]
  Digits_3          WORD    0 [DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING]
  Digits_4          WORD    0 [DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING]
  Digits_5          WORD    0 [DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING]
  Digits_6          WORD    0 [DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING]
  Digits_7          WORD    0 [DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING]
  Digits_8          WORD    0 [DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING]
  Digits_9          WORD    0 [DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING]
  ' A "blank" space, used instead of leading zeros in some cases.
  Digits_Space      WORD    0 [DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING]

  ' Controls the hours format for the display.
  TwelveHour        LONG    1

  ' The stack for the cog that resolves differences between the current state of
  ' the display and the target state.
  ' Start out with 64 entries -- we can probably do with many fewer, perhaps as
  ' few as 16 -- we may need a bunch for arguments to the display write helper
  ' cog, as some of the methods take a lot of arguments.
  ResolutionCog      LONG    0   ' 1 + Cog# for the resolution cog
  ResolutionCogStack LONG    0 [RESOLUTION_COG_STACK_LONGS]

  ' The stack for the cog that advances the time every second and updates the
  ' display.
  ' Start out with 64 entries, but probably needs only a few.
  TimeCog           LONG    0   ' 1 + Cog# for the resolution cog
  TimeCogStack      LONG    0 [TIME_COG_STACK_LONGS]

  '' Because they are "small" we don't dynamically allocate the serial object
  '' buffers.
  'Rx_Buffer         LONG    0[(RX_BUFFER_SIZE + 3) >> 2]
  'Tx_Buffer         LONG    0[(TX_BUFFER_SIZE + 3) >> 2]

  ' The display buffer stores all color layers for all displays.
  ' Someday it will overlap the PASM objects that are initilaized (loaded into
  ' cogs) once at the start of operations, to maximize the amount of memory
  ' available for displays.
  '
  ' One byte of memory stores one color bit layer for two pixels.

  DisplayBufferAddr LONG    0   ' Gets a dynamically-allocated buffer

' ========================================================================
PUB Main | avail, left, needed, firstfree
  ' Collect some numbers for assessing memory use.
  avail := Init.AvailableBytes
  firstfree := Init.GetNextFree

  ' Also fill in the Spin cog stacks for use-tracking at the end
  LONGFILL(@ResolutionCogStack, UNLIKELY_VALUE, RESOLUTION_COG_STACK_LONGS)
  LONGFILL(@TimeCogStack      , UNLIKELY_VALUE, TIME_COG_STACK_LONGS)

  ' First-stage init:
  ' - Collect the digit bitmaps
  ' - reserve memory for the (large-ish) buffers
  ' - load the PASM cogs.
  ' After this call, the init code should no longer be needed, so the reserved
  ' buffers can be overwritten and the objects using the PASM cogs can finish
  ' their initializations.
  needed := Init.Init(@SerialCog, @BufferInfo, @Digits10, Ser.GetInitVars,       {
         }            Panel.GetDisplayBufAddrVar, Panel.GetRefreshClocksAddrHub, {
         }            Helper.GetCmdFifoHeader, @HelperDebugOutput)
  left := Init.AvailableBytes
  'Ser.FinishInit(SerialCog, @Rx_Buffer, RX_BUFFER_SIZE, @Tx_Buffer, TX_BUFFER_SIZE)
  Ser.FinishSecondInit(SerialCog)

  ' How did we do with buffer reservations?
  ReportHex(string("VBASE: "), WORD[8])
  ReportDec(string("Initial memory (bytes):   "), avail)
  ReportDec(string("Recovered from init code: "), avail - Init#RESERVED_SIZE_BYTES)
  ReportHex(string("First free:               "), firstfree)
  ReportHex(string("FiFo buffer size:         "), BufferInfo[0])
  ReportHex(string("FiFo buffer address:      "), BufferInfo[1])
  ReportHex(string("Display buffer size:      "), BufferInfo[2])
  ReportHex(string("Display buffer address:   "), BufferInfo[3])
  ReportDec(string("Memory needed:            "), needed)
  ReportDec(string("Memory remaining (bytes): "), left)

  if needed > 0
    ' We were not able to get as much buffer memory as we wanted.
    if BufferInfo[1] == 0
      ReportStr(string("Not enough memory for Command FiFo or Display Buffer"))
    else
      ReportStr(string("Not enough memory for Display Buffer"))
    'ReportDec(string("Memory needed (bytes):    "), needed)
    ' TODO: (This might not be possible) Estimate if lowering COLOR_DEPTH would
    ' work, and estimate available memory if the main stack is reduced. This
    ' requires a guess as to how much main stack is needed.
  else
    ' We have our buffers -- continue with initialization and run all the tests.
    Main2

  ' Here when done.
  ReportStackUse
  repeat
    ' wait forvever.

PUB Main2 | c, save_style
{{
  Continue top-level init and operation.
}}
  CmdFifoDataAddr   := BufferInfo[1]
  DisplayBufferAddr := BufferInfo[3]
  ReportDec(string("Started display cog: "), DisplayCog - 1)
  ReportDec(string("Started helper cog: "), HelperCog - 1)

  ' Tell the drivers about the two cogs and allocated buffers.
  ' Note that telling the display driver about the address of the display
  ' buffer will unblock that cog _and_ the helper cog at the same time, so
  ' that one should be done last.
  Helper.FinalHelperInit(HelperCog, DisplayBufferAddr, COLOR_DEPTH, {
           }             CmdFifoDataAddr, {
           }             Fifo#HEADER_SIZE_BYTES * COMMAND_FIFO_SIZE)
  Panel.FinalDisplayInit(DisplayCog, DisplayBufferAddr, COLOR_DEPTH)

  ' Prepare the starting display color table.
  SetColorByIndexAndIntensity((@ColorTableYellow - @ColorTable) >> 2, {
    }                          DefaultIntensity)
  ReportDec(string("Maximum number of digits across: "), MAX_DIGIT_COLUMNS)
  ReportDec(string("Bits per color: "), COLOR_DEPTH)
  WipeRandom := 42  ' Init a random number stream.
  'CountDirection := COUNT_DIRECTION_UP

  ' First, do the layout of the digits and colon(s), depending on the display
  ' dimensions.
  ' We can handle display widths of 64 columns or more, and display heights of
  ' 16 rows or more (I think -- I've only tested with 64 rows).
  PrepareDisplayLayout(STYLE_1ROW_BMP_SLIDE_DOWN)
  'PrepareDisplayLayout(STYLE_1ROW_BMP_WIPE_RANDOM)

  ' Now, once a second, advance the time.
  StartTimeUpdateCog
  TimeUpdateEnabled := 1

  ' The time display is now running.

  ' Wait for a command character, and do it.
  ShowHelpText
  c := -1
  repeat while c <> "X" and c <> "x"
    c := Ser.Rx ' Get command character
    case c
      "d":     ' Debugging
        save_style := PauseUpdates
        DumpDigitParams(0)
        DumpDigitParams(1)
        DumpDigitParams(2)
        DumpDigitParams(3)
        if DigitCount > 4
          DumpDigitParams(4)
          DumpDigitParams(5)
      "D":
        ResumeUpdates(save_style)
      "f":
        DumpFifo(string("Fifo state"))
      "h":   ' Advance the hours position
        save_style := PauseUpdates
        if ++Hours > 23
          Hours := 0
        SetHours(Hours)
        ResumeUpdates(save_style)
      "H":   ' Back up the hours position
        save_style := PauseUpdates
        if --Hours < 0
          Hours := 23
        SetHours(Hours)
        ResumeUpdates(save_style)
      "K":
        ReportStackUse
      "c":   ' increment color
        save_style := PauseUpdates
        SetColorByIndex(ColorIndex + 1)
        Refresh
        WaitForNoDifferences
        ResumeUpdates(save_style)
      "i":   ' increment intensity
        save_style := PauseUpdates
        SetIntensity(Intensity + 1)
        Refresh
        WaitForNoDifferences
        ResumeUpdates(save_style)
      "I":   ' Decrement intensity
        save_style := PauseUpdates
        SetIntensity(Intensity - 1)
        Refresh
        WaitForNoDifferences
        ResumeUpdates(save_style)
      "m":   ' Advance the minutes position (and clear seconds)
        save_style := PauseUpdates
        if ++Minutes > 59
          Minutes := 0
        SetMinutes(Minutes)
        SetSeconds(Seconds := 0)
        ResumeUpdates(save_style)
      "M":   ' Back up the minutes position (and clear seconds)
        save_style := PauseUpdates
        if --Minutes < 0
          Minutes := 59
        SetMinutes(Minutes)
        SetSeconds(Seconds := 0)
        ResumeUpdates(save_style)
      "1":   ' 12-hour mode
        save_style := PauseUpdates
        TwelveHour := 1
        SetHours(Hours)
        ResumeUpdates(save_style)
      "2":   ' 24-hour mode
        save_style := PauseUpdates
        TwelveHour := 0
        SetHours(Hours)
        ResumeUpdates(save_style)
      "r", "R":   ' Set "random" animations
        PauseUpdates
        ResumeUpdates(STYLE_1ROW_BMP_WIPE_RANDOM)
      "s":   ' Advance to the next display style
        save_style := PauseUpdates
        if  save_style < MIN_STYLE_1ROW_BMP_WIPE or {
          } save_style => MAX_STYLE_1ROW_BMP_WIPE
          save_style := MIN_STYLE_1ROW_BMP_WIPE
        else
          ++save_style
        ResumeUpdates(save_style)
      "S":   ' go back to previous display style
        save_style := PauseUpdates
        if  save_style =< MIN_STYLE_1ROW_BMP_WIPE or {
          } save_style > MAX_STYLE_1ROW_BMP_WIPE
          save_style := MAX_STYLE_1ROW_BMP_WIPE
        else
          --save_style
        ResumeUpdates(save_style)
      other: ' do nothing

  ReportStr(string("Stopping..."))
  StopTimeUpdateCog
  ReportStr(string("Timing cog stopped"))
  ShutdownDisplay
  ReportStr(string("Resolution cog stopped"))
  Panel.Stop
  ReportStr(string("Display cog stopped"))
  Helper.Stop
  ReportStr(string("Helper cog stopped"))

PRI PauseUpdates | s
{{
  Pause the time update cog, and wait for it to stop.
  Capture the current display animation style, and change it to simple digit
  replace with no pauses, to make time changes ("h" and "m" commands) faster.

  This is used to "freeze" the time display updates, so that style, color,
  intensity, and other display format changes can be made safely. Once those
  changes have been made, that is generally followed by a call to
  ResumeUpdates(previousStyle).

  Arguments:
    None.

  Return Value:
    The display animation style in effect before this call.
}}
  TimeUpdateEnabled := 0  ' Tell the cog to pause
  repeat while TimeUpdateRunning
    ' Stall
  WaitForNoDifferences
  s := DisplayStyle
  DisplayStyle := STYLE_1ROW_BMP_REPLACE
  return s

PRI ResumeUpdates(s) 
{{
  Resume display updates and time tracking, using the supplied digit
  animation style.

  This is usually called after a call to PauseUpdates() and one or more
  changes to the time display state: color, intensity, format, digits.

  Arguments:
    s -- the digit animation style to use going forward.

  Return Value:
    None.
}}
  WaitForNoDifferences
  DisplayStyle := s
  AnimationDelay := GetAnimationDelay
  TimeUpdateEnabled := 1
  repeat until TimeUpdateRunning
    ' Stall

' ------------------------------------------------------------------------
PUB ShowHelpText
  Ser.Str(@HelpText)

DAT
HelpText
    BYTE "Command characters take immediate effect:", CR
    BYTE "Debugging commands:", CR
    BYTE "  d -- Pause the time display, dump the digits state", CR
    BYTE "  D -- Resume the time display after the 'd' command", CR
    BYTE "  f -- Dump the FiFo Header state (insert and remove pointers)", CR
    BYTE "Display operations:", CR
    BYTE "  1 -- Set 12-hour clock display (no AM/PM yet)", CR
    BYTE "  2 -- Set 24-hour clock display", CR 
    BYTE "  c -- Colors: cycle through a set of predefined colors", CR
    BYTE "  K -- Free stacK: report the amount of used and unused stack", CR
    BYTE "  h and H -- Increment or decrement the Hours value", CR
    BYTE "  i and I -- Increase or decrease the digit Intensity", CR
    BYTE "  m and M -- Increment or decrement the Minutes value, which", CR
    BYTE "             also sets the seconds to zero", CR
    BYTE "  r -- Select a Random digit change animation for each change", CR
    BYTE "  s and S -- Iterate in either direction over all the digit", CR
    BYTE "             change animations (Styles), used for all changes", CR
    BYTE "  x or X -- Exit; Shut down.", CR
    BYTE 0

' ========================================================================
'
' The Time-keeping Cog, and some Set functions to make it easier to change
' the time manually.
'
' Below, the local variable "qnr" stands for quotient-and-remainder, from
' a table lookup for values 0..59 to avoid two division operations.

' ------------------------------------------------------------------------
PUB SetSeconds(s) | qnr
{{
  Set the display seconds digits for the specified number of seconds.
  Does nothing if the display is not wide enough to show the seconds.

  Arguments
    s -- new seconds value, 0..59, not range-checked. Odd things may happen
         for values outside that range.

  Return value
    None
}}
  if DigitCount > 4
    ' We only have the seconds digits on wide-enough displays
    qnr := BYTE[@TwoDigits + s]  ' both quotient and remainder for divide-by-10
    SetDigit(4, (qnr >> 4) & $F)
    SetDigit(5, qnr & $F)

' ------------------------------------------------------------------------
PUB SetMinutes(m) | qnr
{{
  Set the display minutes digits for the specified number of minutes.

  Arguments
    m -- new minutes value, 0..59, not range-checked. Odd things may happen
         for values outside that range.

  Return value
    None
}}
  qnr := BYTE[@TwoDigits + m]  ' both quotient and remainder for divide-by-10
  SetDigit(2, (qnr >> 4) & $F)
  SetDigit(3, qnr & $F)

' ------------------------------------------------------------------------
PUB SetHours(h) | qnr
{{
  Set the display hours digits for the specified number of hours.

  Arguments
    h -- new hours value, 0..23, not range-checked. This should always be for
         24-hour format (not 12-hour), and will be converted based on the
         current display mode for 12hour/24hour.

  Return value
    None
}}
  ' Handle 12-hour format and 24-hour format separately
  if TwelveHour <> 0
    ' 12-hour format: map 0..23 to 1..12
    if h == 0
      qnr := $12
    elseif h =< 12
      qnr := BYTE[@TwoDigits + h]       ' 1..12, we want 1..12
    else
      qnr := BYTE[@TwoDigits + h - 12]  ' 13..23, we want 1..11
    ' In 12-hour format, we convert a leading 0 for tens-of-hours to a space.
    SetDigitLeadingSpaces(0, (qnr >> 4) & $F)  ' tens of hours
  else ' 24-hour format
    qnr := BYTE[@TwoDigits + h]
    SetDigit(0, (qnr >> 4) & $F)

  ' In all cases:
  SetDigit(1, qnr & $F)

' ------------------------------------------------------------------------
PUB AdvanceTime(delta_sec)
{{
  Advance the current time by 0 to 60 seconds.

  Arguments
    delta_sec -- the number of seconds to advance, from 0..60 inclusive.
                 The value will be limited to that range.

  Return value
    None
}}
  if (Seconds += ((delta_sec #> 0) <# 60)) => 60
    Seconds -= 60
    if (Minutes += 1) => 60
      Minutes -= 60
      if (Hours += 1) => 24
        Hours := 0
      SetHours(Hours)
    SetMinutes(Minutes)
  SetSeconds(Seconds)

' ------------------------------------------------------------------------
PUB BackUpTime(delta_sec)
{{
  Backs the current time up by 0 to 60 seconds.

  Arguments
    delta_sec -- the number of seconds to back up, from 0..60 inclusive.
                 The value will be limited to that range.

  Return value
    None
}}
  if (Seconds -= ((delta_sec #> 0) <# 60)) < 0
    Seconds += 60
    if (Minutes -= 1) < 0
      Minutes += 60
      if (Hours -= 1) < 0
        Hours := 23
      SetHours(Hours)
    SetMinutes(Minutes)
  SetSeconds(Seconds)

' ------------------------------------------------------------------------
PUB StartTimeUpdateCog
{{
  Launches a cog running the Spin code to track and update the time and the
  flashing colon (which flashes once a second).

  Arguments
    None

  Return value
    None
}}
  StopTimeUpdateCog  ' Just to make sure.
  TimeUpdateEnabled := 0  ' Starts disabled
  TimeCog := 1 + COGNEW(TimeUpdateLoop, @TimeCogStack)
  ReportDec(string("Started timing cog: "), TimeCog - 1)

' ------------------------------------------------------------------------
PUB StopTimeUpdateCog
{{
  Stops the cog running the Spin code to track and update the time and the
  flashing colon.

  Arguments
    None

  Return value
    None
}}
  if TimeCog
    'ReportStr(string("Stopping timing cog"))
    COGSTOP(TimeCog~ - 1)

' ------------------------------------------------------------------------
PUB TimeUpdateLoop | half_sec, current_cnt
{{
  The main routine for the cog measuring time.
  Every half-second, updates the colon state, to get the colon(s) flashing
  once a second.
  Every second, advances the time by one second.

  Arguments
    None

  Return value
    None
}}
  half_sec := clkfreq >> 1
  repeat ' forever, until the cog is stopped
    ' Wait for updates allowed
    repeat until TimeUpdateEnabled
      ' Stall
    TimeUpdateRunning := 1
    current_cnt := CNT
    repeat while TimeUpdateEnabled
      ' Do one update
      waitcnt(current_cnt += half_sec)
      if TimeUpdateEnabled 
        SetColonState(0, COLON_ON)
        SetColonState(1, COLON_ON)
        AdvanceTime(1)
        waitcnt(current_cnt += half_sec)
      if TimeUpdateEnabled 
        SetColonState(0, COLON_OFF)
        SetColonState(1, COLON_OFF)
    ' Here the loop stopped because it was disabled
    TimeUpdateRunning := 0
  
' ========================================================================
' Color and intensity controls for the time display.

' ------------------------------------------------------------------------

PUB SetRgbColorAndIntensity(r, g, b, i) | mi, c, ix
{{
  Set the display color to the RGB values supplied, scaled by the supplied
  intensity.

  Arguments:
    r -- relative value of the Red color
    g -- relative value of the Green color
    b -- relative value of the Blue color
    i -- the intensity to use for the display of the color.

  Return value:
    None

  The three values r, g, and b are interpreted as reltive to the maximum one
  of the three values. The function calculates the RGB values for all the
  intensities with those color ratios, so that the intensity of the display
  can be varied simply by changing the index into the list.
}}
  ' Make the intensity "wrap-around" so that it cycles from 0 to max intensity,
  ' which is limited by the COLOR_DEPTH.
  if i < 0
    i := MaxIntensity
  elseif i > MaxIntensity
    i := 0
  ' "mi" is the max value of any of the three values r, g, or b.
  mi := (r #> g) #> b
  ' Calclulate the color gradient
  ColorGradient[0] := 0  ' Black
  repeat ix from 1 to MAX_COLOR_VALUE
    ColorGradient[ix] := (((ix * r) / mi) << 20) {
      }                | (((ix * g) / mi) << 10) {
      }                |  ((ix * b) / mi)
  ' Set the current display color by selecting the intensity from the
  ' calculated color gradient table.
  SetIntensity(i)

' ------------------------------------------------------------------------
PUB SetColorAndIntensity(c, i)
{{
  Set the display color to the RGB color value supplied, scaled by the supplied
  intensity.

  Arguments:
    c -- the color (ratio of the RGB values) as $00RRGGBB, one color per byte
    i -- the intensity to use for the display of the color.

  Return value:
    None

  The RGB values are interpreted as reltive to the maximum one of those three
  values. The function calculates the RGB values for all the intensities with
  those color ratios, so that the intensity of the display can be varied
  simply by changing the index into the list.
}}
  SetRgbColorAndIntensity((c >> 20) & $3FF, (c >> 10) & $3FF, c & $3FF, i)

' ------------------------------------------------------------------------
PUB SetColor(c)
{{
  Set the display color to the RGB color value supplied, scaled by the current
  intensity.

  Arguments:
    c -- the color (ratio of the RGB values) as $00RRGGBB, one color per byte

  Return value:
    None

  The RGB values are interpreted as reltive to the maximum one of those three
  values.
}}
  SetRgbColorAndIntensity((c >> 20) & $3FF, (c >> 10) & $3FF, c & $3FF, Intensity)

' ------------------------------------------------------------------------
PUB SetColorByIndexAndIntensity(color_index, i)
{{
  Set the display color and intensity via a table of display colors, having the
  supplied intensity.

  Arguments:
    color_index -- a 0-based index into the ColorTable[] of predefined display
                   colors. This is "wrapped around" so that when the index
                   exceeds the max index it is reset to zero, and similarly
                   when the index becomes negative.
    i -- the intensity to use for the display of the color.

  Return value:
    None
}}
  ' Wrap around after increment or decrement
  if color_index < 0
    color_index := ColorTableSize - 1
  elseif color_index => ColorTableSize
    color_index := 0
  ColorIndex := color_index
  SetColorAndIntensity(LONG[@ColorTable + (color_index << 2)], i)

' ------------------------------------------------------------------------
PUB SetColorByIndex(color_index)
{{
  Set the display color and intensity via a table of display colors, having the
  current intensity.

  Arguments:
    color_index -- a 0-based index into the ColorTable[] of predefined display
                   colors. This is "wrapped around" so that when the index
                   exceeds the max index it is reset to zero, and similarly
                   when the index becomes negative.

  Return value:
    None
}}
  SetColorByIndexAndIntensity(color_index, Intensity)

' ------------------------------------------------------------------------
PUB SetIntensity(i)
{{
  Set the intensity of the display to the supplied value, maintaining the current
  display color.

  Arguments:
    i -- the intensity to use for the display.

  Return value:
    None
}}
  ' Make the intensity "wrap-around" so that it cycles from 0 to max intensity,
  ' which is limited by the COLOR_DEPTH.
  if i < 0
    i := MaxIntensity
  elseif i > MaxIntensity
    i := 0
  BoldColorRgb := ColorGradient[i]
  BoldColorDisplay := Panel.ToDisplayColor(BoldColorRgb)
  LONG[@Intensity] := i

' ========================================================================
' Display layout, style selection, and top-level display management functions.
'
' TODO: Provide some discussion of the design, to make the code easier to
' understand. 

PUB PrepareDisplayLayout(style)
{{
  Based on the display column count, select either a 4-digit or 6-digit clock
  layout, prepare the detailed display layout, then start the display
  processing using the supplied display animation style.

  Arguments:
    style -- the constant display animation style to use, from the list of
             STYLE_* constants. Out of range values will instead use the
             style STYLE_1ROW_BMP_REPLACE.

  Return value
    None
}}
  if MAX_DIGIT_COLUMNS => 6
    ReportStr(string("Using 6 digits"))
    SelectStyle(style, 6)
  else
    ReportStr(string("Using 4 digits"))
    SelectStyle(style, 4)

' ------------------------------------------------------------------------
PUB SelectStyle(style, num_digits)
{{
  Stop any active display update cog, then configure the clock display for
  the supplied number of digits and digit animation style.

  Arguments:
    style -- the constant display animation style to use, from the list of
             STYLE_* constants. Out of range values will instead use the
             style STYLE_1ROW_BMP_REPLACE.
    num_digits -- either 4, or 6 -- the seconds digits are only shown for
                  6 digit style.

  Return value
    None
}}
  ' Set things up for the selected style and digit count:
  ' - Prepare the layout of the digits on the display
  ' - Init the display state variables, and start the display update cog.
  if (style <> DisplayStyle) or (num_digits <> DigitCount)
    ' We stop any display update cog that might be running, because we may
    ' want different update code to run depending on the animation style.
    ShutdownDisplay  ' stop the display update cog
    Panel.Clear
    DisplayStyle := style
    DigitCount := num_digits
    ' Select display digit columns and rows, and the colon position(s).
    PrepareLayout
    ' AnimationDelay is the amount of time between animation steps, and is
    ' specific to the animation style selected for the "best" visual
    ' appearance, since different animations need different numbers of steps,
    ' and the animation should take at most about 0.6 seconds or so.
    AnimationDelay := GetAnimationDelay
    ' Show the display and start the display update cog.
    InitDisplay

' DONE TO HERE...

' ------------------------------------------------------------------------
PUB PrepareLayout
  ' For now, all styles can use the same set-up
  PrepareLayout1RowBitmap
  
' ------------------------------------------------------------------------
PUB InitDisplay
  ' For now, all styles can use the same set-up
  InitDisplay1RowBitmap

' ------------------------------------------------------------------------
PUB ShutdownDisplay 
  ShutdownDisplayBase

' ------------------------------------------------------------------------
PUB ShutdownDisplayBase
  if ResolutionCog
    'ReportStr(string("Stopping resolution cog"))
    COGSTOP(ResolutionCog~ - 1)

' ------------------------------------------------------------------------
PUB SetDigit(col, d)
  SetDigit1RowBitmap(col, d)

' ------------------------------------------------------------------------
PUB SetAllDigits(vector) | col
  repeat col from 0 to (DigitCount - 1)
    SetDigit(col, LONG[vector][col])

' ------------------------------------------------------------------------
PUB SetDigitLeadingSpaces(col, d)
  SetDigitSpaces1RowBitmap(col, d)

' ------------------------------------------------------------------------
PUB SetAllDigitsLeadingSpaces(vector) | col
  repeat col from 0 to (DigitCount - 1)
    SetDigitLeadingSpaces(col, LONG[vector][col])
  
' ------------------------------------------------------------------------
'PUB SetCountDirection(dir)
'  SetCountDirection1RowBitmap(dir)

' ------------------------------------------------------------------------
' We might be able to use one routine for all styles here.
' But it depends on how the display is managed...
PUB AnyDifferences
  return AnyDifferencesBase

' ------------------------------------------------------------------------
PUB AnyDifferencesBase | col
  ' Return true if any digit columns still need updates
  repeat col from 0 to (DigitCount - 1)
    if CurrentDigitValue[col] <> TargetDigitValue[col]
      'ReportDec(string("Different Column: "), col)
      'ReportDec(string("   Current digit: "), CurrentDigitValue[col])
      'ReportDec(string("    Target digit: "), TargetDigitValue[col])
      'waitcnt(cnt + clkfreq)
      return 1
  return 0

' ------------------------------------------------------------------------
PUB WaitForNoDifferences
   repeat while AnyDifferences
     ' Stall

' ------------------------------------------------------------------------
PUB Refresh
  ' Cause all the display elements to be redrawn in full
  ' Assumes the current display style is "replace" for fastest update,
  ' but not required.
  ' Does not wait for the refresh to complete before returning.
  Refresh1RowBitmap

' ========================================================================
' ------------------------------------------------------------------------
' Single row display from Bitmap Fonts, all animations

PUB PrepareLayout1RowBitmap | i, width
  ' A single row of digits in the center.
  ' We pick the display row number for the top pixel row of the digits.

  CenterRowTop    := (DISPLAY_ROWS - DIGIT_HEIGHT) / 2
  CenterRowHeight := DIGIT_HEIGHT

  ' Assign columns for the digit positions and colons
  ' the display will be centered horizontally
  ' First, compute the width for 4 columns
  width := constant(4 * DIGIT_COLUMN_DELTA + 2)  ' for a single colon
  ColonCount := 1
  if DigitCount => 6
    DigitCount := 6  ' limit it.
    ' two more digit positions.
    width += constant(2 * DIGIT_COLUMN_DELTA + COLON_WIDTH + SEPARATOR_WIDTH)
    ColonCount := 2
  ' Half of the remaining display width goes on each side
  DigitColumn[0] := (DISPLAY_COLUMNS - width) >> 1
  DigitColumn[1] := DigitColumn[0] + DIGIT_COLUMN_DELTA
  ColonColumn[0] := DigitColumn[1] + DIGIT_COLUMN_DELTA 
  DigitColumn[2] := ColonColumn[0] + COLON_WIDTH + SEPARATOR_WIDTH
  DigitColumn[3] := DigitColumn[2] + DIGIT_COLUMN_DELTA
  CurrentColonState[0] := COLON_OFF
  SetColonState(0, COLON_ON)
  if DigitCount == 6
    ColonColumn[1] := DigitColumn[3] + DIGIT_COLUMN_DELTA 
    DigitColumn[4] := ColonColumn[1] + COLON_WIDTH + SEPARATOR_WIDTH
    DigitColumn[5] := DigitColumn[4] + DIGIT_COLUMN_DELTA
    CurrentColonState[0] := COLON_OFF
    SetColonState(1, COLON_ON)

  ' Init to 00:00:00 or 12:00:00(blinking like a VCR?)
  BYTEFILL(@CurrentDigitValue, " ", DigitCount)
  BYTEFILL(@WorkingDigitValue, " ", DigitCount)
  Hours := 0
  Minutes := 0
  Seconds := 0
  SetHours(Hours)
  SetMinutes(Minutes)
  SetSeconds(Seconds)

PUB InitDisplay1RowBitmap | i
  ' Start the cog that will look for differences between the target digit
  ' value, the working digit value (for transitions-in-progress) and the
  ' current (previous by still partially visible) digit value.
  ' This cog may have "speed limits" to make any digit change animations
  ' visible (or not).
  ResolutionCog := 1 + COGNEW(Update1RowBitmap, @ResolutionCogStack)
  ReportDec(string("Started resolution cog: "), ResolutionCog - 1)


PUB Update1RowBitmap | i, do_delay, prev_cnt
  ' Constantly look for differences between the current and target bitmap
  ' for each digit, and immediately draw the new bitmap.
  ' This runs in a separate cog.
  ' For the "wipe" animation:
  ' - If WorkingDigitValue[i] <> CurrentDigitValue[i], we are inside an ongoing
  '     animation
  ' - Else if TargetDigitValue[i] <> CurrentDigitValue[i], we are starting a new
  '     animation
  ' - otherwise the column is not changing.
  ' As most digit change animations have multiple steps, we need a delay to
  ' spread out the display updates, so each update call can indicate if a
  ' change was made that should use a short animation delay. All the digits
  ' are checked (and one animation step taken), and then if any changed and
  ' requested a delay, one delay is done for all of themm.
  ' We try to time the writes so that one digit is updated in about 0.3 to 0.6
  ' seconds.
  prev_cnt := CNT
  repeat
    do_delay := 0
    repeat i from 0 to (DigitCount - 1)
      if CurrentDigitValue[i] <> WorkingDigitValue[i]
        do_delay |= ContUpd1DigitBmp(i)
      elseif CurrentDigitValue[i] <> TargetDigitValue[i]
        do_delay |= StartNewUpdate1DigitBmp(i)
    repeat i from 0 to (ColonCount - 1)
      if CurrentColonState[i] <> TargetColonState[i]
        ' Blinking colon is not otherwise animated.
        UpdateColon(i)
    ' If any of the digits had an update that needs a delay, do that delay now.
    if (do_delay and AnimationDelay)
      prev_cnt := AnimationDelay1RowBitmap(prev_cnt)
    else
      prev_cnt := CNT

PUB AnimationDelay1RowBitmap(prev_cnt)
  ' delay a short amount of time
  waitcnt(prev_cnt += AnimationDelay)
  return prev_cnt

PUB StartNewUpdate1DigitBmp(digit_col)
  ' This is called when a digit change is first detected. It sets up the
  ' working values for the selected animation.
  '
  ' What to do depends on the digit change animation style.
  ' Returns 1 if a change was made requiring a per-row animation delay
  ' Returns 0 otherwise.
  WorkingDigitValue[digit_col] := TargetDigitValue[digit_col]
  ' Set the internal work information (current pixel row being updated for
  ' example) at the starting pixel row, which gets updated each time
  ' ContUpd1DigitBmp(digit_col) is called.
  ' (Something will be done here once we have animations.)
  WorkingChangeProgress[digit_col] := 0
  ' If we are selecting digit change animations randomly, we select the one
  ' for this digit change here.
  if DisplayStyle == STYLE_1ROW_BMP_WIPE_RANDOM
    WorkingDisplayStyle[digit_col] := {
      }   ((?WipeRandom & $7FFF_FFF) // constant(MAX_STYLE_1ROW_BMP_WIPE - MIN_STYLE_1ROW_BMP_WIPE + 1)) {
      }   + MIN_STYLE_1ROW_BMP_WIPE
  else
    WorkingDisplayStyle[digit_col] := DisplayStyle

  ' Now, do the first (or only) update for the digit
  return ContUpd1DigitBmp(digit_col)

PUB UpdateColon(ix) | colon_color
  ' Either draw or erase the colon at the specified position.
  ' We could use a colon bitmap to draw it, or we could perform two
  ' Helper.Box() operations. A bitmap would be more flexible but two
  ' Box() operations is a bit simpler.
  '
  ' The ix is the colon index: 0 or 1.

  if TargetColonState[ix]
    colon_color := BoldColorDisplay
  else
    colon_color := 0  ' Black
  Helper.Bitmap1Bpp(@ColonBitmap, 1, {
    }               ColonColumn[ix], CenterRowTop, {
    }               COLON_WIDTH, DIGIT_HEIGHT, 0, 0, {
    }               0, colon_color, SKIP_NONE)
  CurrentColonState[ix] := TargetColonState[ix]

PUB ContUpd1DigitBmp(digit_col)
  ' Takes the next digit change animation step for the specific digit.
  '
  ' What to do depends on the digit change animation style.
  ' Returns 1 if a change was made requiring a per-row animation delay
  ' Returns 0 otherwise.
  case WorkingDisplayStyle[digit_col]
    STYLE_1ROW_BMP_REPLACE: return ContUpd1DigitBmpReplace(digit_col)
    STYLE_1ROW_BMP_WIPE_DOWN: return ContUpd1DigitBmpWipeD(digit_col)
    STYLE_1ROW_BMP_WIPE_UP: return ContUpd1DigitBmpWipeUp(digit_col)
    STYLE_1ROW_BMP_WIPE_RIGHT: return ContUpd1DigitBmpWipeR(digit_col)
    STYLE_1ROW_BMP_WIPE_LEFT: return ContUpd1DigitBmpWipeL(digit_col)
    STYLE_1ROW_BMP_WIPE_TB2CTR:
      return ContUpd1DigitBmpWipeTB2C(digit_col)
    STYLE_1ROW_BMP_WIPE_TB_F_CTR:
      return ContUpd1DigitBmpWipeTBFC(digit_col)
    STYLE_1ROW_BMP_WIPE_LR_2_CTR:
      return ContUpd1DigitBmpWipeLR2C(digit_col)
    STYLE_1ROW_BMP_WIPE_LR_F_CTR:
      return ContUpd1DigitBmpWipeLRFC(digit_col)
    STYLE_1ROW_BMP_WIPE_F_CTR:
      return ContUpd1DigitBmpWipeFromCenter(digit_col)
    STYLE_1ROW_BMP_WIPE_TO_CENTER:
      return ContUpd1DigitBmpWipeToCenter(digit_col)
    STYLE_1ROW_BMP_SLIDE_DOWN:
      return ContUpd1DigitBmpSlideDown(digit_col)
    STYLE_1ROW_BMP_SLIDE_UP:
      return ContUpd1DigitBmpSlideUp(digit_col)
    STYLE_1ROW_BMP_SLIDE_LEFT:
      return ContUpd1DigitBmpSlideLeft(digit_col)
    STYLE_1ROW_BMP_SLIDE_RIGHT:
      return ContUpd1DigitBmpSlideRight(digit_col)
    STYLE_1ROW_BMP_SCROLL_LEFT:
      return ContUpd1DigitBmpScrollLeft(digit_col)
    STYLE_1ROW_BMP_SCROLL_RIGHT:
      return ContUpd1DigitBmpScrollRight(digit_col)
    STYLE_1ROW_BMP_SCROLL_UP:
      return ContUpd1DigitBmpScrollUp(digit_col)
    STYLE_1ROW_BMP_SCROLL_DOWN:
      return ContUpd1DigitBmpScrollDown(digit_col)
    STYLE_1ROW_BMP_SCROLL_UP_DN:
      return ContUpd1DigitBmpScrollUpDn(digit_col)
    STYLE_1ROW_BMP_FADE_TO_BLACK:
      return ContUpd1DigitBmpFadeToBlack(digit_col)
    STYLE_1ROW_BMP_FADE_TO_WHITE:
      return ContUpd1DigitBmpFadeToWhite(digit_col)
    other: return ContUpd1DigitBmpReplace(digit_col)

PUB Refresh1RowBitmap
  ' Cause all the display elements to be redrawn in full
  ' Assumes the current display style is "replace" for fastest update,
  ' but not required.
  ' Does not wait for the refresh to complete before returning.
  '
  ' This contains some "magic" logic that works with the way the
  ' display is updated, by manipulating the display state and the
  ' working state so that all the digits are redrawn.

  ' Set all the CurrentDigit values to spaces to force a full refresh.
  ' We also have to override the per-digit style to get immediate
  ' replacement (assuming "replace" display style).
  BYTEFILL(@WorkingDisplayStyle, DisplayStyle, DigitCount)
  BYTEFILL(@CurrentDigitValue, " ", DigitCount)
  if CurrentColonState[0] == COLON_ON
    CurrentColonState[0] := COLON_OFF
    CurrentColonState[1] := COLON_OFF  ' safe even if only 4 digit display 

PUB CalcAnimationDelay(count)
  ' Calculate (clkfreq * 0.55) / count to get an animation step duration that
  ' resuls in the digit change animation taking about 0.55 seconds, so that the
  ' final display is visible "for a while" before it changes again (for the
  ' unit-seconds digit). For a display with only minutes and hours, this is not
  ' that important.
  '
  ' But keep the numerator < 2.1 billion (max signed 32-bit value)
  ' For 80MHz to 100 MHz, that would be up to about " * 21"
  return (clkfreq * 11) / (count * 20)

PUB GetAnimationDelay
  case DisplayStyle
    STYLE_1ROW_BMP_WIPE_RANDOM, {
}   STYLE_1ROW_BMP_SLIDE_DOWN, {
}   STYLE_1ROW_BMP_SLIDE_UP, {
}   STYLE_1ROW_BMP_SCROLL_UP, {
}   STYLE_1ROW_BMP_SCROLL_DOWN, {
}   STYLE_1ROW_BMP_SCROLL_UP_DN:
      ' Use the longest calculation
      return CalcAnimationDelay(DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING)
 
    STYLE_1ROW_BMP_REPLACE:
      return 0  ' no delay needed, since not animated

    STYLE_1ROW_BMP_WIPE_DOWN, {
}   STYLE_1ROW_BMP_WIPE_UP:
      return CalcAnimationDelay(DIGIT_HEIGHT)
 
    STYLE_1ROW_BMP_WIPE_RIGHT, {
}   STYLE_1ROW_BMP_WIPE_LEFT:
      return CalcAnimationDelay(DIGIT_WIDTH << 1)

    STYLE_1ROW_BMP_WIPE_TB2CTR, {
}   STYLE_1ROW_BMP_WIPE_TB_F_CTR:
      return CalcAnimationDelay(DIGIT_HEIGHT) 

    STYLE_1ROW_BMP_WIPE_LR_2_CTR, {
}   STYLE_1ROW_BMP_WIPE_LR_F_CTR, {
}   STYLE_1ROW_BMP_WIPE_F_CTR, {
}   STYLE_1ROW_BMP_WIPE_TO_CENTER:
      return CalcAnimationDelay(DIGIT_WIDTH << 1)

    STYLE_1ROW_BMP_SLIDE_LEFT, {
}   STYLE_1ROW_BMP_SLIDE_RIGHT, {
}   STYLE_1ROW_BMP_SCROLL_LEFT, {
}   STYLE_1ROW_BMP_SCROLL_RIGHT:
      return CalcAnimationDelay(DIGIT_WIDTH + DIGIT_SCROLL_H_SPACING)

    STYLE_1ROW_BMP_FADE_TO_BLACK, {
}   STYLE_1ROW_BMP_FADE_TO_WHITE:
      ' In order to work well with the "random" style selection, the scaling
      ' should depend on the number of color bits. For example:
      '   Color Bits   Max Fade steps   Scaling      (Maybe)
      '       1              2           => 8 (3)       16
      '       2              4           => 4 (2)        8
      '       3              8           => 2 (1)        4
      '       4             16           => 1 (0)        2
      '       5+            32+             1 (0)        1
      '
      ' This would be ((4 - COLOR_DEPTH) #> 0)
      return CalcAnimationDelay((Intensity - 1) << (FADE_ANIMATION_SHIFT + 1))
 
    other: return CalcAnimationDelay(DIGIT_HEIGHT) ' just a guess

PUB SetDigitSpaces1RowBitmap(col, d)
  ' To handle leading zeros becoming spaces, whenever we set a column
  ' digit to zero, we look at the column to the left: if it is a space
  ' and not zero, we use a space for the digit in this column as well.
  ' This only works if digits are changed from left to right when updating
  ' several at once.
  if col => DigitCount
    return
  if d < 10
    ' must be 0..9
    d += "0"

  if d == "0" and col <> (DigitCount - 1)
    ' if this is the leftmost column, or if the column to the left contains
    ' a space, then we use a space, 
    if (col == 0) or (TargetDigitValue[col - 1] == " ")
      ' leading spaces to the left -- use a space here as well
      TargetDigitValue[col] := " "
      'TargetDigitBitmapAddr[col] := @Digits0To9_Space
      return
  ' Otherwise, normal digit display
  SetDigit1RowBitmap(col, d)

' All the Wipe styles could use this (and Replace, too)
PUB SetDigit1RowBitmap(col, d)
  if col => DigitCount
    return
  if d < 10
    ' must be 0..9
    d += "0"
  'TargetDigitBitmapAddr[col] := GetScrollUpBitmapAddr(d)
  TargetDigitValue[col] := d

PUB SetColonState(colon, state)
  TargetColonState[colon] := state

'' All the Wipe styles could use this (and Replace, too)
'PUB SetCountDirection1RowBitmap(dir)
'  ' Since there is no scrolling for this display, style, we have nothing to
'  ' do here.

PUB GetBitmap(c)
  ' Find the bitmap to use for the specific character, return space bitmap
  ' if c is not a digit character
  case c
    0..9: return @Digits_0[c * constant(DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING)]
    "0".."9": return @Digits_0[(c - "0") * constant(DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING)]
    '"0".."9": return @Digits_0 + (((c - "0") * constant(DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING)) << 1)
    other: return @Digits_Space

' ========================================================================
' ------------------------------------------------------------------------
' Single row display, bitmap Font, immediate update of digits (no animation).

PUB ContUpd1DigitBmpReplace(digit_col)
  ' Simply display the new full digit from rom.
  ' Since there is no animation, return 0 to mean no animation delay is needed.
  Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
    }               DigitColumn[digit_col], CenterRowTop, {
    }               DIGIT_WIDTH, CenterRowHeight, 0, 0,  {
    }               0, BoldColorDisplay, SKIP_NONE)
  CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]
  return 0  ' Never needs an animation delay.

' ------------------------------------------------------------------------
' Single row display, bitmap font, wipe down form the top.

PUB ContUpd1DigitBmpWipeD(digit_col) | pixel_row
  ' The intended behavior:
  ' For the current pixel row in the digit:
  ' - display that horizontal pixel row from the WorkingDigitValue[i]
  ' - increment the working horizontal pizel row
  ' - if the pixel row has reached the end of the digit bitmap:
  '   - Set the CurrentDigitValue[i] := WorkingDigitValue[i]
  ' return 1 to mean that an animation delay is wanted after this operation.

  ' Write a new digit pixel row based on WorkingChangeProgress[digit_col].
  pixel_row := WorkingChangeProgress[digit_col]++
  Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
    }                 DigitColumn[digit_col], CenterRowTop + pixel_row, {
    }                 DIGIT_WIDTH, 1, 0, pixel_row, {
    }                 0, BoldColorDisplay, SKIP_NONE)
  if WorkingChangeProgress[digit_col] => DIGIT_HEIGHT
    ' We just finished the animation, so change the current digit value to
    ' be the working digit value
    CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]

  return 1  ' we made a change
  
' ------------------------------------------------------------------------
' Single row display, bitmap font, wipe up form the bottom.

PUB ContUpd1DigitBmpWipeUp(digit_col) | pixel_row
  ' The intended behavior:
  ' For the current pixel row in the digit:
  ' - display that horizontal pixel row from the
  '   DIGIT_HEIGHT - WorkingDigitValue[digit_col]
  ' - decrement the working horizontal pizel row
  ' - if the pixel row has reached the top of the digit bitmap:
  '   - Set the CurrentDigitValue[i] := WorkingDigitValue[i]
  ' return 1 to mean that an animation delay is wanted after this operation.

  ' Write a new digit pixel row based on WorkingChangeProgress[digit_col].
  pixel_row := DIGIT_HEIGHT - (++WorkingChangeProgress[digit_col])
  Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
    }                 DigitColumn[digit_col], CenterRowTop + pixel_row, {
    }                 DIGIT_WIDTH, 1, 0, pixel_row, {
    }                 0, BoldColorDisplay, SKIP_NONE)
  if WorkingChangeProgress[digit_col] => DIGIT_HEIGHT
    ' We just finished the animation, so change the current digit value to
    ' be the working digit value
    CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]

  return 1  ' we made a change

' ------------------------------------------------------------------------
' Single row display, bitmap font, wipe from left to right

PUB ContUpd1DigitBmpWipeR(digit_col) | pixel_col
  ' The intended behavior:
  ' For the current pixel row in the digit:
  ' - display that vertical pixel column from the WorkingDigitValue[i]
  ' - increment the working vertical pixel column
  ' - if the pixel column has reached the end of the digit bitmap:
  '   - Set the CurrentDigitValue[i] := WorkingDigitValue[i]
  ' return 1 to mean that an animation delay is wanted after this operation.

  ' Write a new digit pixel row based on WorkingChangeProgress[digit_col].
  pixel_col := WorkingChangeProgress[digit_col]++
  if (pixel_col & 1) == 0
    ' the odd steps -- do nothing
    return 1  ' we still want a delay
  pixel_col >>= 1
  Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
    }               DigitColumn[digit_col] + pixel_col, CenterRowTop, {
    }               1, DIGIT_HEIGHT, pixel_col, 0, {
    }               0, BoldColorDisplay, SKIP_NONE)
  if pixel_col => constant(DIGIT_WIDTH - 1)
    ' We just finished the animation, so change the current digit value to
    ' be the working digit value
    CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]

  return 1  ' we made a change
  
' ------------------------------------------------------------------------
' Single row display, bitmap font, wipe from right to left

PUB ContUpd1DigitBmpWipeL(digit_col) | pixel_col, p
  ' The intended behavior:
  ' For the current pixel row in the digit:
  ' - display that vertical pixel column from the WorkingDigitValue[i]
  ' - increment the working vertical pixel column
  ' - if the pixel column has reached the end of the digit bitmap:
  '   - Set the CurrentDigitValue[i] := WorkingDigitValue[i]
  ' return 1 to mean that an animation delay is wanted after this operation.

  ' Write a new digit pixel row based on WorkingChangeProgress[digit_col].
  p := WorkingChangeProgress[digit_col]++
  if (p & 1) == 0
    ' the odd steps -- do nothing
    return 1  ' we still want a delay
  p >>= 1
  pixel_col := DIGIT_WIDTH - ++p
  Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
    }               DigitColumn[digit_col] + pixel_col, CenterRowTop, {
    }               1, DIGIT_HEIGHT, pixel_col, 0, {
    }               0, BoldColorDisplay, SKIP_NONE)
  if p => DIGIT_WIDTH
    ' We just finished the animation, so change the current digit value to
    ' be the working digit value
    CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]

  return 1  ' we made a change
  
' ------------------------------------------------------------------------
' Single row display, bitmap font, wipe down from the top and up from bottom

PUB ContUpd1DigitBmpWipeTB2C(digit_col) | top_pixel_row, bottom_pixel_row
  ' Write a new digit pixel row based on WorkingChangeProgress[digit_col].
  top_pixel_row := WorkingChangeProgress[digit_col]++
  if (top_pixel_row & 1) == 0
    ' the odd steps -- do nothing
    return 1  ' we still want a delay
  top_pixel_row >>= 1
  bottom_pixel_row := DIGIT_HEIGHT - top_pixel_row - 1
  Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
    }               DigitColumn[digit_col], CenterRowTop + top_pixel_row, {
    }               DIGIT_WIDTH, 1, 0, top_pixel_row, {
    }               0, BoldColorDisplay, SKIP_NONE)
  if top_pixel_row <> bottom_pixel_row
    Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
      }               DigitColumn[digit_col], CenterRowTop + bottom_pixel_row, {
      }               DIGIT_WIDTH, 1, 0, bottom_pixel_row, {
      }               0, BoldColorDisplay, SKIP_NONE)
  if top_pixel_row => constant((DIGIT_HEIGHT - 1) >> 1)
    ' We just finished the animation, so change the current digit value to
    ' be the working digit value
    CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]

  return 1  ' we made a change
  
' ------------------------------------------------------------------------
' Single row display, bitmap font, wipe form the center down to the bottom and
' up to the top

PUB ContUpd1DigitBmpWipeTBFC(digit_col) | top_pixel_row, bottom_pixel_row
  ' Write a new digit pixel row based on WorkingChangeProgress[digit_col].
  top_pixel_row := WorkingChangeProgress[digit_col]++
  if (top_pixel_row & 1) == 0
    ' the odd steps -- do nothing
    return 1  ' we still want a delay
  top_pixel_row >>= 1
  top_pixel_row := constant((DIGIT_HEIGHT - 1) >> 1) - top_pixel_row
  bottom_pixel_row := DIGIT_HEIGHT - top_pixel_row - 1
  Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
    }               DigitColumn[digit_col], CenterRowTop + top_pixel_row, {
    }               DIGIT_WIDTH, 1, 0, top_pixel_row, {
    }               0, BoldColorDisplay, SKIP_NONE)
  if top_pixel_row <> bottom_pixel_row
    Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
      }               DigitColumn[digit_col], CenterRowTop + bottom_pixel_row, {
      }               DIGIT_WIDTH, 1, 0, bottom_pixel_row, {
      }               0, BoldColorDisplay, SKIP_NONE)
  if top_pixel_row =< 0
    ' We just finished the animation, so change the current digit value to
    ' be the working digit value
    CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]

  return 1  ' we made a change
  
' ------------------------------------------------------------------------
' Single row display, bitmap font, wipe from both left and right to center

PUB ContUpd1DigitBmpWipeLR2C(digit_col) | left_pixel_col, right_pixel_col
  ' Write a new digit pixel row based on WorkingChangeProgress[digit_col].
  left_pixel_col := WorkingChangeProgress[digit_col]++
  if (left_pixel_col & 3) == 0
    ' the odd steps -- do nothing
    return 1  ' we still want a delay
  left_pixel_col >>= 2
  right_pixel_col := DIGIT_WIDTH - left_pixel_col - 1
  Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
    }               DigitColumn[digit_col] + left_pixel_col, CenterRowTop, {
    }               1, DIGIT_HEIGHT, left_pixel_col, 0, {
    }               0, BoldColorDisplay, SKIP_NONE)
  if left_pixel_col <> right_pixel_col
    Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
      }               DigitColumn[digit_col] + right_pixel_col, CenterRowTop, {
      }               1, DIGIT_HEIGHT, right_pixel_col, 0, {
      }               0, BoldColorDisplay, SKIP_NONE)
  if left_pixel_col > constant((DIGIT_WIDTH - 1) >> 1)
    ' We just finished the animation, so change the current digit value to
    ' be the working digit value
    CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]

  return 1  ' we made a change
  
' ------------------------------------------------------------------------
' Single row display, bitmap font, wipe from center to both left and right

PUB ContUpd1DigitBmpWipeLRFC(digit_col) | left_pixel_col, right_pixel_col
  ' Write a new digit pixel row based on WorkingChangeProgress[digit_col].
  left_pixel_col := WorkingChangeProgress[digit_col]++
  if (left_pixel_col & 3) == 0
    ' the odd steps -- do nothing
    return 1  ' we still want a delay
  left_pixel_col := constant((DIGIT_WIDTH - 1) >> 1) - (left_pixel_col >> 2)
  right_pixel_col := DIGIT_WIDTH - left_pixel_col - 1
  Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
    }               DigitColumn[digit_col] + left_pixel_col, CenterRowTop, {
    }               1, DIGIT_HEIGHT, left_pixel_col, 0, {
    }               0, BoldColorDisplay, SKIP_NONE)
  if left_pixel_col <> right_pixel_col
    Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
      }               DigitColumn[digit_col] + right_pixel_col, CenterRowTop, {
      }               1, DIGIT_HEIGHT, right_pixel_col, 0, {
      }               0, BoldColorDisplay, SKIP_NONE)
  if left_pixel_col == 0
    ' We just finished the animation, so change the current digit value to
    ' be the working digit value
    CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]

  return 1  ' we made a change
  
' ------------------------------------------------------------------------
' Single row display, bitmap font, wipe from center out in all directions
'
' This is the easier case than wiping in towards the center, since we can
' do it with a single drawing primitive for the center portion of the digit,
' expanding by two rows and columns per operation.
'
' Firstly, we assume that DIGIT_WIDTH <= DIGIT_HEIGHT, so we calculate the
' additional number of rows for the ((DIGIT_WIDTH + 1) >> 1) column updates.
' Which would be:
'    extra_rows := ((DIGIT_HEIGHT + 1) >> 1) - (((DIGIT_WIDTH + 1) >> 1) << 1)
' extra_rows will always be => 0 if DIGIT_HEIGHT => DIGIT_WIDTH.
' Then we calculate the starting and ending column, and the starting and
' ending row from the column range by adding the extra_rows.
' Then, we just draw the "center" of the new digit over the old.
' This will redundantly write the center as it grows, but it is faster than
' doing four writes (I think).
  
PUB ContUpd1DigitBmpWipeFromCenter(digit_col) | left_pixel_col, l
  ' Write a new digit pixel row based on WorkingChangeProgress[digit_col].
  l := WorkingChangeProgress[digit_col]++
  if (l & 3) == 0
    ' the intermediate steps -- do nothing
    return 1  ' we still want a delay
  left_pixel_col := constant((DIGIT_WIDTH - 1) >> 1) - (l >> 2)
  ' top_pixel row will be the same.
  ' The width written will be:
  '   DIGIT_WIDTH - (left_pixel_col << 1)
  ' The height written will be
  '   DIGIT_HEIGHT - ((left_pixel_col << 1) + WIPE_BOTH_EXTRA_ROWS
  Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
    }               DigitColumn[digit_col] + left_pixel_col, CenterRowTop + left_pixel_col, {
    }               DIGIT_WIDTH - (left_pixel_col << 1), {
    }               DIGIT_HEIGHT - ((left_pixel_col << 1)), {
    }               left_pixel_col, left_pixel_col, {
    }               0, BoldColorDisplay, SKIP_NONE)
  if left_pixel_col == 0
    ' We just finished the animation, so change the current digit value to
    ' be the working digit value
    CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]

  return 1  ' we made a change
  
' ------------------------------------------------------------------------
' Single row display, bitmap font, wipe from all edges to the center
'
' This is the messiest wipe so far: each incremental update, we draw the
' hollow box of pixels starting at the upper-left corner. This takes two to
' four write operations, each one row high or one column wide. At the end,
' we may have only the center to draw, which could be simpler.
'
' Firstly, we assume that DIGIT_WIDTH <= DIGIT_HEIGHT, so we calculate the
' additional number of rows for the ((DIGIT_WIDTH + 1) >> 1) column updates.
' Which would be:
'    extra_rows := DIGIT_HEIGHT - (((DIGIT_WIDTH + 1) >> 1) << 1)
' extra_rows will always be => 0 if DIGIT_HEIGHT => DIGIT_WIDTH.
' Then we calculate the starting and ending column, and the starting and
' ending row from the column range by adding the extra_rows.
' Then, we just draw the "center" of the new digit over the old.
' This will redundantly write the center as it grows, but it is faster than
' doing four writes (I think).

PUB ContUpd1DigitBmpWipeToCenter(digit_col) | left_pixel_col, write_width
  ' Write a new digit pixel row based on WorkingChangeProgress[digit_col].
  left_pixel_col := WorkingChangeProgress[digit_col]++
  if (left_pixel_col & 3) == 0
    ' the odd steps -- do nothing
    return 1  ' we still want a delay
  left_pixel_col >>= 2
  ' top_pixel_row will be the same.
  ' The width written will be:
  write_width := DIGIT_WIDTH - (left_pixel_col << 1)
  ' The height written will be
  '  write_width + WIPE_BOTH_EXTRA_ROWS
  ' Assuming the width < height, when the width written is 1 or two, we can
  ' just use a single write of that solid box in the center of the digit.
  ' Otherwise, we have to write all the edges in four write operations.
  if write_width =< 2
    ' We can do a single write
    Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
      }               DigitColumn[digit_col] + left_pixel_col, {
      }                 CenterRowTop + left_pixel_col, {
      }               write_width, DIGIT_HEIGHT - (left_pixel_col << 1), {
      }               left_pixel_col, left_pixel_col, {
      }               0, BoldColorDisplay, SKIP_NONE)
    ' We just finished the animation, so change the current digit value to
    ' be the working digit value
    CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]
  else
    ' We need to do four writes for the four sides of the hollow box of pixels.
    ' Top row
    Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
      }               DigitColumn[digit_col] + left_pixel_col, CenterRowTop + left_pixel_col, {
      }               write_width, 1, left_pixel_col, left_pixel_col, {
      }               0, BoldColorDisplay, SKIP_NONE)
    ' Bottom row
    Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
      }               DigitColumn[digit_col] + left_pixel_col, {
      }                 CenterRowTop + DIGIT_HEIGHT - left_pixel_col - 1, {
      }               write_width, 1, {
      }               left_pixel_col, constant(DIGIT_HEIGHT - 1) - left_pixel_col, {
      }               0, BoldColorDisplay, SKIP_NONE)
    '' Left edge (this will rewrite the top and bottom row pixels)
    Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
      }               DigitColumn[digit_col] + left_pixel_col, {
      }                 CenterRowTop + left_pixel_col, {
      }               1, DIGIT_HEIGHT - (left_pixel_col << 1), {
      }               left_pixel_col, left_pixel_col, {
      }               0, BoldColorDisplay, SKIP_NONE)
    '' Right edge
    Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
      }               DigitColumn[digit_col] + left_pixel_col + write_width - 1, {
      }                 CenterRowTop + left_pixel_col, {
      }               1, DIGIT_HEIGHT - (left_pixel_col << 1), {
      }               left_pixel_col + write_width - 1, left_pixel_col, {
      }               0, BoldColorDisplay, SKIP_NONE)

  return 1  ' we made a change
  
' ------------------------------------------------------------------------
' Single row display, bitmap font, slide from top to bottom
' The new digit "slides over" the old one, with a few blank lines to make
' it look good.
' We need to allow some blank columns between the digit leaving and the
' digit arriving -- we just use "blank" rows on the bottom end of the
' digit arriving.
' There will be DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING steps to complete
' the operation.

PUB ContUpd1DigitBmpSlideDown(digit_col) | drop_rows ' , new_digit_height
  drop_rows := ++WorkingChangeProgress[digit_col]
  ' We have to make sure that we don't try to draw too many rows, so we limit
  ' the height.
  Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
    }               DigitColumn[digit_col], CenterRowTop, {
    }               DIGIT_WIDTH, drop_rows <# DIGIT_HEIGHT, {
    }               0, constant(DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING) - drop_rows, {
    }               0, BoldColorDisplay, SKIP_NONE)
  if drop_rows => constant(DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING)
    ' We just finished the animation, so change the current digit value to
    ' be the working digit value
    CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]
    ' WorkingChangeProgress[digit_col] := 0  ' Just in case

  return 1  ' we made a change

' ------------------------------------------------------------------------
' Single row display, bitmap font, slide from bottom to top
' The new digit "slides over" the old one, with a few blank lines to make
' it look good.
' We need to allow some blank columns between the digit leaving and the
' digit arriving -- we just use "blank" rows on the top end of the
' digit arriving.
' There will be DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING steps to complete
' the operation.

PUB ContUpd1DigitBmpSlideUp(digit_col) | drop_rows ' , new_digit_height
  drop_rows := ++WorkingChangeProgress[digit_col]
  if drop_rows =< DIGIT_HEIGHT
    ' We are drawing the new digit bitmap, along with a few blank rows, and it
    ' will be truncated on the bottom.
    ' We draw "drop_rows" rows, on the row CenterRowTop + DIGIT_HEIGHT - drop_rows
    Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]) - constant(DIGIT_SCROLL_V_SPACING << 1), 2, {
      }               DigitColumn[digit_col], CenterRowTop + DIGIT_HEIGHT - drop_rows, {
      }               DIGIT_WIDTH, drop_rows, {
      }               0, 0, {
      }               0, BoldColorDisplay, SKIP_NONE)
  else
    ' We have completely overwritten the previous digit, but we still have
    ' DIGIT_SCROLL_V_SPACING rows to draw.
    Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
    }               DigitColumn[digit_col], CenterRowTop + DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING - drop_rows, {
    }               DIGIT_WIDTH, drop_rows - DIGIT_SCROLL_V_SPACING + 1, {
    }               0, 0, {
    }               0, BoldColorDisplay, SKIP_NONE)
    if drop_rows => constant(DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING)
      ' We just finished the animation, so change the current digit value to
      ' be the working digit value
      CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]
      ' WorkingChangeProgress[digit_col] := 0  ' Just in case

  return 1  ' we made a change

' ------------------------------------------------------------------------
' Single row display, bitmap font, slide from right to left.
' 
' The new digit "slides over" the old one, with a few blank columns to make
' it look good.
'
' We do not have blank columns in the digits the way they are stored so we
' have to draw the blank pixel columns separately.

PUB ContUpd1DigitBmpSlideLeft(digit_col) | hstep
  hstep := ++WorkingChangeProgress[digit_col]
  if hstep =< DIGIT_WIDTH
    ' We need to erase (in advance) the next column of the existing digit.
    Helper.VerticalLine(DigitColumn[digit_col] + DIGIT_WIDTH - hstep, CenterRowTop, {
    }                   DIGiT_HEIGHT, 0) 
  if hstep > DIGIT_SCROLL_H_SPACING
    ' The new character has visible columns.
    ' There are hstep - DIGIT_SCROLL_H_SPACING columns visible from the left
    ' edge of the new digit, shown on the right edge of the digit area.
    Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
    }                 DigitColumn[digit_col] + DIGIT_WIDTH + DIGIT_SCROLL_H_SPACING - hstep, {
    }                   CenterRowTop, {
    }                 hstep - DIGIT_SCROLL_H_SPACING, DIGIT_HEIGHT, {
    }                 0, 0, {
    }                 0, BoldColorDisplay, SKIP_NONE)
    if hstep => constant(DIGIT_WIDTH + DIGIT_SCROLL_H_SPACING)
      ' We just finished the animation, so change the current digit value to
      ' be the working digit value
      CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]
  return 1  ' We made changes.

' ------------------------------------------------------------------------
' Single row display, bitmap font, slide from left to right.
' 
' The new digit "slides over" the old one, with a few blank columns to make
' it look good.
'
' We do not have blank columns in the digits the way they are stored so we
' have to draw the blank pixel columns separately.

PUB ContUpd1DigitBmpSlideRight(digit_col) | hstep
  hstep := ++WorkingChangeProgress[digit_col]
  if hstep =< DIGIT_WIDTH
    ' We need to erase (in advance) the next column of the existing digit.
    Helper.VerticalLine(DigitColumn[digit_col] + hstep - 1, CenterRowTop, {
    }                   DIGiT_HEIGHT, 0) 
  if hstep > DIGIT_SCROLL_H_SPACING
    ' The new character has visible columns.
    ' There are hstep - DIGIT_SCROLL_H_SPACING columns visible from the right
    ' edge of the new digit, shown on the left edge of the digit area, so we
    ' use the ltrim parameter.
    Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
    }                 DigitColumn[digit_col], CenterRowTop, {
    }                 hstep - DIGIT_SCROLL_H_SPACING, DIGIT_HEIGHT, {
    }                 constant(DIGIT_WIDTH + DIGIT_SCROLL_H_SPACING) - hstep, 0, {
    }                 0, BoldColorDisplay, SKIP_NONE)
    if hstep => constant(DIGIT_WIDTH + DIGIT_SCROLL_H_SPACING)
      ' We just finished the animation, so change the current digit value to
      ' be the working digit value
      CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]
  return 1  ' We made changes.

' ------------------------------------------------------------------------
' Single row display, bitmap font, scroll from right to left
' We need to allow some blank columns between the digit leaving and the
' digit arriving -- we just use "blank" columns on the right end of the
' digit leaving.
' There will be DIGIT_WIDTH + DIGIT_SCROLL_H_SPACING steps to complete
' the operation.

PUB ContUpd1DigitBmpScrollLeft(digit_col) | drop_cols, new_pixel_col
  drop_cols := ++WorkingChangeProgress[digit_col]
  ' If the departing digit (on the left) is still visible, we need to
  ' draw the visible portion, and any spacing blank columns.
  if drop_cols =< DIGIT_WIDTH
    ' Draw the CurrentDigitValue[digit_col] skipping drop_cols columns,
    ' on the left portion of the display area for the digit.
    Helper.Bitmap1Bpp(GetBitmap(CurrentDigitValue[digit_col]), 2, {
      }               DigitColumn[digit_col], CenterRowTop, {
      }               constant(DIGIT_WIDTH + DIGIT_SCROLL_H_SPACING) - drop_cols, {
      }                 DIGIT_HEIGHT, {
      }               drop_cols, 0, {
      }               0, BoldColorDisplay, SKIP_NONE)
  if drop_cols > DIGIT_SCROLL_H_SPACING
    ' The newly arriving digit (on the right) is visible.
    new_pixel_col := constant(DIGIT_WIDTH + DIGIT_SCROLL_H_SPACING) - drop_cols
    Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
      }               DigitColumn[digit_col] + new_pixel_col, CenterRowTop, {
      }               DIGIT_WIDTH - new_pixel_col, DIGIT_HEIGHT, 0, 0, {
      }               0, BoldColorDisplay, SKIP_NONE)
  if drop_cols => constant(DIGIT_WIDTH + DIGIT_SCROLL_H_SPACING)
    ' We just finished the animation, so change the current digit value to
    ' be the working digit value
    CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]

  return 1  ' we made a change
  
' ------------------------------------------------------------------------
' Single row display, bitmap font, scroll from left to right
' We need to allow some blank columns between the digit leaving and the
' digit arriving -- we just use "blank" columns on the right end of the
' digit arriving.
' There will be DIGIT_WIDTH + DIGIT_SCROLL_H_SPACING steps to complete
' the operation.

PUB ContUpd1DigitBmpScrollRight(digit_col) | drop_cols, new_digit_width
  drop_cols := ++WorkingChangeProgress[digit_col]
  ' If the departing digit (on the right) is still visible, we need to
  ' draw the visible portion.
  if drop_cols < DIGIT_WIDTH
    ' Draw the CurrentDigitValue[digit_col] skipping drop_cols columns,
    ' on the right portion of the display area for the digit.
    Helper.Bitmap1Bpp(GetBitmap(CurrentDigitValue[digit_col]), 2, {
      }               DigitColumn[digit_col] + drop_cols, CenterRowTop, {
      }               DIGIT_WIDTH - drop_cols, DIGIT_HEIGHT, 0, 0, {
      }               0, BoldColorDisplay, SKIP_NONE)
  ' We always draw the digit arriving on the left, so that we also get the
  ' blank columns. We have to make sure that we don't try to draw too many
  ' columns, so we check for drop_cols => DIGIT_WIDTH, in which case we limit
  ' the width.
  new_digit_width := drop_cols <# DIGIT_WIDTH
  Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
    }               DigitColumn[digit_col], CenterRowTop, {
    }               new_digit_width, DIGIT_HEIGHT, {
    }               constant(DIGIT_WIDTH + DIGIT_SCROLL_H_SPACING) - drop_cols, 0, {
    }               0, BoldColorDisplay, SKIP_NONE)
  if drop_cols => constant(DIGIT_WIDTH + DIGIT_SCROLL_H_SPACING)
    ' We just finished the animation, so change the current digit value to
    ' be the working digit value
    CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]

  return 1  ' we made a change
  
' ------------------------------------------------------------------------
' Single row display, bitmap font, scroll from bottom to top
' We need to allow some blank rows between the digit leaving and the
' digit arriving -- we just use "blank" rows on the bottom end of the
' digit leaving.
' There will be DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING steps to complete
' the operation.

PUB ContUpd1DigitBmpScrollUp(digit_col) | drop_rows, new_pixel_row
  drop_rows := ++WorkingChangeProgress[digit_col]
  ' If the departing digit (on the top) is still visible, we need to
  ' draw the visible portion, and also when it just disappears from the
  ' top of the screen, to draw the top blank line.
  if drop_rows =< DIGIT_HEIGHT
    ' Draw the CurrentDigitValue[digit_col] skipping drop_rows rows,
    ' on the top portion of the display area for the digit.
    Helper.Bitmap1Bpp(GetBitmap(CurrentDigitValue[digit_col]), 2, {
      }               DigitColumn[digit_col], CenterRowTop, {
      }               DIGIT_WIDTH, constant(DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING) - drop_rows, {
      }               0, drop_rows, {
      }               0, BoldColorDisplay, SKIP_NONE)
  if drop_rows > DIGIT_SCROLL_V_SPACING
    ' The newly arriving digit (on the bottom) is visible.
    new_pixel_row := constant(DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING) - drop_rows
    Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
      }               DigitColumn[digit_col], CenterRowTop + new_pixel_row, {
      }               DIGIT_WIDTH, DIGIT_HEIGHT - new_pixel_row, 0, 0, {
      }               0, BoldColorDisplay, SKIP_NONE)
  if drop_rows => constant(DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING)
    ' We just finished the animation, so change the current digit value to
    ' be the working digit value
    CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]
    ' WorkingChangeProgress[digit_col] := 0  ' Just in case

  return 1  ' we made a change
  
' ------------------------------------------------------------------------
' Single row display, bitmap font, scroll from top to bottom
' We need to allow some blank columns between the digit leaving and the
' digit arriving -- we just use "blank" rows on the bottom end of the
' digit arriving.
' There will be DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING steps to complete
' the operation.

PUB ContUpd1DigitBmpScrollDown(digit_col) | drop_rows ' , new_digit_height
  drop_rows := ++WorkingChangeProgress[digit_col]
  ' If the departing digit (on the bottom) is still visible, we need to
  ' draw the visible portion.
  if drop_rows < DIGIT_HEIGHT
    ' Draw the CurrentDigitValue[digit_col] skipping drop_rows rows,
    ' on the bottom portion of the display area for the digit.
    Helper.Bitmap1Bpp(GetBitmap(CurrentDigitValue[digit_col]), 2, {
      }               DigitColumn[digit_col], CenterRowTop + drop_rows, {
      }               DIGIT_WIDTH, DIGIT_HEIGHT - drop_rows, 0, 0, {
      }               0, BoldColorDisplay, SKIP_NONE)
  ' We always draw the digit arriving on the top, so that we also get the
  ' blank rows. We have to make sure that we don't try to draw too many
  ' rows, so we check for drop_rows => DIGIT_HEIGHT, in which case we limit
  ' the height.
  'new_digit_height := drop_rows <# DIGIT_HEIGHT
  'Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
  '  }               DigitColumn[digit_col], CenterRowTop, {
  '  }               DIGIT_WIDTH, new_digit_height, {
  '  }               0, constant(DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING) - drop_rows, {
  '  }               0, BoldColorDisplay, SKIP_NONE)
  Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
    }               DigitColumn[digit_col], CenterRowTop, {
    }               DIGIT_WIDTH, drop_rows <# DIGIT_HEIGHT, {
    }               0, constant(DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING) - drop_rows, {
    }               0, BoldColorDisplay, SKIP_NONE)
  if drop_rows => constant(DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING)
    ' We just finished the animation, so change the current digit value to
    ' be the working digit value
    CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]
    ' WorkingChangeProgress[digit_col] := 0  ' Just in case

  return 1  ' we made a change

' ------------------------------------------------------------------------
' Single row display, bitmap font, scroll up and down in alternate pixel columns

PUB ContUpd1DigitBmpScrollUpDn(digit_col) | drop_rows, new_pixel_row, new_digit_height
  drop_rows := ++WorkingChangeProgress[digit_col]
  ' If the departing digit (on the top) is still visible, we need to
  ' draw the visible portion.
  if drop_rows =< DIGIT_HEIGHT
    ' Draw the CurrentDigitValue[digit_col] skipping drop_rows rows,
    ' on the top portion of the display area for the digit.
    Helper.Bitmap1Bpp(GetBitmap(CurrentDigitValue[digit_col]), 2, {
      }               DigitColumn[digit_col], CenterRowTop, {
      }               DIGIT_WIDTH, constant(DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING) - drop_rows, {
      }               0, drop_rows, {
      }               0, BoldColorDisplay, SKIP_UP)
  ' If the departing digit (on the bottom) is still visible, we need to
  ' draw the visible portion.
  if drop_rows < DIGIT_HEIGHT
    ' Draw the CurrentDigitValue[digit_col] skipping drop_rows rows,
    ' on the bottom portion of the display area for the digit.
    Helper.Bitmap1Bpp(GetBitmap(CurrentDigitValue[digit_col]), 2, {
      }               DigitColumn[digit_col], CenterRowTop + drop_rows, {
      }               DIGIT_WIDTH, DIGIT_HEIGHT - drop_rows, 0, 0, {
      }               0, BoldColorDisplay, SKIP_DOWN)
  if drop_rows > DIGIT_SCROLL_V_SPACING
    ' The newly arriving digit (on the bottom) is visible.
    new_pixel_row := constant(DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING) - drop_rows
    Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
      }               DigitColumn[digit_col], CenterRowTop + new_pixel_row, {
      }               DIGIT_WIDTH, DIGIT_HEIGHT - new_pixel_row, 0, 0, {
      }               0, BoldColorDisplay, SKIP_UP)
  ' We always draw the digit arriving on the top, so that we also get the
  ' blank rows. We have to make sure that we don't try to draw too many
  ' rows, so we check for drop_rows => DIGIT_HEIGHT, in which case we limit
  ' the height.
  new_digit_height := drop_rows <# DIGIT_HEIGHT
  Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
    }               DigitColumn[digit_col], CenterRowTop, {
    }               DIGIT_WIDTH, new_digit_height, {
    }               0, constant(DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING) - drop_rows, {
    }               0, BoldColorDisplay, SKIP_DOWN)
  if drop_rows => constant(DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING)
    ' We just finished the animation, so change the current digit value to
    ' be the working digit value
    CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]

  return 1  ' we made a change

' ------------------------------------------------------------------------
' Single row display, bitmap font, Fade to black and back up
' This fades the current digit to black, then fades up the new one.

PUB ContUpd1DigitBmpFadeToBlack(digit_col) | ix
  ix := WorkingChangeProgress[digit_col]++
  if (ix & constant((|< FADE_ANIMATION_SHIFT) - 1)) == 0
    ' the odd steps -- do nothing
    return 1  ' we still want a delay
  ix >>= FADE_ANIMATION_SHIFT

  ' ix == 0..Intensity-1: we show the current digit with
  '   the foreground color as ColorGradient[Intensity - ix + 1]
  ' ix == Intensity..2*(Intensity-1) - 1: we show the new digit with
  '   the foreground color as ColorGradient[ix - Intensity + 1]
  ' ix => 2 * (Intensity - 1): show the new digit at normal intensity
  '   and stop.
  if ix < Intensity
    ' Simply display the new full digit from rom.
    ' Since there is no animation, return 0 to mean no animation delay is needed.
    Helper.Bitmap1Bpp(GetBitmap(CurrentDigitValue[digit_col]), 2, {
      }               DigitColumn[digit_col], CenterRowTop, {
      }               DIGIT_WIDTH, CenterRowHeight, 0, 0,  {
      }               0, ColorGradient[Intensity - ix + 1], SKIP_NONE)
  elseif ix =< (Intensity << 1) - 3
    Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
      }               DigitColumn[digit_col], CenterRowTop, {
      }               DIGIT_WIDTH, CenterRowHeight, 0, 0,  {
      }               0, ColorGradient[ix - Intensity + 1], SKIP_NONE)
  else
    ' Last time -- just show the digit at full brightness
    Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
      }               DigitColumn[digit_col], CenterRowTop, {
      }               DIGIT_WIDTH, CenterRowHeight, 0, 0,  {
      }               0, BoldColorDisplay, SKIP_NONE)
    CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]
  return 1
' ------------------------------------------------------------------------
' Single row display, bitmap Font, Fade to white and back
' Similar to fade-to-black, but instead we fade the background color up to
' full brightness, then back down revealing the new digit.

PUB ContUpd1DigitBmpFadeToWhite(digit_col) | ix
  ix := WorkingChangeProgress[digit_col]++
  if (ix & constant((|< FADE_ANIMATION_SHIFT) - 1)) == 0
    ' the odd steps -- do nothing
    return 1  ' we still want a delay
  ix >>= FADE_ANIMATION_SHIFT

  ' ix == 0..Intensity-1: we show the current digit with
  '   the background color as ColorGradient[ix + 1]
  ' ix == Intensity .. 2 * (Intensity-1) - 1: we show the new digit with
  '   the background color as ColorGradient[2 * Intensity - ix + 1]
  ' ix => 2 * (Intensity - 1): show the new digit at normal intensity
  '   and stop.
  if ix < Intensity
    ' Simply display the new full digit from rom.
    ' Since there is no animation, return 0 to mean no animation delay is needed.
    Helper.Bitmap1Bpp(GetBitmap(CurrentDigitValue[digit_col]), 2, {
      }               DigitColumn[digit_col], CenterRowTop, {
      }               DIGIT_WIDTH, CenterRowHeight, 0, 0,  {
      }               ColorGradient[ix + 1], BoldColorDisplay, SKIP_NONE)
  elseif ix =< (Intensity << 1) - 2
    Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
      }               DigitColumn[digit_col], CenterRowTop, {
      }               DIGIT_WIDTH, CenterRowHeight, 0, 0,  {
      }               ColorGradient[(Intensity << 1) - ix + 1], BoldColorDisplay, SKIP_NONE)
  else
    ' Last time -- just show the digit at full brightness
    Helper.Bitmap1Bpp(GetBitmap(WorkingDigitValue[digit_col]), 2, {
      }               DigitColumn[digit_col], CenterRowTop, {
      }               DIGIT_WIDTH, CenterRowHeight, 0, 0,  {
      }               0, BoldColorDisplay, SKIP_NONE)
    CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]
  return 1

  
' ========================================================================

PUB CheckStack(n, p, l) | lused, pend
  ' Examine the stack starting at p, that contains L longs, and determine
  ' how much of it was used, and how much was untouched.
  pend := p + (l << 2) - 4
  repeat while (pend => p) and (long[pend] == UNLIKELY_VALUE)
    pend -= 4

  ' Here pend is either p - 4, or the address of the last value
  ' used.
  '   longs_used = ((pend + 4) - p) >> 2
  '   longs_free = l - longs_used
  lused := ((pend + 4) - p) >> 2
  Ser.Str(n)
  ReportDec(string(" used longs: "), lused)
  Ser.Str(n)
  ReportDec(string(" free longs: "), l - lused)

PUB CheckMainStack
  CheckStack(string("Main"), WORD[STACK_BASE_ADDR], {
    }        (END_OF_STACK - WORD[STACK_BASE_ADDR]) >> 2)
 
' ========================================================================
' Routines to help with testing.

PUB udec32tostr(v, buf)
  ' Returns the length of the string written into buf
  return Umath.udec64tostr(0, v, buf)

PUB sdec32tostr(v, buf) | p, len, c, i
  ' Returns the length of the string written into buf
  ' One negative value cannot be converted to a positive value by
  ' negation: $8000_0000, but it is its own negation, so we can just
  ' handle it like any other negative number
  if v < 0
    buf := "-"
    return Umath.udec64tostr(0, -v, buf + 1) + 1
  return Umath.udec64tostr(0, v, buf)

PUB ReportStr(s)
  Ser.Str(s)
  Ser.Tx(CR)

PUB ReportDec(label, v)
  Ser.Str(label)
  Ser.Dec(v)
  Ser.Tx(CR)

PUB ReportHex(label, v)
  Ser.Str(label)
  Ser.Hex(v, 8)
  Ser.Tx(CR)

PUB ReportBin(label, v)
  Ser.Str(label)
  Ser.Bin(v, 32)
  Ser.Tx(CR)

Pub ReportTenths(label, v)
  Ser.Str(label)
  Ser.Dec(v / 10)
  Ser.Tx(".")
  Ser.Tx("0" + (v // 10))
  Ser.Tx(CR)

PUB DumpFifo(label) | p
  ReportStr(label)
  p := Helper.GetCmdFifoHeader
  
  ReportHex(string("FiFo   ptr: "), p)
  ReportHex(string("Insert ptr: "), word[p + Fifo#INSERT_HUBADDR_OFFSET])
  ReportHex(string("Remove ptr: "), word[p + Fifo#REMOVE_HUBADDR_OFFSET])
  ReportHex(string("Start  ptr: "), word[p + Fifo#START_HUBADDR_OFFSET])
  ReportHex(string("Limit  ptr: "), word[p + Fifo#LIMIT_HUBADDR_OFFSET])
  ReportDec(string("Entry size: "), word[p + Fifo#ENTRY_SIZE_BYTES_OFFSET])

PUB ReportBitmapDebugOutput(entry)
  ReportHex(string("result address:  "), entry)
  ReportHex(string("Requested background color: "), long[entry + 20])
  ReportHex(string("Seen      background color: "), long[entry +  8])
  ReportBin(string("Converted background color: "), long[entry +  0])
  ReportHex(string("Requested foreground color: "), long[entry + 24])
  ReportHex(string("Seen      foreground color: "), long[entry + 12])
  ReportBin(string("Converted foreground color: "), long[entry +  4])

PUB ReportRomCharDebugOutput(entry)
  ReportHex(string("result address:  "), entry)
  ReportHex(string("Rom char base address: "), long[entry +  0])
  ReportHex(string("Rom char seen:         "), long[entry +  4])
  ReportHex(string("Rom char requested:    "), word[entry + 16])
  ReportHex(string("Expected base address: "), ((word[entry + 16] & $FE) << 6) + $8000)

PUB ReportSetColorDebugOutput(entry) | c
  c := long[entry + 8]
  ReportHex(string("Original      color value: "), c)
  ReportBin(string("Converted raw color value: "), long[entry + 4])
  ReportBin(string("Expected  raw color value: "), {
    } Panel.DisplayColor(((c >> 20) & $3FF), ((c >> 10) & $3FF), c & $3FF))
  'ReportBin(string("Expected2 raw color value: "), Panel.DisplayColor(6, 6, 0))

PUB ReportDigitRowUpdateSpeed(startcnt, endcnt) | clks
  clks := endcnt - startcnt
  ReportDec(string("Sysclks to update all digits: "), clks)
  ' calculate and report update rate, to tenths.
  ReportTenths(string("Update rate, per second: "), {
    }          Umath.multdiv_rounded(CLKFREQ, 10, clks))
  'ReportDec(string("Update rate, in tenths: "), {
  '  }          Umath.multdiv_rounded(CLKFREQ, 10, clks))

PUB ReportStackUse
  CheckMainStack  ' So we can see if we can shrink the main stack, and increase
                  ' the amount of memory for dynamic buffers.
  CheckStack(string("Resolution"), @ResolutionCogStack, RESOLUTION_COG_STACK_LONGS)
  CheckStack(string("Time"), @TimeCogStack, TIME_COG_STACK_LONGS)

PUB PowerOfTen(p) | r
  r := 1
  repeat while p--
    r *= 10
  return r

PUB DumpDigitParams(d)
  ReportDec(string("Digit position: "), d)
  ReportDec(string("  Current Digit:    "), CurrentDigitValue[d])
  ReportDec(string("  Working Digit:    "), WorkingDigitValue[d])
  ReportDec(string("  Target  Digit:    "), TargetDigitValue[d])
  ReportDec(string("  Working Progress: "), WorkingChangeProgress[d])
