{{
  Initialization code for test_clock_bitmap.spin

Author:         Tom Dinger (Ding-Batty)
Release Date:   2024/04/10

Copyright 2024 Thomas Dinger

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

--------------------------------------------------------------------------

This file contains init code for the test program, that can be recovered
later for additional buffer space.


Revisions:
==========
- 2024/04/10 -- Change for FDS in the init code;
- 2024/02/27 -- Original release

}}

OBJ
  ' Pick up the static configuration for the display(s).
  Config : "hub75_config.spin"
  Fifo   : "FifoBig_Header.spin"

  ' The following objects will be "reclaimed" after initial use, along with
  ' most of this object.
  SerPasm    : "Db_FullDuplexSerial_pasm"  ' uses a cog
  Ser        : "Db_FullDuplexSerial_Init"
  HelperPasm : "hub75_helper_pasm.spin"

  ' Select exactly one of these for the display refresh implementation. 
  'PanelPasm  : "hub75_display_pasm_repeat.spin"
  PanelPasm  : "hub75_display_pasm_video_4clock.spin"
  'PanelPasm  : "hub75_display_pasm_video_8clock.spin"
  'PanelPasm  : "hub75_display_pasm_waitcnt.spin"

CON
  { Configuration for serial connection. }
  BAUDRATE = 115_200
  RXPIN    = 31    ' The usual pins.
  TXPIN    = 30
  ' For this version of FDS, the sizes do not need to be powers-of-two. The
  ' Spin code is a little slower, but the PASM code is not slowed by this.
  RX_BUFFER_SIZE = 28
  TX_BUFFER_SIZE = 200
  CR = $0D

  ' Import some CONstants
  DISPLAY_COLUMNS = Config#DISPLAY_COLUMNS
  REFRESH_ROWS    = Config#REFRESH_ROWS

  ' The number of color bits for this program
  COLOR_DEPTH        =  4   ' Number of bits of resolution per color, up to
                            ' 8 bits, which is the limit for the implementation
                            ' details. The limits are available memory and
                            ' flickering from slower refresh rates.

  ' Constants for capturing the ROM Font digits into 1 bpp bitmaps.
  DIGIT_WIDTH  = 13
  DIGIT_HEIGHT = 22
  ' The number of pixels to trim a Rom Font digit to skip over the empty
  ' pixel rows and columns on the left and top.
  ROM_FONT_TRIM_TOP  = 5
  ROM_FONT_TRIM_LEFT = 2

  ' For scrolling, we need to decide how many blank rows or columns to use
  ' between the outgoing and incoming digits. For capturing the digit bitmaps,
  ' we only need the row spacing.
  DIGIT_SCROLL_V_SPACING = 3  ' three rows between scrolling digits.

  ' This is the distance from the last local variable address in the
  ' function that fills main memory that is safe to fill with the
  ' "unlikely" value.
  OFFSET_FOR_MAIN_STACK_FILL = $0010
  END_OF_STACK = $0000_8000      ' Highest hub RAM address
  UNLIKELY_VALUE = $DEADBEEF

  ' The initial amount of reserved memory, before the rest of the program
  ' that will be considered "available".
  ' This "recovers" 2888 bytes from init functions and PASM code
  ' images, about 8.8% of all hub memory (with BST)
  ' But NOTE: This value may need to change as the INI code size changes.
  ' Specifially, the different lengths of the HUB75 display refresh PASM
  ' code makes this vary.
  ' This value works with BST, with unused function removal.
  ' Leave about 225 longs for the main stack, which uses about 50. 
  'RESERVED_SIZE_BYTES = 1024 * 11 + (1700 * 4)
  ' This value works for PropTool
  RESERVED_SIZE_BYTES = 1024 * 11 + (700 * 4)

  GLOBAL_VBASE_ADDR = $0008  ' WORD variable

  ' Instead of a single mailbox, we use a FiFo of command (a mailbox queue)
  ' to enqueue multiple commands for the display buffer write helper cog to
  ' execute for the main program. We have a large queue because for some
  ' operations, we want the updates to go as fast as possible with little
  ' time for stalling.
  '
  ' If we need even faster display update, we could use multiple writer
  ' cogs, each with a separate command FiFo (or even a shared FiFo with the
  ' proper locks) to process write requests in parallel. There can be
  ' write interactions, however, because display buffer updates are
  ' read-modify-write, and are not locked, so two writers is best avoided.  
  COMMAND_FIFO_SIZE = 32

DAT
  ' Because they are "small" we don't dynamically allocate the serial object
  ' buffers. But they must go before FreeMemStart, or they will be overwritten
  ' after init time.
  Rx_Buffer          LONG    0[(RX_BUFFER_SIZE + 3) >> 2]
  Tx_Buffer          LONG    0[(TX_BUFFER_SIZE + 3) >> 2]

  ' NextFree contains the lowest address currently available for a buffer.
  ' The maximum size available is WORD[GLOBAL_VBASE_ADDR] - NextFree bytes.
  ' When a buffer is "allocated" the value of NextFree is increased by the
  ' buffer size, rounded up to a long boudary.
  NextFree       LONG 0  ' @FreeMemStart

  ' This is the actual start of available memory. We need a real allocation
  ' here, in order to push the VBASE value as high in memory as we can.
  FreeMemStart   LONG 0 [(RESERVED_SIZE_BYTES + 3) >> 2]

' ========================================================================

PUB Init(pCogIds, pBuffers, pDigits, pSerialConfig, pDisplayBufAddr, {
      }  pRefreshClocks, pCmdFifoHeader, pHelperDebugOutput) | needed
{{
  Perform all the "early" initialization operations for the program, so that
  this code can be recovered/reused for runtime buffers.

  The main program and this object _must_ agree on the layout and use of
  the buffers whose pointers are passed in. BEWARE: It is easy for these
  to get out of sync between this code and the main program.

  Arguments:
    pCogIds -- a list of three WORD values (TOOD: BYTE, WORD or LONG? They only
               need to be BYTEs based on the value range) that get the
               CogId + 1 of the three PASM cogs started for this program
                 pCogIds + 0 -- FullDuplexSerial cog
                 pCogIds + 2 -- Display refresh cog
                 pCogIds + 4 -- Display buffer write helper cog
               If any are zero on return, that cog could not be started.
    pBuffers -- a list of four WORD values, that get the addresses and sizes
                reserved for the two buffers:
                  pBuffers + 0 -- the size in bytes of the Command FiFo buffer
                                  for the display buffer write helper
                  pBuffers + 2 -- the reserved address of the above FiFo.
                  pBuffers + 4 -- the size in bytes of the Display Buffer
                  pBuffers + 6 -- the reserved address of the Display Buffer
                If either of the address entries are zero on return, then there
                was not enough available memory to satisfy the amount needed
                for the current configuration.
    pDigits -- the base address of the captured ROM Font digit bitmaps, from
               0 to 9 -- see the structure in the main program. This should be
               completely filled in during init time.
    pSerialConfig -- the configuration structure used to initialize the
                     FullDuplexSerial PASM cog, and used by the Spin routines
                     that read and write bytes to the serial connection.
                     The rx and tx buffer addresses and sizes are _not_ filled
                     in here.
    pDisplayBufAddr -- The address of the location in hub memory that the display
                       cog and the helper cog will watch for the actual display
                       buffer address, so that they will continue init.
    pRefreshClocks -- The address of the location for the display cog to put
                      its calculated or measured count of system clocks to
                      complete one full display refresh for all color bits, so
                      that the refresh rate can be calculated.
    pCmdFifoHeader -- The address (in the Spin write helper object) of the
                      command FiFo header used to send commands from the Spin
                      functions to the helper cog.
    pHelperDebugOutput -- The address of a buffer that will recieve debug
                          output from write helper cog operations.


    Return value:
      == 0 -- buffer memory reservation/allocation succeeded.
      <> 0 -- not enough available memory for our buffers, the value returned
              is the number of bytes additional that is needed.
}}
  waitcnt(clkfreq * 4 + cnt)   ' Wait 4 seconds at startup, to allow time to
                               ' change focus to the serial window.
  ' Fill the main stack area so we can determine how much was used during
  ' runtime.
  FillMainStack
  WORDFILL(pCogIds, 0, 3)
  WORDFILL(pBuffers, 0, 4)
  CaptureAllDigits(pDigits)

  ' Since we are not using "dynamic" allocation of the serial buffers, we
  ' can start the serial cog now, because the caller can use the FDS object
  ' to report any problems with the rest of init.
  WORD[pCogIds + 0] := SerPasm.LoadCog(RXPIN, TXPIN, %0000, BAUDRATE, SerPasm#NO_CTS, {
                                 }     pSerialConfig)
  ' If that failed, not much we can do here. (Although we could just hang...)

  ' Initialize our init-time serial support. This will be discarded after the
  ' end of initialization.\
  ' Hmmm.... we need the rx and tx buffers and sizes. We can move them here,
  ' I suupose...
  'PUB FinishInit(vars, cid, rx_bufptr, rx_size, tx_bufptr, tx_size)
  Ser.FinishInit(pSerialConfig, WORD[pCogIds + 0], {
        }        @Rx_Buffer, RX_BUFFER_SIZE, @Tx_Buffer, TX_BUFFER_SIZE)
  Ser.Str(string("Init-time output", CR))

  ' We could now use the serial rountines to get the color depth, for
  ' example.

  ' Prepare to reserve our (large) buffers
  if NextFree == 0
    ' First time initialization.
    NextFree := @FreeMemStart
  WORD[pBuffers + 0] := Fifo#HEADER_SIZE_BYTES * COMMAND_FIFO_SIZE
  WORD[pBuffers + 4] := DISPLAY_COLUMNS * REFRESH_ROWS * COLOR_DEPTH
  needed := MultiAllocate(pBuffers, 2) ' allocate two buffers.
  if needed <> 0
    return needed  ' We failed to reserve enough memory -- give up.

  ' Finally, start the display refresh cog and the writer cog.
  WORD[pCogIds + 2] := PanelPasm.LoadDisplayCog {
    }                    (COLOR_DEPTH, pDisplayBufAddr, pRefreshClocks)

  WORD[pCogIds + 4] := HelperPasm.LoadHelperCog {
    }                    (COLOR_DEPTH, pDisplayBufAddr, pCmdFifoHeader, {
    }                     pHelperDebugOutput)

  return 0

' ========================================================================
' Memory reservation/allocation routines.

PUB Allocate(size_in_bytes) | ptr
{{
  Try to reserve/allocate a buffer of the requested size.

  Arguments:
    size_in_bytes -- the size of the desired buffer.

  Return value:
    <> 0 -- the starting address of the reserved buffer; there was enough
            memory to satisfy an allocation of this amount.
    == 0 -- there is not enough memory to satisfy the allocation.
            The caller can use AdditionalNeeded(size_in_bytes) to find out
            the shortfall. 
}}
  ptr := AdditionalNeeded(size_in_bytes)
  if ptr > 0
    return 0   ' not enough space
  ' There is enough room -- get the pointer, and advance NextFree
  ptr := NextFree
  NextFree += (size_in_bytes + 3) & constant(!$3)
  return ptr

' ------------------------------------------------------------------------
PUB MultiAllocate(word_ptr, len) | p
{{
  Try to reserve/allocate a list of buffers of the requested sizes.

  Arguments:
    word_ptr -- the address of a list of pairs of WORDs, one pair for each
                buffer to allocate/reserve. The first WORD of the pair is
                the desired size of that buffer, and the second WORD
                receives the address of the allocated buffer. If the second
                word is non-zero on entry, that pair of WORDs is skipped,
                assuming that the buffer was already allocated.

                There can be one additional WORD at the end, with the value
                zero, which is used as a sentinel to stop repeated
                allocations when the "len" argument is zero.

    len -- The number of WORD pairs in the list "word_ptr", i.e. buffers" to
           allocation/reserve. This may be zero, in which case the
           allocations stop when an entry in the "word_ptr" list has a zero
           size to allocation.

  Return value:
    == 0 -- All the buffers in the "word_ptr" list were allocated.
    <> 0 -- There was not enough memory to allocate all the buffers.
            Any buffers that were allocated will have starting addresses
            in the "word_ptr" entries, and ones not allocated will (still)
            have 0 addresses. The return value is the shortfall of the
            amount of needed memory. 
}}
  ' len may be == 0, in which case this stops when we find an entry with
  ' a size == 0
  if len == 0
    len := 32767   ' no limit by count -- only use a sentinel
  p := word_ptr
  repeat while WORD[p] <> 0 and len-- > 0
    if WORD[p + 2] == 0  ' We skip entries already allocated.
      WORD[p + 2] := Allocate(WORD[p])
      if WORD[p + 2] == 0
        ' This allocation failed -- return how much more is needed.
        return MultiAdditionalNeeded(word_ptr, len)  ' Out of memory
    p += 4  ' advance
  return 0   ' all allocated

' ------------------------------------------------------------------------
PUB AdditionalNeeded(size_in_bytes) | available
{{
  Calculate how much more reserved memory is needed to satisfy an allocation
  request for some bytes of memory.

  If there is not enough memory, this returns the (positive) number of
  additional bytes needed. This value should be added to RESERVED_SIZE_BYTES
  to get the needed value for RESERVED_SIZE_BYTES (which may not be possible
  without shrinking the main stack too much).

  Arguments:
    size_in_bytes -- the size of the desired buffer.

  Return value:
    0   -- there is enough memory to satisfy an allocation of this amount.
    > 0 -- there is not enough memory to satisfy the allocation. The value
           is the number of bytes of additional memory needed (
           "shortfall"). 

  The space available, in bytes, goes from NextFree to VBASE, i.e.
    WORD[GLOBAL_VBASE_ADDR] - NextFree
}}
  available := AvailableBytes
  ' Round size_in_bytes up to a multiple of LONGs
  size_in_bytes := (size_in_bytes + 3) & constant(!$3)
  if available => size_in_bytes
    return 0  ' Enough room
  return size_in_bytes - available  ' how much more room is needed.

' ------------------------------------------------------------------------
PUB MultiAdditionalNeeded(word_ptr, len) | wanted 
{{
  Determine if there is enough available memory to reserve/allocate a list
  of buffers with individual sizes.

  Arguments:
    word_ptr -- the address of a list of pairs of WORDs, one pair for each
                buffer to allocate/reserve. The first WORD of the pair is
                the desired size of that buffer, and the second WORD
                receives the address of the allocated buffer. If the second
                word is non-zero on entry, that pair of WORDs is skipped,
                assuming that the buffer was already allocated.

                There can be one additional WORD at the end, with the value
                zero, which is used as a sentinel to stop repeated
                allocations when the "len" argument is zero.

    len -- The number of WORD pairs in the list "word_ptr", i.e. buffers" to
           allocation/reserve. This may be zero, in which case the
           allocations stop when an entry in the "word_ptr" list has a zero
           size to allocation.

  Return value:
    == 0 -- There is enough memory to allocate/reserve all the buffers.
    <> 0 -- There was not enough memory to allocate all the buffers.
            The return value is the shortfall of the amount of needed memory. 
}}
  ' len may be == 0, in which case this stops when we find an entry with
  ' a size == 0
  if len == 0
    len := 32767   ' no limit by count -- only use a sentinel
  wanted := 0
  repeat while WORD[word_ptr] <> 0 and len-- > 0
    if WORD[word_ptr + 2] == 0  ' Skip ones already allocated
      wanted += (WORD[word_ptr] + 3) & constant(!$3)
    word_ptr += 4
  return AdditionalNeeded(wanted)

' ------------------------------------------------------------------------
PUB AvailableBytes
{{
  Returns the number of bytes available for allocation/reservation.
}}
  if NextFree == 0
    ' First time initialization.
    NextFree := @FreeMemStart
  return WORD[GLOBAL_VBASE_ADDR] - NextFree

' ------------------------------------------------------------------------
PUB GetNextFree
{{
  Returns the hub address that would be the starting address of the next
  buffer to be allocated/reserved.

  Arguments:
    None

  Return value:
    The base addres that is the start of the next allocation.
}}
  return NextFree

' ========================================================================
PUB FillMainStack | p
{{
  Fills almost all of the unused main stack area with an "unlikely" value;
  When the program ends, that area can be examined to see how much was used
  and how much was not used.

  Arguments:
    None

  Return value:
    None
}}
  ' The OFFSET_FOR_MAIN_STACK_FILL is some extra distance into the stack to
  ' start writing, so that the LONGFILL does not overwrite any short-term
  ' stack it might be using.
  p := @p + OFFSET_FOR_MAIN_STACK_FILL
  LONGFILL(p, UNLIKELY_VALUE, (END_OF_STACK - p) >> 2)

' ========================================================================
' Capture all the ROM Font digit bitmaps for easier use as 1bpp bitmaps.
' This assumes the layout of the buffer to store all the bitmaps, so it
' should change in sync with the storage in the main program.

PUB CaptureAllDigits(p)
{{
  Copy the bitmaps for the digits 0..9 into 1bpp bitmaps (in WORD vectors).
  Also creates a "blank" digit.

  This assumes the memory layout has that digits consecutive from 0 to 9
  in memory with  WORD gaps for the DIGIT_SCROLL_V_SPACING rows.
  This must agree with the layout and use in the main program.

  Arguments:
    p - the address of the WORD vector that will store the bitmaps.
        It should be at least DIGIT_HEIGHT * 11 + DIGIT_SCROLL_V_SPACING * 12
        words long.

  Return value:
    None.
}}
  ' 
  WORDFILL(p, 0, constant(DIGIT_HEIGHT * 11 + DIGIT_SCROLL_V_SPACING * 12))
  CaptureDigitPixels("O", p += (               DIGIT_SCROLL_V_SPACING) << 1) ' Zero has a slash, "O" does not.
  CaptureDigitPixels("1", p += (DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING) << 1)
  CaptureDigitPixels("2", p += (DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING) << 1)
  CaptureDigitPixels("3", p += (DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING) << 1)
  CaptureDigitPixels("4", p += (DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING) << 1)
  CaptureDigitPixels("5", p += (DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING) << 1)
  CaptureDigitPixels("6", p += (DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING) << 1)
  CaptureDigitPixels("7", p += (DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING) << 1)
  CaptureDigitPixels("8", p += (DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING) << 1)
  CaptureDigitPixels("9", p += (DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING) << 1)

PRI CaptureDigitPixels(d, p) | vp, v, rdbit, wrbit, o
  ' Capture the pixels from the ROM character font for d (ASCII character)
  ' into the supplied WORD-based buffer, keeping only the 13 columns of
  ' interest 2..14 (this leaves off the empty columns), and the 22 rows
  ' of interest 5..26 (this leaves off some of the empty rows).

  ' Get the pixels for the pair of characters, skipping over ROM_FONT_TRIM_TOP
  ' rows of pixels, each row is a LONG, i.e. 4 bytes
  ' In constant terms, we skip over ROM_FONT_TRIM_TOP << 2 bytes
  vp := $8000 + ((d & $FE) << 6) + constant(ROM_FONT_TRIM_TOP << 2)
  repeat DIGIT_HEIGHT  ' rows
    v := LONG[vp] >> constant(ROM_FONT_TRIM_LEFT << 1)  ' drop blank columns on the left
    rdbit := 1 << (d & 1)
    wrbit := 1
    o := 0
    repeat DIGIT_WIDTH  ' bits wide
      if (v & rdbit) <> 0
        o |= wrbit
      wrbit <<= 1
      rdbit <<= 2
    WORD[p] := o
    p += 2
    vp += 4

