{{
Test counting up and down using the ROM font, for one or more HUB75 displays.

Author:         Tom Dinger (Ding-Batty)
Release Date:   2024/04/10

Copyright 2024 Thomas E. Dinger

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

--------------------------------------------------------------------------

This program can be configured for one or several daisy-chained HUB75 displays,
and it can be configured for several digit counts, and several ways to display
the digits. It also measures the time for various display updates, to report the
minimum time for those types of display updates, which can be useful when
configuring a particular application's display update approach.

Ways to show and update digits:
- single row of digits:
  - simple digit replace (fast replace) -- SimpleDigitsReplace()
  - slow digit replace, next digit replaces the previous one line at a
    time -- SlowDigitsReplace(), or Wipe
    - wipe directions:
      - up, down, left, right, verticalfrommiddle, verticaltomiddle, etc.
  - fade to black old digit, fade in new. (LATER)
  - fade to white (gray?) old digit, fade in new. (LATER)
- columns of digits, uniform color/intensity
  - scroll down from the top
  - scroll up from the bottom (same timing as scroll down)
  - interleaved scroll of pixel columns in the digit, half scroll up, half
    scroll down
- columns of digits, bold/different color for the "middle" digits.
  - scroll down from the top
  - scroll up from the bottom
  - interleaved scroll of pixel columns
    - single color during scroll, intensity/color change at end.
    - maintain three "layers" of colors.

How to measure time for updates:
- We need to use the feature of the helper opreations to get an indication that
  an operation is done, which we do for the final digit update.
- Advance all digits once, measure Sysclks from enqueuing the first write to the
  final write completing.
- Advance all the digits N times, measure the Sysclks a similar way:
  - They can all use the same completion status variable, as long as we only
    check it after the final command has been started/completed, so when the
    command FiFo is empty. (Needs a helper in the hub75_helper.spin code.) 
    - This is a problem -- instead, use a helper function that does a NoOp
      and waits for "done".
- Sysclks can be converted into an update rate per second, similarly to the
  calculation for refreshes per second for the display code.
  

Revisions:
==========
- 2024/04/10 -- Change from 8-bit color max, to 10-bit color max;
- 2024/02/27 -- Original release

}}

OBJ
  ' Pick up the static display configuration first.
  Config : "hub75_config.spin"

  Ser    : "Db_FullDuplexSerial"
  Umath  : "umath2.spin"
  Fifo   : "FifoBig.spin"
  Panel  : "hub75_display.spin"
  Helper : "hub75_helper.spin"

  SerPasm    : "Db_FullDuplexSerial_pasm"
  HelperPasm : "hub75_helper_pasm.spin"

  ' Pick exactly one of these to do the display refreshes.
  'PanelPasm  : "hub75_display_pasm_repeat.spin"
  PanelPasm  : "hub75_display_pasm_video_4clock.spin"
  'PanelPasm  : "hub75_display_pasm_video_8clock.spin"
  'PanelPasm  : "hub75_display_pasm_waitcnt.spin"


CON { General constants }
  _CLKMODE = XTAL1 + PLL16X 
  _XINFREQ = 5_000_000

  { Configuration for serial connection. }
  BAUDRATE = 115_200
  RXPIN    = 31    ' The usual pins.
  TXPIN    = 30
  RX_BUFFER_SIZE = 28
  TX_BUFFER_SIZE = 200

  CR = $0D
  LF = $0A

  { LED panel(s) configuration }
  COLOR_DEPTH     =  3   ' number of bits of resolution per color.
  MAX_COLOR_VALUE = (|< COLOR_DEPTH) - 1

  DISPLAY_ROWS       = Config#DISPLAY_ROWS
  REFRESH_ROWS       = Config#REFRESH_ROWS
  DISPLAY_COLUMNS    = Config#DISPLAY_COLUMNS
  MAX_DISPLAY_ROW    = Config#MAX_DISPLAY_ROW
  MAX_DISPLAY_COLUMN = Config#MAX_DISPLAY_COLUMN

  ' We use the digits from the ROM Font. They are 16 x 32, but some of the
  ' rows and columns are blank, so we only keep 13 x 20 -- every digit (and
  ' the capital letter "O", used for zero) fits into 13-pixels wide, and
  ' 20 pixels tall. Blank rows and columns are handled by the display
  ' management code.
  { Digit "shapes" }
  DIGIT_WIDTH  = 13
  DIGIT_HEIGHT = 22

  ' The number of pixels to trim a Rom Font digit to skip over the empty
  ' pixel rows and columns.
  ROM_FONT_TRIM_TOP  = 5
  ROM_FONT_TRIM_LEFT = 2

  ' We use a large command FiFo for sending commands to the display write helper
  ' cog, because we may want to enqueue quite a few commands at times. Things
  ' really depend on how fast the helper cog can process the commands. We could
  ' always get a speed-up by using more than one display writer cog, because
  ' they can write to the display buffer in parallel. That is an optimization
  ' for another day, if needed.
  COMMAND_FIFO_SIZE = 32

  ' Calculate some limits on digit display
  '
  ' COLUMNS per digit must be at least 13 -- there will be no blank pixels
  ' between columns of digits in that case.
  ' Note that for values > 13, the blank pixel columns will be on the _right_
  ' side of the digit column. Beware -- values < DIGIT_WIDTH might not work
  ' correctly without additional code changes.
  COLUMNS_PER_DIGIT     = 16  ' four across on a 64x64 panel, eight for two panels
  BLANK_PIXELS_ON_RIGHT = COLUMNS_PER_DIGIT - DIGIT_WIDTH ' 
  MAX_DIGIT_COLUMNS     = DISPLAY_COLUMNS / COLUMNS_PER_DIGIT
  'MAX_DIGIT_COLUMNS = 2 ' For initial testing

  ' Styles for the display.
  ' This specifies the layout and update style of the digit display.
  STYLE_1ROW_ROM_WIPE_RANDOM   = 1  ' Select each digit animation randomly from
                                    ' the ones below.

  MIN_STYLE_1ROW_ROM_WIPE            =  2
  STYLE_1ROW_ROM_REPLACE             =  2  ' immediate replacement
  STYLE_1ROW_ROM_WIPE_DOWN           =  3
  STYLE_1ROW_ROM_WIPE_UP             =  4
  STYLE_1ROW_ROM_WIPE_RIGHT          =  5
  STYLE_1ROW_ROM_WIPE_LEFT           =  6
  STYLE_1ROW_ROM_WIPE_TB_2_CTR       =  7
  STYLE_1ROW_ROM_WIPE_TB_F_CTR       =  8
  STYLE_1ROW_ROM_WIPE_LR_2_CTR       =  9
  STYLE_1ROW_ROM_WIPE_LR_F_CTR       = 10
  STYLE_1ROW_ROM_WIPE_FROM_CTR       = 11
  STYLE_1ROW_ROM_WIPE_TO_CENTER      = 12
  ' "Scrolls" are variations on "Wipes"
  STYLE_1ROW_ROM_SCROLL_LEFT         = 13
  STYLE_1ROW_ROM_SCROLL_RIGHT        = 14
  STYLE_1ROW_ROM_SCROLL_UP           = 15
  STYLE_1ROW_ROM_SCROLL_DOWN         = 16
  MAX_STYLE_1ROW_ROM_WIPE            = 16

  ' More wipe styles:
  ' down/up/left/right in halves
  ' Horizontal/vertical edge-to-center, center_to_edge
  ' out from center (both directions), in to center (both directions).
  '   this could have variations because the two directions are different
  '   distances.
  ' random column order, random row order
  ' diagonal wipes (not so easy due to rectilinear bias in implementation)
  ' WipeLRToCenter WipeLRFromCenter
  ' WipeTBToCenter WipeTBToCenter
  ' WipeFromCenter, WipeToCenter (both directions).

  ' - Scrolling changes
  '   - up/down/left-right
  '   - alternate columns (not so easy for ROM Font operations).
  ' More replacement styles:
  ' - Variations on sparkle.
  '   These would have a different animation delay behavior:
  '   - random order of updating the DIGIT_HEIGHT * DIGIT_WIDTH pixels
  '   - random sparkles before the per-pixel updates

  ' Counting direction.
  ' Changing direction of counting for some display styles can be complicated.
  COUNT_DIRECTION_UP   = 0
  COUNT_DIRECTION_DOWN = 1

VAR
  ' Variables used in Spin, that are initialized at runtime (not by the compiler)
  ' Display management
  ' The approach to display update is to have the display update routine maintain
  ' the current state of the display, including partial updates, and have
  ' the other code set the "target" or desired "final" state of the display.
  '
  ' The display update routine "owns" the current display state variables, and
  ' updates them as the only writer for them. Other code may read them, but as
  ' they can change at any time, they are advisory.
  '
  ' The target display state is only read by the display update code, and can
  ' change at any time during update, so the display update code has to be able
  ' to handle changes at arbitrary times.

  ' The leftmost pixel column for each digit (or column of digits).
  ' Set when the display layout is initialized, based on the desired number
  ' of digit positions on the display.
  WORD    DigitColumn [MAX_DIGIT_COLUMNS]

  ' These describe the position of the "middle" row of digits: the display
  ' row for the top of the digits, and the height of the digits.
  BYTE    CenterRowTop
  BYTE    CenterRowHeight

  ' The style of display: how to update the display, the number of digits,
  ' and the direction for counting-style scrolling (not in this version of
  ' the test program).
  BYTE    DisplayStyle
  BYTE    DigitCount
  BYTE    CountDirection

  ' The list of the currently displayed values for each digit.
  ' The values may be 0..0 or ASCII codes, depending on the style of update.
  BYTE    CurrentDigitValue [MAX_DIGIT_COLUMNS]

  ' The list of the values that should be displayed for each digit.
  ' The display update cog will make the display changes based on the
  ' transformation of the current value displayed to the target value.
  BYTE    TargetDigitValue [MAX_DIGIT_COLUMNS]

  ' The following are the working (private) storage for managing the digit
  ' display updates.
  '
  ' We can have a different update style for each digit
  BYTE    WorkingDisplayStyle [MAX_DIGIT_COLUMNS]
  ' The target digit value, captured to keep it constant during the update
  BYTE    WorkingDigitValue [MAX_DIGIT_COLUMNS]
  ' The following tracks the progress through the digit update animation.
  LONG    WorkingChangeProgress [MAX_DIGIT_COLUMNS]

  ' The number of sysclks to delay per pixel row update for digit change
  ' animations
  LONG    AnimationDelay
  ' For random selection of Wipe style per digit change.
  LONG    WipeRandom

  ' Working (private) storage for the tests

  LONG    TestDigitValue [MAX_DIGIT_COLUMNS]
  BYTE    TestDigitString[12]

DAT
' Define the variables that contain the operating parameters.
' We use a DAT section so we can initialize them, which means there can be only
' one instance of this object; if we used a VAR section, we'd have to devote
' code to initializing the VARs, which might happen later, depending on how the
' code develops.

' The storage for the command Fifo entries, used in the helper object to
' communicate with the helper PASM cog.
' We could allocate this any number of ways: VAR, DAT, "dynamically" like the
' display buffer.
CmdFifoData      LONG    0 [Fifo#HEADER_SIZE_LONGS * COMMAND_FIFO_SIZE]

' Storage for debug output from the helper driver
' See that driver for details on the contents.
' Should be at least 10 longs
HelperDebugOutput LONG    0[16]

' Color values used in the tests.
BoldColorRgb      LONG    (MAX_COLOR_VALUE << 20) | (MAX_COLOR_VALUE << 10) | 0  ' bright yellow
BoldColorDisplay  LONG    0  ' Filled in at runtime.

DimColorRgb       LONG    (0 << 20) | (1 << 10) | 0  ' dim green
DimColorDisplay   LONG    0  ' Filled in at runtime
               
' The stack for the cog that resolves differences between the current state of
' the display and the target state.
' Start out with 64 entries -- we can probably do with many fewer, perhaps as
' few as 8.
ResolutionCog      LONG    0   ' 1 + Cog# for the resolution cog
ResolutionCogStack LONG    0 [64]

Rx_Buffer          LONG    0[(RX_BUFFER_SIZE + 3) >> 2]
Tx_Buffer          LONG    0[(TX_BUFFER_SIZE + 3) >> 2]

' The display buffer stores all color layers for all displays.
' Someday it will overlap the PASM objects that are initilaized (loaded into
' cogs) once at the start of operations, to maximize the amount of memory
' available for displays.
'
' One byte of memory stores one color bit layer for two pixels.

DisplayBuffer   BYTE 0 [DISPLAY_COLUMNS * REFRESH_ROWS * COLOR_DEPTH]

' ========================================================================
PUB Main | dcog, hcog  ' , r, c, color, dimcolor, rt
  waitcnt(clkfreq * 4 + cnt)   ' Wait 4 seconds at startup.
  InitFDS

  ' Load the display refresh cog and the write-helper cog, and then tell
  ' the drivers about the two cogs. Note that telling the display driver
  ' about the address of the display buffer will unblock that cog _and_
  ' the helper cog at the same time, so that should be done last.
  dcog := StartDisplayCog
  hcog := StartHelperCog
  Helper.FinalHelperInit(hcog, @DisplayBuffer, COLOR_DEPTH, @CmdFifoData, {
           }             Fifo#HEADER_SIZE_BYTES * COMMAND_FIFO_SIZE)
  Panel.FinalDisplayInit(dcog, @DisplayBuffer, COLOR_DEPTH)
  ReportDec(string("Started display cog: "), dcog - 1)
  ReportDec(string("Started helper cog: "), hcog - 1)

  ' Get the display colors
  BoldColorDisplay := Panel.ToDisplayColor(BoldColorRgb)
  DimColorDisplay := Panel.ToDisplayColor(DimColorRgb)
  ReportDec(string("Maximum number of digits across: "), MAX_DIGIT_COLUMNS)
  ReportDec(string("Bits per color: "), COLOR_DEPTH)
  WipeRandom := 42  ' Init a random number stream.

  ' TODO:
  ' - Compute the display layout, for the type of digit display and update.
  ' - Set the initial display conditions.
  ' - Start the display of digits, and the (possible) cogs that update the
  '   digit display based on any changes.
  ' - Continue testing digit display changes, until the user says to stop.

  'SelectStyle(STYLE_1ROW_ROM_REPLACE, 2)
  ' This does the layout prep and update cog init for that number of
  ' digits (if it will fit onto the display).

  ' Change a random digit to a random digit value, once a second.
  ' Stop when the user presses a key.
  'TestRandomDigits
  '
  'TestCountUpFromZero(29)
  'TestCountDownToZero(29)

  'SelectStyle(STYLE_1ROW_ROM_WIPE_DOWN, 4)
  'TestCountUpFromZero(299)
  'TestCountDownToZero(299)
  'TestCountUpFromZeroJumps(1999, 23)

  'SelectStyle(STYLE_1ROW_ROM_WIPE_UP, 4)
  'TestCountUpFromZeroJumps(1999, 23)

  'SelectStyle(STYLE_1ROW_ROM_WIPE_TO_CENTER, 4)
  'TestCountUpFromZeroJumps(1999, 23)
  ''TestSlowCountUpFromZero(12)

  if MAX_DIGIT_COLUMNS => 5
    SelectStyle(STYLE_1ROW_ROM_WIPE_RANDOM, 5)
    TestCountUpFromZeroJumps(49999, 383)
  else
    SelectStyle(STYLE_1ROW_ROM_WIPE_RANDOM, 4)
    TestCountUpFromZeroJumps(4999, 37)

  ' Stall, waiting for input from the user
  ReportStr(string("Press a key to stop the display"))
  Ser.Rx  ' Stall until a key press, and discard the character
  ShutdownDisplay
  Panel.Stop
  ReportStr(string("Display cog stopped"))
  Helper.Stop
  ReportStr(string("Helper cog stopped"))
  repeat  ' wait forvever.

PRI InitFDS | SerialCog
  SerialCog := SerPasm.LoadCog(RXPIN, TXPIN, %0000, BAUDRATE, Ser#NO_CTS, Ser.GetInitVars)
  if SerialCog == 0
    ' Could not start the serial cog -- just hang.
    repeat
      ' Stall
  Ser.FinishInit(serialcog, @Rx_Buffer, RX_BUFFER_SIZE, {
        }                   @Tx_Buffer, TX_BUFFER_SIZE)

PUB StartDisplayCog
  return PanelPasm.LoadDisplayCog(COLOR_DEPTH, Panel.GetDisplayBufAddrVar, {
           } Panel.GetRefreshClocksAddrHub)

PUB StartHelperCog
  return HelperPasm.LoadHelperCog(COLOR_DEPTH, Panel.GetDisplayBufAddrVar, {
           } Helper.GetCmdFifoHeader, @HelperDebugOutput)

' ------------------------------------------------------------------------

PUB TestRandomDigits | rg, next_cnt, change_delay, r, d, digit_col
  ' Once a second, change one of the digits to another (random) value.
  ReportStr(string("Change random digits"))
  ReportStr(string("Press a key to stop the digit changes"))
  ' Initialize the display to all zeros
  repeat digit_col from 0 to (DigitCount - 1)
    SetDigit(digit_col, 0)
  change_delay := CLKFREQ >> 1
  rg := 42
  next_cnt := CNT
  repeat while Ser.RxCheck < 0
    waitcnt(next_cnt += change_delay)
    r := (?rg) & $7FFFFFFF
    d := r // 10
    digit_col := (r / 10) // DigitCount
    SetDigit(digit_col, d)

PUB TestCountDownToZero(startvalue) | digit_col, next_cnt, change_delay, s
  startvalue <#= PowerOfTen(DigitCount) - 1
  ReportDec(string("Count down to zero from "), startvalue)
  ReportStr(string("Press a key to stop the test early"))
  change_delay := CLKFREQ  ' one second delay

  ' Initialize the digits in the columns
  s := startvalue
  ' We really want to set digits most significant first, to get the
  ' right leading space behavior. We get around the problem by converting
  ' directly into DigitValues[] without updating the target bitmap, then
  ' go back and update all the digits, left-to-right
  ' all the digits into a table, then 
  repeat digit_col from (DigitCount - 1) to 0
    'SetDigitLeadingSpaces(digit_col, s // 10)
    TestDigitValue[digit_col] := s // 10
    s /= 10
  SetAllDigitsLeadingSpaces(@TestDigitValue)
  WaitForNoDifferences
  SetCountDirection(COUNT_DIRECTION_DOWN)
  next_cnt := CNT
  repeat while (startvalue-- > 0) and (Ser.RxCheck < 0)
    waitcnt(next_cnt += change_delay)
    ' Subtract one from the Digit values, but update the digits
    ' left-to-right, which allows leading spaces to be displayed
    ' instead of leading zeros.
    ' Algorithm:
    ' - Find the Rightmost digit that is not zero.
    ' - subtract one from that digit
    ' - For all digits to the right of that one, set them to 9.
    ' Because startvalue was non-zero, we can assume that we will find
    ' a non-zero digit.
    digit_col := DigitCount - 1
    repeat while TestDigitValue[digit_col] == 0 and digit_col > 0
      --digit_col
    SetDigitLeadingSpaces(digit_col, --TestDigitValue[digit_col])
    'ReportDec(string("Column decremented: "), digit_col)
    'ReportDec(string("Column new value: "), TargetDigitValue[digit_col])
    repeat while ++digit_col < DigitCount
      SetDigitLeadingSpaces(digit_col, TestDigitValue[digit_col] := 9)
      'ReportDec(string("Column decremented: "), digit_col)
      'ReportDec(string("Column new value: "), TargetDigitValue[digit_col])

PUB TestCountUpFromZero(startvalue) | digit_col, next_cnt, change_delay
  startvalue <#= PowerOfTen(DigitCount) - 1
  ReportDec(string("Count up from zero to "), startvalue)
  ReportStr(string("Press a key to stop the test early"))
  change_delay := CLKFREQ  ' one second delay

  ' Initialize the digits in the columns
  repeat digit_col from 0 to (MAX_DIGIT_COLUMNS - 1)
    SetDigitLeadingSpaces(digit_col, 0)
    TestDigitValue[digit_col] := 0
  WaitForNoDifferences
  SetCountDirection(COUNT_DIRECTION_UP)
  next_cnt := CNT
  repeat while (startvalue-- > 0) and (Ser.RxCheck < 0)
    waitcnt(next_cnt += change_delay)
    ' Add one to the Digit values, but update the digits
    ' left-to-right, which allows leading spaces to be displayed
    ' instead of leading zeros.
    ' Algorithm:
    ' - Find the Rightmost digit that is not nine.
    ' - Add one to that digit
    ' - For all digits to the right of that one, set them to 0.
    digit_col := DigitCount - 1
    repeat while TestDigitValue[digit_col] == 9 and digit_col > 0
      --digit_col
    SetDigitLeadingSpaces(digit_col, ++TestDigitValue[digit_col])
    'ReportDec(string("Column decremented: "), digit_col)
    'ReportDec(string("Column new value: "), TargetDigitValue[digit_col])
    repeat while ++digit_col < DigitCount
      SetDigitLeadingSpaces(digit_col, TestDigitValue[digit_col] := 0)
      'ReportDec(string("Column decremented: "), digit_col)
      'ReportDec(string("Column new value: "), TargetDigitValue[digit_col])


PUB TestSlowCountUpFromZero(startvalue) | digit_col, next_cnt, change_delay
  startvalue <#= PowerOfTen(DigitCount) - 1
  ReportDec(string("Count up from zero to "), startvalue)
  ReportStr(string("Press a key to stop the test early"))
  change_delay := CLKFREQ * 2  ' two second delay

  ' Initialize the digits in the columns
  repeat digit_col from 0 to (MAX_DIGIT_COLUMNS - 1)
    SetDigitLeadingSpaces(digit_col, 0)
    TestDigitValue[digit_col] := 0
  WaitForNoDifferences
  SetCountDirection(COUNT_DIRECTION_UP)
  ReportStr(string("Set count direction"))
  'next_cnt := CNT
  repeat while (startvalue-- > 0) and (Ser.RxCheck < 0)
    WaitForNoDifferences
    ReportStr(string("Incrementing the value"))
    waitcnt(cnt + change_delay)
    'waitcnt(next_cnt += change_delay)
    ' Add one to the Digit values, but update the digits
    ' left-to-right, which allows leading spaces to be displayed
    ' instead of leading zeros.
    ' Algorithm:
    ' - Find the Rightmost digit that is not nine.
    ' - Add one to that digit
    ' - For all digits to the right of that one, set them to 0.
    digit_col := DigitCount - 1
    repeat while TestDigitValue[digit_col] == 9 and digit_col > 0
      --digit_col
    SetDigitLeadingSpaces(digit_col, ++TestDigitValue[digit_col])
    'ReportDec(string("Column decremented: "), digit_col)
    'ReportDec(string("Column new value: "), TargetDigitValue[digit_col])
    repeat while ++digit_col < DigitCount
      SetDigitLeadingSpaces(digit_col, TestDigitValue[digit_col] := 0)
      'ReportDec(string("Column decremented: "), digit_col)
      'ReportDec(string("Column new value: "), TargetDigitValue[digit_col])

PUB TestCountUpFromZeroJumps(endvalue, incr) | digit_col, next_cnt, change_delay, v
  endvalue <#= PowerOfTen(DigitCount) - 1
  ReportDec(string("Count up from zero to "), endvalue)
  ReportDec(string("In steps of "), incr)
  ReportStr(string("Press a key to stop the test early"))
  change_delay := CLKFREQ  ' one second delay

  ' Initialize the digits in the columns
  v := 0
  SetTestDigits(v)
  repeat digit_col from 0 to (DigitCount - 1)
    SetDigitLeadingSpaces(digit_col, TestDigitValue[digit_col])
    'TestDigitValue[digit_col] := 0
  WaitForNoDifferences
  SetCountDirection(COUNT_DIRECTION_UP)
  next_cnt := CNT
  repeat while ((v += incr) < endvalue) and (Ser.RxCheck < 0)
    waitcnt(next_cnt += change_delay)
    ' Add the increment to v, and redisplay the digits
    SetTestDigits(v)
    repeat digit_col from 0 to (DigitCount - 1)
      SetDigitLeadingSpaces(digit_col, TestDigitValue[digit_col])


PUB SetTestDigits(v) | len,  prefix_len, digit_col
  ' Fill in the TestDigitValue[] vector from the representation of the
  ' number. Currently only non-negative numbers.
  'v <#= (PowerOfTen(DigitCount) - 1)
  'v #>= 0
  'ReportDec(string("Value: "), v)

  len := udec32tostr(v, @TestDigitString)
  prefix_len := DigitCount - len
  'ReportDec(string("Prefix length: "), prefix_len)
  digit_col := 0
  repeat while digit_col < prefix_len
    TestDigitValue[digit_col++] := 0
  repeat while digit_col < DigitCount
    'ReportDec(string("Column: "), digit_col)
    'ReportDec(string("Column ASCII: "), TestDigitString[digit_col - prefix_len])
    TestDigitValue[digit_col] := TestDigitString[digit_col - prefix_len] - "0"
    'ReportDec(string("Column value: "), TestDigitValue[digit_col])
    ++digit_col
    
' ------------------------------------------------------------------------

PUB SelectStyle(style, num_digits)
  ' Set things up for the selected style and digit count:
  ' - Prepare the layout of the digits on the display
  ' - Init the display state variables, and start the display update cog.
  if (style <> DisplayStyle) or (num_digits <> DigitCount)
    ShutdownDisplay  ' stop the display update cog
    Panel.Clear
    DisplayStyle := style
    DigitCount := num_digits
    PrepareLayout
    InitDisplay

PUB PrepareLayout
  ' Based on the selected style, set up the initial display layout and
  ' conditions.
  'case DisplayStyle
  '  STYLE_1ROW_ROM_REPLACE: PrepareLayout1RowRom
  '  STYLE_1ROW_ROM_WIPE_DOWN: PrepareLayout1RowRom
  '  STYLE_1ROW_ROM_WIPE_UP: PrepareLayout1RowRom
  '  STYLE_1ROW_ROM_WIPE_RIGHT: PrepareLayout1RowRom
  '  STYLE_1ROW_ROM_WIPE_LEFT: PrepareLayout1RowRom
  '  STYLE_1ROW_ROM_WIPE_RANDOM: PrepareLayout1RowRom
  '  ' ...
  '  other: PrepareLayout1RowRom
  ' For now, all styles can use the same set-up
  PrepareLayout1RowRom
  
PUB InitDisplay
  ' Based on the selectoed style, present the initial digit display.
  'case DisplayStyle
  '  STYLE_1ROW_ROM_REPLACE: InitDisplay1RowRom
  '  STYLE_1ROW_ROM_WIPE_DOWN: InitDisplay1RowRom
  '  STYLE_1ROW_ROM_WIPE_UP: InitDisplay1RowRom
  '  STYLE_1ROW_ROM_WIPE_RIGHT: InitDisplay1RowRom
  '  STYLE_1ROW_ROM_WIPE_LEFT: InitDisplay1RowRom
  '  STYLE_1ROW_ROM_WIPE_RANDOM: InitDisplay1RowRom
  '  other: InitDisplay1RowRom
  ' For now, all styles can use the same set-up
  InitDisplay1RowRom

PUB ShutdownDisplay 
  ShutdownDisplayBase

PUB ShutdownDisplayBase
  if ResolutionCog
    ReportStr(string("Stopping resolution cog"))
    COGSTOP(ResolutionCog~ - 1)

PUB SetDigit(col, d)
  'case DisplayStyle
  '  STYLE_1ROW_ROM_REPLACE: SetDigit1RowRom(col, d)
  '  STYLE_1ROW_ROM_WIPE_DOWN: SetDigit1RowRom(col, d)
  '  STYLE_1ROW_ROM_WIPE_UP: SetDigit1RowRom(col, d)
  '  STYLE_1ROW_ROM_WIPE_RIGHT: SetDigit1RowRom(col, d)
  '  STYLE_1ROW_ROM_WIPE_LEFT: SetDigit1RowRom(col, d)
  '  STYLE_1ROW_ROM_WIPE_RANDOM: SetDigit1RowRom(col, d)
  '  other: SetDigit1RowRom(col, d)
  SetDigit1RowRom(col, d)

PUB SetAllDigits(vector) | col
  repeat col from 0 to (DigitCount - 1)
    SetDigit(col, LONG[vector][col])

PUB SetDigitLeadingSpaces(col, d)
  'case DisplayStyle
  '  STYLE_1ROW_ROM_REPLACE: SetDigitSpaces1RowRom(col, d)
  '  STYLE_1ROW_ROM_WIPE_DOWN: SetDigitSpaces1RowRom(col, d)
  '  STYLE_1ROW_ROM_WIPE_UP: SetDigitSpaces1RowRom(col, d)
  '  STYLE_1ROW_ROM_WIPE_RIGHT: SetDigitSpaces1RowRom(col, d)
  '  STYLE_1ROW_ROM_WIPE_LEFT: SetDigitSpaces1RowRom(col, d)
  '  STYLE_1ROW_ROM_WIPE_RANDOM: SetDigitSpaces1RowRom(col, d)
  '  other: SetDigitSpaces1RowRom(col, d)
  SetDigitSpaces1RowRom(col, d)

PUB SetAllDigitsLeadingSpaces(vector) | col
  repeat col from 0 to (DigitCount - 1)
    SetDigitLeadingSpaces(col, LONG[vector][col])
  
PUB SetCountDirection(dir)
  'case DisplayStyle
  '  STYLE_1ROW_ROM_REPLACE: SetCountDirection1RowRom(dir)
  '  STYLE_1ROW_ROM_WIPE_DOWN: SetCountDirection1RowRom(dir)
  '  STYLE_1ROW_ROM_WIPE_UP: SetCountDirection1RowRom(dir)
  '  STYLE_1ROW_ROM_WIPE_RIGHT: SetCountDirection1RowRom(dir)
  '  STYLE_1ROW_ROM_WIPE_LEFT: SetCountDirection1RowRom(dir)
  '  STYLE_1ROW_ROM_WIPE_RANDOM: SetCountDirection1RowRom(dir)
  '  other: SetCountDirection1RowRom(dir)
  SetCountDirection1RowRom(dir)

' We might be able to use one routine for all styles here.
' But it depends on how the display is managed...
PUB AnyDifferences
  return AnyDifferencesBase

PUB AnyDifferencesBase | col
  ' Return true if any digit columns still need updates
  repeat col from 0 to (DigitCount - 1)
    if CurrentDigitValue[col] <> TargetDigitValue[col]
      'ReportDec(string("Different Column: "), col)
      'ReportDec(string("   Current digit: "), CurrentDigitValue[col])
      'ReportDec(string("    Target digit: "), TargetDigitValue[col])
      'waitcnt(cnt + clkfreq)
      return 1
  return 0

PUB WaitForNoDifferences
   repeat while AnyDifferences
     ' Stall

' ------------------------------------------------------------------------
' Single row display from ROM Fonts, all animations

' All the Wipe styles could use this (and Replace, too)
PUB PrepareLayout1RowRom | i
  ' A single row of digits in the center.
  ' We pick the display row number for the top pixel row of the digits.

  CenterRowTop    := (DISPLAY_ROWS - DIGIT_HEIGHT) / 2
  CenterRowHeight := DIGIT_HEIGHT

  ' Assign columns for the digit positions
  DigitColumn[0] := {
    }  (DISPLAY_COLUMNS - ((COLUMNS_PER_DIGIT * (DigitCount - 1)) + DIGIT_WIDTH)) / 2
  WorkingDigitValue[0] := CurrentDigitValue[0] := " "
  SetDigitSpaces1RowRom(0, 0)
  if DigitCount > 1
    repeat i from 1 to (DigitCount - 1)
      DigitColumn[i] := DigitColumn[i - 1] + COLUMNS_PER_DIGIT
      WorkingDigitValue[i] := CurrentDigitValue[i] := " "
      SetDigitSpaces1RowRom(i, 0)

  ' Calculate the per-pixel-row delay for digit change animations.
  ' One digit animation should take at most about 1/2 second.
  ' The following calculates a time for complete digit change to be
  ' at most 5/9 of a second.
  ' 
  AnimationDelay := (CLKFREQ * 5) / (DIGIT_HEIGHT * 9)
  'AnimationDelay := CLKFREQ  ' for slow testing, to see what's happening.


PUB InitDisplay1RowRom | i
  ' Start the cog that will look for differences between the target digit
  ' value, the working digit value (for transitions-in-progress) and the
  ' current (previous by still partially visible) digit value.
  ' This cog may have "speed limits" to make any digit change animations
  ' visible (or not).
  ResolutionCog := 1 + COGNEW(Update1RowRom, @ResolutionCogStack)
  ReportDec(string("Started resolution cog: "), ResolutionCog - 1)

'' All styles might be able to share this.
'PUB ShutdownDisplaySingleRowWipeDown
'  if ResolutionCog
'    Ser.str(TESTPORT, string("Stopping resolution cog", CR))
'    COGSTOP(ResolutionCog~ - 1)


PUB Update1RowRom | i, do_delay, prev_cnt
  ' Constantly look for differences between the current and target bitmap
  ' for each digit, and immediately draw the new bitmap.
  ' This runs in a separate cog.
  ' For the "wipe" animation:
  ' - If WorkingDigitValue[i] <> CurrentDigitValue[i], we are inside an ongoing animation
  ' - Else if TargetDigitValue[i] <> CurrentDigitValue[i], we are starting a new animation
  ' - otherwise the column is not changing.
  ' Each changing column in get a new character pixel row written, then there will be a
  ' delay to "spread out" the row writes. We only delay if any rows were written on the pass.
  ' We try to time the writes so that one digit is updated in about 0.3 to 0.5 seconds.
  ' TODO: Implement the above
  prev_cnt := CNT
  repeat
    do_delay := 0
    repeat i from 0 to (DigitCount - 1)
      if CurrentDigitValue[i] <> WorkingDigitValue[i]
        do_delay |= ContinueUpdate1DigitRom(i)
      elseif CurrentDigitValue[i] <> TargetDigitValue[i]
        do_delay |= StartNewUpdate1DigitRom(i)
      'if CurrentDigitValue[i] <> TargetDigitValue[i]
      '  Helper.RomCharacter(TargetDigitValue[i], DigitColumn[i], CenterRowTop, {
      '    }                 DIGIT_WIDTH, CenterRowHeight, ROM_FONT_TRIM_LEFT,  {
      '    }                 ROM_FONT_TRIM_TOP, 0, BoldColorDisplay)
      '  CurrentDigitValue[i] := TargetDigitValue[i]
    ' If any of the digits had an update that needs a delay, do that delay now.
    if (do_delay)
      prev_cnt := AnimationDelay1RowRom(prev_cnt)
    else
      prev_cnt := CNT

PUB AnimationDelay1RowRom(prev_cnt)
  ' delay a short amount of time
  waitcnt(prev_cnt += AnimationDelay)
  return prev_cnt

PUB StartNewUpdate1DigitRom(digit_col)
  ' What to do depends on the digit change animation style.
  ' Returns 1 if a change was made requiring a per-row animation delay
  ' Returns 0 otherwise.
  WorkingDigitValue[digit_col] := TargetDigitValue[digit_col]
  ' Set the internal work information (current pixel row being updated for
  ' example) at the starting pixel row, which gets updated each time
  ' ContinueUpdate1DigitRom(digit_col) is called.
  ' (Something will be done here once we have animations.)
  WorkingChangeProgress[digit_col] := 0
  ' We could also select the digit-specific display style here, e.g.
  ' randomly selected from a range of styles.
  if DisplayStyle == STYLE_1ROW_ROM_WIPE_RANDOM
    WorkingDisplayStyle[digit_col] := {
      }   ((?WipeRandom & $7FFF_FFF) // constant(MAX_STYLE_1ROW_ROM_WIPE - MIN_STYLE_1ROW_ROM_WIPE + 1)) {
      }   + MIN_STYLE_1ROW_ROM_WIPE
  else
    WorkingDisplayStyle[digit_col] := DisplayStyle

  ' Now, do the first (or only) update for the digit
  return ContinueUpdate1DigitRom(digit_col)

PUB ContinueUpdate1DigitRom(digit_col)
  ' What to do depends on the digit change animation style.
  ' Returns 1 if a change was made requiring a per-row animation delay
  ' Returns 0 otherwise.
  case WorkingDisplayStyle[digit_col]
    STYLE_1ROW_ROM_REPLACE: return ContUpd1DigitRomReplace(digit_col)
    STYLE_1ROW_ROM_WIPE_DOWN: return ContUpd1DigitRomWipeDown(digit_col)
    STYLE_1ROW_ROM_WIPE_UP: return ContUpd1DigitRomWipeUp(digit_col)
    STYLE_1ROW_ROM_WIPE_RIGHT: return ContUpd1DigitRomWipeRight(digit_col)
    STYLE_1ROW_ROM_WIPE_LEFT: return ContUpd1DigitRomWipeLeft(digit_col)
    STYLE_1ROW_ROM_WIPE_TB_2_CTR:
      return ContUpd1DigitRomWipeTB2C(digit_col)
    STYLE_1ROW_ROM_WIPE_TB_F_CTR:
      return ContUpd1DigitRomWipeTBFC(digit_col)
    STYLE_1ROW_ROM_WIPE_LR_2_CTR:
      return ContUpd1DigitRomWipeLR2C(digit_col)
    STYLE_1ROW_ROM_WIPE_LR_F_CTR:
      return ContUpd1DigitRomWipeLRFC(digit_col)
    STYLE_1ROW_ROM_WIPE_FROM_CTR:
      return ContUpd1DigitRomWipeFromCenter(digit_col)
    STYLE_1ROW_ROM_WIPE_TO_CENTER:
      return ContUpd1DigitRomWipeToCenter(digit_col)
    STYLE_1ROW_ROM_SCROLL_LEFT:
      return ContUpd1DigitRomScrollLeft(digit_col)
    STYLE_1ROW_ROM_SCROLL_RIGHT:
      return ContUpd1DigitRomScrollRight(digit_col)
    STYLE_1ROW_ROM_SCROLL_UP:
      return ContUpd1DigitRomScrollUp(digit_col)
    STYLE_1ROW_ROM_SCROLL_DOWN:
      return ContUpd1DigitRomScrollDown(digit_col)
    other: return ContUpd1DigitRomReplace(digit_col)

PUB SetDigitSpaces1RowRom(col, d)
  ' To handle leading zeros becoming spaces, whenever we set a column
  ' digit to zero, we look at the column to the left: if it is a space
  ' and not zero, we use the address for the space.
  ' This only works if digits are changed from left to right when updating
  ' several at once.
  if d < 10
    ' must be 0..9
    d += "0"

  if d == "0" and col <> (DigitCount - 1)
    ' if this is the leftmost column, or if the column to the left contains
    ' a space, then we use a space, 
    if (col == 0) or (TargetDigitValue[col - 1] == " ")
      ' leading spaces to the left -- use a space here as well
      TargetDigitValue[col] := " "
      'TargetDigitBitmapAddr[col] := @Digits0To9_Space
      return
  ' Otherwise, normal digit display
  SetDigit1RowRom(col, d)

' All the Wipe styles could use this (and Replace, too)
PUB SetDigit1RowRom(col, d)
  if d < 10
    ' must be 0..9
    d += "0"
  'TargetDigitBitmapAddr[col] := GetScrollUpBitmapAddr(d)
  TargetDigitValue[col] := d

' All the Wipe styles could use this (and Replace, too)
PUB SetCountDirection1RowRom(dir)
  ' Since there is no scrolling for this display, style, we have nothing to
  ' do here.

'PUB AnyDifferences1RowRomWipeDown | col
'  ' Return true if any digit columns still need updates
'  repeat col from 0 to constant(MAX_DIGIT_COLUMNS - 1)
'    if CurrentDigitBitmapAddr[col] <> TargetDigitBitmapAddr[col]
'    return 1
'  return 0

' ------------------------------------------------------------------------
' Single row display, ROM Font, immediate update of digits (no animation).

PUB ContUpd1DigitRomReplace(digit_col)
  ' Simply display the new full digit from rom.
  ' Since there is no animation, return 0 to mean no animation delay is needed.
  Helper.RomCharacter(WorkingDigitValue[digit_col], {
    }                 DigitColumn[digit_col], CenterRowTop, {
    }                 DIGIT_WIDTH, CenterRowHeight, ROM_FONT_TRIM_LEFT,  {
    }                 ROM_FONT_TRIM_TOP, 0, BoldColorDisplay)
  CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]
  return 0

' ------------------------------------------------------------------------
' Single row display, ROM font, wipe down form the top.

PUB ContUpd1DigitRomWipeDown(digit_col) | pixel_row
  ' The intended behavior:
  ' For the current pixel row in the digit:
  ' - display that horizontal pixel row from the WorkingDigitValue[i]
  ' - increment the working horizontal pizel row
  ' - if the pixel row has reached the end of the digit bitmap:
  '   - Set the CurrentDigitValue[i] := WorkingDigitValue[i]
  ' return 1 to mean that an animation delay is wanted after this operation.

  ' Write a new digit pixel row based on WorkingChangeProgress[digit_col].
  pixel_row := WorkingChangeProgress[digit_col]++
  Helper.RomCharacter(WorkingDigitValue[digit_col], DigitColumn[digit_col], {
    }                 CenterRowTop + pixel_row, DIGIT_WIDTH, 1, {
    }                 ROM_FONT_TRIM_LEFT, ROM_FONT_TRIM_TOP + pixel_row, {
    }                 0, BoldColorDisplay)
  if WorkingChangeProgress[digit_col] => DIGIT_HEIGHT
    ' We just finished the animation, so change the current digit value to
    ' be the working digit value
    CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]

  return 1  ' we made a change
  
' ------------------------------------------------------------------------
' Single row display, ROM font, wipe up form the bottom.

PUB ContUpd1DigitRomWipeUp(digit_col) | pixel_row
  ' The intended behavior:
  ' For the current pixel row in the digit:
  ' - display that horizontal pixel row from the
  '   DIGIT_HEIGHT - WorkingDigitValue[digit_col]
  ' - decrement the working horizontal pizel row
  ' - if the pixel row has reached the top of the digit bitmap:
  '   - Set the CurrentDigitValue[i] := WorkingDigitValue[i]
  ' return 1 to mean that an animation delay is wanted after this operation.

  ' Write a new digit pixel row based on WorkingChangeProgress[digit_col].
  pixel_row := DIGIT_HEIGHT - (++WorkingChangeProgress[digit_col])
  Helper.RomCharacter(WorkingDigitValue[digit_col], DigitColumn[digit_col], {
    }                 CenterRowTop + pixel_row, DIGIT_WIDTH, 1, {
    }                 ROM_FONT_TRIM_LEFT, ROM_FONT_TRIM_TOP + pixel_row, {
    }                 0, BoldColorDisplay)
  if WorkingChangeProgress[digit_col] => DIGIT_HEIGHT
    ' We just finished the animation, so change the current digit value to
    ' be the working digit value
    CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]

  return 1  ' we made a change

' ------------------------------------------------------------------------
' Single row display, ROM font, wipe from left to right

PUB ContUpd1DigitRomWipeRight(digit_col) | pixel_col
  ' The intended behavior:
  ' For the current pixel row in the digit:
  ' - display that vertical pixel column from the WorkingDigitValue[i]
  ' - increment the working vertical pixel column
  ' - if the pixel column has reached the end of the digit bitmap:
  '   - Set the CurrentDigitValue[i] := WorkingDigitValue[i]
  ' return 1 to mean that an animation delay is wanted after this operation.

  ' Write a new digit pixel row based on WorkingChangeProgress[digit_col].
  pixel_col := WorkingChangeProgress[digit_col]++
  Helper.RomCharacter(WorkingDigitValue[digit_col], {
    }                 DigitColumn[digit_col] + pixel_col, CenterRowTop, {
    }                 1, DIGIT_HEIGHT, {
    }                 ROM_FONT_TRIM_LEFT + pixel_col, ROM_FONT_TRIM_TOP, {
    }                 0, BoldColorDisplay)
  if WorkingChangeProgress[digit_col] => DIGIT_WIDTH
    ' We just finished the animation, so change the current digit value to
    ' be the working digit value
    CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]

  return 1  ' we made a change
  
' ------------------------------------------------------------------------
' Single row display, ROM font, wipe from right to left

PUB ContUpd1DigitRomWipeLeft(digit_col) | pixel_col
  ' The intended behavior:
  ' For the current pixel row in the digit:
  ' - display that vertical pixel column from the WorkingDigitValue[i]
  ' - increment the working vertical pixel column
  ' - if the pixel column has reached the end of the digit bitmap:
  '   - Set the CurrentDigitValue[i] := WorkingDigitValue[i]
  ' return 1 to mean that an animation delay is wanted after this operation.

  ' Write a new digit pixel row based on WorkingChangeProgress[digit_col].
  pixel_col := DIGIT_WIDTH - (++WorkingChangeProgress[digit_col])
  Helper.RomCharacter(WorkingDigitValue[digit_col], {
    }                 DigitColumn[digit_col] + pixel_col, CenterRowTop, {
    }                 1, DIGIT_HEIGHT, {
    }                 ROM_FONT_TRIM_LEFT + pixel_col, ROM_FONT_TRIM_TOP, {
    }                 0, BoldColorDisplay)
  if WorkingChangeProgress[digit_col] => DIGIT_WIDTH
    ' We just finished the animation, so change the current digit value to
    ' be the working digit value
    CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]

  return 1  ' we made a change
  
' ------------------------------------------------------------------------
' Single row display, ROM font, wipe down from the top and up from bottom

PUB ContUpd1DigitRomWipeTB2C(digit_col) | top_pixel_row, bottom_pixel_row
  ' Write a new digit pixel row based on WorkingChangeProgress[digit_col].
  top_pixel_row := WorkingChangeProgress[digit_col]++
  bottom_pixel_row := DIGIT_HEIGHT - top_pixel_row - 1
  Helper.RomCharacter(WorkingDigitValue[digit_col], DigitColumn[digit_col], {
    }                 CenterRowTop + top_pixel_row, DIGIT_WIDTH, 1, {
    }                 ROM_FONT_TRIM_LEFT, ROM_FONT_TRIM_TOP + top_pixel_row, {
    }                 0, BoldColorDisplay)
  if top_pixel_row <> bottom_pixel_row
    Helper.RomCharacter(WorkingDigitValue[digit_col], DigitColumn[digit_col], {
      }                 CenterRowTop + bottom_pixel_row, DIGIT_WIDTH, 1, {
      }                 ROM_FONT_TRIM_LEFT, ROM_FONT_TRIM_TOP + bottom_pixel_row, {
      }                 0, BoldColorDisplay)
  if top_pixel_row => constant((DIGIT_HEIGHT + 1) >> 1)
    ' We just finished the animation, so change the current digit value to
    ' be the working digit value
    CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]

  return 1  ' we made a change
  
' ------------------------------------------------------------------------
' Single row display, ROM font, wipe form the center down to the bottom and up to the top

PUB ContUpd1DigitRomWipeTBFC(digit_col) | top_pixel_row, bottom_pixel_row
  ' Write a new digit pixel row based on WorkingChangeProgress[digit_col].
  top_pixel_row := constant((DIGIT_HEIGHT - 1) >> 1) - (WorkingChangeProgress[digit_col]++)
  bottom_pixel_row := DIGIT_HEIGHT - top_pixel_row - 1
  Helper.RomCharacter(WorkingDigitValue[digit_col], DigitColumn[digit_col], {
    }                 CenterRowTop + top_pixel_row, DIGIT_WIDTH, 1, {
    }                 ROM_FONT_TRIM_LEFT, ROM_FONT_TRIM_TOP + top_pixel_row, {
    }                 0, BoldColorDisplay)
  if top_pixel_row <> bottom_pixel_row
    Helper.RomCharacter(WorkingDigitValue[digit_col], DigitColumn[digit_col], {
      }                 CenterRowTop + bottom_pixel_row, DIGIT_WIDTH, 1, {
      }                 ROM_FONT_TRIM_LEFT, ROM_FONT_TRIM_TOP + bottom_pixel_row, {
      }                 0, BoldColorDisplay)
  if top_pixel_row =< 0
    ' We just finished the animation, so change the current digit value to
    ' be the working digit value
    CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]

  return 1  ' we made a change
  
' ------------------------------------------------------------------------
' Single row display, ROM font, wipe from both left and right to center

PUB ContUpd1DigitRomWipeLR2C(digit_col) | left_pixel_col, right_pixel_col
  ' Write a new digit pixel row based on WorkingChangeProgress[digit_col].
  left_pixel_col := WorkingChangeProgress[digit_col]++
  right_pixel_col := DIGIT_WIDTH - left_pixel_col - 1
  Helper.RomCharacter(WorkingDigitValue[digit_col], {
    }                 DigitColumn[digit_col] + left_pixel_col, CenterRowTop, {
    }                 1, DIGIT_HEIGHT, {
    }                 ROM_FONT_TRIM_LEFT + left_pixel_col, ROM_FONT_TRIM_TOP, {
    }                 0, BoldColorDisplay)
  if left_pixel_col <> right_pixel_col
    Helper.RomCharacter(WorkingDigitValue[digit_col], {
      }                 DigitColumn[digit_col] + right_pixel_col, CenterRowTop, {
      }                 1, DIGIT_HEIGHT, {
      }                 ROM_FONT_TRIM_LEFT + right_pixel_col, ROM_FONT_TRIM_TOP, {
      }                 0, BoldColorDisplay)
  if left_pixel_col > constant((DIGIT_WIDTH + 1) >> 1)
    ' We just finished the animation, so change the current digit value to
    ' be the working digit value
    CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]

  return 1  ' we made a change
  
' ------------------------------------------------------------------------
' Single row display, ROM font, wipe from center to both left and right

PUB ContUpd1DigitRomWipeLRFC(digit_col) | left_pixel_col, right_pixel_col
  ' Write a new digit pixel row based on WorkingChangeProgress[digit_col].
  left_pixel_col := constant((DIGIT_WIDTH - 1) >> 1) - WorkingChangeProgress[digit_col]++
  right_pixel_col := DIGIT_WIDTH - left_pixel_col - 1
  Helper.RomCharacter(WorkingDigitValue[digit_col], {
    }                 DigitColumn[digit_col] + left_pixel_col, CenterRowTop, {
    }                 1, DIGIT_HEIGHT, {
    }                 ROM_FONT_TRIM_LEFT + left_pixel_col, ROM_FONT_TRIM_TOP, {
    }                 0, BoldColorDisplay)
  if left_pixel_col <> right_pixel_col
    Helper.RomCharacter(WorkingDigitValue[digit_col], {
      }                 DigitColumn[digit_col] + right_pixel_col, CenterRowTop, {
      }                 1, DIGIT_HEIGHT, {
      }                 ROM_FONT_TRIM_LEFT + right_pixel_col, ROM_FONT_TRIM_TOP, {
      }                 0, BoldColorDisplay)
  if left_pixel_col == 0
    ' We just finished the animation, so change the current digit value to
    ' be the working digit value
    CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]

  return 1  ' we made a change
  
' ------------------------------------------------------------------------
' Single row display, ROM font, wipe from center out in all directions
'
' This is the easier case than wiping in towards the center, since we can
' do it with a single drawing primitive for the center portion of the digit,
' expanding by two rows and columns per operation.
'
' Firstly, we assume that DIGIT_WIDTH <= DIGIT_HEIGHT, so we calculate the
' additional number of rows for the ((DIGIT_WIDTH + 1) >> 1) column updates.
' Which would be:
'    extra_rows := ((DIGIT_HEIGHT + 1) >> 1) - (((DIGIT_WIDTH + 1) >> 1) << 1)
' extra_rows will always be => 0 if DIGIT_HEIGHT => DIGIT_WIDTH.
' Then we calculate the starting and ending column, and the starting and
' ending row from the column range by adding the extra_rows.
' Then, we just draw the "center" of the new digit over the old.
' This will redundantly write the center as it grows, but it is faster than
' doing four writes (I think).
CON
  ' We will be writing (DIGIT_WIDTH + 1) >> 1 times, for the values
  '   0 .. ((DIGIT_WIDTH - 1) >> 1)
  ' So every write will need extra rows:
  '   DIGIT_HEIGHT - (((DIGIT_WIDTH + 1) >> 1) << 1)
  WIPE_BOTH_EXTRA_ROWS = DIGIT_HEIGHT - (((DIGIT_WIDTH + 1) >> 1) << 1)
  
PUB ContUpd1DigitRomWipeFromCenter(digit_col) | left_pixel_col
  ' Write a new digit pixel row based on WorkingChangeProgress[digit_col].
  left_pixel_col := constant((DIGIT_WIDTH - 1) >> 1) - WorkingChangeProgress[digit_col]++
  ' top_pixel row will be the same.
  ' The width written will be:
  '   DIGIT_WIDTH - (left_pixel_col << 1)
  ' The height written will be
  '   DIGIT_HEIGHT - ((left_pixel_col << 1) + WIPE_BOTH_EXTRA_ROWS
  Helper.RomCharacter(WorkingDigitValue[digit_col], {
    }                 DigitColumn[digit_col] + left_pixel_col, CenterRowTop + left_pixel_col, {
    }                 DIGIT_WIDTH - (left_pixel_col << 1), {
    }                 DIGIT_HEIGHT - ((left_pixel_col << 1)), {
    }                 ROM_FONT_TRIM_LEFT + left_pixel_col, {
    }                 ROM_FONT_TRIM_TOP + left_pixel_col, {
    }                 0, BoldColorDisplay)
  if left_pixel_col == 0
    ' We just finished the animation, so change the current digit value to
    ' be the working digit value
    CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]

  return 1  ' we made a change
  
' ------------------------------------------------------------------------
' Single row display, ROM font, wipe from all edges to the center
'
' This is the messiest wipe so far: each incremental update, we draw the
' hollow box of pixels starting at the upper-left corner. This takes two to
' four write operations, each one row high or one column wide. At the end,
' we may have only the center to draw, which could be simpler.
'
' Firstly, we assume that DIGIT_WIDTH <= DIGIT_HEIGHT, so we calculate the
' additional number of rows for the ((DIGIT_WIDTH + 1) >> 1) column updates.
' Which would be:
'    extra_rows := ((DIGIT_HEIGHT + 1) >> 1) - (((DIGIT_WIDTH + 1) >> 1) << 1)
' extra_rows will always be => 0 if DIGIT_HEIGHT => DIGIT_WIDTH.
' Then we calculate the starting and ending column, and the starting and
' ending row from the column range by adding the extra_rows.
' Then, we just draw the "center" of the new digit over the old.
' This will redundantly write the center as it grows, but it is faster than
' doing four writes (I think).

PUB ContUpd1DigitRomWipeToCenter(digit_col) | left_pixel_col, write_width
  ' Write a new digit pixel row based on WorkingChangeProgress[digit_col].
  left_pixel_col := WorkingChangeProgress[digit_col]++
  ' top_pixel_row will be the same.
  ' The width written will be:
  write_width := DIGIT_WIDTH - (left_pixel_col << 1)
  ' The height written will be
  '  write_width + WIPE_BOTH_EXTRA_ROWS
  ' Assuming the width < height, when the width written is 1 or two, we can
  ' just use a single write of that solid box in the center of the digit.
  ' Otherwise, we have to write all the edges in four write operations.
  if write_width =< 2
    ' We can do a single write
    Helper.RomCharacter(TargetDigitValue[digit_col], {
      }                 DigitColumn[digit_col] + left_pixel_col, {
      }                   CenterRowTop + left_pixel_col, {
      }                 write_width, DIGIT_HEIGHT - (left_pixel_col << 1), {
      }                 ROM_FONT_TRIM_LEFT + left_pixel_col, {
      }                 ROM_FONT_TRIM_TOP + left_pixel_col, {
      }                 0, BoldColorDisplay)
    ' We just finished the animation, so change the current digit value to
    ' be the working digit value
    CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]
  else
    ' We need to do four writes for the four sides of the hollow box of pixels.
    ' Top row
    Helper.RomCharacter(TargetDigitValue[digit_col], {
      }                 DigitColumn[digit_col] + left_pixel_col, CenterRowTop + left_pixel_col, {
      }                 write_width, 1, {
      }                 ROM_FONT_TRIM_LEFT + left_pixel_col, {
      }                 ROM_FONT_TRIM_TOP + left_pixel_col, {
      }                 0, BoldColorDisplay)
    ' Bottom row
    Helper.RomCharacter(TargetDigitValue[digit_col], {
      }                 DigitColumn[digit_col] + left_pixel_col, {
      }                   CenterRowTop + DIGIT_HEIGHT - left_pixel_col - 1, {
      }                 write_width, 1, {
      }                 ROM_FONT_TRIM_LEFT + left_pixel_col, {
      }                 constant(ROM_FONT_TRIM_TOP + DIGIT_HEIGHT - 1) - left_pixel_col, {
      }                 0, BoldColorDisplay)
    '' Left edge (this will rewrite the top and bottom row pixels)
    Helper.RomCharacter(TargetDigitValue[digit_col], {
      }                 DigitColumn[digit_col] + left_pixel_col, {
      }                   CenterRowTop + left_pixel_col, {
      }                 1, DIGIT_HEIGHT - (left_pixel_col << 1), {
      }                 ROM_FONT_TRIM_LEFT + left_pixel_col, {
      }                 ROM_FONT_TRIM_TOP + left_pixel_col, {
      }                 0, BoldColorDisplay)
    '' Right edge
    Helper.RomCharacter(TargetDigitValue[digit_col], {
      }                 DigitColumn[digit_col] + left_pixel_col + write_width - 1, {
      }                   CenterRowTop + left_pixel_col, {
      }                 1, DIGIT_HEIGHT - (left_pixel_col << 1), {
      }                 ROM_FONT_TRIM_LEFT + left_pixel_col + write_width - 1, {
      }                 ROM_FONT_TRIM_TOP + left_pixel_col, {
      }                 0, BoldColorDisplay)

  return 1  ' we made a change
  
' ------------------------------------------------------------------------
' Single row display, ROM font, scroll from right to left
' We need to allow some blank columns between the digit leaving and the
' digit arriving -- we just use "blank" columns on the right end of the
' digit leaving.
' There will be DIGIT_WIDTH + DIGIT_SCROLL_H_SPACING steps to complete
' the operation.

CON
  DIGIT_SCROLL_H_SPACING = 2

PUB ContUpd1DigitRomScrollLeft(digit_col) | drop_cols, new_pixel_col
  drop_cols := ++WorkingChangeProgress[digit_col]
  ' If the departing digit (on the left) is still visible, we need to
  ' draw the visible portion, and any spacing blank columns.
  if drop_cols =< DIGIT_WIDTH
    ' Draw the CurrentDigitValue[digit_col] skipping drop_cols columns,
    ' on the left portion of the display area for the digit.
    Helper.RomCharacter(CurrentDigitValue[digit_col], {
      }                 DigitColumn[digit_col], CenterRowTop, {
      }                 constant(DIGIT_WIDTH + DIGIT_SCROLL_H_SPACING) - drop_cols, {
      }                   DIGIT_HEIGHT, {
      }                 ROM_FONT_TRIM_LEFT + drop_cols, ROM_FONT_TRIM_TOP, {
      }                 0, BoldColorDisplay)
  if drop_cols > DIGIT_SCROLL_H_SPACING
    ' The newly arriving digit (on the right) is visible.
    new_pixel_col := constant(DIGIT_WIDTH + DIGIT_SCROLL_H_SPACING) - drop_cols
    Helper.RomCharacter(WorkingDigitValue[digit_col], {
      }                 DigitColumn[digit_col] + new_pixel_col, CenterRowTop, {
      }                 DIGIT_WIDTH - new_pixel_col, DIGIT_HEIGHT, {
      }                 ROM_FONT_TRIM_LEFT, ROM_FONT_TRIM_TOP, {
      }                 0, BoldColorDisplay)
  if drop_cols => constant(DIGIT_WIDTH + DIGIT_SCROLL_H_SPACING)
    ' We just finished the animation, so change the current digit value to
    ' be the working digit value
    CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]

  return 1  ' we made a change
  
' ------------------------------------------------------------------------
' Single row display, ROM font, scroll from left to right
' We need to allow some blank columns between the digit leaving and the
' digit arriving -- we just use "blank" columns on the right end of the
' digit arriving.
' There will be DIGIT_WIDTH + DIGIT_SCROLL_H_SPACING steps to complete
' the operation.

PUB ContUpd1DigitRomScrollRight(digit_col) | drop_cols, new_digit_width
  drop_cols := ++WorkingChangeProgress[digit_col]
  ' If the departing digit (on the right) is still visible, we need to
  ' draw the visible portion.
  if drop_cols < DIGIT_WIDTH
    ' Draw the CurrentDigitValue[digit_col] skipping drop_cols columns,
    ' on the right portion of the display area for the digit.
    Helper.RomCharacter(CurrentDigitValue[digit_col], {
      }                 DigitColumn[digit_col] + drop_cols, CenterRowTop, {
      }                 DIGIT_WIDTH - drop_cols, DIGIT_HEIGHT, {
      }                 ROM_FONT_TRIM_LEFT, ROM_FONT_TRIM_TOP, {
      }                 0, BoldColorDisplay)
  ' We always draw the digit arriving on the left, so that we also get the
  ' blank columns. We have to make sure that we don't try to draw too many
  ' columns, so we check for drop_cols => DIGIT_WIDTH, in which case we limit
  ' the width.
  new_digit_width := drop_cols <# DIGIT_WIDTH
  Helper.RomCharacter(WorkingDigitValue[digit_col], {
    }                 DigitColumn[digit_col], CenterRowTop, {
    }                 new_digit_width, DIGIT_HEIGHT, {
    }                 constant(ROM_FONT_TRIM_LEFT + DIGIT_WIDTH + DIGIT_SCROLL_H_SPACING) - drop_cols, ROM_FONT_TRIM_TOP, {
    }                 0, BoldColorDisplay)
  if drop_cols => constant(DIGIT_WIDTH + DIGIT_SCROLL_H_SPACING)
    ' We just finished the animation, so change the current digit value to
    ' be the working digit value
    CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]

  return 1  ' we made a change
  
' ------------------------------------------------------------------------
' Single row display, ROM font, scroll from bottom to top
' We need to allow some blank rows between the digit leaving and the
' digit arriving -- we just use "blank" rows on the bottom end of the
' digit leaving.
' There will be DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING steps to complete
' the operation.

CON
  ' NOTE: Due to the code using the blank rows before and after the digits
  ' in the ROM area, this number is limited to five or less.
  DIGIT_SCROLL_V_SPACING = 3

PUB ContUpd1DigitRomScrollUp(digit_col) | drop_rows, new_pixel_row
  drop_rows := ++WorkingChangeProgress[digit_col]
  ' If the departing digit (on the top) is still visible, we need to
  ' draw the visible portion.
  if drop_rows =< DIGIT_HEIGHT
    ' Draw the CurrentDigitValue[digit_col] skipping drop_rows rows,
    ' on the top portion of the display area for the digit.
    Helper.RomCharacter(CurrentDigitValue[digit_col], {
      }                 DigitColumn[digit_col], CenterRowTop, {
      }                 DIGIT_WIDTH, constant(DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING) - drop_rows, {
      }                 ROM_FONT_TRIM_LEFT, ROM_FONT_TRIM_TOP + drop_rows, {
      }                 0, BoldColorDisplay)
  if drop_rows > DIGIT_SCROLL_V_SPACING
    ' The newly arriving digit (on the bottom) is visible.
    new_pixel_row := constant(DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING) - drop_rows
    Helper.RomCharacter(WorkingDigitValue[digit_col], {
      }                 DigitColumn[digit_col], CenterRowTop + new_pixel_row, {
      }                 DIGIT_WIDTH, DIGIT_HEIGHT - new_pixel_row, {
      }                 ROM_FONT_TRIM_LEFT, ROM_FONT_TRIM_TOP, {
      }                 0, BoldColorDisplay)
  if drop_rows => constant(DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING)
    ' We just finished the animation, so change the current digit value to
    ' be the working digit value
    CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]

  return 1  ' we made a change
  
' ------------------------------------------------------------------------
' Single row display, ROM font, scroll from top to bottom
' We need to allow some blank columns between the digit leaving and the
' digit arriving -- we just use "blank" rows on the bottom end of the
' digit arriving.
' There will be DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING steps to complete
' the operation.

PUB ContUpd1DigitRomScrollDown(digit_col) | drop_rows, new_digit_height
  drop_rows := ++WorkingChangeProgress[digit_col]
  ' If the departing digit (on the bottom) is still visible, we need to
  ' draw the visible portion.
  if drop_rows < DIGIT_HEIGHT
    ' Draw the CurrentDigitValue[digit_col] skipping drop_rows rows,
    ' on the bottom portion of the display area for the digit.
    Helper.RomCharacter(CurrentDigitValue[digit_col], {
      }                 DigitColumn[digit_col], CenterRowTop + drop_rows, {
      }                 DIGIT_WIDTH, DIGIT_HEIGHT - drop_rows, {
      }                 ROM_FONT_TRIM_LEFT, ROM_FONT_TRIM_TOP, {
      }                 0, BoldColorDisplay)
  ' We always draw the digit arriving on the top, so that we also get the
  ' blank rows. We have to make sure that we don't try to draw too many
  ' rows, so we check for drop_rows => DIGIT_HEIGHT, in which case we limit
  ' the height.
  new_digit_height := drop_rows <# DIGIT_HEIGHT
  Helper.RomCharacter(WorkingDigitValue[digit_col], {
    }                 DigitColumn[digit_col], CenterRowTop, {
    }                 DIGIT_WIDTH, new_digit_height, {
    }                 ROM_FONT_TRIM_LEFT, constant(ROM_FONT_TRIM_TOP + DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING) - drop_rows, {
    }                 0, BoldColorDisplay)
  if drop_rows => constant(DIGIT_HEIGHT + DIGIT_SCROLL_V_SPACING)
    ' We just finished the animation, so change the current digit value to
    ' be the working digit value
    CurrentDigitValue[digit_col] := WorkingDigitValue[digit_col]

  return 1  ' we made a change
  
' ========================================================================
' Routines to help with testing.

PUB udec32tostr(v, buf)
  ' Returns the length of the string written into buf
  return Umath.udec64tostr(0, v, buf)

PUB sdec32tostr(v, buf) | p, len, c, i
  ' Returns the length of the string written into buf
  ' One negative value cannot be converted to a positive value by
  ' negation: $8000_0000, but it is its own negation, so we can just
  ' handle it like any other negative number
  if v < 0
    buf := "-"
    return Umath.udec64tostr(0, -v, buf + 1) + 1
  return Umath.udec64tostr(0, v, buf)

PUB ReportStr(s)
  Ser.Str(s)
  Ser.Tx(CR)

PUB ReportDec(label, v)
  Ser.Str(label)
  Ser.Dec(v)
  Ser.Tx(CR)

PUB ReportHex(label, v)
  Ser.Str(label)
  Ser.Hex(v, 8)
  Ser.Tx(CR)

PUB ReportBin(label, v)
  Ser.Str(label)
  Ser.Bin(v, 24)
  Ser.Tx(CR)

Pub ReportTenths(label, v)
  Ser.Str(label)
  Ser.Dec(v / 10)
  Ser.Tx(".")
  Ser.Tx("0" + (v // 10))
  Ser.Tx(CR)

PUB DumpFifo(label) | p
  Ser.Str(label)
  Ser.Tx(CR)
  p := Helper.GetCmdFifoHeader
  
  ReportHex(string("FiFo   ptr: "), p)
  ReportHex(string("Insert ptr: "), word[p + Fifo#INSERT_HUBADDR_OFFSET])
  ReportHex(string("Remove ptr: "), word[p + Fifo#REMOVE_HUBADDR_OFFSET])
  ReportHex(string("Start  ptr: "), word[p + Fifo#START_HUBADDR_OFFSET])
  ReportHex(string("Limit  ptr: "), word[p + Fifo#LIMIT_HUBADDR_OFFSET])
  ReportDec(string("Entry size: "), word[p + Fifo#ENTRY_SIZE_BYTES_OFFSET])

PUB ReportBitmapDebugOutput(entry)
  ReportHex(string("result address:  "), entry)
  ReportHex(string("Requested background color: "), long[entry + 20])
  ReportHex(string("Seen      background color: "), long[entry +  8])
  ReportBin(string("Converted background color: "), long[entry +  0])
  ReportHex(string("Requested foreground color: "), long[entry + 24])
  ReportHex(string("Seen      foreground color: "), long[entry + 12])
  ReportBin(string("Converted foreground color: "), long[entry +  4])

PUB ReportRomCharDebugOutput(entry)
  ReportHex(string("result address:  "), entry)
  ReportHex(string("Rom char base address: "), long[entry +  0])
  ReportHex(string("Rom char seen:         "), long[entry +  4])
  ReportHex(string("Rom char requested:    "), word[entry + 16])
  ReportHex(string("Expected base address: "), ((word[entry + 16] & $FE) << 6) + $8000)

PUB ReportSetColorDebugOutput(entry) | c
  c := long[entry + 8]
  ReportHex(string("Original      color value: "), c)
  ReportBin(string("Converted raw color value: "), long[entry + 4])
  ReportBin(string("Expected  raw color value: "), {
    } Panel.DisplayColor(((c >> 20) & $3FF), ((c >> 10) & $3FF), c & $3FF))
  'ReportBin(string("Expected2 raw color value: "), Panel.DisplayColor(6, 6, 0))

PUB ReportDigitRowUpdateSpeed(startcnt, endcnt) | clks
  clks := endcnt - startcnt
  ReportDec(string("Sysclks to update all digits: "), clks)
  ' calculate and report update rate, to tenths.
  ReportTenths(string("Update rate, per second: "), {
    }          Umath.multdiv_rounded(CLKFREQ, 10, clks))
  'ReportDec(string("Update rate, in tenths: "), {
  '  }          Umath.multdiv_rounded(CLKFREQ, 10, clks))

PUB PowerOfTen(p) | r
  r := 1
  repeat while p--
    r *= 10
  return r

' ========================================================================

