{{
Test one or more HUB75-based RGB panels in series.

Author:         Tom Dinger (Ding-Batty)
Release Date:   2024/04/10

Copyright 2024 Thomas E. Dinger

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

--------------------------------------------------------------------------

This test program can use one or more RGB LED panels, all of which are the
same size, and that use the HUB75 convention for wiring and control.

This can be used to test both Spin-based routines and PASM-based write
routines. The requirements for actually sending serial data to the display
to control the color and intensity of the LEDs is demanding enough that
the display driver currently used is PASM-based. (I have similar code
entirely in Spin, but it is quite slow, so there is a lot of flicker in
a 64 x 64 display; smaller displays _might_ have acceptable performance
with a pure Spin display driver. The PASM driver is 50x to 120x faster than
the pure Spin driver.)

The test program tests various aspects of behavior, and exercises the write
routines in various ways.

NOTE: Not all of the tests adapt to the configure screen size -- they were
originally written for a single 64x64 panel with COLOR_DEPTH = 3, so they
still have some built-in assumptions about those settings.

For more information:
---------------------

See the file "hub75_info.txt" for some details on the display interface
called "HUB75", with some useful links.


Revisions:
==========
- 2024/04/10 -- Change from 8-bit color max, to 10-bit color max;
- 2024/02/23 -- Original release

}}
' Define the operating constants:

OBJ
  ' Pick up the static configuration for the display(s).
  Config: "hub75_config.spin"

  Ser    : "Db_FullDuplexSerial"
  Umath  : "umath2.spin"
  ' The FiFo is used to enqueue write commands for the Helper to send to
  ' a dedicated PASM writer cog that updates the Display buffer.
  Fifo   : "FifoBig.spin"
  ' The Panel is the wrapper for the PASM cog that send the data to the
  ' display, and contains Spin-based write routines, such as Point, Box,
  ' HorizontalLine, etc. 
  Panel  : "hub75_display.spin"
  ' The Helper contains wrapper functions for writing to the display buffer
  ' by sending the commands to a helper PASM cog, in order to have faster
  ' display updates. It has the same write functions, e.g. Point, Box, etc.
  Helper : "hub75_helper.spin"

  ' The init object (below) and all objects it includes, will be used as
  ' additional buffer allocation space after they are finished during init.
  ' This helps to maximize hub memory available for large buffer use.
  Init : "test_panels_init.spin"


CON { General constants }
  _CLKMODE = XTAL1 + PLL16X 
  _XINFREQ = 5_000_000

  ' NOTE: Some of the configuration information is actually in the 
  ' test_panels_init.spin file, because they then can be imported here,
  ' rather than duplicating the definitions.

  { Configuration for test serial connection. }
  BAUDRATE = Init#BAUDRATE  ' 115_200
  RXPIN    = Init#RXPIN     ' 31    ' The usual pins.
  TXPIN    = Init#TXPIN     ' 30
  ' For this version of FDS, the sizes do not need to be powers-of-two. The
  ' Spin code is a little slower, but the PASM code is not slowed by this.
  RX_BUFFER_SIZE = 28
  TX_BUFFER_SIZE = 200

  CR = $0D
  LF = $0A

  { LED panel(s) configuration }
  COLOR_DEPTH = Init#COLOR_DEPTH
                            ' Number of bits of resolution per color, up to
                            ' 8 bits, which is the limit for the implementation
                            ' details. The limits are available memory and
                            ' flickering from slower refresh rates.
  MAX_COLOR_VALUE = (|< COLOR_DEPTH) - 1

  ' Save some typing later, by reflecting a few constants here.
  DISPLAY_ROWS       = Config#DISPLAY_ROWS
  REFRESH_ROWS       = Config#REFRESH_ROWS
  DISPLAY_COLUMNS    = Config#DISPLAY_COLUMNS
  MAX_DISPLAY_ROW    = Config#MAX_DISPLAY_ROW
  MAX_DISPLAY_COLUMN = Config#MAX_DISPLAY_COLUMN

  RED_BIT            = Config#RED_BIT
  GRN_BIT            = Config#GREEN_BIT
  BLU_BIT            = Config#BLUE_BIT  

  ' Instead of a single mailbox, we use a FiFo of command (a mailbox queue)
  ' to enqueue multiple commands for the display buffer write helper cog to
  ' execute for the main program. We have a large queue because for some
  ' operations, we want the updates to go as fast as possible with little
  ' time for stalling.
  '
  ' If we need even faster display update, we could use multiple writer
  ' cogs, each with a separate command FiFo (or even a shared FiFo with the
  ' proper locks) to process write requests in parallel. There can be
  ' write interactions, however, because display buffer updates are
  ' read-modify-write, and are not locked, so two writers is best avoided.  
  COMMAND_FIFO_SIZE = Init#COMMAND_FIFO_SIZE   ' 32

  ' For some display tests showing and changing digits, we need bitmaps for
  ' the digits themselves. The ROM stores two characters interleaved in one
  ' group of 32 LONG values, but for our purposes it is better to store the
  ' digits separately. We also can treat a column of digits as an extended
  ' height bitmap, which can be used for a column-of-digits scrolling effect
  ' with very little code.
  '
  ' We extract the digits from the ROM Font. They are 16 x 32, but some of the
  ' rows and columns are blank, so we only keep 13 x 22 -- every digit (and
  ' the capital letter "O", used for zero) fits into 13-pixels wide, and
  ' using 26 rows is enough for a few blank pixel rows between rows of
  ' digits.
  { Digit "shapes" }
  'DIGIT_WIDTH  = 13
  'DIGIT_HEIGHT = 22
  ' The number of pixels to trim a Rom Font digit to skip over the empty
  ' pixel rows and columns.
  'ROM_FONT_TRIM_TOP  = 5
  'ROM_FONT_TRIM_LEFT = 2
  DIGIT_WIDTH  = Init#DIGIT_WIDTH
  DIGIT_HEIGHT = Init#DIGIT_HEIGHT
  ROM_FONT_TRIM_TOP  = Init#ROM_FONT_TRIM_TOP
  ROM_FONT_TRIM_LEFT = Init#ROM_FONT_TRIM_LEFT

  ' For a few tests, on a 64-row display we have three rows of digits:
  ' one row centered vertically, and two more partial rows of digits, one
  ' above and one below the center row. This is also true for a 32-row
  ' display, since we are only using 26 pixel rows for the middle digit row.
  ' These constant define the display layout.
  ' We always have a row of digits centered vertially on the display.
  TOP_DIGIT_TOP_ROW       = 0
  CENTER_DIGIT_TOP_ROW    = (DISPLAY_ROWS - DIGIT_HEIGHT) / 2
  'TOP_DIGIT_ROW_HEIGHT    = CENTER_DIGIT_TOP_ROW - TOP_DIGIT_TOP_ROW
  'BOTTOM_DIGIT_TOP_ROW    = CENTER_DIGIT_TOP_ROW + DIGIT_HEIGHT
  'BOTTOM_DIGIT_ROW_HEIGHT = DISPLAY_ROWS - BOTTOM_DIGIT_TOP_ROW

  ' Constants for stack tracking.

  ' The address of the location that stores the base address of the stack,
  ' which runs up form there to $7FFF
  STACK_BASE_ADDR = $000A
  END_OF_STACK    = Init#END_OF_STACK  ' $0000_8000

  ' This is the distance from the last local variable address in the
  ' function that fills main memory that is safe to fill with the
  ' UNLIKELY_VALUE.
  OFFSET_FOR_MAIN_STACK_FILL = Init#OFFSET_FOR_MAIN_STACK_FILL  ' $0010
  UNLIKELY_VALUE             = Init#UNLIKELY_VALUE   ' $DEADBEEF

VAR
  ' These must be consecutive in this order, so that the Init object can fill
  ' them in.
  WORD    SerialCog   ' 0 if not started, else CogID + 1
  WORD    DisplayCog  ' 0 if not started, else CogID + 1
  WORD    HelperCog   ' 0 if not started, else CogID + 1

  ' We allocate two buffers from the free space:
  ' - Command FiFo for display write-helper cog
  ' - Display Buffer
  ' They are allocated from a combination of reserved space and reuse of the
  ' initialization code including the PASM cog code from Hub memory.
  ' Since the largest buffer is the display buffer, and it (usually) is much
  ' larger than the init code that is recovered, we don't need to dynamically
  ' allocate the serial driver buffers, which makes init a bit simpler.
  WORD    BufferInfo [2 * 2]  ' two buffers to allocate.

  ' Storge for the Stars test
  ' Each value is a random number, that is decomposed into row and column.
  ' We have a limit on the number of "stars" that can be displayed at once.
  LONG    StarPositions[100]

' Define the variables that contain the operating parameters.
' We use a DAT section so we can initialize them, which means there can be only
' one instance of this object; if we used a VAR section, we'd have to devote
' code to initializing the VARs, which might happen later, depending on how the
' code develops.
DAT
' We allocate the storage for the command FiFo entries dynamically, overlaying
' the PASM (and Spin) code at the end of the program, which is only used
' during early initialization.
CmdFifoDataAddr  LONG    0

' Storage for debug output from the helper driver
' See that driver for details on the contents.
' Should be at least 10 longs
HelperDebugOutput LONG    0[16]

' Because they are "small" we don't dynamically allocate the serial object
' buffers.
Rx_Buffer          LONG    0[(RX_BUFFER_SIZE + 3) >> 2]
Tx_Buffer          LONG    0[(TX_BUFFER_SIZE + 3) >> 2]

' Storge for the Stars test
' Each value is a random number, that is decomposed into row and column.
' We have a limit on the number of "stars" that can be displayed at once.
' StarPositions   LONG    0[100]

' The display buffer stores all color layers for all displays.
' For some applications, it may overlap the PASM object loading code once
' the Cogs have been started, to maximize the amount of memory available
' for the display(s).
'
' One byte of memory stores one color bit layer for two pixels.
'
' For the initial set of tests, we compile-time initialize the display
' buffer with a known pattern of colors, to verify display refresh
' without depending on any of the buffer write routines.
'
' This buffer assumes REFRESH_ROWS = 32 (the current maximum supoprted
' by the HUB75 interface, AFAIK). Shorter displays could also display this
' correctly, but only for one-bit color depth; otherwise the other color
' layers are in the wrong place.
'
' For testing, each layer is shown in sections, and only one section is
' visible for each layer. The sections are 1/8th of the display width.
' They show for all rows.

DisplayBufferAddr  LONG    0   ' gets a dynamically-allocated buffer

' Storage for the digit bitmaps.
' The initialization code "assumes" this order for the bitmap layout.

Digits10
Digits10_0      WORD    0[DIGIT_HEIGHT]
Digits10_1      WORD    0[DIGIT_HEIGHT]
Digits10_2      WORD    0[DIGIT_HEIGHT]
Digits10_3      WORD    0[DIGIT_HEIGHT]
Digits10_4      WORD    0[DIGIT_HEIGHT]
Digits10_5      WORD    0[DIGIT_HEIGHT]
Digits10_6      WORD    0[DIGIT_HEIGHT]
Digits10_7      WORD    0[DIGIT_HEIGHT]
Digits10_8      WORD    0[DIGIT_HEIGHT]
Digits10_9      WORD    0[DIGIT_HEIGHT]

' The Init Colors here are used to initialize the display buffer before it is
' first displayed. The code that fills it in "knows" the memory layout for
' the display buffer, and fills in a different region of the display buffer
' for each color bit layer, so each region shows a different intensity for
' all the colors.
DisplayBufferInitColors
                BYTE    ((                          0) << 3) | (                          0)  ' Black
                BYTE    ((RED_BIT                    ) << 3) | (RED_BIT                    )  ' Red
                BYTE    ((          GRN_BIT          ) << 3) | (          GRN_BIT          )  ' Green
                BYTE    ((                    BLU_BIT) << 3) | (                    BLU_BIT)  ' Blue  
                BYTE    ((RED_BIT           | BLU_BIT) << 3) | (RED_BIT           | BLU_BIT)  ' Magenta
                BYTE    ((RED_BIT | GRN_BIT          ) << 3) | (RED_BIT | GRN_BIT          )  ' Yellow
                BYTE    ((          GRN_BIT | BLU_BIT) << 3) | (          GRN_BIT | BLU_BIT)  ' Cyan
                BYTE    ((RED_BIT | GRN_BIT | BLU_BIT) << 3) | (RED_BIT | GRN_BIT | BLU_BIT)  ' White

' ========================================================================
PUB Main | avail, left, needed, firstfree
  ' Collect some numbers for assessing memory use.
  avail := Init.AvailableBytes
  firstfree := Init.GetNextFree

  ' First-stage init:
  ' - Collect the digit bitmaps
  ' - reserve memory for the (large-ish) buffers
  ' - load the PASM cogs.
  ' After this call, the init code should no longer be needed, so the reserved
  ' buffers can be overwritten and the objects using the PASM cogs can finish
  ' their initializations.
  needed := Init.Init(@SerialCog, @BufferInfo, @Digits10, Ser.GetInitVars,       {
         }            Panel.GetDisplayBufAddrVar, Panel.GetRefreshClocksAddrHub, {
         }            Helper.GetCmdFifoHeader, @HelperDebugOutput)
  left := Init.AvailableBytes
  Ser.FinishInit(serialcog, @Rx_Buffer, RX_BUFFER_SIZE, @Tx_Buffer, TX_BUFFER_SIZE)

  ' How did we do with buffer reservations?
  ReportHex(string("VBASE: "), WORD[8])
  ReportDec(string("Initial memory (bytes):   "), avail)
  ReportDec(string("Recovered from init code: "), avail - Init#RESERVED_SIZE_BYTES)
  ReportHex(string("First free:               "), firstfree)
  ReportHex(string("FiFo buffer size:         "), BufferInfo[0])
  ReportHex(string("FiFo buffer address:      "), BufferInfo[1])
  ReportHex(string("Display buffer size:      "), BufferInfo[2])
  ReportHex(string("Display buffer address:   "), BufferInfo[3])
  ReportDec(string("Memory needed:            "), needed)
  ReportDec(string("Memory remaining (bytes): "), left)

  if needed > 0
    ' We were not able to get as much buffer memory as we wanted.
    if BufferInfo[1] == 0
      ReportStr(string("Not enough memory for Command FiFo or Display Buffer"))
    else
      ReportStr(string("Not enough memory for Display Buffer"))
    ReportDec(string("Memory needed (bytes):    "), needed)
    ' TODO: (This might not be possible) Estimate if lowering COLOR_DEPTH would
    ' work, and estimate available memory if the main stack is reduced. This
    ' requires a guess as to how much main stack is needed.
  else
    ' We have our buffers -- continue with initialization and run all the tests.
    Main2

  ' Here when done.
  CheckMainStack  ' So we can see if we can shrink the main stack, and increase
                  ' the amount of memory for dynamic buffers.
  repeat
    ' wait forvever.

PUB Main2 | r, c, color, dimcolor, rt
  ' Complete initialization once the buffers are reserved and the PASM cogs started.
  CmdFifoDataAddr   := BufferInfo[1]
  DisplayBufferAddr := BufferInfo[3]
  ReportDec(string("Started display cog: "), DisplayCog - 1)
  ReportDec(string("Started helper cog:  "), HelperCog - 1)

  ' Pre-fill the display buffer with a known pattern.
  InitDisplayBuffer(WORD[@DisplayBufferAddr], COLOR_DEPTH)

  ' Tell the drivers about the two cogs and allocated buffers.
  ' Note that telling the display driver about the address of the display
  ' buffer will unblock that cog _and_ the helper cog at the same time, so
  ' that one should be done last.
  Helper.FinalHelperInit(HelperCog, DisplayBufferAddr, COLOR_DEPTH, {
    }                    CmdFifoDataAddr, {
    }                    Fifo#HEADER_SIZE_BYTES * COMMAND_FIFO_SIZE)
  Panel.FinalDisplayInit(DisplayCog, DisplayBufferAddr, COLOR_DEPTH)
    
  'DumpFifo(string("Initial FiFo state"))  ' For debugging
  ShowRefreshRate

  ' All the single arguments are how many seconds to leave the test on the
  ' display before continuing.
  TestShowInitialScreen(4)
  TestColorConversions(1)

  ' These tests all use the Spin-based display write functions.
  TestPoints(4)
  TestSparkles(10000, 50)  ' 10K random points displayed, up to 50 displayed at once.
  TestHorizontalLines(3)
  TestVerticalLines(3)
  TestAllColorsInBoxes(3)
  TestRomFontDigits(5)
  TestStaticClockDisplay(5)

  ' These tests are similar but using the helper object and a PASM cog for
  ' writing.

  TestPointsWHelper(4)
  TestOutOfRangePointsWHelper(4)
  TestSparklesWHelper(10000, 80)  ' 10K random points display, up to 80 displayed at once. 
  TestPointsWHelperAs1BitBitmaps(4)
  TestHorizontalLinesWHelper(3)
  TestVerticalLinesWHelper(3)
  TestAllColorsInBoxesWHelper(3)
  TestRomFontDigitsWHelper(4)
  TestStaticClockDisplayWHelper(4)

  'TestVerticalLinesBmp(4)
  'TestHorizontalLinesBmp(4)

  ' Stall, waiting for input from the user
  ReportStr(string("Press a key to stop the display"))
  Ser.Rx  ' Wait, and discard the character

  Panel.Stop
  ReportStr(string("Display cog stopped"))
  Helper.Stop
  ReportStr(string("Helper cog stopped"))


PUB ShowRefreshRate | rc
{{
  Wait for the display cog to report the clock cycles for one refresh, and
  then report the refresh rate, rounded to tenths.
}}
  ' Wait for the first report of the refresh time.
  rc := Panel.GetRefreshClocks
  ReportDec(string("One refresh, approx. sysclks: "), rc)
  ReportTenths(string("Refresh rate, per second: "), {
    }          Umath.multdiv_rounded(CLKFREQ, 10, rc))

PUB InitDisplayBuffer(bp, depth) | ls, w
{{
  Fill in the display buffer with a known color pattern, divided into
  separate regions for each color bit layer.
  - For one bit, the entire display is uniform
  - For two bits, the top half is dim, and the bottom is bright
  - For three and four bits, the display is in quadrants.
  - For five to eight bits, the display is in eight parts: four on the top,
    and four on the bottom.

  Arguments:
    bp -- ("buffer pointer") the address of the display buffer.
    depth -- the color depth: number of bits per color.

  Return value:
    None.
}}
  ls := DISPLAY_COLUMNS * REFRESH_ROWS  ' Layer Size, in bytes
  'ReportHex(string("Filling display buffer at: "), bp)
  LONGFILL(bp, 0, (ls >> 2) * depth)
  case depth
    1:
      ' Top and bottom half are the same colors for the one layer
      FillLayer(bp, %111_111, DISPLAY_COLUMNS)
    2:
      ' Top half is one layer, bottom half is the other
      FillLayer(bp + (ls * 0), %000_111, DISPLAY_COLUMNS)
      FillLayer(bp + (ls * 1), %111_000, DISPLAY_COLUMNS)
    3, 4:
      'First two layers are on top, next two are on the bottom
      w := DISPLAY_COLUMNS >> 1
      FillLayer(bp               , %000_111, w)
      FillLayer(bp +  ls      + w, %000_111, w)
      FillLayer(bp + (ls * 2)    , %111_000, w)
      if depth > 3
        FillLayer(bp + (ls * 3) + w, %111_000, w)
    5..8:
      ' First four layers are on top, the rest on the bottom
      w := DISPLAY_COLUMNS >> 2
      FillLayer(bp                 , %000_111, w)
      FillLayer(bp +  ls      + w  , %000_111, w)
      FillLayer(bp + (ls * 2) + w*2, %000_111, w)
      FillLayer(bp + (ls * 3) + w*3, %000_111, w)
      FillLayer(bp + (ls * 4)      , %111_000, w)
      if depth > 5
        FillLayer(bp + (ls * 5) + w  , %111_000, w)
        if depth > 6
          FillLayer(bp + (ls * 6) + w*2, %111_000, w)
          if depth > 7
            FillLayer(bp + (ls * 7) + w*3, %111_000, w)

PUB FillLayer(bp, mask, width) | r
{{
  Fill the layer for width columns starting at bute address bp.
  This always writes REFRESH_ROWS rows, repeatedly using the masked
  values from DisplayBufferInitColors
}}
  repeat r from 0 to constant(REFRESH_ROWS - 1)
    BYTEFILL(bp, BYTE[@DisplayBufferInitColors + (r & 7)] & mask, width)
    bp += DISPLAY_COLUMNS  

' ========================================================================
' All the test routines.

PUB TestShowInitialScreen(s)
  ' The initial screen is already displayed.
  ReportStr(string("Test: Initial screen"))
  waitcnt(clkfreq * s + cnt)

PUB TestColorConversions(s) | display, rgb  ', old
  ReportStr(string("Test: Color conversions"))
  rgb := Panel.RgbColor(6, 0, 6)
  ReportHex(string("RgbColor(6, 0, 6) --> "), rgb)
  display := Panel.DisplayColor(6, 0, 6)
  ReportBin(string("DisplayColor(6, 0, 6) --> "), display)
  display := Panel.ToDisplayColor(rgb)
  ReportBin(string("ToDisplayColor(RgbColor(6, 0, 6)) --> "), display)
  'old := Panel.BuildColor(6, 0, 6)
  'ReportBin(string("BuildColor(6, 0, 6) --> "), old)
  ' no need to wait  

PUB TestPoints(s) | c, r
  ReportStr(string("Test: Single pixel writes"))
  ' This also tests display using both Display color and RGB color values.
  Panel.Clear
  Panel.Pixel(1, 1, Panel.DisplayColor(MAX_COLOR_VALUE, 0, 0))
  Panel.Pixel(2, 2, Panel.DisplayColor(0, MAX_COLOR_VALUE, 0))
  Panel.Pixel(3, 3, Panel.DisplayColor(0, 0, MAX_COLOR_VALUE))
  Panel.Pixel(4, 4, Panel.DisplayColor(MAX_COLOR_VALUE, MAX_COLOR_VALUE, {
    }                                  MAX_COLOR_VALUE))

  ' Bottom half, right half
  c := DISPLAY_COLUMNS >> 1
  r := DISPLAY_ROWS >> 1
  Panel.Pixel(c + 1, r + 1, Panel.RgbColor(MAX_COLOR_VALUE, MAX_COLOR_VALUE, 0))
  Panel.Pixel(c + 2, r + 2, Panel.RgbColor(MAX_COLOR_VALUE, 0, MAX_COLOR_VALUE))
  Panel.Pixel(c + 3, r + 3, Panel.RgbColor(0, MAX_COLOR_VALUE, MAX_COLOR_VALUE))
  Panel.Pixel(c + 4, r + 4, Panel.RgbColor(MAX_COLOR_VALUE, MAX_COLOR_VALUE, {
    }                                   MAX_COLOR_VALUE))

  waitcnt(clkfreq * s + cnt)

PUB TestPointsWHelper(s) | c, r
  ReportStr(string("Test: Single pixel writes with PASM helper"))
  ' This also tests display using both Display color and RGB color values.
  Panel.Clear
  Helper.Pixel(1, 1, Panel.DisplayColor(MAX_COLOR_VALUE, 0, 0))
  Helper.Pixel(2, 2, Panel.DisplayColor(0, MAX_COLOR_VALUE, 0))
  Helper.Pixel(3, 3, Panel.DisplayColor(0, 0, MAX_COLOR_VALUE))
  Helper.Pixel(4, 4, Panel.DisplayColor(MAX_COLOR_VALUE, MAX_COLOR_VALUE, {
    }                                  MAX_COLOR_VALUE))

  ' Bottom half, right half
  c := DISPLAY_COLUMNS >> 1
  r := DISPLAY_ROWS >> 1
  Helper.Pixel(c + 1, r + 1, Panel.RgbColor(MAX_COLOR_VALUE, MAX_COLOR_VALUE, 0))
  Helper.Pixel(c + 2, r + 2, Panel.RgbColor(MAX_COLOR_VALUE, 0, MAX_COLOR_VALUE))
  Helper.Pixel(c + 3, r + 3, Panel.RgbColor(0, MAX_COLOR_VALUE, MAX_COLOR_VALUE))
  Helper.Pixel(c + 4, r + 4, Panel.RgbColor(MAX_COLOR_VALUE, MAX_COLOR_VALUE, {
    }                                    MAX_COLOR_VALUE))

  waitcnt(clkfreq * s + cnt)

PUB TestOutOfRangePointsWHelper(s)
  ReportStr(string("Test: Out-of-range single pixel writes with PASM helper"))
  ReportStr(string("      Nothing should be shown on the display panel"))
  Panel.Clear
  Helper.Pixel(1, -1, Panel.DisplayColor(6, 0, 0))
  Helper.Pixel(-2, 2, Panel.DisplayColor(0, 6, 0))
  Helper.Pixel(3, DISPLAY_ROWS+3, Panel.DisplayColor(0, 0, 6))
  Helper.Pixel(DISPLAY_COLUMNS+4, 4, Panel.DisplayColor(6, 6, 6))

  ' Bottom half
  Helper.Pixel(1, -100000, Panel.RgbColor(6, 6, 0))
  Helper.Pixel(-200000, 34, Panel.RgbColor(6, 0, 6))
  Helper.Pixel(3, 300035, Panel.RgbColor(0, 6, 6))
  Helper.Pixel(400004, 36, Panel.RgbColor(6, 6, 6))

  'ReportHex(string("Address expected for pixel: "), WORD[@DisplayBufferAddr] + 64 + 1)
  'waitcnt(clkfreq + cnt)    ' 1 second
  'DumpFifo(string("FiFo state after eight commands"))
  'DumpHelperDebug(string("Helper debug fifo data"))
  waitcnt(clkfreq * s + cnt)

PUB TestSparkles(num_iterations, max_stars) | color, rg, ix, v, r, c
  ' Show random "sparkles" (white pixels) for a short amount of time.
  ' We use a circular buffer (StarPositions[]) that stores the random
  ' coordinates of currently displayed stars.

  ReportStr(string("Test: Sparkles"))
  color := Panel.DisplayColor(MAX_COLOR_VALUE, MAX_COLOR_VALUE, MAX_COLOR_VALUE)
  max_stars <#= 100   ' the max we can display
  ' Prepare random number generator
  rg := 42
  ' Display the first max_stars sparkles
  Panel.Clear
  repeat ix from 0 to max_stars - 1
    StarPositions[ix] := v := ((?rg) & $7FFFFFF)
    r := v // DISPLAY_ROWS
    c := (v / DISPLAY_ROWS) // DISPLAY_COLUMNS
    'ReportColRow(string("Star at: "), c, r)
    Panel.Pixel(c, r, color)
    'waitcnt(clkfreq + cnt) ' wait 1 second
  ' Now, for each additional sparkle, erase one, then write one.
  ix := 0
  repeat num_iterations - max_stars
    v := StarPositions[ix]
    r := v // DISPLAY_ROWS
    c := (v / DISPLAY_ROWS) // DISPLAY_COLUMNS
    Panel.Pixel(c, r, 0)  ' Erase the previous pixel
    StarPositions[ix] := v := ((?rg) & $7FFFFFF)
    r := v // DISPLAY_ROWS
    c := (v / DISPLAY_ROWS) // DISPLAY_COLUMNS
    'ReportColRow(string("Star at: "), c, r)
    Panel.Pixel(c, r, color)
    if ++ix => max_stars
      ix := 0
    'waitcnt(clkfreq + cnt) ' wait 1 second
  Panel.Clear

PUB TestSparklesWHelper(num_iterations, max_stars) | color, rg, ix, v, r, c
  ' Show random "sparkles" (white pixels) for a short amount of time.
  ' We use a circular buffer (StarPositions[]) that stores the random
  ' coordinates of currently displayed stars.

  ReportStr(string("Test: Sparkles, with PASM helper"))
  color := Panel.DisplayColor(MAX_COLOR_VALUE, MAX_COLOR_VALUE, MAX_COLOR_VALUE)
  max_stars <#= 100   ' the max we can display
  ' Prepare random number generator
  rg := 42
  ' Display the first max_stars sparkles
  Panel.Clear
  repeat ix from 0 to max_stars - 1
    StarPositions[ix] := v := ?rg
    r := v // DISPLAY_ROWS
    c := (v / DISPLAY_ROWS) // DISPLAY_COLUMNS
    Helper.Pixel(c, r, color)
  ' Now, for each additional sparkle, erase one, then write one.
  ix := 0
  repeat num_iterations - max_stars
    v := StarPositions[ix]
    Helper.Pixel((v / DISPLAY_ROWS) // DISPLAY_COLUMNS, {
      }         v // DISPLAY_ROWS, 0)
    StarPositions[ix] := v := ?rg
    Helper.Pixel((v / DISPLAY_ROWS) // DISPLAY_COLUMNS, {
      }         v // DISPLAY_ROWS, color)
    if ++ix => max_stars
      ix := 0
  Panel.Clear

PUB TestPointsWHelperAs1BitBitmaps(s) | c, r
{{
  Displays single pixels as if they were 1x1 1bpp bitmaps.
  This is an early test of the PASM bitmap output routine.
  Very inefficient, but a good test.
}}
  ReportStr(string("Test: Single pixel writes with PASM helper as bitmaps"))
  ' This also tests display using both Display color and RGB color values.
  Panel.Clear
  Helper.DrawPixelWithBitmap(1, 1, Panel.DisplayColor(MAX_COLOR_VALUE, 0, 0))
  Helper.DrawPixelWithBitmap(2, 2, Panel.DisplayColor(0, MAX_COLOR_VALUE, 0))
  Helper.DrawPixelWithBitmap(3, 3, Panel.DisplayColor(0, 0, MAX_COLOR_VALUE))
  Helper.DrawPixelWithBitmap(4, 4, Panel.DisplayColor(MAX_COLOR_VALUE, MAX_COLOR_VALUE, {
    }                                                 MAX_COLOR_VALUE))

  ' Bottom half
  c := DISPLAY_COLUMNS >> 1
  r := DISPLAY_ROWS >> 1
  Helper.DrawPixelWithBitmap(c + 1, r + 1, Panel.RgbColor(MAX_COLOR_VALUE, MAX_COLOR_VALUE, 0))
  Helper.DrawPixelWithBitmap(c + 2, r + 2, Panel.RgbColor(MAX_COLOR_VALUE, 0, MAX_COLOR_VALUE))
  Helper.DrawPixelWithBitmap(c + 3, r + 3, Panel.RgbColor(0, MAX_COLOR_VALUE, MAX_COLOR_VALUE))
  Helper.DrawPixelWithBitmap(c + 4, r + 4, Panel.RgbColor(MAX_COLOR_VALUE, MAX_COLOR_VALUE, {
    }                                                     MAX_COLOR_VALUE))

  waitcnt(clkfreq * s + cnt)

PUB TestHorizontalLines(s)| color
  ReportStr(string("Test: Horizontal lines"))
  ' TODO: Adapt for different display configurations (rows and columns)
  Panel.Clear
  color := Panel.DisplayColor(MAX_COLOR_VALUE, 0, 0)  ' Red
  Panel.HorizontalLine(12, 12, 16, color)
  Panel.HorizontalLine(13, 13, 16, color)
  Panel.HorizontalLine(14, 14, 16, color)
  Panel.HorizontalLine(15, 15, 16, color)
  ' Draw some longer lines

  Panel.HorizontalLine(0, 23, DISPLAY_COLUMNS, Panel.RgbColor(0, MAX_COLOR_VALUE, 0))
  ' A line that extends off the screen.
  Panel.HorizontalLine(0, 24, DISPLAY_COLUMNS + 4, {
    }                  Panel.RgbColor(0, MAX_COLOR_VALUE, 0))
  if DISPLAY_ROWS > 32
    Panel.HorizontalLine(0, 40, DISPLAY_COLUMNS, Panel.RgbColor(0, 0, MAX_COLOR_VALUE))
    Panel.HorizontalLine(0, 41, DISPLAY_COLUMNS + 4, {
      }                  Panel.RgbColor(0, 0, MAX_COLOR_VALUE))
  waitcnt(clkfreq * s + cnt)

PUB TestHorizontalLinesWHelper(s)| color
  ReportStr(string("Test: Horizontal lines with PASM helper"))
  ' TODO: Adapt for different display configurations (rows and columns)
  Panel.Clear
  color := Panel.DisplayColor(MAX_COLOR_VALUE, 0, 0)  ' Red
  Helper.HorizontalLine(12, 12, 16, color)
  Helper.HorizontalLine(13, 13, 16, color)
  Helper.HorizontalLine(14, 14, 16, color)
  Helper.HorizontalLine(15, 15, 16, color)
  ' Draw some longer lines

  Helper.HorizontalLine(0, 23, DISPLAY_COLUMNS, Panel.RgbColor(0, MAX_COLOR_VALUE, 0))
  ' A line that extends off the screen.
  Helper.HorizontalLine(0, 24, DISPLAY_COLUMNS + 4, {
    }                   Panel.RgbColor(0, MAX_COLOR_VALUE, 0))
  if DISPLAY_ROWS > 32
    Helper.HorizontalLine(0, 40, DISPLAY_COLUMNS, Panel.RgbColor(0, 0, MAX_COLOR_VALUE))
    Helper.HorizontalLine(0, 41, DISPLAY_COLUMNS + 4, {
      }                   Panel.RgbColor(0, 0, MAX_COLOR_VALUE))
  waitcnt(clkfreq * s + cnt)

PUB TestVerticalLines(s) | wh, ye, gr
  ReportStr(string("Test: Vertical lines"))
  ' TODO: Adapt for different display configurations (rows and columns)
  Panel.Clear
  wh := Panel.DisplayColor((MAX_COLOR_VALUE + 1) >> 1, {
          }                (MAX_COLOR_VALUE + 1) >> 1, {
          }                (MAX_COLOR_VALUE + 1) >> 1)
  ye := Panel.DisplayColor((MAX_COLOR_VALUE + 1) >> 1, {
          }                (MAX_COLOR_VALUE + 1) >> 1, {
          }                0)
  gr := Panel.DisplayColor(0, MAX_COLOR_VALUE, 0)
  ' Black line in the top.
  Panel.VerticalLine(34,  1, 20, wh)  ' white
  Panel.VerticalLine(35,  1, 20, wh)
  Panel.VerticalLine(36,  1, 20, wh)
  Panel.VerticalLine(40, 34, 20, ye)  ' yellow
  Panel.VerticalLine(41, 34, 20, ye)
  Panel.VerticalLine(42, 34, 20, ye)
  Panel.VerticalLine(16, 16, 32, gr) ' green
  Panel.VerticalLine(17, 16, 32, gr)
  waitcnt(clkfreq * s + cnt)

PUB TestVerticalLinesWHelper(s) | wh, ye, gr
  ReportStr(string("Test: Vertical lines with PASM helper"))
  ' TODO: Adapt for different display configurations (rows and columns)
  Panel.Clear
  wh := Panel.DisplayColor((MAX_COLOR_VALUE + 1) >> 1, {
          }                (MAX_COLOR_VALUE + 1) >> 1, {
          }                (MAX_COLOR_VALUE + 1) >> 1)
  ye := Panel.DisplayColor((MAX_COLOR_VALUE + 1) >> 1, {
          }                (MAX_COLOR_VALUE + 1) >> 1, {
          }                0)
  gr := Panel.DisplayColor(0, MAX_COLOR_VALUE, 0)
  ' Black line in the top.
  Helper.VerticalLine(34,  1, 20, wh)  ' white
  Helper.VerticalLine(35,  1, 20, wh)
  Helper.VerticalLine(36,  1, 20, wh)
  Helper.VerticalLine(40, 34, 20, ye)  ' yellow
  Helper.VerticalLine(41, 34, 20, ye)
  Helper.VerticalLine(42, 34, 20, ye)
  Helper.VerticalLine(16, 16, 32, gr) ' green
  Helper.VerticalLine(17, 16, 32, gr)
  waitcnt(clkfreq * s + cnt)

PUB TestAllColorsInBoxes(s) | color, r, c
  ' Draw a grid of all colors and intensities.
  ' This is coded with knowledge of the COLOR_DEPTH, and the display geometry,
  ' and is "optimized" for COLOR_DEPTH=3 and 64 x 64, but it should "work" for
  ' other configurations, with varying amounts of color range.
  '
  ' We have 9 bits of color, which is 512 colors, 16 x 32
  ' so use two rows and 4 columns per color combination.
  ' (There is probably a better organization for the colors, but for
  ' now, this will do.)
  ReportStr(string("Test: All colors, in boxes"))
  color := 0
  repeat r from 0 to MAX_DISPLAY_ROW step 2 ' every-two-rows
    repeat c from 0 to MAX_DISPLAY_COLUMN step 4 ' every-4-columns
      Panel.Box(c, r, 4, 2, Panel.DisplayColor(color >> 6, color >> 3, color))
      ++color
  waitcnt(clkfreq * s + cnt)

PUB TestAllColorsInBoxesWHelper(s) | color, r, c
  ' Draw a grid of all colors and intensities.
  ' This is coded with knowledge of the COLOR_DEPTH, and the display geometry,
  ' and is "optimized" for COLOR_DEPTH=3 and 64 x 64, but it should "work" for
  ' other configurations, with varying amounts of color range.
  '
  ' We have 9 bits of color, which is 512 colors, 16 x 32
  ' so use two rows and 4 columns per color combination.
  ' (There is probably a better organization for the colors, but for
  ' now, this will do.)
  ReportStr(string("Test: All colors, in boxes with PASM helper"))
  color := 0
  repeat r from 0 to MAX_DISPLAY_ROW step 2 ' every-two-rows
    repeat c from 0 to MAX_DISPLAY_COLUMN step 4 ' every-4-columns
      Helper.Box(c, r, 4, 2, Panel.DisplayColor(color >> 6, color >> 3, color))
      ++color
  waitcnt(clkfreq * s + cnt)

PUB TestRomFontDigits(s)
  ' Clear the display, and show digits on the display.
  ' These are somewhat oriented toward a 64 x 64 display, and COLOR_DEPTH = 3
  ' Other configurations should work as well, but won't demonstrate quite
  ' all the features. 
  ' TODO: Adapt for different display configurations (rows and columns, colorbits)
  ReportStr(string("Test: ROM font digits"))
  Panel.Clear
  Panel.Bitmap1Bpp(@Digits10_4, 2, 0, 3, 13, 22, 0, 0, 0, Panel.DisplayColor(0, 3, 0))
  Panel.RomCharacter("1",        13,  0, 14, 32, 2, 0, 0, Panel.DisplayColor(3, 0, 0))
  Panel.RomCharacter("2",        32, -6, 16, 32, 0, 0, 0, Panel.DisplayColor(0, 0, 3))
  Panel.RomCharacter("3",        54,  0, 16, 32, 0, 0, 0, Panel.DisplayColor(3, 3, 0))
  Panel.RomCharacter("4",        -2, 32, 16, 32, 0, 0, 0, Panel.DisplayColor(0, 3, 3))
  Panel.RomCharacter("5",        14, 32, 16, 32, 0, 0, 0, Panel.DisplayColor(3, 0, 3))
  Panel.RomCharacter("6",        33, 32, 16, 32, 0, 0, 0, Panel.DisplayColor(3, 3, 3))
  Panel.RomCharacter("7",        49, 40, 16, 32, 0, 0, 0, Panel.DisplayColor(3, 3, 1))
  waitcnt(clkfreq * s + cnt)

PUB TestRomFontDigitsWHelper(s) | d
  ' Clear the display, and show digits on the display.
  ReportStr(string("Test: ROM font digits with PASM helper"))
  Panel.Clear
  d := Helper.RomCharacter("4",  0,  3, 13, 26, 2, 3, 0, Panel.RgbColor(0, 2, 0))
  'ReportRomCharDebugOutput(d)
  Helper.RomCharacter("1", 13,  0, 14, 32, 2, 0, 0, Panel.RgbColor(2, 0, 0))
  Helper.RomCharacter("2", 32, -6, 16, 32, 0, 0, 0, Panel.RgbColor(0, 0, 2))
  Helper.RomCharacter("3", 54,  0, 16, 32, 0, 0, 0, Panel.RgbColor(2, 2, 0))
  Helper.RomCharacter("4", -2, 32, 16, 32, 0, 0, 0, Panel.RgbColor(0, 2, 2))
  Helper.RomCharacter("5", 14, 32, 16, 32, 0, 0, 0, Panel.RgbColor(2, 0, 2))
  Helper.RomCharacter("6", 33, 32, 16, 32, 0, 0, 0, Panel.RgbColor(2, 2, 2))
  Helper.RomCharacter("7", 49, 40, 16, 32, 0, 0, 0, Panel.RgbColor(3, 2, 1))
  waitcnt(clkfreq * s + cnt)

PUB TestStaticClockDisplay(s) | color, dimcolor
  ' The proposed clock display
  ' TODO: Adapt for different display configurations (rows and columns)
  ReportStr(string("Test: Static clock display"))
  Panel.Clear
  color := Panel.DisplayColor(MAX_COLOR_VALUE, MAX_COLOR_VALUE, 0)  ' Bright yellow
  dimcolor := Panel.DisplayColor(1, 1, 0)  ' dim yellow
  Panel.Bitmap1Bpp(@Digits10_1, 2,  0, 21, 13, 22, 0, 0, 0, color)
  Panel.RomCharacter("2",          14, 21, 13, 22, 2, 5, 0, color)
  Panel.Bitmap1Bpp(@Digits10_4, 2, 37, 21, 13, 22, 0, 0, 0, color)
  Panel.RomCharacter("5",          51, 21, 13, 22, 2, 5, 0, color)
  ' Write the colon
  Panel.Box(31, 26, 2, 2, color)
  Panel.Box(31, 36, 2, 2, color)
  ' "Offscreen" digits above and below.
  Panel.Bitmap1Bpp(@Digits10_2, 2,  0, -4, 13, 22, 0, 0, 0, dimcolor)
  Panel.RomCharacter("3",          14, -4, 13, 22, 2, 5, 0, dimcolor)
  Panel.Bitmap1Bpp(@Digits10_5, 2, 37, -4, 13, 22, 0, 0, 0, dimcolor)
  Panel.RomCharacter("6",          51, -4, 13, 22, 2, 5, 0, dimcolor)
  '
  Panel.Bitmap1Bpp(@Digits10_0, 2,  0, 46, 13, 22, 0, 0, 0, dimcolor)
  Panel.RomCharacter("1",          14, 46, 13, 22, 2, 5, 0, dimcolor)
  Panel.Bitmap1Bpp(@Digits10_3, 2, 37, 46, 13, 22, 0, 0, 0, dimcolor)
  Panel.RomCharacter("4",          51, 46, 13, 22, 2, 5, 0, dimcolor)
  waitcnt(clkfreq * s + cnt)

PUB TestStaticClockDisplayWHelper(s) | color, dimcolor, rawcolor, rawdimcolor, e
  ' The proposed clock display
  ReportStr(string("Test: Static clock display via 1bpp bitmap"))
  Panel.Clear
  color := Panel.RgbColor(6, 6, 0)  ' Bright yellow
  rawcolor := Panel.ToDisplayColor(color)
  dimcolor := Panel.RgbColor(1, 1, 0)  ' dim yellow
  rawdimcolor := Panel.ToDisplayColor(dimcolor)
  e := Helper.Bitmap1Bpp(@Digits10_1, 2,  0, 21, 13, 22, 0, 0, 0, color, 0)
  'ReportBitmapDebugOutput(e)
  'ReportBin(string("Expected  foreground color: "), rawcolor)
  Helper.Bitmap1Bpp(@Digits10_2, 2, 14, 21, 13, 22, 0, 0, 0, color, 0)
  Helper.Bitmap1Bpp(@Digits10_4, 2, 37, 21, 13, 22, 0, 0, 0, color, 0)
  Helper.Bitmap1Bpp(@Digits10_5, 2, 51, 21, 13, 22, 0, 0, 0, color, 0)
  ' Write the colon
  'Panel.Box(31, 26, 2, 2, rawcolor)
  'Panel.Box(31, 36, 2, 2, rawcolor)
  Helper.Box(31, 26, 2, 2, color)
  Helper.Box(31, 36, 2, 2, color)
  ' "Offscreen" digits above and below.
  e := Helper.Bitmap1Bpp(@Digits10_2, 2,  0, -4, 13, 22, 0, 0, 0, dimcolor, 0)
  'ReportBitmapDebugOutput(e)
  'ReportBin(string("Expected  foreground color: "), rawdimcolor)
  Helper.Bitmap1Bpp(@Digits10_3, 2, 14, -4, 13, 22, 0, 0, 0, dimcolor, 0)
  Helper.Bitmap1Bpp(@Digits10_5, 2, 37, -4, 13, 22, 0, 0, 0, dimcolor, 0)
  Helper.Bitmap1Bpp(@Digits10_6, 2, 51, -4, 13, 22, 0, 0, 0, dimcolor, 0)
  '
  Helper.Bitmap1Bpp(@Digits10_0, 2,  0, 46, 13, 22, 0, 0, 0, dimcolor, 0)
  Helper.Bitmap1Bpp(@Digits10_1, 2, 14, 46, 13, 22, 0, 0, 0, dimcolor, 0)
  Helper.Bitmap1Bpp(@Digits10_3, 2, 37, 46, 13, 22, 0, 0, 0, dimcolor, 0)
  Helper.Bitmap1Bpp(@Digits10_4, 2, 51, 46, 13, 22, 0, 0, 0, dimcolor, 0)
  waitcnt(clkfreq * s + cnt)

PUB TestSinglePixelPasmWritesBmp(s) | d
  ReportStr(string("Test: Single pixel writes via 1bpp bitmap with PASM helper"))
  Panel.Clear
  d := Helper.DrawPixelWithBitmap(1, 1, Panel.RgbColor(6, 0, 0)) ' red
  'return long[entry + 4]
  'ReportHex(string("Pixel write via bitmap result address difference: "), d)
  'ReportBitmapDebugOutput(d)
  'ReportDec(string("Fifo is empty boolean: "), d)

  Helper.DrawPixelWithBitmap(2, 2, Panel.RgbColor(0, 6, 0))
  Helper.DrawPixelWithBitmap(3, 3, Panel.RgbColor(0, 0, 6))
  Helper.DrawPixelWithBitmap(4, 4, Panel.RgbColor(6, 6, 6))

  '' Bottom half
  Helper.DrawPixelWithBitmap(1, 33, Panel.RgbColor(6, 6, 0))
  Helper.DrawPixelWithBitmap(2, 34, Panel.RgbColor(6, 0, 6))
  Helper.DrawPixelWithBitmap(3, 35, Panel.RgbColor(0, 6, 6))
  Helper.DrawPixelWithBitmap(4, 36, Panel.RgbColor(6, 6, 6))
  waitcnt(clkfreq * s + cnt)

PUB TestVerticalLinesBmp(s)
  ReportStr(string("Test: Vertical lines via 1bpp bitmap"))
  Panel.Clear
  Helper.VerticalLine(34,  1, 20, Panel.RgbColor(2, 2, 2))  ' white
  Helper.VerticalLine(35,  1, 20, Panel.RgbColor(2, 2, 2))
  Helper.VerticalLine(36,  1, 20, Panel.RgbColor(2, 2, 2))
  Helper.VerticalLine(40, 34, 20, Panel.RgbColor(0, 0, 3)) ' blue
  Helper.VerticalLine(41, 34, 20, Panel.RgbColor(0, 0, 3))
  Helper.VerticalLine(42, 34, 20, Panel.RgbColor(0, 0, 3))
  Helper.VerticalLine(16, 16, 32, Panel.RgbColor(0, 3, 0)) ' green
  Helper.VerticalLine(17, 16, 32, Panel.RgbColor(0, 3, 0))
  waitcnt(clkfreq * s + cnt)

' ========================================================================

PUB CheckStack(n, p, l) | lused, pend
  ' Examine the stack starting at p, that contains L longs, and determine
  ' how much of it was used, and how much was untouched.
  pend := p + (l << 2) - 4
  repeat while (pend => p) and (long[pend] == UNLIKELY_VALUE)
    pend -= 4

  ' Here pend is either p - 4, or the address of the last value
  ' used.
  '   longs_used = ((pend + 4) - p) >> 2
  '   longs_free = l - longs_used
  lused := ((pend + 4) - p) >> 2
  Ser.Str(n)
  ReportDec(string(" used longs: "), lused)
  Ser.Str(n)
  ReportDec(string(" free longs: "), l - lused)

PUB CheckMainStack
  CheckStack(string("Main"), WORD[STACK_BASE_ADDR], (END_OF_STACK - WORD[STACK_BASE_ADDR]) >> 2)
 
' ========================================================================

PUB ReportStr(s)
  Ser.Str(s)
  Ser.Tx(CR)

PUB ReportDec(label, v)
  Ser.Str(label)
  Ser.Dec(v)
  Ser.Tx(CR)

PUB ReportHex(label, v)
  Ser.Str(label)
  Ser.Hex(v, 8)
  Ser.Tx(CR)

PUB ReportBin(label, v)
  Ser.Str(label)
  Ser.Bin(v, 32)
  Ser.Tx(CR)

Pub ReportTenths(label, v)
  Ser.Str(label)
  Ser.Dec(v / 10)
  Ser.Tx(".")
  Ser.Tx("0" + (v // 10))
  Ser.Tx(CR)

PUB ReportColRow(label, c, r)
  Ser.Str(label)
  Ser.Dec(c)
  Ser.Tx(string(", "))
  Ser.Dec(r)
  Ser.Tx(CR)

PUB DumpFifo(label) | p
  ReportStr(label)
  p := Helper.GetCmdFifoHeader
  
  ReportHex(string("FiFo   ptr: "), p)
  ReportHex(string("Insert ptr: "), word[p + Fifo#INSERT_HUBADDR_OFFSET])
  ReportHex(string("Remove ptr: "), word[p + Fifo#REMOVE_HUBADDR_OFFSET])
  ReportHex(string("Start  ptr: "), word[p + Fifo#START_HUBADDR_OFFSET])
  ReportHex(string("Limit  ptr: "), word[p + Fifo#LIMIT_HUBADDR_OFFSET])
  ReportDec(string("Entry size: "), word[p + Fifo#ENTRY_SIZE_BYTES_OFFSET])

PUB DumpHelperDebug(label)
  ' Not currently up-to-date.
  ReportStr(label)
  ReportHex(string("FiFo base: "), HelperDebugOutput[0])
  ReportHex(string("FiFo insert ptr addr: "), HelperDebugOutput[1])
  ReportHex(string("FiFo remove ptr addr: "), HelperDebugOutput[2])
  ReportHex(string("FiFo remove count: "), HelperDebugOutput[5])
  ReportHex(string("FiFo remove ptr: "), HelperDebugOutput[6])
  ReportHex(string("FiFo start ptr: "), HelperDebugOutput[7])
  ReportHex(string("FiFo limit ptr: "), HelperDebugOutput[8])
  ReportHex(string("FiFo entry size: "), HelperDebugOutput[9])
  ReportHex(string("Display buf addr: "), HelperDebugOutput[10])
  ReportHex(string("last pixel addr: "), HelperDebugOutput[11])

PUB ReportBitmapDebugOutput(entry)
  ReportHex(string("result address:  "), entry)
  ReportHex(string("Requested background color: "), long[entry + 20])
  ReportHex(string("Seen      background color: "), long[entry +  8])
  ReportBin(string("Converted background color: "), long[entry +  0])
  ReportHex(string("Requested foreground color: "), long[entry + 24])
  ReportHex(string("Seen      foreground color: "), long[entry + 12])
  ReportBin(string("Converted foreground color: "), long[entry +  4])

PUB ReportRomCharDebugOutput(entry)
  ReportHex(string("result address:  "), entry)
  ReportHex(string("Rom char base address: "), long[entry +  0])
  ReportHex(string("Rom char seen:         "), long[entry +  4])
  ReportHex(string("Rom char requested:    "), word[entry + 16])
  ReportHex(string("Expected base address: "), ((word[entry + 16] & $FE) << 6) + $8000)

PUB ReportSetColorDebugOutput(entry) | c
  c := long[entry + 8]
  ReportHex(string("Original      color value: "), c)
  ReportBin(string("Converted raw color value: "), long[entry + 4])
  ReportBin(string("Expected  raw color value: "), {
    } Panel.DisplayColor(((c >> 20) & $3FF), ((c >> 10) & $3FF), c & $3FF))
  'ReportBin(string("Expected2 raw color value: "), Panel.DisplayColor(6, 6, 0))

